package ZipTie::Adapters::Cisco::IOS::AddressTranslationParser;

use strict;
use warnings;
use v5.10;

use ZipTie::Logger;
use ZipTie::Model::AddressTranslation;
use ZipTie::Addressing::Subnet;
use ZipTie::Adapters::Utils qw(get_crep int_port invert_mask mask_to_bits get_any_address_encoding);
# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

use Exporter 'import';

our @EXPORT_OK =
    qw(parseAddressTranslation);

sub parseAddressTranslation
{
    my ($args) = @_;
    my $addrTranslation;
    my $addrTranslationList; 
    my $ipAddrRe = get_crep('cipm') . "|any"; # allow for "any" address here
    my $doPAT;
    my $ifSubnets;
    my $ruleDirection;
    
 
    my $in = $args->{in};
    my $xlateDirection = $args->{direction} // "both";
    my $ifMapperFcn = $args->{interfaceMapper};

   # make a separate copy of running_config for searches embedded within the overall $in->{running_config} search
    my $rCfg = $in->{running_config};
	
	# Realistically, there are two "zones" in IOS, inside and outside.  We could use the keys
	# of the $args->{interfaces} hash.
	my @natSides = ( "inside", "outside");
	
	foreach my $i (0..1)
	{

		foreach my $eRuleDirection ("source", "dest[inaton]*")
		{
	    	NAT_RULE: while ( $in->{running_config} =~ /(^ip nat\s+$natSides[$i]\s+($eRuleDirection)\s+([a-z]+)\s+(.+?)\n)/mig)
	    	{
	    		my $line = $1;
	    		my $natDefinition = $3;
				my $addressingBlob = $4;
				if ($addressingBlob =~ /\sextendable$|\sextendable\s/){
					$LOGGER->debug("Cisco IOS NAT Translation; dropping rule; $line ; extendable rules not supported by QRM.");
					next NAT_RULE;
				}
				
				my $ruleDirection;
	    		if ($eRuleDirection =~ /^dest/i) 
	    		{ 
	    			$ruleDirection = "dest";
	    		}
	    		else
	    		{
	    			$ruleDirection = $eRuleDirection;
	    		}  # for labelling args

				if (not defined $addrTranslation)
				{
					$addrTranslation = new ZipTie::Model::AddressTranslation($natSides[$i], $natSides[$i-1]);
				    # Set interface zone mapper for address translator
				    $addrTranslation->setInterfaceZoneMapper ( sub { return $args->{interfaces} } );
				    # Common routine to add policy for IOS "overload" PAT ranges
				    _assignPATRanges($addrTranslation);					
				    # Throttle any translation that may come about from direction
				    given($xlateDirection) {
				    	when ("source") { $addrTranslation->setSourceTranslateOnly(); }
				    	when ("destination") { $addrTranslation->setDstTranslationOnly(); }
				    }

				    $addrTranslation->setTranslationName($natSides[$i] . "_" . $ruleDirection . "_NATRules" );				    
				}	    
	    		
	    		if ($natDefinition ~~ "static")
	    		{    			
	    			if (($addressingBlob) && ($addressingBlob =~ /^(udp|tcp)?\s*([\w\s\/\.\-]+)$/mig))
	    			{
		    			my $protocol = $1;
	    				$addressingBlob = $2;
					 	my $firstAddr;
						my $firstPort;
						my $secondAddr;
						my $secondPort;
						my $netPrefix;
						my $staticNATRouteFilter;
					
						# addresing blob for static is <localaddr> <localport> <globaladdr> <secondPort>
						# "esp" - special case for navigation across IPSec VPN tunnels, added in 12.2(13)T.  I'm not entirely
						# sure if it's this simple--the documentation (most of it unofficial) is very confusing
						if ($addressingBlob =~ /^(esp\s+)?($ipAddrRe)\s+(.+)$/)
						{
							$protocol = "esp" if defined ($1);
							$firstAddr = $2;
							$addressingBlob = $3;				
						}
						elsif ($addressingBlob =~ /^network\s+($ipAddrRe)\s+($ipAddrRe)\s\/(\d+)(\s+(.*))?/)
						{
							# The unique case of static network-based NAT.  This basically says, map network a to network
							# b statically, using the governing mask of the numeric prefix length as generated.
							$firstAddr = $1;
							$secondAddr = $2;
							$netPrefix = $3;
							$addressingBlob = (defined $4) ? $5 : "";
							
						}
						elsif ($addressingBlob =~ /^interface\s+(\S+)\s+(.+)$/)
						{
							$addressingBlob = $2;
							$firstAddr = $ifMapperFcn->($1, $args->{interfaceMapperContext});						
						}
						else
						{
							$LOGGER->debug("Cisco IOS NAT Translation; unable to detect source address at $addressingBlob");
							next;
						}
					
						if ($addressingBlob)
						{
							# Start with the possibility of a static mapping constrained by route map
							if ($addressingBlob =~ /^\s*($ipAddrRe)\s+(route\-map\s+(\S+))/)
							{
								$secondAddr = $1;
								if (defined $2)
								{
									$staticNATRouteFilter = $3;
									# +++wayne Do something with route map -- analyze it against $in->{routeMapFilters}
								}
							}
							# Account for the possibility of a port here
							elsif ($addressingBlob =~ /^((\w+)\s+)?($ipAddrRe)(\s+(\w+))?(.*)/)			
							{
								if (defined $1)
								{
									$firstPort = int_port($2);
								}
								if (defined $4)
								{
									$secondPort = int_port($5);
								}
								$secondAddr = $3;
								$addressingBlob = $6;
							}
							elsif ($addressingBlob =~ /^interface\s+(\S+)\s+(.*)/)
							{
								$addressingBlob = $2;
								$secondAddr =  $ifMapperFcn->($1, $args->{interfaceMapperContext});
							}							
						}
						
												
						# Note that if prefix was dug up, it is used identically on the mapped and criterion addresses 
						# (static network-based NAT), so its being passed once for each of the following calls in intentional.
						
						my $ruleHandle = $addrTranslation->addRule( {$ruleDirection . "AddrMapped" => $secondAddr, 
															$ruleDirection . "PortMapped" => $secondPort,
															$ruleDirection . "AddrMaskMapped" => $netPrefix,
															isStaticNATMapping => 1,
															protocol => $protocol // "ip"} );
						$addrTranslation->addRuleAddressCriterion( {rule => $ruleHandle, 
															$ruleDirection . "Ip" => $firstAddr,
															$ruleDirection . "Prefix" => $netPrefix,	
															$ruleDirection . "ServicePort" => $firstPort});
															
						# Handle "extendable" rules--we handle them as exemptions
						if (($addressingBlob) && ($addressingBlob =~ /\s*extendable/))
						{
							# Handle this as an exemption basically
							$addrTranslation->setExemptionRule($ruleHandle);
						}
	    			}
	    		}
	    		else
	    		{ # +++wayne all of dynamic NAT is going here--no support for route maps yet.
	    			my $criterionList;
	    			my $translationList;
	    			
		    		if ($natDefinition ~~ "list") 
		    		{
						# Dynamic policy NAT
						if ($addressingBlob =~ /^(\S+)\s+(.+)/)
						{
							my $listName = $1; 
							$addressingBlob = $2; # Move on up the list
							$criterionList = _findAndParseNATAccessList($rCfg, $listName, $ruleDirection);
							
							if (not defined $criterionList)
							{
								$LOGGER->debug("ZipTie::Adapters::Cisco::IOS::AddressTranslationParser::parseAddressTranslation ; " .
								 "unable to find or parse NAT access list \'$listName\'");
								 next;  # next IP NAT statement
							}
						}
						elsif ($natDefinition ~~ "route-map")
						{
							# +++wayne implement me
						}
							
						if ($addressingBlob =~ /^interface\s+(\S+)(\s+(.+))?$/)
						{
							my $ifName = $1;
							$addressingBlob = $3;  #which may be undef
							# Call caller's passed interface name -> addr function
							my $ifAddr = $ifMapperFcn->($ifName, $args->{interfaceMapperContext});
							
							if (defined $ifAddr)
							{
								push(@{$translationList},  { $ruleDirection . "AddrMapped" =>  $ifAddr} );
							}
							else
							{
								$LOGGER->debug("ZipTie::Adapters::Cisco::IOS::AddressTranslationParser::parseAddressTranslation ; " .
								 "unable to find address for interface \'$ifName\'");
								 next;  # next IP NAT statement							
							}
						}
						elsif ($addressingBlob =~ /^pool\s+(\S+)(\s+(.*))?$/)
						{
							my $poolName = $1;
							$addressingBlob = $3;
							$translationList = _findAndParseNATPool($rCfg, $poolName, $ruleDirection);						
						}				
		
						$doPAT = (($addressingBlob) && ($addressingBlob =~ /\s*overload/));
		
						# Now we are in a position to allocate the rule and apply the address criteria to it.
						# This may end up resulting in a different numbering and ordering of generated rule entries,
						# but the point of the abstraction maintained into AddressTranslation is to shield the 
						# per-config-regime caller from having to sweat that.
						
						my $ruleHandle = $addrTranslation->addRule( {
										protocol => "ip"} );
	
						# Set dynamic NAT to not allow reverse connection opens.  This is explicitly controlled
						# Through this method
						$addrTranslation->setEstablishedReverseConnectionTrafficOnly($ruleHandle, "true");
															
						# We have separate sets of elements to add - address translation and criterion rule data.									
						foreach my $atRule (@{$translationList})
						{
							$atRule->{"rule"} = $ruleHandle;
							$atRule->{"usePATOverload"} = $doPAT;
							$addrTranslation->addRuleAddressTranslationMapping($atRule );
						}
										
						foreach my $critHash (@{$criterionList})
						{
							$critHash->{"rule"} = $ruleHandle;
							$addrTranslation->addRuleAddressCriterion( $critHash );
						}															
		    		}    			
	    		}
	      	}
	      	if (defined $addrTranslation)
	      	{
				push ( @{$addrTranslationList}, $addrTranslation);
				undef $addrTranslation;  # it's in the list now....      		
	      	}
		}
	}	
	return $addrTranslationList;
}

sub _assignPATRanges
{
	my $addrTranslationInst = shift;
	
	$addrTranslationInst->addPortTranslationRange(  (0,511) );
	$addrTranslationInst->addPortTranslationRange( (512,1023) );
	$addrTranslationInst->addPortTranslationRange( (1024,65535) );	
}

#  Parsing for the pool by name, entry of its data into a list suitable for calling 
# AddressTranslation::addRuleTranslationMapping with
sub _findAndParseNATPool
{
	my ($rc, $poolName, $rDirection) = @_;
    my $ipAddrRe = get_crep('cipm');
    my $pEntry;
    my $pEntryList;
	
	# There are at least two addresses in each pool entry
	while ($rc =~ /^ip\s+nat\s+pool\s+$poolName\s+($ipAddrRe)\s+($ipAddrRe)(\s+(.*))?/mig)
	{
		undef $pEntry;
		my $natListBlob = $4;
		($pEntry->{$rDirection . "AddrMapped"}, $pEntry->{$rDirection . "AddrEndMapped"}) = ($1, $2);
		
		# See if there are any trailing arguments to the nat pool entry
		if (defined $natListBlob and $natListBlob =~ /^(\S+)\s+($ipAddrRe|\d+)(\s+(.*))?/)
		{
			$natListBlob = (defined $3) ? $4 : undef;
			
			given ($1)
			{
				# turns out both of these map to the same param, but I'm not entirely sure if will
				# always be that way
				when ("netmask") {$pEntry->{$rDirection . "AddrMaskMapped"} = $2;}				
				when ("prefix-length")  {$pEntry->{$rDirection . "AddrMaskMapped"} = $2;}
			}
			
			if (defined $natListBlob and $natListBlob =~ /type\s+([\w\-]+)/)
			{
				given ($1)
				{
					# Both of the following map to an applied addressing usage.
					when ("match-host") {$pEntry->{"useAppliedAddressRules"} = "true"; }
				}
			}	
		}
		push (@{$pEntryList}, $pEntry);
	}
	
	return $pEntryList;
}						

sub _findAndParseNATAccessList
{
	my ($rc, $listName, $direction) = @_;
	my $servicePort;
	my $alCriterionList;
	my $alCurrentCriterion = undef;
    my $ipAddrRe = get_crep('cipm');
	
	# while  ($in =~ /^(ip )?access-list\s+(extended|dynamic|standard)\s+$listName\s+(permit|deny)\s+((tcp|udp|ip)\s+)?([\w\s\.]+)/sig )
	# This could be a single line entry, or a multiline extended entry.  So, let's start by iterating on all access lists
	# so identified, and it is safe to snarf that up to create a blob for the line(s) which follow.  Pick up EVERYTHING
	# which follows up to the next new line which *isn't* followed by white space.  Make that last check ungreedy.
	
	while  ($rc =~ /^(ip )?access\-list\s+((extended|dynamic|standard)\s+)?$listName\s+(.+?)(?=\n\S)/smig) 
	 {
	 	my $alType = $3 // _tryToResolveAccessListType($listName);
		my $alMBlob = $4;
		
		while ($alMBlob =~ /(permit|deny)\s+((tcp|udp|ip|esp|icmp)\s+)?([\w\s\.]+?)$/mig)
		{
			# +++wayne if this is an "extended", we need a parsing of the block within
			my ($alEntryDisp, $alBlob) = ($1, $4);
			my $ipProtocol = (defined $2) ? $3 : undef;
			my $pNum;
			
			if ($alEntryDisp =~ /deny/)
			{
				$alCurrentCriterion->{"negateMatch"} = 1;
			}
			
			$alCurrentCriterion->{"ipProtocol"} = $ipProtocol // undef;
		
			# Now, as I understand extended access lists and how they apply here, "any" is
			# something of a noise word if it's seen just *once*.  Which is to say, 
			# "permit any any" is an indication that we should encode the "0.0.0.0/0" encoding
			if ($alType and $alType ~~ "extended")
			{
				if ($alBlob =~ /^any\s+([\w\s\.]+)/)
				{
					$alBlob = $1;
					
					# we've moved things up to the point where the non-extended parsing code can deal with it
				}
				elsif ($alBlob =~ /^host\s+($ipAddrRe)(\s+(.+))?/)
				{
					$alCurrentCriterion->{$direction . "Ip"} = $1;
					$alBlob = $3;  # pick up what we can here...
					
					if ($alBlob =~ /^any\s*(.*)/)
					{
						$alBlob = $1;  # trim off "any"
					} elsif ($alBlob =~ /^host\s+($ipAddrRe)/){
						$alCurrentCriterion->{"dest" . "Ip"} = $1;
					}
				}
			}
			
			if ($alBlob and (not defined $alCurrentCriterion->{$direction . "Ip"}) )
			{
				# Pick up ip (the one required element here)
				if ($alBlob =~ /^($ipAddrRe)(\s+($ipAddrRe))?((\s+[\w\s\.]*))?/)
				{
					$alCurrentCriterion->{$direction . "Ip"} = $1;
					if (defined $2)
					{
						my $pf = mask_to_bits($3);
						# Cisco inverse order?
						if ($pf =~ /$ipAddrRe/)
						{
							$pf = mask_to_bits(invert_mask($pf));
						}
						$alCurrentCriterion->{$direction . "Prefix"} = $pf;
					}
					$alBlob = $5;
				}
				elsif ($alBlob =~ /^(\w+)((\s+[\w\s\.]*))?/)
				{
					if ($1 ~~ "any")
					{
						$alCurrentCriterion->{$direction . "Ip"} = $1;  # AddressTranslation can in fact deal with this
						$alBlob = $2;
					}
					elsif ($1 ~~ "host")
					{
						if ($2 =~ /^\s*($ipAddrRe)\s+(.*)/)
						{
							$alCurrentCriterion->{$direction . "Ip"} = $1;
							$alBlob = $2;
						}
					}
					else
					{
						$alCurrentCriterion->{$direction . "IpObjectGroup"} = $1;
						$alBlob = $3;							
					}
				}
				else
				{
					$LOGGER->debug("ZipTie::Adapters::Cisco::IOS::AddressTranslationParser::parseAddressTranslation; " .
					" unable to pick out access-list NAT source at \'$alBlob\'");
					next;
				}				
			}
			# Now go for service/port
			if ((defined $alBlob) and( $alBlob =~ /(eq)\s+(\S+)$/))
			{
				$pNum = int_port($2);
				$alBlob = $3;
				
				if ($pNum =~ /^\d+$/)
				{
					$alCurrentCriterion->{$direction . "ServicePort"} = $pNum;
				}
				else 
				{
					$LOGGER->debug('Unable to parse or map presumed NAT filter entry source port/service: ' . $pNum );
				}
			} 				
			
			push (@{$alCriterionList}, $alCurrentCriterion);
			# clear it for reinit
			undef $alCurrentCriterion;
		} # end of iterate over $alMBlob		
	}	
	return $alCriterionList;
}

#  This is mostly used to coalesce the rules for "implied" type of numerically
#  identified access list.  It should probably be somewhere more "IOS common",
#   but no such place exists now...
#
sub _tryToResolveAccessListType
{
	my $accessListID = shift;
	my $extended;
	$accessListID = $accessListID + 0;
	
	for  ($accessListID)
	{
		if (1<= $_ && $_ <= 99) { $extended = 0; last; }
		if (100 <= $_ && $_ <= 199) {$extended = 1; last; }
		if (1300<= $_ && $_ <= 1399) {$extended = 0; last; }
		if (2600<= $_ && $_ <= 2699) {$extended = 1; last; }
		 return undef;  # Out of range or nonnumeric
 	}
	
	return ($extended) ? "extended" : "simple";
}

=head1 NAME

    ZipTie::Adapters::Cisco::IOS::AddressTranslationParser - Parser for IOS
    NAT and PAT.

=head1 VERSION

    This document refers to version 1.0 of ZipTie::Adapters::Cisco::IOS::AddressTranslationParser, released 31 May, 2012.

=head1 SYNOPSIS

  use ZipTie::Adapters::Cisco::IOS::AddressTranslationParser;
  my $addressTranslation = Ziptie::Model::AddressTranslation->new();
  parseAddressTranslation( 
    in                 => $in,
    interfaces         => parse_interfaces($in),
    addressTranslation => $addressTranslation);
  $addressTranslation->generateTranslationList($out);

=head1 DESCRIPTION

    This module scans the elements of Cisco NAT, and handles "ip nat" 
    configuration statements, nat pools, and application to interfaces.
    It sets those up by allocating and acting as the compositional 
    interface point to Ziptie::Model::AddressTranslation.

=head2 Overview

   Mostly provides the interface to do NAT parsing, plus callbacks and
   interface points to the Ziptie::Model::AddressTranslation it gets
    at parse time.

=head2 Constructor and initialization 

    None; this is an interface package

=head2 Class and object methods

=over

=item "parseAddressTranslation"
    routine which will parse a running config for address translation 
    attributes. It can be limited to looking for source or destination 
    address and port translations.

    Parameters:  (passed as named interfaces in hash, "$args")

=over 6

=item * in => running configuration of device (required)

=item * interfaces => output from parse_interfaces() in IOS parser (required)

=item * addressTranslation => reference to ZipTie::Model::AddressTranslation object constructed by caller (required).

=item * direction => "source" or "destination" will cause the parsing and generation to be focused entirely on source NAT or destination NAT. (optional).

=item * interfaceMapped => reference to function which will be passed through/called back as intefaceZoneMapper for AddressTranslation object

=back
    
    Returns: list of AddressTranslation object instances reflecting NAT configuration detected
    in $args->{in}->{running-config}.  There is roughly one for all "ip nat source..." rules and
    one for all "ip nat [inside|outside] [source | destination]..." rule set permutation, leading
    to a max of four AddressTranslation instances in the list.

=back

=head1 ENVIRONMENT

=head1 DIAGNOSTICS

=over

=item "error message that may appear"
    Explanation of error message

=back

=head1 BUGS
    Description of known bugs (and any work-arounds).
    Usually also includes an invitation to send the author bug reports.

=head1 FILES
     Ziptie::Model::AddressTranslation 

=head1 SEE ALSO

    http://wiki.q1labs.inc/display/QRM/Address+Translation+Support+in+ZipTie+and+CSM

=head1 AUTHOR(S)
    Wayne Tackabury
    wtackabury@us.ibm.com

=head1 COPYRIGHT
    Copyright (c) 2012, IBM Corporation. All Rights Reserved.
    This module is free software. It may be used, redistributed
    and/or modified under the same terms as Perl itself.
=cut

1;
