package ZipTie::Adapters::Cisco::Nexus;

use strict;
use warnings;
use v5.10;

use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::Cisco::Nexus::AutoLogin;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);

use ZipTie::Adapters::Cisco::Nexus::Disconnect qw(disconnect);
use ZipTie::ConnectionPath;
use ZipTie::Typer;
use ZipTie::Model::XmlPrint;
use ZipTie::Logger;

use ZipTie::Adapters::Cisco::Nexus::Parsers qw(create_config parse_system parse_chassis parse_access_ports parse_object_groups parse_access_lists 
			parse_interfaces parse_static_routing parse_routing parse_dynamic_routes open_vdc_section close_vdc_section
			get_default_vrf_name parse_vlans);
use ZipTie::Adapters::Cisco::Nexus::ResponseHandler qw(populate_virtual_device_context delete_instance_data netconf_elements_extract  netconf_response_elements_extract netconf_text_response_elements_extract get_virtual_device_contexts nexus_command_cli_response cli_xml_command);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

# List the command invocations we can make here under "show" that don't require any sort
# of synchronization.

my $Nexus_Filesystems = [ "bootflash", "slot0", "volatile", "log", "logflash",  "system" ];
my $Nexus_XML_Show_Commands = [ "module",  "interface brief", "interface snmp-ifindex",  
		"ip access-lists", "vlan", "vtp status", "spanning-tree summary", "object-group" ];
my $Nexus_Text_Show_Commands = [ "snmp", "running-config", "startup-config" ];
my $Nexus_Feature_Commands = [ "hsrp", "vrrp", "vtp", "glbp" ];
my $Nexus_IP_Routing_Protocols = [ "eigrp", "ospf", "rip" ];

# These are manifest constants to direct how we interpret XML command issuance

# If this is XML over SSH, we issue commands with the " | xml" CLI designator
use constant NEXUS_XML_COMMANDS_XML_SSH => 1;
# if this is a real Netconf connection, then we issue Netconf requests to the port requested
use constant NEXUS_XML_COMMANDS_NETCONF => 2;
# The real kicker is that at least one device (the Nexus 4000) in fact does not support Netconf or
# XML command formatting *at all*.  We actually have to back off to a pure CLI text response mode of
# parsing, and the command as issued has no XML form or request designator (although we'll try the latter
# and glean if we got an error to see if we're in this predicament
use constant NEXUS_XML_NO_XML_TEXT_ONLY => 3;

# If we need to explicitly map where a command result will be placed in terms of its results, we can use this as an override.
# Ordinarily, the rule is the bucket strips off "ip", then uses only one word thereafter.  For some commands (e.g., interface
# subcommands), this won't be sufficient.
my $AltSBktName = {	"interface snmp-ifindex" => "interface_ifindex" };


sub backup
{
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;
	my $responses  = {};
	my $deviceContextList; 
	my $localVDC;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ( $cli_protocol, $exec_prompt_regex ) = _connect($connection_path);

	# Collect a few up-front responses from the device.  The first few here are a little awkward in that we 
	# haven't got everything ginned up yet to do this by calling populate_virtual_device_context(), 
	# which will really get the ball rolling on command responses and setting them up.  
	# So, we still have to sweat Netconf/CLI code paths,for example.
	my $mHostName;
	# Only set terminal length if netconf is not our CLI protocol.  
	if ($cli_protocol->get_protocol_name() =~ /Netconf/i)
	{
		$responses->{xmlForm} = NEXUS_XML_COMMANDS_NETCONF;
		# Since we're here, add the known element namespace(s) for later strip
		$cli_protocol->add_netconf_element_namespace("mod");
		# need to extract a bare host name.  It may be domain qualified.  The first one is an odd text command which
		# requires a special containm{}ent tag.
		$mHostName = netconf_text_response_elements_extract($cli_protocol->send_netconf_exec_msg("show hostname"), "mod:hostname");		
		$responses->{version}        = nexus_command_cli_response($responses, $cli_protocol, "show version",  $exec_prompt_regex);
	}
	else
	{	
		# need to extract a bare host name.  It may be domain qualified.  We're *also* going to try this
		# using a XML form to see if we get an error.  If we do, we know we need to fall back to
		# purse CLI.
		
		$mHostName = netconf_elements_extract( "show hostname | xml", $exec_prompt_regex, $cli_protocol );
		
		if ($mHostName)
		{
			# fine, there's a conformant SSH -> XML formatter
			$responses->{xmlForm} = NEXUS_XML_COMMANDS_XML_SSH;
			# get the hostname out of the XML
			if ($mHostName =~ /\<hostname\>(\S+?)\<\/hostname/im)
			{
				$mHostName = $1;  # just the host, at the most
			}
		}
		else
		{
			$LOGGER->fatal_error_code("$DEVICE_ISSUE", $connection_path->get_ip_address(),undef,"No XML/NETCONF formatted response available which is required for successful backup.
								\nPossible reasons:
								\nUnsupported device model/OS
								\nDevice XML agent not enabled");
			return;
		}
		
		# The version is only retrieved at the admin VDC level....
		$responses->{version}        = nexus_command_cli_response($responses, $cli_protocol, "show version", $exec_prompt_regex   );
	}

	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle( "Nexus", $connection_path->get_ip_address() );

	$mHostName =~ s/.*?\n([\w_\-\.]+)[\s\r\n]+.*/$1/sm;
	$LOGGER->debug("Cisco::Nexus::backup : local hostname identified as \'$mHostName\'");
	$responses->{hostname} = $mHostName;
	
	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'cisco', "http://www.ziptie.org/model/cisco/1.0 cisco.xsd" );
	$printer->open_model();

	# unless netconf protocol get current vdc
	my $isAdminContext = is_admin_context ($exec_prompt_regex, $cli_protocol);
	
	# Initialize our contexts with a null (root) value for the system	
	# Get VDContexts on device / the list after that
	# Only look for vdcs if this is the "admin" context
	if ($isAdminContext){ 
		$deviceContextList = get_virtual_device_contexts($responses, $mHostName, $cli_protocol, $exec_prompt_regex);
	} else {
		$LOGGER->debug("We are not in the 'Admin' context so skipping vdc lookup.");
	}
	
	# We have to walk the device list twice, once just to populate, and then once to parse.
	# Reason being, we have to have all of the context configs read in by the time we hit "create_config" for the
	# global section.  Since that's a "light" traversal, use the implied "map" array enumeration with embedeed block 
	# side effect. 
	
	map {populate_virtual_device_context({responses => $responses, cliProtocolInst => $cli_protocol, promptRegex => $exec_prompt_regex,
				vdcName => $_, filesystems => $Nexus_Filesystems, xmlShowCommands => $Nexus_XML_Show_Commands, textShowCommands => $Nexus_Text_Show_Commands,
				featureCommands => $Nexus_Feature_Commands, ipRoutingProtocols => $Nexus_IP_Routing_Protocols, altBucketName => $AltSBktName,
				additionalInterfaceCommands => 1})} ("", @{$deviceContextList}); 		

	my $adminContextVlanXML = '';
	
	foreach $localVDC ("", @{$deviceContextList})
	{
		my $rs; 
		
		# Now, populate all responses for the current VDC (which might be global)
		# have to patch in version...
		if ($localVDC)
		{
			if (not defined $responses->{contexts}->{"$localVDC"})
			{
				# populate_virtual_device_context couldn't deal with it
				$LOGGER->debug("Cisco::Nexus::backup : error encountered in populating virtual device context \'$localVDC\'");
				next;				
			}
			
			$responses->{contexts}->{$localVDC}->{version} = $responses->{version};
			$rs = $responses->{contexts}->{$localVDC}; 	
			
			open_vdc_section($responses, $printer, $deviceContextList);	

			parse_system($rs, $printer);
		}
		else
		{
			$rs = $responses;

			# Parse the top level elements, most notably, the chassis level
			parse_system($rs, $printer);
			parse_chassis($rs, $printer);
			
			#Serialize the configuration.  This is the end of the core ZED section.
			create_config ($rs, $printer);
		}
		
		# These are things we parse for both global contexts and
		# VDC's.
		parse_access_ports($rs, $printer);
		parse_dynamic_routes($rs, $printer);
	
		# Parse object group data
		parse_object_groups($rs, $printer);
		
		# Filter lists want to come after the object groups they reference but
		# before the interface definitions that reference them....fortunately, the 
		# schema sets us up for that
		
		parse_access_lists($rs, $printer);
		
		# Commence parsing interface data...store away subnets hash which is created for us as
		# a side-effect
		
		($rs->{subnets}) = parse_interfaces($rs, $printer);
	
		parse_routing($rs, $printer);
		
		parse_static_routing($rs, $printer);

		# vlan parsing is difficult because in the admin device (non VDC), the vlans element comes after the subelement element.
		# If we are parsing a VDC we can simply call parse_vlans as usual, if we are parsing the admin device, we still need to
		# call parse_vlans, but print the XML to a buffer to be printed after all the other VDCs have been parsed.
		
		if ($localVDC)
		{
			parse_vlans($rs, $printer);

			close_vdc_section($responses, $printer, $localVDC);
		}		
		else
		{
			# Parse the vlan information but store the output into a buffer to be printed at a later time.
			open(my $vlanFh, '>', \$adminContextVlanXML);
			my $vlanPrinter = ZipTie::Model::XmlPrint->new($vlanFh, 'cisco', 'http://www.ziptie.org/model/cisco/1.0 cisco.xsd');
			parse_vlans($rs, $vlanPrinter);
			close($vlanFh);
		}
	}

	# Now that all of the subelements have been printed, we can output the admin context vlan information.
	$printer->print_string($adminContextVlanXML);
	
	# remove instance data
	delete_instance_data({responses => $responses, textShowCommands => $Nexus_Text_Show_Commands, xmlShowCommands => $Nexus_XML_Show_Commands,
			featureCommands => $Nexus_Feature_Commands, ipRoutingProtocols => $Nexus_IP_Routing_Protocols});
		
	$printer->close_model;

	# Close the model output file handle
	close_model_filehandle($filehandle);

	disconnect($cli_protocol);
}

sub restore
{
	# +++waynet we don't support this, and this is crazy, but here's my entry, so write me maybe...
}

sub _connect
{

	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the IOS device
	my $exec_prompt_regex = ZipTie::Adapters::Cisco::Nexus::AutoLogin::execute( $cli_protocol, $connection_path );
	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name( "execPrompt", $exec_prompt_regex );

	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $exec_prompt_regex );
}

1;


sub is_admin_context {
	my $exec_prompt_regex = shift;
	my $cli_protocol      = shift;

	if ( $cli_protocol->get_protocol_name() =~ /Netconf/i ) {
		return 0; # vdc modelling not supported for Netconf
	}
		
	my $currentVdc;
	my $currentVdcResponse = $cli_protocol->send_and_wait_for( "show vdc current-vdc", $exec_prompt_regex );
	if ( $currentVdcResponse =~ /Current\svdc\sis\s\d+\s-\s(\S+)/g ) {
		$currentVdc = $1;
		# find context type
		my $contextType = $cli_protocol->send_and_wait_for( "show vdc $currentVdc", $exec_prompt_regex );
		if ( $contextType =~ /\s+Admin\s+/g ) {
			return 1;
		}
	}
	return 0;
}
__END__

=head1 NAME

ZipTie::Adapters::Cisco::Nexus - Adapter for performing various operations against Cisco Nexus switches.

=head1 SYNOPSIS

    use ZipTie::Adapters::Cisco::Nexus;
	ZipTie::Adapters::Cisco::Nexus::backup( $backup_document );

=head1 DESCRIPTION

This module represents an adapter that can be used to perform various operations against against Cisco Nexus switches and routers.

Generally you would run this module through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document.  This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question.  This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retreived, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2012, IBM Corp.

=head1 AUTHOR

Contributor(s): Wayne Tackabury (wtackabury@us.ibm.com)
Date: Novemder 10, 2012

=cut
