package ZipTie::Adapters::Cisco::Nexus::Parsers;

use strict;
use warnings;
use v5.10;

use ZipTie::Adapters::Utils qw(trim get_crep parseCIDR create_network_element get_any_address_encoding get_port_number get_mask);
use ZipTie::Addressing::Subnet;
use ZipTie::Addressing::RouteFilter;
use ZipTie::Ranges;

# Not sure if we want this import path, but it works for me for now

use XML::Twig;
use XML::Simple;
use MIME::Base64;

use Exporter 'import';
our @EXPORT_OK = qw(parse_system parse_chassis create_config parse_access_ports parse_interfaces 
		parse_object_groups parse_access_lists parse_routing parse_static_routing expand_nexus_interface_name 
		parse_dynamic_routes parse_routing_neighbors open_vdc_section close_vdc_section get_default_vrf_name
		parse_vlans);

my $LOGGER = ZipTie::Logger::get_logger();

use constant SUPPORTED_ROUTING_PROTOCOLS => qw(eigrp rip);
use constant SUPPORTED_ROUTING_SECTIONS => qw(rip cisco:eigrp);


# +++wayne modullarize these
# If this is XML over SSH, we issue commands with the " | xml" CLI designator
use constant NEXUS_XML_COMMANDS_XML_SSH => 1;
# if this is a real Netconf connection, then we issue Netconf requests to the port requested
use constant NEXUS_XML_COMMANDS_NETCONF => 2;
# The real kicker is that at least one device (the Nexus 4000) in fact does not support Netconf or
# XML command formatting *at all*.  We actually have to back off to a pure CLI text response mode of
# parsing, and the command as issued has no XML form or request designator (although we'll try the latter
# and glean if we got an error to see if we're in this predicament
use constant NEXUS_XML_NO_XML_TEXT_ONLY => 3;

use constant DEFAULT_VRF_NAME => "default";
		
# parse_systen
#
# Parse the ziptie core sections describing top level system attributes in the ZED.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  
#
sub parse_system{
	my ($in, $out) = @_;
	my $sysData = _parse_system($in);
	my $lastReboot = undef;

	# The following vairables are required by the schema, however if they are missing it's
	# not important enough to fail the backup.  Report debug errors.
	if (!defined($sysData->{version}))
	{
		$LOGGER->debug('Failed to parse required element core:version from \'show version\'.');
	}
	if (defined($sysData->{utDays}) && defined($sysData->{utHrs}) && defined($sysData->{utMins}) && defined($sysData->{utSecs}))
	{
		# Calculate lastreboot (based of of Cisco IOS)
		$lastReboot = time();
		$lastReboot -= $sysData->{utDays} * 24 * 60 * 60 if (defined($sysData->{utDays}));
		$lastReboot -= $sysData->{utHrs} * 60 * 60 if (defined($sysData->{utHrs}));
		$lastReboot -= $sysData->{utMins} * 60 if (defined($sysData->{utMins}));
		$lastReboot -= $sysData->{utSecs} if (defined($sysData->{utSecs}));
	}else{
		$LOGGER->debug('Failed to parse and calculate required element core:lastReboot from \'show system\'.');
	}

	$out->print_element('core:systemName', $sysData->{systemName});

	$out->open_element('core:osInfo');
	$out->print_element('core:fileName', $sysData->{fileName}) if (defined($sysData->{fileName}));
	$out->print_element('core:make', 'Cisco');
	$out->print_element('core:version', $sysData->{version}) if (defined($sysData->{version}));
	$out->print_element('core:osType', 'NX-OS');

	# Here is where packetProcessingOrdering would go.  Because Cisco Nexus currently doesn't support NAT
	# the packetProcessingOrdering is not currently required.

	$out->close_element('core:osInfo');

	$out->print_element('core:biosVersion', $sysData->{biosVersion}) if (defined($sysData->{biosVersion}));
	$out->print_element('core:deviceType', 'Switch');
	$out->print_element('core:contact', $sysData->{contact}) if (defined($sysData->{contact}));
	$out->print_element('core:lastReboot', $lastReboot) if (defined($lastReboot));
}

sub _parse_system{
	my ($in) = @_;
	my $sysData = {};
	my $parseSystemReturn;

	given ($in->{xmlForm})
	{
		when (NEXUS_XML_COMMANDS_XML_SSH) {$parseSystemReturn =  _parse_system_xml($in, $sysData);};
		when (NEXUS_XML_COMMANDS_NETCONF ) {$parseSystemReturn =  _parse_system_xml($in, $sysData);};
		when (NEXUS_XML_NO_XML_TEXT_ONLY) {$parseSystemReturn = _parse_system_cli_text($in, $sysData)};
	}

	# Avoid partial assignment
	if (!$parseSystemReturn){
		$sysData = undef;
	}
	
	return $sysData;
}

sub _parse_system_xml{
	my ($in, $sysData) = @_;

	# Check for version data.  If not present, then this is a VDC and
	# get what we need to where we can.
	if (not $in->{version}){
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_version() : no version data found in device response.');
	}else{
		my $twig = XML::Twig->new(twig_roots => {
					host_name => _createSimpleTwigSetter(\$sysData->{systemName}),
					isan_file_name => _createSimpleTwigSetter(\$sysData->{fileName}),
					sys_ver_str => _createSimpleTwigSetter(\$sysData->{version}),
					bios_ver_str => _createSimpleTwigSetter(\$sysData->{biosVersion}),
					kern_uptm_days => _createSimpleTwigSetter(\$sysData->{utDays}),
					kern_uptm_hrs => _createSimpleTwigSetter(\$sysData->{utHrs}),
					kern_uptm_mins => _createSimpleTwigSetter(\$sysData->{utMins}),
					kern_uptm_secs => _createSimpleTwigSetter(\$sysData->{utSecs})
				} );
		$twig->parse($in->{version});
	
		# We will overwrite what is in the version host_name with the definition of hostname in the
		# running-config, if present.  This is necessary for subcontexts
		if ($in->{running_config} =~ /^(host|switch)name (\S+)/m)
		{
			my $hostname = $2;
			$sysData->{systemName} = $hostname;
		}
	
		# This looks convoluted, but it has to handle some five different cases
		if ($in->{snmp} && $in->{snmp} =~ /^sys contact: +(\w)/m)
		{
				($sysData->{contact}) = $in->{snmp} =~ /^sys contact: +(.*)$/m;
				chomp($sysData->{contact});		
		}
	}
	
	return scalar (keys ( %{$sysData}));
}

sub _parse_system_cli_text
{
	my ($in, $sysData) = @_;

	return scalar (keys ( %{$sysData}));
}

# parse_chassis
#
# Parse the ziptie core sections describing asset data and cards (and their asset data)
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  
#
sub parse_chassis
{
	my ($in, $out) = @_;

	# Check for a lack of version data
	if (not $in->{version})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_chassis() : no version data found in device response.');
		return;		
	}

	my $modules = _parse_chassis($in);
	
	if (not defined $modules)
	{
		# _parse_chassis returned error and logged it
		return;
	}
	
	# Use the first two tokens in the description (chassis_id) as the chassis modelNumber.
	my $modelNumber = undef;
	if ($modules->{_chassis}->{description} =~ /^\s*(\S+(?:\s+\S+)?)/){
		$modelNumber = $1;
	}else{
		$modelNumber = 'Unknown';
	}

	$out->open_element('chassis');

	$out->open_element('core:asset');
	$out->print_element('core:assetType', 'Chassis');

	$out->open_element('core:factoryinfo');
	$out->print_element('core:make', 'Cisco');
	$out->print_element('core:modelNumber', $modelNumber);

	# Serial numbers can be the same on multiple virtual devices, QRM needs unique serial numbers so appending hostname
	my $serial_number = $modules->{_chassis}->{serialNumber};
	if($in->{hostname}){
		$serial_number .= '_' . $in->{hostname};
	}
	$out->print_element('core:serialNumber', $serial_number);

	$out->close_element('core:factoryinfo');

	$out->close_element('core:asset');

	$out->print_element('core:description', $modules->{_chassis}->{description}) if (defined($modules->{_chassis}->{description}));

	foreach my $card (values(%{$modules->{_cards}}))
	{
		if (!defined($card->{modelNumber}))
		{
			# modelNumber must be defined (required by schema).  If it's not defined this isn't
			# serious enough to stop the adapter.
			$LOGGER->debug('Required element core:modelNumber missing from card with id: ' . $card->{id});
			next;
		}

		$out->open_element('card');

		$out->open_element('core:asset');
		$out->print_element('core:assetType', 'Card');
		$out->open_element('core:factoryinfo');
		$out->print_element('core:hardwareVersion', $card->{hardwareVersion}) if (defined($card->{hardwareVersion}));
		$out->print_element('core:make', 'Cisco');  # +++waynet need to parse this somehow for third party cards
		$out->print_element('core:modelNumber', $card->{modelNumber});
		$out->print_element('core:serialNumber', $card->{serialNumber}) if (defined($card->{serialNumber}));
		$out->close_element('core:factoryinfo');
		$out->close_element('core:asset');

		$out->print_element('core:description', $card->{description}) if (defined($card->{description}));
		$out->print_element('portCount', $card->{portCount}) if (defined($card->{portCount}));
		$out->print_element('softwareVersion', $card->{softwareVersion}) if (defined($card->{softwareVersion}));
		$out->print_element('status', $card->{status}) if (defined($card->{status}));

		$out->close_element('card');
	}

	if (defined($modules->{_chassis}->{cpuType}))
	{
		trim($modules->{_chassis}->{cpuType});

		$out->open_element('cpu');
		$out->print_element('cpuType', $modules->{_chassis}->{cpuType});
		$out->close_element('cpu');
	}

	# TODO: Hardcoded dir paths?  Probably need to improve this.
	# _parseDirOutput($in, $out, 'dir_slot0:');
	# _parseDirOutput($in, $out, 'dir_volatile:');
	# _parseDirOutput($in, $out, 'dir_log:');
	# _parseDirOutput($in, $out, 'dir_logflash:');
	# _parseDirOutput($in, $out, 'dir_debug:');

	if (defined($modules->{_chassis}->{memInKb}))
	{
		$out->open_element('memory');
		$out->print_element('kind', 'RAM');
		$out->print_element('size', ($modules->{_chassis}->{memInKb} * 1024));  # Memory is sed-encoded in bytes
		$out->close_element('memory');
	}

	$out->close_element('chassis');
}

# This is the interior routine for parsing chassis data.  It will build up a hash in the return ref, keyed on "global"
# attributes and cards
sub _parse_chassis
{
	my $in = shift;
	my $modules = {};
	my $parseChassisReturn = 0;

	given ($in->{xmlForm})
	{
		when (NEXUS_XML_COMMANDS_XML_SSH) {$parseChassisReturn =  _parse_chassis_xml($in, $modules);};
		when (NEXUS_XML_COMMANDS_NETCONF ) {$parseChassisReturn =  _parse_chassis_xml($in, $modules);};
		when (NEXUS_XML_NO_XML_TEXT_ONLY) {$parseChassisReturn = _parse_chassis_cli_text($in, $modules)};
	} 

	# mark error for return to caller if nothing was set
	$modules = undef unless ($parseChassisReturn);
	
	return $modules;
}

sub _parse_chassis_xml
{
	my ($in, $modules) = @_;
	
	my $twig = XML::Twig->new(twig_roots => {
				proc_board_id => _createSimpleTwigSetter(\$modules->{_chassis}->{serialNumber}),
				chassis_id => _createSimpleTwigSetter(\$modules->{_chassis}->{description}),
				cpu_name => _createSimpleTwigSetter(\$modules->{_chassis}->{cpuType}),
				memory => _createSimpleTwigSetter(\$modules->{_chassis}->{memInKb})
			} );
	$twig->parse($in->{version});

	
	# Check for a lack of chassis data
	if (not $modules->{_chassis})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_chassis() : no chassis data found in device response.');
		return 0;		
	}
	# Serial Number is very important to topology.
	elsif (!defined($modules->{_chassis}->{serialNumber}))
	{
		$LOGGER->debug('Failed to parse required element core:serialNumber for chassis from \'show system\'.');
		return 0;
	}

	$twig = XML::Twig->new(twig_roots => {
				ROW_modinfo => sub {
						my ($twig, $element) = @_;
	
						my $id = $element->first_child('modinf')->text();
						$modules->{_cards}->{$id} = {};
						$modules->{_cards}->{$id}->{id} = $id;

						$modules->{_cards}->{$id}->{portCount} = $element->first_child('ports')->text();
						$modules->{_cards}->{$id}->{description} = $element->first_child('modtype')->text();
						$modules->{_cards}->{$id}->{modelNumber} = $element->first_child('model')->text();
						$modules->{_cards}->{$id}->{status} = $element->first_child('status')->text();
	
						return 1;
					},
				ROW_modwwninfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('modwwn')->text();

						$modules->{_cards}->{$id}->{softwareVersion} = $element->first_child('sw')->text();
						$modules->{_cards}->{$id}->{hardwareVersion} = $element->first_child('hw')->text();

						return 1;
					},
				ROW_modmacinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('modmac')->text();

						$modules->{_cards}->{$id}->{serialNumber} = $element->first_child('serialnum')->text();

						return 1;
					},
				ROW_xbarinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarinf')->text();
						$id .= '_fabric';
						$modules->{_cards}->{$id} = {};
						$modules->{_cards}->{$id}->{id} = $id;

						$modules->{_cards}->{$id}->{portCount} = $element->first_child('xbarports')->text();
						$modules->{_cards}->{$id}->{description} = $element->first_child('xbartype')->text();
						$modules->{_cards}->{$id}->{modelNumber} = $element->first_child('xbarmodel')->text();
						$modules->{_cards}->{$id}->{status} = $element->first_child('xbarstatus')->text();

						return 1;
					},
				ROW_xbarwwninfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarwwn')->text();
						$id .= '_fabric';

						$modules->{_cards}->{$id}->{softwareVersion} = $element->first_child('xbarsw')->text();
						$modules->{_cards}->{$id}->{hardwareVersion} = $element->first_child('xbarhw')->text();

						return 1;
					},
				ROW_xbarmacinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarmac')->text();
						$id .= '_fabric';

						$modules->{_cards}->{$id}->{serialNumber} = $element->first_child('xbarserialnum')->text();

						return 1;
					}
			} );
	$twig->parse($in->{module});
	
	return scalar (keys ( %{$modules}));
}

sub _parse_chassis_cli_text
{
	my ($in, $modules) = @_;
	
	# We need to parse through sections: cards/ports, hardware/software, serial #, then the same three for "xbar"  
	# Each has a pattern of a header, a dashed line, and the data directly beneath for each module, 1-n.
	# it closes with a blank line for each section (xbar sw closes it)
	
	if ($in->{version} =~ /^Hardware(.+?)\n\S/gism)
	{
		my $hardwareVersionBlob = $1;
		# first line is description
		if ($hardwareVersionBlob =~ /^\s+(.+?)$/ig)
		{
			$modules->{_chassis}->{description} = $1;
		}
		
		if ($hardwareVersionBlob =~ /\s+Processor Board ID (\S+)/ig)
		{
			$modules->{_chassis}->{serialNumber} = $1;
		}
		else
		{
			# This is an error, since it will create an unusable SED
			$LOGGER->debug('Failed to parse required element core:serialNumber for chassis from \'show version\'.');
			return 0;
		}
		
		# first mention of CPU is cpu type
		if ($hardwareVersionBlob =~ /^\s+(.+? CPU)\s+with (\d+)/ig)
		{
			$modules->{_chassis}->{cpuType} = $1;
			$modules->{_chassis}->{memInKB} = $2;
		}
	}
	else
	{
		# This is an error, since it will create an unusable SED
		$LOGGER->debug('Failed to parse required element core:serialNumber (hardware info) for chassis from \'show version\'.');
		return 0;		
	}
	
	# parse board-level serial $, description, CPU name, and memory out of $in->{module}
	
	if ($in->{module} =~ /^Mod\s+Ports\s+Module\-Type\s+Model\s+Status(.+?)Mod/mis)
	{
		my $cardPortsBlob = $1;
		
		# We can parse this line by line. The difference between the description and model is tough since they can run together.
		# I'm having to just use the fact that the module is always uppercase, digits, and dashes, e.g., N7K-M108X2-12
		while ($cardPortsBlob =~ /^(\d+)\s+(\d+)\s+(.+?)\s+([A-Z\d\-])\s+(\S+)/gi)
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{portCount} = $2;
			$modules->{_cards}->{$id}->{description} = $3;
			$modules->{_cards}->{$id}->{modelNumber} = $4;
			$modules->{_cards}->{$id}->{status} = $5;			
		}
		
	}

	# parse hw, sw out of $in->{version}
	
	if ($in->{module} =~ /^Mod\s+Sw\s+Hw.(.+?)Mod/mis)
	{
		my $hwSwBlob = $1;
		
		while ($hwSwBlob =~ /^(\d+)\s+([\d\.\(\)]+)\s+([\d\.]+)/gi)
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{softwareVersion} = $2;
			$modules->{_cards}->{$id}->{hardwareVersion} = $3;
		}
		
	}

	# parse serial # out of the MAC addr/serial # section in $in->{version}
	
	if ($in->{module} =~ /^Mod\s+Mac\-Address\(es\)\s+Serial\-Num(.*?)Mod/mis)
	{
		my $serialBlob = $1;
		
		# move right to the serial number here
		while ($serialBlob =~ /^(\d+)\s+([\da-z\- ])+([A-Z\d+]+)/gi)		
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{serialNum} = $3;
		}
		
	}

	# Parse crossbar ports, modujle, type, status
	if ($in->{module} =~ /^Xbar\s+Ports\s+Module\-Type\s+Model\s+Status(.+?)Xbar/mis)
	{
		my $xbarPortsBlob =  $1;
		
		# We can parse this line by line. The difference between the description and model is tough since they can run together.
		# I'm having to just use the fact that the module is always uppercase, digits, and dashes, e.g., N7K-M108X2-12
		while ($xbarPortsBlob =~ /^(\d+)\s+(\d+)\s+(.+?)\s+([A-Z\d\-])\s+(\S+)/gi)
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{portCount} = $2;
			$modules->{_cards}->{$id}->{description} = $3;
			$modules->{_cards}->{$id}->{modelNumber} = $4;
			$modules->{_cards}->{$id}->{status} = $5;			
		}
		
	}

	# parse hw, sw out of $in->{version}
	
	if ($in->{module} =~ /^Xbar\s+Sw\s+Hw.(.+?)Mod/mis)
	{
		my $xbarHwSwBlob = $1;
		
		while ($xbarHwSwBlob =~ /^(\d 	+)\s+([\d\.\(\)]+)\s+([\d\.]+)/gi)
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{softwareVersion} = $2;
			$modules->{_cards}->{$id}->{hardwareVersion} = $3;
		}
		
	}
	# parse xbar serial # out of the MAC addr/serial # section in $in->{version}.  This closes that section, so 
	# just scan to the end of it
	
	if ($in->{module} =~ /^Xbar\s+Mac\-Address\(es\)\s+Serial\-Num(.*)/mis)
	{
		my $xbarSerialBlob = $1;
		
		# move right to the serial number here
		while ($xbarSerialBlob =~ /^(\d+)\s+([\da-z\- ])+([A-Z\d+]+)/gi)		
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{serialNum} = $3;
		}
		
	}

	return scalar (keys ( %{$modules}));

};

# parse_interfaces
#
# Go through the interface section(s) parsing of the running config.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  CAN BE undef, in which case parsing and subnet formation will occur
#    without XML output.
#
# RETURNS:
#   a list:
#   	$subnets - a hash of Ziptie::Addressing::Subnets instances, indexed by interface name
#       $filterAssignments - a hash indexed by interface name, indicating assigned ingress and egress filters.
#

sub parse_interfaces
{
	my ($in, $out) = @_;
	my $locInterfaces = {};
	my $ifIndices;
	my $twig;
	my $fAssignments;
	my $ipAddrRe = get_crep('cipm');
	my $ipCidrRe = get_crep('cidr');
	my $ipSubnets = ();

	_parse_vlan_information($in);

	if ($in->{interface_ifindex})
	{
		# To most efficiently use the various twigs we have to pull up here, create a simple hash of
		# the standalone snmp index interface element
		$twig = XML::Twig->new(twig_roots => {
					ROW_interface => sub {
							my ($twig, $element) = @_;
		
							my $ifName = $element->first_child('interface')->text();
							if (defined $element->first_child("ifindex-dec"))
							{
								$ifIndices->{"$ifName"} = $element->first_child("ifindex-dec")->text();
							}
							return 1;
						}
		} );
		if (_is_conforming_xml($in->{interface_ifindex}))
		{
			$twig->parse($in->{interface_ifindex});
		}
	}
	else
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_interfaces() : no ifIndex data found in device response.'); # nonfatal
	}
	
	
	# Running config interfaces have been scanned for meaningful interface relevance.and are within the 
	# keys of the {interfaces} bucket.  If interfaces aren't found here, game over, really.
	if (not $in->{interface})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_interfaces() : no interfaces data found in device response.');
		return;
	}
	
	$twig = XML::Twig->new(twig_roots => {
			ROW_interface => sub {
					my ($twig, $element) = @_;
					my $ifName = $element->first_child('interface')->text();

					if (_parse_interface_brief($in, $ifName, $element, $locInterfaces))
					{
						# assign if index
						if (defined $ifIndices->{"$ifName"})
						{
							$locInterfaces->{$ifName}->{ifIndex} = $ifIndices->{"$ifName"};  
						}

						# Now, dive into the running-config section for this adapter to pull up layer 3 properties.
						# We also need to collect subnets and filter assignments here.
						if ($in->{running_config} =~ /^interface\s+($ifName)(.+?)\n(?=\w)/smi)
						{
							my $ifCfgBlob = $2;
							my $ipOptions = undef;
							
							if ($ifCfgBlob =~ /^\s+description\s+([\S ]*)/mi)
							{
								$locInterfaces->{"$ifName"}->{description} = $1;
							}
							
							while ($ifCfgBlob =~ /^\s+ip\s+(\S.*)$/mig)
							{
								my $ipLine = $1;
								given ($ipLine) 
								{
									when (/^directed\-broadcast/i)	{$ipOptions->{directedBroadcast} = "true";}
									when (/^proxy\-arp/i) {$ipOptions->{localProxyArp} = "true";}
									when (/^route-cache/i) {$ipOptions->{routeCache} = "true";}
									when (/^dhcp\s+relay\s+address/)
									{
										if ($ipLine =~ /^dhcp\s+relay\s+address\s+($ipAddrRe)/)
										{
											push (@{$ipOptions->{udpForwarder}}, $1);
										}
									}
									when (/^address\s+$ipCidrRe/)
									{
										# assign ipConfiguration											
										if ($ipLine =~ /^address\s+($ipAddrRe)\/(\d+)/)
										{
											my $sn = new ZipTie::Addressing::Subnet($1, $2);
											push (@{$ipOptions->{ipConfiguration}}, 
												{ ipAddress => $sn->getHostIp(), mask => $sn->getMask()});
											push (@{$ipSubnets->{$ifName}}, $sn);
										}
									}
									when (/^access\-group\s+(\S+)\s+(\w+)/)
									{
										# access list interface application 
										given ($2)
										{
											when ("in") {$locInterfaces->{$ifName}->{'ingressFilter'} = $1;}
											when ("out") {$locInterfaces->{$ifName}->{'egressFilter'} = $1;}
											default {$LOGGER->debug("Cisco::Nexus::Parsers::parse_interfaces: unable to parse access list direction \'$2\'");}; 												
										}
									}										
								}
								$locInterfaces->{"$ifName"}->{interfaceIp} = $ipOptions;									
							}
						}
					}												
					return 1;
				}
	} );
	# need to clean interface due to bug in 'show interfaces brief' XML response
	my ($cleanInterfaces) = clean_interfaces ($in);
	$twig->parse($cleanInterfaces);
	#   Look up from config:
	#      ip address, mask
	#      directed broadcast, ip helper, proxy arp, route cache
	

	if (defined $out)
	{
		$out->open_element("interfaces") if (defined $out);
		
		for my $nextIf (keys(%{$locInterfaces}))
		{
			_printInterface( $locInterfaces->{"$nextIf"}, $out);
		}
		$out->close_element("interfaces") if (defined $out);		
	}
	
	return ($ipSubnets, undef);  
		
}

sub create_config
{
	my ($in, $out) = @_;

	my $repository = {};
	$repository->{'core:config'} = [];

	# The "root folder" of where we put things is dependent on whether we have contexts we'll need
	# to rake over or not.
	$repository->{'core:name'} =  "/";

	$LOGGER->debug("Cisco::Nexus::Parsers::create_config: setting storage for admin context'...");
	# Back up global context....
	my $dcAdminContextFolder;
	$dcAdminContextFolder->{'core:name'} = "system space";
	
	push (@{$dcAdminContextFolder->{'core:config'}}, _createSingleContextConfigFromCommand($in, 'running-config', 'running_config',
			'text/plain', 'active', 'false'));

	if ($in->{startup_config})
	{
		push (@{$dcAdminContextFolder->{'core:config'}}, _createSingleContextConfigFromCommand($in, 'startup-config', 'startup_config',
				'text/plain', 'boot', 'true'));
	}
	push ( @{ $repository->{'core:folder'} }, $dcAdminContextFolder);		 
	
	if ($in->{contexts})
	{
		for my $vdcName (keys %{$in->{contexts}})
		{
			$LOGGER->debug("Cisco::Nexus::Parsers::create_config: setting storage for context \'" . $vdcName . "\'...");
			my $dcFolder;
			$dcFolder->{'core:name'} = $vdcName;
			
			push ( @{ $dcFolder->{'core:config'} },  
				_createSingleContextConfigFromCommand($in->{contexts}->{$vdcName},'running-config', 'running_config', 
					'text/plain',	'active', 'false')); 
			push ( @{ $repository->{'core:folder'} }, $dcFolder);		 
		}			
	}

	$out->print_element('core:configRepository', $repository);
}

# parse_access_ports
#
# Parse the ziptie core sections describing diagnostic and chassis-level ports to access the 
# system at console level.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  
#
#

sub parse_access_ports
{
	my ($in, $out) = @_;

	my $accessPortsOpen = 0;
	while ($in->{running_config} =~ /^line (\S+)/mg)
	{
		my $type = $1;

		if (!$accessPortsOpen)
		{
			$out->open_element('accessPorts');
			$accessPortsOpen = 1;
		}

		$out->open_element('accessPort');
		$out->print_element('type', $type);
		$out->close_element('accessPort');
	}
	if ($accessPortsOpen)
	{
		$out->close_element('accessPorts');
	}
}

# parse_object_groups
#
# Parses the output of the show object groups commands and generates the objectGroups section of the SED.
#
# PARAMS: 
# $in - input context, featuring a {object_group} section
# $out - output XML writer context.
sub parse_object_groups
{
	my ($in, $out) = @_;

	# Currently the Nexus only supports network groups and service groups
	my $objectGroups = {};
	$objectGroups->{networkGroups} = [];
	$objectGroups->{serviceGroups} = [];

	# If no object groups, no worries.  Totally optional element.  But don't pass it to Twig::parse, or it
	# will NPE.  
	if (not $in->{object_group})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_object_groups() : no object group data found in device response (which is perfectly acceptable).');
		return;
	}	

	# This hash maps the group type integers found in the Nexus XML to the correct parser to use to parse each object group.
	# The first argument of each parser will be the XML::Twig element to parse, and the second argument to each parser will
	# be the $objectGroups hash, as seen in the XML::Twig handler bellow for the ROW_ogroup element.
	my %GROUP_PARSERS = (

			# ipv4 object group parser
			100 => \&_networkGroupParser,

			# port object group parser
			102 => \&_serviceGroupParser
		);

	my $twig = XML::Twig->new(twig_roots => {
				ROW_ogroup => sub {
						my ($twig, $element) = @_;

						my $groupTypeElement = $element->first_child('group_type');
						if (!defined($groupTypeElement))
						{
							$LOGGER->debug("parse_object_groups: Failed to parse element group_type from object group");
							return 1;
						}

						my $groupType = $groupTypeElement->text();
						if ($groupType == "101")
						{
							$LOGGER->debug("parse_object_groups: Dropping IPv6 object group");
						}
						elsif (defined($GROUP_PARSERS{$groupType}))
						{
							$GROUP_PARSERS{$groupType}->($element, $objectGroups);
						}
						else
						{
							$LOGGER->debug("parse_object_groups: Unknown object group type: $groupType");
						}

						return 1;
					} });
	$twig->parse($in->{object_group});

	# This is where we actually output the XML, if there is anything to output
	if (scalar(@{$objectGroups->{networkGroups}}) || scalar(@{$objectGroups->{serviceGroups}}))
	{
		$out->open_element('objectGroups');

		foreach my $networkGroup (@{$objectGroups->{networkGroups}})
		{
			$out->open_element('networkGroup');
			$out->print_element('id', $networkGroup->{id});

			if (defined($networkGroup->{host}))
			{
				foreach my $host (@{$networkGroup->{host}})
				{
					$out->print_element('host', $host);
				}
			}

			if (defined($networkGroup->{subnet}))
			{
				foreach my $subnet (@{$networkGroup->{subnet}})
				{
					$out->print_element('subnet', $subnet);
				}
			}

			$out->close_element('networkGroup');
		}

		foreach my $serviceGroup (@{$objectGroups->{serviceGroups}})
		{
			$out->open_element('serviceGroup');
			$out->print_element('id', $serviceGroup->{id});

			if (defined($serviceGroup->{portExpression}))
			{
				foreach my $portExpression (@{$serviceGroup->{portExpression}})
				{
					$out->print_element('portExpression', $portExpression);
				}
			}

			if (defined($serviceGroup->{portRange}))
			{
				foreach my $portRange (@{$serviceGroup->{portRange}})
				{
					$out->print_element('portRange', $portRange);
				}
			}

			$out->close_element('serviceGroup');
		}

		$out->close_element('objectGroups');
	}
}


# parse_access_lists
#
# Parses the access lists declared within the Nexus configuration and "show ip access-lists expanded"
# command
#
# PARAMS: 
# $in - input context, featuring a {access_lists} section
# $out - output XML writer context.  ALSO assigns the generated_access_lists $in 
#        hash bucket
sub parse_access_lists
{
	my ($in, $out) = @_;
	my $fListsDeclared = 0;
	my $fEntryName;
	my $fEntryDesc;	
	my $fEntryDescNext;
	my $alNID;
	my @alSpecTerms; 
	my $alSignatureAssembly;

	# Back away unless actual access lists were declared as part of the 
	# configuration
	return unless ($in->{access_lists});

	# Look over $in->{access_lists}, Enumerate on
	#     ROW_ip_ipv6_mac entries
	#     +++waynet MAC acl including ethertype (*not* supported on layer 3 interfaces)
	#     +++waynet VLAN access map
	
	my $twig = XML::Twig->new(twig_roots => {
				ROW_ip_ipv6_mac => sub {
					my ($twig, $element) = @_;
					@alSpecTerms = ();
						
					if (not $fListsDeclared)
					{ 
						$out->open_element("filterLists");
						$fListsDeclared = 1;
					} 
					# pick up the acl name here...
					$fEntryName = $element->first_child_text('acl_name');
					$fEntryDesc = "";

					$out->open_element('filterList configPath="running-config"');
									
					# Now, specs are sequenced by priority.  It sure looks like they come in order,
					# but I'm not one to say necessarily.  To that end, put them in an array which will 
					# index by the eval order number

					# if this as applied to an i/f doesn't match traffic, default is deny...
					$out->print_element("defaultAction", "deny");

					# Filter signature.  I don't recall why, but there's a reason that we need to put
					# this after default signature.
					$alSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");

					# Process the TABLE_seqno/ROW_seqno subpaths
					for ($element->first_child('TABLE_seqno'))
					{
						if(defined($_)){
							foreach my $feElement ($_->children('ROW_seqno'))
							{
								my $alSpecTerm = {};
								($alNID, $fEntryDescNext) = _parse_alSpecTerm($in, $fEntryName, $feElement, $alSpecTerm);

								# Description is to be found in "remark" element.  We latch the first such thing found to be the
								# overall filter description.  This often comes unterminated by the netconf *ssh CLI,*
								if ($fEntryDescNext and not $fEntryDesc)
								{
									$fEntryDesc = $feElement->first_child_text("remark");
									chomp($fEntryDesc);
								}

								$alSpecTerms[$alNID] = $alSpecTerm;
							}
						}
					}
					# was there a description to elevate?
					if ($fEntryDesc)
					{
						$out->print_element("description", $fEntryDesc);											
					}
					
					# Now iterate our list

					my $alIndex = 1;
					for $alNID (0 .. $#alSpecTerms)
					{
						if (defined $alSpecTerms[$alNID])
						{
							# assign the "name" element as being the index here
							$alSpecTerms[$alNID]->{'name'} = $alNID;
							$alSpecTerms[$alNID]->{'processOrder'} = $alIndex++;
							$out->print_element("filterEntry", $alSpecTerms[$alNID] );
						}
					}
					
					$out->print_element("mode", "stateful");
					$out->print_element("name", $fEntryName);
					
					$out->close_element('filterList');
					return 1;
				}
		}
	);
	$twig->parse($in->{access_lists});
		
	$out->close_element("filterLists") if ($fListsDeclared); 
	
}

sub parse_routing
{
	my ($in, $out) = @_;
	my @vrfList;
	
	# so as not to have the running-config in mid-parse when these are called,
	# just do a scan for vrfs in the config
	
	while ($in->{running_config} =~ /^vrf\s+context\s+(\S+)/mig)
	{
		push (@vrfList, $1);
	}
	
	# Now, just iterate the list (plus the DEFAULT_VRF_NAME), doing the route
	# calculations for each
	
	map { _parse_vrf_routing($in, $out, $_); } (DEFAULT_VRF_NAME, @vrfList);
		_check_routing_sections_close($in, $out);    
	
}

sub _parse_vrf_routing {
    my ( $in, $out, $vrfName ) = @_;

    # call common scalar routine to set up %IOSRoutingProcesses
    _parse_routing_processes($in, $vrfName);

	# will there be routing processes to emit?
	    
    # The results are now in $in->{routingProcesses}->{"$vrfName"}
    # This next list is very deliberately ordered.
    
    for (SUPPORTED_ROUTING_SECTIONS)
    {
    	if (defined $in->{routingProcesses}->{"$vrfName"}->{"$_"})
    	{
    		_check_routing_sections_open($in, $out);
    		$out->print_element("$_", $in->{routingProcesses}->{"$vrfName"}->{"$_"})
    	}
    }
}

sub parse_static_routing
{
	my ( $in, $out) = @_;
	
	my $staticRoutes = _parse_static_route_data($in);
	
	if (defined $staticRoutes)
	{
		$out->open_element( "staticRoutes");
		$out->print_element( 'staticRoute', $staticRoutes );
		$out->close_element("staticRoutes");		
	}
}

sub parse_vlans
{
	my ($in, $out) = @_;

	if (!defined($in->{_vlanHash}))
	{
		return;
	}

	my @vlans = values(%{$in->{_vlanHash}});
	if (scalar(@vlans))
	{  
		$out->open_element('vlans');
		$out->print_element('vlan', \@vlans);
		$out->close_element('vlans');
	}
}

sub _parse_vlan_information
{
	my ($in) = @_;

	my %vlanHash = ();
	my %interfaceVlanHash= ();  # Make sure this is initialized to something a reference will resolve

	my $twig = XML::Twig->new(twig_roots => {
		ROW_vlanbrief => sub {
				my ($twig, $element) = @_;
				my $vlanEl = $element->first_child('vlanshowbr-vlanid-utf') // $element->first_child('vlanshowbr-vlanid');
				
				if (not defined $vlanEl)
				{
					return;  # no VLAN to identify
				}

				my $vlanId = $vlanEl->text();

				if (!defined($vlanHash{$vlanId}))
				{
					$vlanHash{$vlanId} = {
							number => $vlanId,
							interfaceMember => [] # This array is populated in parse_interfaces
						};
				}

				if ($element->first_child('vlanshowbr-shutstate')->text() eq 'noshutdown')
				{
					$vlanHash{$vlanId}->{enabled} = 'true';
				}
				else
				{
					$vlanHash{$vlanId}->{enabled} = 'false';
				}

				my $vlanNameElement = $element->first_child('vlanshowbr-vlanname');
				if (defined($vlanNameElement))
				{
					$vlanHash{$vlanId}->{name} = $vlanNameElement->text();
				}

				foreach my $vlanPortListElement ($element->children('vlanshowplist-ifidx'))
				{
					foreach my $portRange (split(/,/, $vlanPortListElement->text()))
					{
						my $ports = _expand_port_range($portRange);

						foreach my $port (@{$ports})
						{
							if (!defined($interfaceVlanHash{$port}))
							{
								$interfaceVlanHash{$port} = ZipTie::Ranges->new();
							}
							$interfaceVlanHash{$port}->add_range($vlanId);
						}
					}
				}
			}});
	
	# The VLAN definition could fail for a variety of reasons, mostly if it's not valid XML.  Hopefully, this will take
	# care of the implications for all and any....

	if (_is_conforming_xml($in->{vlan}))
	{
		$twig->parse($in->{vlan});
	}				

	$in->{_vlanHash} = \%vlanHash;
	$in->{_interfaceVlanHash} = \%interfaceVlanHash;
}

sub _expand_port_range
{
	my ($portRange) = @_;

	my $ports = [];

	if ($portRange =~ /^(\S+?)(\d+)-(\d+)$/)
	{
		my $prefix = $1;
		my $startPort = $2;
		my $endPort = $3;

		for (my $i = $startPort;$i <= $endPort;$i++)
		{
			push(@{$ports}, ($prefix . $i));
		}
	}
	else
	{
		push(@{$ports}, $portRange);
	}

	return $ports;
}

# parse_dynamic_routes
#
# Create the common:dynamic_routes section of the ZED by looking at the routing state data
#
# PARAMS: 
# $in - input context, featuring a {<routing protocols>->{routes}} section.  if there is no $in->{subnets}
#          bucket in place, it will be created by a call to parse_interfaces.  
# $out - output XML writer context.  
#

sub parse_dynamic_routes
{
	my ($in, $out) = @_;
	my @vrfList;
	
	#+++wayne cribbed this from parse_routing -- should probably just have a parse_vrfs that can be called
	# and assign this to a $in bucket.
	
	# so as not to have the running-config in mid-parse when these are called,
	# just do a scan for vrfs in the config
	
	while ($in->{running_config} =~ /^vrf\s+context\s+(\S+)/mig)
	{
		push (@vrfList, $1);
	}
	
	# Now, just iterate the list (plus the DEFAULT_VRF_NAME), doing the route
	# calculations for each
	
	map { _parse_dynamic_routes_data($in, $out, $_); } (DEFAULT_VRF_NAME, @vrfList);
	
	# asymmetric interface -- see if we need to close the dynamic routes section
	_check_dynamic_routes_close($in, $out);	
}

sub _parse_dynamic_routes_data
{
	my ($in, $out, $vrfName) = @_;
	my $rProt;
	my $redistributionMap = {};
	my $dRoutesEmittedHere = 0;
	
	# Make sure we have subnets with which to work
	_assert_subnets_present($in);
	# AND we have to make sure we've parsed the static routes, in case they're distributed.
	_parse_static_route_data($in);
	
    # For starters, we need to have the routing process context available...
    # +++this will have to change for context support
    _parse_routing_processes($in, $vrfName);
	
	foreach $rProt (SUPPORTED_ROUTING_PROTOCOLS)
	{
		my $rNS = ($rProt ~~ "eigrp") ? "cisco:" : "";  # namespace qualifier, where necessary for hash keys and such
		
		my $emittedRB = {};  # +++waynet should be Net::CIDR::Set
		# inject routes from sets which are redistributed and match route filters
		if (_redistribution_sources($in, $redistributionMap, $vrfName, $rProt))
		{
			foreach my $rdSource (keys (%{$redistributionMap->{$rProt}}))
			{
				#  pull from that source in declared routes, match route filters
				foreach my $rdRoute (@{$in->{declaredRoutes}->{$rdSource}})
				{
					if (defined $emittedRB->{$rProt}->{_make_routeNework_CIDR($rdRoute->{routeNetwork})})
					{
						# We've already generated this route from another source
						next;
					}
					
					my $routeRedistSource = _check_route_redist($redistributionMap->{$rProt}, 
						$rdSource, _make_routeNework_CIDR($rdRoute->{routeNetwork}));					

					if ($routeRedistSource)
					{
						# Emit route, but mark redistribution first.  Try to coerce the advertising routing process name from
						# a process ID, if we have one in the reams of routing process info we've been collecting.						
						$rdRoute->{advertisingRoutingProcess} = $in->{routingProcesses}->{"$vrfName"}->{$rNS . $rProt}->{$rNS . "asNumber"} // $rProt;
						
						if ($rdSource =~ /^(\w+)(\-(\d+))?/)
						{
							$rdRoute->{routeRedistributionSourceProcess}->{routeOriginProcessName} = $1;
							$rdRoute->{routeRedistributionSourceProcess}->{routeOriginProcessId} = $2 if ($2);							
						}
						
						_check_dynamic_routes_open($in, $out);
						$out->print_element("dynamicRoute", $rdRoute);
						# Mark as emitted
						$emittedRB->{$rProt}->{_make_routeNework_CIDR($rdRoute->{routeNetwork})} = 1; 
								# Assignment could be anything, we're using this as a set.
					}
				}
			}
		}
		
		# Emit routes which are in the VRF-declared routing base for this protocol
		for (keys %{$in->{dRoutes}->{$rNS . $rProt}->{$vrfName}})
		{
			_check_dynamic_routes_open($in, $out);

			$out->print_element("dynamicRoute",  $in->{dRoutes}->{$rNS . $rProt}->{$vrfName}->{$_});
			$emittedRB->{$rProt}->{_make_routeNework_CIDR($in->{dRoutes}->{$rNS . $rProt}->{$vrfName}->{$_}->{routeNetwork})} 
									= 1;									
			
		}
		
		my $twig = XML::Twig->new(twig_roots => { ROW_vrf => 
			sub {
				my ($twig, $element) = @_;		
				
				# Now, let's look for the routes which should be presumably most if not all about routes learned
				# from neighbor systems.
				if ($element->first_child_text('vrf-name-out') !~ /^$vrfName/i)
				{
					return;  # wait for next VRF
				}
				
				# Now, we have a lot of navigation to do here.  First, we want to be sure we're on the IPv4
				# routing table
				
				for ($element->first_child('TABLE_addrf'))
				{					
					foreach my $afElement  ($_->children('ROW_addrf'))
					{
						if ($afElement->first_child_text('addrf') ne 'ipv4')
						{
							next;
						}
						# if we have routes...                                                                                                                             
						if (not	$afElement->first_child('TABLE_prefix'))
                        {
                        	next;
						}
						
						# Within this, there is a single TABLE_prefix, with ROW_prefix rows.
						for ($afElement->first_child('TABLE_prefix'))
						{
							foreach my $prefixElt ($_->children('ROW_prefix'))
							{
								my $dRouteElt;
								my $nextElt;
								my $nh;
								
								my $tblNH = $prefixElt->first_child('TABLE_path');
								
								($dRouteElt->{routeNetwork}->{address}, $dRouteElt->{routeNetwork}->{mask}) =
									($prefixElt->first_child_text('ipprefix') =~ /^([\d\.]+)\/(\d+)/);

								# Have we heard about/emitted this route already?
								if (defined $emittedRB->{$rProt}->{_make_routeNework_CIDR($dRouteElt->{routeNetwork})})
								{
									next;  # Move to next network then
								}
								
								
								# dive down to find next hop and where we heard about it from
								if ($tblNH)
								{
									foreach my $pathElt ($tblNH->children('ROW_path'))
									{
										undef $nh;  # Clear this for next-hop recalc.
										
										# If that didn't get assigned, then just make the process the entire
										# process name (which would be $rProt)
										if ($pathElt->first_child_text('clientname') =~ /(.+?)(\-(\d+))?$/)
										{
											$dRouteElt->{advertisingRoutingProcess} = $3 // $1;
										}
										else
										{
											# just to to rProt
											$dRouteElt->{advertisingRoutingProcess} = $rProt;
										}
										

										$nextElt = $pathElt->first_child_text('metric');
										if ($nextElt)
										{
											$nh->{adjacencyMetric} = $nextElt;
										}
	
										$nextElt = $pathElt->first_child_text('ipnexthop');
										if ($nextElt)
										{
											$nh->{adjacencyNode} = $nextElt;
										}

										$nextElt = $pathElt->first_child_text('ifname');
																				
										if ($nextElt)
										{
											$nh->{adjacencyInterface} = expand_nexus_interface_name($nextElt);
										}
										else
										{
											# We have to dig up a next hop out of our own subnets
											# This could be a lot more optimized if this was stored 
											# as a hash trie, AND/OR had some LRU caching to the paths
											
											foreach my $snK ( keys(%{$in->{subnets}}))
											{
												foreach my $sn (@{$in->{subnets}->{$snK}})
												{
													if ($sn->contains($nh->{adjacencyNode}))
													{
														$nh->{adjacencyInterface} = $snK; # already expanded
														last;
													}
												}	
											}
											
											if (not defined $nh->{adjacencyInterface})
											{
												$LOGGER->debug("Cisco::Nexus::Parsers::parse_dynamic_routes : unable to find interface for next hop " ,
												 $nh->{adjacencyNode});
												
											}
										}																		
										
										if ($nh)
										{
											push (@{$dRouteElt->{routeNextHop}}, $nh);									
										}
										else
										{
											$LOGGER->debug("Cisco::Nexus::Parsers::parse_dynamic_routes : unable to parse next hop for network " ,
												 $prefixElt->first_child_text('ipprefix'));
										}
									}
								}
								else # tblNH not defined in XML input.  This is, I think, a Nexus bug.
								{
									$LOGGER->debug("Cisco::Nexus::Parsers::parse_dynamic_routes : " .
										"not able to find next hop information in Nexus data for route " . 
										$dRouteElt->{routeNetwork}->{address} .  "/"  . $dRouteElt->{routeNetwork}->{mask});
									next; # drop this route									
								}
								
								if ($dRouteElt) # should be
								{
									_check_dynamic_routes_open($in, $out);

									$out->print_element("dynamicRoute", $dRouteElt);
									$emittedRB->{$rProt}->{_make_routeNework_CIDR($dRouteElt->{routeNetwork})} = 1;									
								}	
							} # end of iteration across ROW_prefixes
						} # end of scan for TABLE_prefix]
					} # End of each ROW_addrf
				} # End of scan for TABLE_addrf
			} # end of ROW_vrf sub 
		});				
		
		$twig->parse($in->{$rProt}->{routes}) if $in->{$rProt}->{routes};
	}	
}												 


# Takes an abbreviated interface name and returns the full interface name.
sub expand_nexus_interface_name
{
	my ($ifShortName) = @_;

	my $interface;

	if ($ifShortName =~ /Eth((?:\d+\/?)+)/)
	{
		$interface = "Ethernet$1";
	}
	elsif ($ifShortName =~ /Po(\d+)/)
	{
		$interface = "port-channel$1";
	}
	else
	{
		$interface = $ifShortName;
	}

	return $interface;
}


sub open_vdc_section
{
	my ($in, $out, $subelementSet)  = @_;
	
	if (defined $subelementSet) 
	{
		if (not defined $in->{subelementContexts})
		{
			# Track our generation of vdc's / subelements this way
			map { $in->{subelementContexts}->{"$_"} = 1; } @{$subelementSet};
			$out->open_element("subelements");
		}
	}
	else
	{
		# this is for testing
		$out->open_element("subelements");
	}		

	$out->open_element("subelement");
}

sub close_vdc_section
{
	my ($in, $out, $thisSubelement) = @_;
	
	# Close out the subelement.  that's the easy part.  We also need to identify the 
	# subelement type as  a mandatory element
	$out->print_element("subelementType", "virtual");
	$out->close_element("subelement");
	
	if (defined $thisSubelement)
	{
		# Now, remove the subelement name...
		if (exists $in->{subelementContexts}->{"$thisSubelement"})
		{
			delete $in->{subelementContexts}->{"$thisSubelement"};
			
			# Now, check if there's a nonempty hash key
			for (keys %{$in->{subelementContexts}})
			{
				return if ($_ =~ /\w/);   # if this is a nonempty key, then we've still got contexts to generate
			}
			
			# If we get here, then we can close out the entire 'subelements' section
			$out->close_element("subelements");
		}		
	}
	else
	{
		# this is for testing
		$out->close_element("subelements");
		
	}
}

sub get_default_vrf_name { return DEFAULT_VRF_NAME; }

##### Functions private to this module #################

sub _createSingleContextConfigFromCommand
{
	my ($in,  $name, $command, $mediaType, $context, $promotable) = @_;

	my $config = {};
	# The listing format encodes the time of query into it, which is a bad thing for
	# persisting our Nexus configs, since it means all of them will display differences.
	# So, make a copy and eliminate that line for storage
	my $configText = $in->{$command};
	$configText =~ s/^\s*!Time[\w\s\d:]+\n//m;
	
	$config->{'core:name'} = $name;
	$config->{'core:textBlob'} = encode_base64($in->{$command});
	$config->{'core:mediaType'} = $mediaType;
	$config->{'core:context'} = $context;
	$config->{'core:promotable'} = $promotable;

	return $config;
}

#
# _parse_routing_processes can be called many times to assure that routing process
# data is available for explicit SED generation and for usage by other parsing
# subsystems, but only parses the process data once and then stores it in a 
# private bucket off of the $in response context
#
sub _parse_routing_processes
{
	my ($in, $vrfName) = @_;
	my $routingSection;
	my $rNS;
	
   if (scalar keys %{$in->{routingProcesses}->{"$vrfName"}}) {
		return;
    }
    
    # Look for supported routing protocols
    for my $rProt (SUPPORTED_ROUTING_PROTOCOLS)
    {
    	my $protRBlob;
   		my $rProcOptionsVector;
   		my $rProtDef;
   		my $protRProcID;    	

   		# The field name for recording enabled/active interfaces is different from protocol to protocol
  		# in the ZED (thank you).  This stores it for common assignment code.
   		my $activeInterfacesField;

    	if ($in->{running_config} !~  /^feature\s+$rProt/msg)
    	{
    		next;   # There is no enabling of the protocol
    	} 
    	# Game on...we have to collect the rblobs separately, since it's yet another scan of the 
    	# running-config and we can't nest those.
    	while ( $in->{running_config} =~
	    	/^router\s+$rProt(?:\s+(\d+))?\s*\n((?:\s+.*?\n)*)/msgi )
    	{
    		my $rBlob = $2;
    		my $rProcID = $1 // "global";
    		
    		$rNS = "";  # may be overriden for other namespace

			if ($vrfName eq DEFAULT_VRF_NAME)
			{
				# this is default vrf, make sure this isn't assigned to a SPECIFIC nondef VRF.
				if ($rBlob =~ /^\s+vrf\s+\S+/mig)
				{
					next;  # this is for another VRF.
				} 				

			}
			elsif ($rBlob !~ /^\s+vrf\s+($vrfName)/mig)
			{
				next;  # this is for another VRF.
			} # otherwise, we have a matched VRF and requested VRF so we should move ahead.

			 
			# we should only assign this once, even though we'll let this spin through
			# +++wayne n.b!  This is a key place which will have to change once we can allows
			# multiple instances of a routing protocol per VRF, e.g., OSPF.
			
			if ($protRBlob)
			{
				$LOGGER->debug("Cisco::Nexus::Parsers::parse_routing : I think I found duplicate routing process " .
					"entries for protocol $rProt in VRF \'$vrfName\'.  Tossing this second one: $rBlob");				
			}
			else
			{
				# it's just the blob, the blob u want (insert Devo disco groove here)
				$protRBlob = $rBlob;
				$protRProcID = $rProcID;   						
			}
    	}
    	
    	if ($protRBlob)
    	{
    		given ($rProt)
    		{
    			when ("eigrp") 
    			{ 
    				$rNS = "cisco:";
    				$rProcOptionsVector = "_parse_eigrp_routing_options";    				

    				# determining active and passive interfaces requires a scan of all interfaces in the running config, not
    				# the VRF/routing declaration. In fact, passive interfaces don't seem to show up for "show ip eigrp 
    				# interfaces"
    				while ($in->{running_config} =~ /^interface\s+(\S+)(.+?)\n(?=\w)/smig)
					{
						my $ifName = $1;
						my $ifCfgBlob = $2;
						
						# if this is a nondefault VRF, make sure the interface is assigned to that
						# VRF
						if ($vrfName eq DEFAULT_VRF_NAME) 
						{
							if ($ifCfgBlob =~ /^\s+vrf\s+member\s+\S+/mi)
							{
								next;							
							}
						}
						elsif ($ifCfgBlob !~ /^\s+vrf\s+member\s+$vrfName/mi)
						{
							next;
						}
						# OK, we match vrf name and/or default VRF if we get here.
						
						if ($ifCfgBlob =~ /^\s+ip passive-interface eigrp (\d+)/mi)
						{
							# we could check AS here, but there are a number of places where we're assuming
							# one EIGRP instance per device/VDC
							push (@{$rProtDef->{"cisco:passiveInterface"}}, $ifName);								
						}
					}
					$activeInterfacesField = "activeInterface";
							
					# initialize AS in case it's not in the routing config (which it generally won't be)
				    $rProtDef->{"cisco:asNumber"} = $protRProcID;																			    				
    			} # End of EIGRP definition
    			when ("rip") 
    			{ 
    				$rProcOptionsVector = "_parse_rip_routing_options";
    				$activeInterfacesField = "enabledInterface";  
    			}
    		} # end of "given $rProt"
    		
    		# call options vector for protocol
    		if (defined $rProcOptionsVector)
    		{
    			my @procOptionsArgs = ($protRBlob, $rProtDef);
    			eval ( "$rProcOptionsVector(\@procOptionsArgs)" );
    		}
    		
    		# Scan interfaces for protocol at routing protocol level
    		
    		if ($activeInterfacesField)
    		{
	    		my $rProtInterfaces = _parse_routing_protocol_interfaces ($in, $rProt, $vrfName);
	    		
	    		for (@{$rProtInterfaces})
	    		{
	    			push (@{$rProtDef->{$rNS . $activeInterfacesField}}, $_);	
	    		}
    		}    		
    		
    		# Now, search for any referenced VRF routes.  *So near as I can tell*, the exact same
    		# options which are present for a given protocol *but including a potential routing
    		# base* are present at the vrf.

			# +++waynet need to have a non-Hogans device against which to test this interface\s+($ifName)(.+?)\n(?=\w)/smi
			if ($vrfName ne DEFAULT_VRF_NAME)
			{
				foreach (@{_scan_for_vrf_routes($in, $vrfName, $rNS . $rProt, $protRProcID)})
				{
					push (@{$rProtDef->{$rNS . "network"}}, $_);
				}
			}    		
    		
   				# redistribution
   				# what was "redistribute connected" in IOS is "redistribute direct" in NX-OS
   				# redistribution in NX-OS is always governed by a route-map, but we'll pick that up
   				# later.
    		while ($protRBlob =~ /^\s+redistribute\s+(\S+)\s+((\d+)\s+)?route\-map/mig)
    		{
    			my $rdst;
    			$rdst->{targetProtocol} = ($1 eq "direct") ? "connected" : $1;
    			$rdst->{processId} = $3 if $3;
    						
    			push (@{$rProtDef->{$rNS . "redistribution"}}, $rdst);
    		}
     		
    		# Add definition to output
    		# +++waynet need to expand for multiple routing process instance defs, e.g., ospf or bgp
    		$in->{routingProcesses}->{"$vrfName"}->{$rNS . "$rProt"} = $rProtDef;
    		
    	}	# end of "if $protRBlob"..." 
    }
}

# scan for "ip route's" within a VRF
# +++wayne find a way to combine this with parse_static_route_data
sub _scan_for_vrf_routes
{
	my ($in, $vrfName, $effectiveRProt, $effectiveRProc) = @_;
	my $routeSet = ();
	my $nextRoute;
    my $ipAddrRe = get_crep('cipm');
    my $ipCidrRe = get_crep('cidr');
	
	if ($in->{running_config} =~ /^vrf\s+context\s+$vrfName(.+?)\n(?=\S)/smi)
	{
		my $vrfDefinitionBlob = $1;
		
		while ($vrfDefinitionBlob =~ /\s+(ip route ([\d\.\/]+)\s+(.*))/mig)
		{
			my $routeIPData = $2;
			my $routeBlob = $3;
			my $routeCfLine = $1;  # Save for diagnostic purposes
			
			# route: dest addr/mask
			if ($routeIPData =~ /$ipCidrRe/)
			{
				my ($routeIP, $routeMask) = split(/\//, $routeIPData);
	
				if (($vrfName eq DEFAULT_VRF_NAME) or ($routeIP !~ /^0\.0\.0\.0/))
				{
					$nextRoute = { address => $routeIP, mask => $routeMask};
					push (@{$routeSet}, $nextRoute);	
					
					_assert_subnets_present($in);
					
					# scan for next hop
					if ($routeBlob =~ /^($ipAddrRe)(\s+|$)(.*?)/)
					{
						$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->{adjacencyNode} = $1;

						for my $netIfNetKey ( keys %{$in->{subnets}})
						{
							my $foundIf = 0;
							# This is an array of networks off of each interface
							for  (@{$in->{subnets}->{$netIfNetKey}})
							{
								if ($_->contains($in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->{adjacencyNode}))
								{
									$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->
										{adjacencyInterface} = $netIfNetKey;
									$foundIf = 1;
									last;					
								}
							}
							last if ($foundIf);							
						}
					}
					else
					{
						if ($routeBlob =~ /^(\S+)(\s+\d+\/\d+)?\s*(.*)/)
						{
							my $nhIF = $1; $nhIF .= $2 // "";
							$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->
										{adjacencyInterface} = $nhIF;							
						}						
					}  																
					if (not defined $in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->
								{adjacencyInterface})
					{
						$LOGGER->debug("Cisco::Nexus::Parsers::parse_routing_processes : unable to resolve ip route " .
							$routeIP . " to an interface address or network: directing to default gateway");
						$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->{adjacencyNode}  = 
							$in->{defaultGatewayRoute}->{destinationAddress};	
						$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNextHop}->{adjacencyInterface} = 
							$in->{defaultGatewayRoute}->{interface};																
					}
					# Add it as a route with a next hop since we've resolved the gateway i/f
					$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{routeNetwork} = $nextRoute;
					$in->{dRoutes}->{$effectiveRProt}->{$vrfName}->{"$routeIP"}->{advertisingRoutingProcess} =
						$effectiveRProc;
				}
			}
		}

		if ($vrfDefinitionBlob =~ /\s+default\-information originate/mi)
		{
			# this is a special option to indicate that default gateway info is to be placed into this RB.  
			# let's find it
			if ($in->{defaultGatewayRoute})
			{
				$nextRoute = { address => $in->{defaultGatewayRoute}->{destinationAddress}, 
					mask => $in->{defaultGatewayRoute}->{destinationMask}};
				push (@{$routeSet}, $nextRoute);
			}
		}
		
	}
	return $routeSet;
}

sub _parse_static_route_data
{
	my $in = shift;
	my $staticRouteData;
	my $ipAddrRe = get_crep('cipm');
	my $ipCidrRe = get_crep('cidr');

	# Clear the static "declared routes" in case it's present.
	$in->{declaredRoutes}->{static} = ();

	while ($in->{running_config} =~ /^(ip route ([\d\.\/]+)\s+(.*))/mig)
	{
		my $staticRoute;
		my $routeIPData = $2;
		my $routeBlob = $3;
		my $routeCfLine = $1;  # Save for diagnostic purposes
		
		my $nextIfNetKey;
		my $nextIfNet;

		# route: dest addr/mask
		if ($routeIPData =~ /$ipCidrRe/)
		{
			my ($routeIP, $routeMask) = split(/\//, $routeIPData);

			$staticRoute = { destinationAddress => $routeIP, destinationMask => $routeMask};

			$staticRoute->{defaultGateway} = ($routeIP eq '0.0.0.0' && $routeMask == 0) ? 'true' : 'false';
		}
		else
		{
			$LOGGER->debug("Cisco::Nexus::Parsers::parse_static_routes : unable to interpret \'ip route $routeIPData\'");
			next;
		}

		# what if there is no explicit default gateway?  Nexus doesn't make this very clear
		
		# following element is either an interface or a next hop.  This is matching an IP address followed
		# *not* directly by a nonspace charactger (which we're presuming might be a CIDR address, to be 
		# matched in the next "elsif" clause)

		if ($routeBlob =~ /^($ipAddrRe)(\s+|$)(.*?)/)
		{
			$staticRoute->{gatewayAddress} = $1;
			$routeBlob = $3;
			_assert_subnets_present($in);

			foreach $nextIfNetKey ( keys %{$in->{subnets}})
			{
				# This is an array of networks off of each interface
				for $nextIfNet (@{$in->{subnets}->{$nextIfNetKey}})
				{
					if ($nextIfNet->contains($staticRoute->{gatewayAddress}))
					{
						$staticRoute->{interface} = $nextIfNetKey;
						last;
					}
				}
				
				if (defined $staticRoute->{interface})
				{
					last;
				}
			}
		}
			# This is either an interface or a next hop network.  There are three variations really.  One is the
			# identification of a next hop network.  The ZED doesn't really support that, so we have to match
			# it up with an interface matching the network, if we can.
		elsif ($routeBlob =~ /^($ipCidrRe)/)
		{
			if ($routeBlob =~ /^($ipAddrRe)\/(\d+)\s*(.*?)/)
			{
				# definitionally, this has to match
				my $nhNet = new ZipTie::Addressing::Subnet($1, $2);
				$routeBlob = $3;
				_assert_subnets_present($in);
				foreach $nextIfNetKey ( keys %{$in->{subnets}})
				{
					# This is an array of networks off of each interface
					for $nextIfNet (@{$in->{subnets}->{$nextIfNetKey}})
					{
						if ($nextIfNet->containsNet($nhNet))
						{
							$staticRoute->{interface} = $nextIfNetKey;
							# This is a stretch goal, but bug #42869 makes an empassioned case.  If this is a 
							# gateway host disguised as a net, get all opportunistic and encode it as a 
							# gateway address.
							if ($nhNet->getMask() ~~ 32)
							{
								$staticRoute->{gatewayAddress} = $nhNet->getHostIp();
							}
							last;
						}
					}

					if (defined $staticRoute->{interface})
					{
						last;
					}
				}

				# Did that get assigned?
				if (not defined $staticRoute->{interface})
				{
					$LOGGER->debug("Cisco::Nexus::Parsers::parse_static_routes : unable to resolve network " .
						$nhNet->to_string() . " to an interface address or network");
				}
			}

		}
		else # Else, we're going with an interface here
		{
			if ($routeBlob =~ /^(\S+)(\s+\d+\/\d+)?\s*(.*)/)
			{
				my $nhIF = $1; $nhIF .= $2 // "";
				$routeBlob = $3;
				
				# See if this is just a discard thing
				if ($nhIF =~ /^null/i)
				{
					$staticRoute->{discard} = 'true';
				}
				# Either way, we always set interface - required element
				$staticRoute->{interface} = $nhIF;
				
				# See if a next hop was specified here as well as an IP addr
				if ($routeBlob =~ /^($ipAddrRe)(\s+(.+))?/)
				{
					$staticRoute->{gatewayAddress} = $1;
					$routeBlob = $3 // "";
				}
			}
		}
		
		# See if we can infer a route pref or metric  it will presumably just be a 
		if ($routeBlob =~ /\S+/)
		{
			if ($routeBlob =~ /^(\d+)/)
			{cisco:
				$staticRoute->{routePreference} = $1;
			}
		}
		
		if (defined $staticRoute)
		{
			# put this to a static route/dynamic route entry, to be able to mark
			# for any processes doing redistribution
			my $declRoute = { routeNetwork => { address=> $staticRoute->{destinationAddress}, mask => $staticRoute->{destinationMask}} };
			
			# assign adjacency fields
			$declRoute->{routeNextHop}->{adjacencyInterface} = $staticRoute->{interface} if $staticRoute->{interface};
			$declRoute->{routeNextHop}->{adjacencyMetric} = $staticRoute->{routePreference} if $staticRoute->{routePreference};
			# no next hop node--we're it			
		
			push ( @{$in->{declaredRoutes}->{static}}, $declRoute);
			push(@{$staticRouteData}, $staticRoute);
			
			if ($staticRoute->{defaultGateway} eq 'true')
			{
				# this route gets squirreled away in a special place
				$in->{defaultGatewayRoute} = $staticRoute;
			}
		}
		else
		{
			# sanity here
			$LOGGER->debug("Cisco::Nexus::Parsers::parse_static_routes :unable to parse a static route from " . $routeCfLine );
		}		
	}
	
	# be kind, rewind....since those searches through $in->{running_config} can cause a match to terminate before
	# the string is fully searched, reset it.
	pos($in->{running_config}) = 0;
	return $staticRouteData;
}

#  _redistribution_sources
#
# Check if route redistribution is in play for the indicated protocol.  If so, set up the
# internal redistributuion map for the protocol.
#
# PARAMS:
#  $in - response context for entire parse
#  $rdMap - redistribtion results to this point.  This may already have the map set up for this protocol.  If
#   so, do not recompute it, since that's expensive.
#  $vrfName - Virtual Route Forwarding table name scope caller is interested in
#  $routeProcess - the "routing process ID" as known by the ZEd and indexed in $in->{routeProtocolName}.  This is
#    sort of hokey since this could be a real protocol key or a process ID +++wayne fix this!
#
#  RETURNS
#    boolean - if true, then route redist is in play, and the key $rdMap->{$routeProcess} is assigned
#
sub _redistribution_sources
{
	my ($in, $rdMap, $vrfName, $routeProcess) = @_;
	my $routingProtocolName;
	my $rpKey;
	my $rpNS;
	
	foreach $rpKey (keys (%{$in->{routingProcesses}->{"$vrfName"}}))
	{
		my $rpComp;
		# Make sure to strip the namespace
		($rpNS, $rpComp) = ($rpKey =~ /(\w*:?)(\w+)/);
		
		if ($routeProcess =~ /^$rpComp/)
		{
			$routingProtocolName = $rpComp;  # un process indexed route id
			last;			
		}
	}

	if (not $routingProtocolName)
	{
		# routing protocol not configured, for example
		return 0;		
	}
	
	# Now, we know what to scan for redistribution
	# Need to create a key for the redistribution here, plus any route filtering in play
	if ( $in->{running_config} =~
    		/^router\s+$routingProtocolName(?:\s+(\d+))?\s*\n((?:\s+.*?\n)*)/m)
    {
    	my $rBlob = $2;

		while ($rBlob =~ /^\s+redistribute (\S+) route\-map (\S+)/gm)
		{
			my $rmName = $2;
			
			$rdMap->{$routeProcess}->{$1} = _create_route_map_from($in, $rmName);			
		}
    }
    
    return (defined $rdMap->{$routeProcess});
}

#
# RETURNS: process name of redistribtution
sub _check_route_redist
{
	my ($rdMap, $rProt, $network) = @_;
	
	# Check all redist specs in $rdMap
	if ($network =~ /^([\d\.]+)\/(\d+)/)
	{
		return $rdMap->{$rProt}->exerciseFilterOn($1, $2);
	}	
}

#
# returns: route map from route map, coalescing all prefix list elements
sub _create_route_map_from
{
	my ($in, $routeMapName) = @_;
	my $rfOut;
	
	if ($in->{running_config} =~ /^route\-map\s+$routeMapName\s+(\w+)(?:\s+(\d+))?\s*\n((?:\s+.*?\n)*)/msi)
	{
		my $routeMapDisp = ($1 eq 'deny') ? 0 : 1;
		my $rmBlob = $3;  # $2 is the index which is a dontcare for us
		
		if (not defined $rfOut)
		{
			$rfOut = ZipTie::Addressing::RouteFilter->new(0); # default: deny all
		}
		
		# Now scan for prefix list references
		while ($rmBlob =~ /^\s+match ip address prefix\-list (\S+)/mgi)
		{
			my $pfListName = $1;
			if (0 == _fold_prefix_list_into_route_filter($in, $rfOut, $pfListName, $routeMapDisp))
			{
				# We weren't able to find any prefix list by that name...
				$LOGGER->debug("Cisco::Nexus::Parsers::parse_dynamic_routes : unable to find prefix-list $pfListName " .
					"referenced in route-map $routeMapName");
			}
		}
	}

	if (not defined $rfOut)
	{
		$LOGGER->debug("Cisco::Nexus::Parsers::parse_dynamic_routes : unable to find route-map for  " . $routeMapName);
		# better create a "allow all" on the fly
		$rfOut = ZipTie::Addressing::RouteFilter->new(1);
	}
	
	return $rfOut;

}

sub _fold_prefix_list_into_route_filter
{
	my ($in, $rFilter, $pfListName, $routeMapDisp) = @_;
	my $listSpecs = 0;
	my $ipCidrRe = get_crep('cidr');
	
	
	while ($in->{running_config} =~ /^ip prefix\-list $pfListName (seq \d+ )?(permit|deny) ($ipCidrRe)/mg)
	{
		# see if we have to flip sense
		my $plDisp = ($2 eq 'permit') ? $routeMapDisp : not $routeMapDisp;
		# break up the cidr mask...
		my $plNetCidr = $3;
		
		if ($plNetCidr =~ /([\d\.]+)\/(\d+)/)
		{
			$rFilter->addConstraint($1, $2, $plDisp);
			$listSpecs++;
		}
	}
	
	return $listSpecs;
}

# This just ends up being easiest to abstract
sub _make_routeNework_CIDR
{
	my $routeNw = shift;  # a ZED IPNetwork encoding
	return ( $routeNw->{address} . '/' . $routeNw->{mask});
}

sub _check_dynamic_routes_open
{
	my ($in, $out) = @_;
	
	if (not $in->{dynamicRoutesEmitted})
	{
		$out->open_element("dynamicRoutes");
		$in->{dynamicRoutesEmitted} = 1;
	}	
}

sub _check_dynamic_routes_close
{
	my ($in, $out) = @_;
	
	if ($in->{dynamicRoutesEmitted})
	{
		$out->close_element("dynamicRoutes");
	}
}


sub _check_routing_sections_open
{
	my ($in, $out) = @_;
	
	if (not $in->{routingDeclEmitted})
	{
		$out->open_element("cisco:routing");
		$in->{routingDeclEmitted} = 1;
	}	
}

sub _check_routing_sections_close
{
	my ($in, $out) = @_;
	
	if ($in->{routingDeclEmitted})
	{
		$out->close_element("cisco:routing");
	}
}

########  Routing process option handlers ######################
#
#  These all follow a similar pattern, but are abstracted out because
#  they can be called across a pattern space which can be (so near as 
#  I can tell) either a routing process declaration or a VRF
#  which is defining the options for a routing protocol instance
#
#  INPUTS
#   $inputSpace : The options for the routing protocol, either at the routing process
#       definition or the VRF.  All options can be assume space-prepended on their own
#       lines.  For ease of passage, the "router" or "vrf" header can be missing, it won't
#       be searched for
#   $rProcOptions : reference to hash which will receive ZED-compatible values for the routing
#        process in question.
#
#  RETURNS
#    $rProcOptions, updated
#
sub _parse_eigrp_routing_options
{
	my ($inputSpace, $rProcOptions) = @_;
	
    if ($inputSpace =~ /^\s+autonomous\-system\s+(\d+)/mi)
    {
    	$rProcOptions->{"cisco:asNumber"} = $1; # I can't tell if this comes out in running_config "show"
	}

	# autosummarization not supported on Nexus
    $rProcOptions->{"cisco:autoSummarization"} = 0;
 	# default is active interface configuration, unless overridden at the interface level
    $rProcOptions->{"cisco:passiveInterfaceDefault"} = 0;
    				
    if ($inputSpace =~ /^\s+eigrp\s+router\-id\s+(\S+)/mi)
    {
    	$rProcOptions->{"cisco:routerId"} = $1;
 	}
    				
	return $rProcOptions;
}

sub _parse_rip_routing_options
{
	my ($inputSpace, $rProcOptions) = @_;
	
	# We only have one option to set here, and it's not dependent upon content
	$rProcOptions->{version} = 2;
	return $rProcOptions;
}

sub _assert_subnets_present
{
	my $in = shift;
	if (not defined $in->{subnets})
	{
		# only do this once
		($in->{subnets}) = parse_interfaces($in);  # Do *not* pass an out param
	}	
}
	
########  Twig subhandlers and utilities ######################
#
#  This takes the element context of the ROW_Interface 
#  element for a single element of "show interfaces brief".  The
#  interface name, as passed, is already parsed.
#
#  All element context is filled in in $zInterfaces->{$ifName}
#
#  RETURNS: 0 or nonzero - if nonzero, then the interface has been
#   acknowledged as pertinent to the ZED (something was filled in zInterfaces)
#   and parsing/processing on this interface should continue.  This routine will
#   ONLY set up applicable interface-first level and interfaceEthernet properties.

sub _parse_interface_brief
{
	my ($in, $ifName, $tElement, $zInterfaces) = @_;

	#  if we find an explicit interface entry for it

	if ($in->{interfaces}->{"$ifName"})
	{
		#    add to Interfaces hash, marking admin status from interface brief.
		$zInterfaces->{"$ifName"}->{adminStatus} = $tElement->first_child('state')->text();
		# Now, for speed.  The default reporting is in megabits it appears, unless appended by "G"
		
		my $ifSpeed = 0;
		
		if ($tElement->first_child_text('speed') =~ /^(\d+)(\w?)/)
		{
			if ($2)
			{
				given ($2)
				{
					when ('G') { $ifSpeed = 1073741824; } # 1 Gb in bits
					when ('M') { $ifSpeed = 1048576; }
					default { 
						$LOGGER->debug("ZipTie::Adapters::Cisco::Nexus::parse_interfaces : Unexpected speed modifier $2");
						$ifSpeed = 1048576; # for some basis		
					}
				}
			}
			else
			{
				# base 1 MB/sec basis
				$ifSpeed = 1048576; 
			}
			
		}
		else
		{
			# This is one place we pick this up
			if ($tElement->first_child_text('speed') =~ /auto/i)
			{
				$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "true";
			}
		}
		
		if ($ifSpeed)
		{
			# Now, just scale appropriately
			$ifSpeed *= $1;
			$zInterfaces->{"$ifName"}->{speed} = $ifSpeed;
		}
		elsif (not defined  $zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed})
		{
			# There's not a great way with the data we have to determine, "virtual", so this is a blunt one...
			if ($ifName !~ /^(loopback|port\-channel|Vlan)/)
			{
				$LOGGER->debug("ZipTie::Adapters::Cisco::Nexus::parse_interfaces : Unable to determine speed on interface $ifName");						
			}
		}
		
		if ($tElement->first_child("desc"))
		{
			$zInterfaces->{"$ifName"}->{description} = $tElement->first_child_text("desc");
			# for some reason, this often remains with a trailing eol in the xml output
			chomp($zInterfaces->{"$ifName"}->{description});
		}
		
		my $portmode = $tElement->first_child_text("portmode");
		if ($portmode =~ /^route/i)
		{
			push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "routing");			
		}
		elsif ($portmode eq 'trunk' && defined($in->{_interfaceVlanHash}->{$ifName}))
		{
			foreach my $vlanRange (@{$in->{_interfaceVlanHash}->{$ifName}->get_ranges()})
			{
				if ($vlanRange->{lowInt} == $vlanRange->{highInt})
				{
					push(@{$zInterfaces->{$ifName}->{interfaceVlanTrunks}}, { startVlan => $vlanRange->{lowInt} });
				}
				else
				{
					push(@{$zInterfaces->{$ifName}->{interfaceVlanTrunks}}, { startVlan => $vlanRange->{lowInt}, endVlan => $vlanRange->{highInt} });
				}
			}
		}

		# Record vlan membership information, used in parse_vlans
		my $vlan = undef;
		my $vlanElement = $tElement->first_child("vlan");
		if (defined($vlanElement))
		{
			$vlan = $vlanElement->text();
		}
		elsif ($ifName =~ /^Vlan(\d+)$/)
		{
			$vlan = $1;
		}

		if (defined($vlan) && ($vlan =~ /^\d+$/) && defined($in->{_vlanHash}->{$vlan}))
		{
			push(@{$in->{_vlanHash}->{$vlan}->{interfaceMember}}, $ifName);
		}
		
		# Pull what we can out of the full interface entry for this device for Ethernet address and other properties
		my $fTwig = XML::Twig->new(twig_roots => {
				ROW_interface => sub {
						my ($twig, $element) = @_;

						$zInterfaces->{"$ifName"}->{name} = $element->first_child_text("interface");

						# We need to make our searches for interfaces for virtual ethernet ports be prepended with "veth", since the 
						# tags will be different.
						my $ethernetPropertiesTagPrefix = ($ifName =~ /^Vether/i) ? "veth" : "eth";

						# MTU can come from two different places
						my $mtu = $element->first_child_text($ethernetPropertiesTagPrefix . "_mtu") // $element->first_child_text("svi_mtu");

						if ($mtu)
						{
							# See bug 83444, this is fix for a bug in the Nexus
							# https://tools.cisco.com/quickview/bug/CSCtt41070
							if ($mtu =~ /^[^\d]+(\d+)[^\d]+$/)
							{
								$mtu = $1;
							}

							$zInterfaces->{"$ifName"}->{mtu} = $mtu;
						}
						
						if ($element->first_child_text($ethernetPropertiesTagPrefix . "_hw_addr") =~ /^([\w\d\.]+)/ )
						{
							# Get rid of the delimiting dots
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{macAddress} = $1;
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{macAddress} =~ s/\.//g;
						}
						
						if ( $element->first_child($ethernetPropertiesTagPrefix . "_duplex"))
						{
							if ($element->first_child_text($ethernetPropertiesTagPrefix . "_duplex") =~ /^(full|half)/)
							{
								# if interface is down, this can also just contain the value "auto", which
								# isn't a legal assigment to operational duplex, you see....
								$zInterfaces->{"$ifName"}->{interfaceEthernet}->{operationalDuplex} = $1;
							}

							# Nexus default
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoDuplex} = "true";
						}

						# Going to base the loopback determination here as to whether octet-level stats are returned in
						# an element named "loopXXX"
						if ($element->first_child("loop_in_bytes"))
						{
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{loopBack} = "network";
							$zInterfaces->{"$ifName"}->{interfaceType} = "softwareLoopback";
							$zInterfaces->{"$ifName"}->{physical} = "false";
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "switching");
						}
						elsif ($ifName =~ /^mgmt/)  # Only way I can figure this...
						{
							# management interface
							# Disabling admin behavior here.  If this is the only L3 interface (which it will be, on a Nexus 1000v,
							# ) then topology would drop the device.
							# push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "administrative");
							
							# Adding this on to provide ability to form profile for bridged forwarding of L2 traffic
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "bridging");							
							$zInterfaces->{"$ifName"}->{physical} = "true";
							$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";
						}
						else
						{
							# make some bold determinations here
							$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";
							$zInterfaces->{"$ifName"}->{physical} = "true";
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "switching");
						}

						# we may override some of this by looking at port mode.
						if ($portmode eq "access")
						{
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "bridging");														
						}

						given ($tElement->first_child_text("type"))
						{
							when ('virt')
							{
								$zInterfaces->{"$ifName"}->{physical} = "false";
								$zInterfaces->{"$ifName"}->{interfaceType} = "virtual";								
							}
							
							when ('eth')
							{
								$zInterfaces->{"$ifName"}->{physical} = "true";
								$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";																
							}
						}
												
						if ($element->first_child("svi_arp_type"))
						{
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{encapsulation} =
								$element->first_child_text("svi_arp_type"); 
						}

						# This is a second place to assess auto_speed
						given ($element->first_child_text($ethernetPropertiesTagPrefix . "_autoneg"))
						{
							when ('on') { 	$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "true"; } 
							when ('off'){ 	$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "false"; }

						}

						# Now, let's deal with trunking vs. access (VLAN) and EtherChannel trunking
						# relationships.
						# +++wayne not sure this is meaningful, or if it is (which it probably is), what this
						# means to core and how to uniquely associate port-channel 
#						if ($zInterfaces->{"$ifName"}->{interfaceType} ~~ "ethernet")
#						{
#							if ($element->first_child_text($ethernetPropertiesTagPrefix . "_hw_desc") =~ /^Port\-channel/i)
#							{
#								push ( @{$zInterfaces->{"$ifName"}->{interfaceGrouping}}, 
#			   						{ifGroupName => $ifName,
#			    				ifGroupingCharacteristic => "primaryInterface"});
#							}
#						}

				}
		} );
		if ($in->{interfaces}->{"$ifName"})
		{
			$fTwig->parse($in->{interfaces}->{"$ifName"}); # full interface entry for this i/f		
		}
		else
		{
			$LOGGER->debug('Cisco::Nexus::Parsers::parse_interfaces() : no interface specific data found in device response for interface $iName.'); 
		}

		return 1;
	}
	else
	{
		# This isn't an explicitly configured interface,.  This is normal for things like Vlan 1 (the "all vlans" meta-vlan)
		return 0;
	}
}

# This subroutine parses both ipv6 and ipv4 network groups from the Nexus
sub _networkGroupParser
{
	my ($element, $objectGroups) = @_;

	my $networkGroup = {};

	my $nameElement = $element->first_child('group_name');
	if (!defined($nameElement))
	{
		$LOGGER->debug("parse_object_groups: Failed to parse name from network object group.");
		return;
	}
	$networkGroup->{id} = $nameElement->text();

	my $groupMembersTableElement = $element->first_child('TABLE_seqno');
	if (!defined($groupMembersTableElement))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for network object group: " . $networkGroup->{id});
		return;
	}

	# Parses an ipv4 or an ipv6 host
	my $parseHost = sub {
			my ($host) = @_;

			if (!defined($networkGroup->{host}))
			{
				$networkGroup->{host} = [];
			}
			push(@{$networkGroup->{host}}, $host);

			return 1;
		};

	# Parses an ipv4 or an ipv6 subnet
	my $parseSubnet = sub {
			my ($subnet) = @_;

			my $address = undef;
			my $mask = undef;
			if ($subnet =~ /^([^\/]*)\/(\d+)/)
			{
				$address = $1;
				$mask = $2;
			}
			else
			{
				return 0;
			}

			if (!defined($networkGroup->{subnet}))
			{
				$networkGroup->{subnet} = [];
			}
			push(@{$networkGroup->{subnet}}, { address => $address, mask => $mask });
		};

	foreach my $groupMember ($groupMembersTableElement->children('ROW_seqno'))
	{
		# Parse out any ipv4 hosts if they exist
		my $hostElement = $groupMember->first_child('hostaddr');
		if (defined($hostElement))
		{
			my $success = $parseHost->($hostElement->text());

			if (!$success)
			{
				$LOGGER->debug("parse_object_groups: Failed to parse ipv4 host: " . $hostElement->text());
			}
		}

		# Parse out any ipv4 subnets if they exist
		my $netElement = $groupMember->first_child('net_ip');
		if (defined($netElement))
		{
			my $success = $parseSubnet->($netElement->text());

			if (!$success)
			{
				$LOGGER->debug("parse_object_groups: Failed to parse ipv4 subnet: " . $netElement->text());
			}
		}
	}
	if (!defined($networkGroup->{host}) && !defined($networkGroup->{subnet}))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for network object group: " . $networkGroup->{id});
		return;
	}

	push(@{$objectGroups->{networkGroups}}, $networkGroup);
}

# This subroutine parses port object groups from the Nexus (stored as service groups in the SED).
sub _serviceGroupParser 
{
	my ($element, $objectGroups) = @_;

	my $serviceGroup = {};

	my $nameElement = $element->first_child('group_name');
	if (!defined($nameElement))
	{
		$LOGGER->debug("parse_object_groups: Failed to parse name from service object group.");
		return;
	}
	$serviceGroup->{id} = $nameElement->text();

	my $groupMembersTableElement = $element->first_child('TABLE_seqno');
	if (!defined($groupMembersTableElement))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for service object group: " . $serviceGroup->{id});
		return;
	}

	foreach my $groupMember ($groupMembersTableElement->children('ROW_seqno'))
	{
		# Parse any port operators
		my $portOpElement = $groupMember->first_child('_port_op');
		if (defined($portOpElement))
		{
			# When a port operator is used we expect a single port value stored in port0_num
			my $portElement = $groupMember->first_child('port0_num');
			if (defined($portElement))
			{
				my $operator = $portOpElement->text();
				my $port = $portElement->text();

				# Nexus operators match operators avaiable in the SED schema, except for the not equals operator.
				if ($operator eq 'neq')
				{
					$operator = 'ne';
				}

				if (!defined($serviceGroup->{portExpression}))
				{
					$serviceGroup->{portExpression} = [];
				}
				push(@{$serviceGroup->{portExpression}}, { port => $port, operator => $operator });
			}
			else
			{
				$LOGGER->debug("parse_object_groups: Faild to parse port0_num for service object group: " . $serviceGroup->{id});
			}
		}

		# Parse any port ranges
		my $portRangeElement = $groupMember->first_child('_port_range');
		if (defined($portRangeElement))
		{
			# When a port range is used we expect two port values storted in port1_num and port2_num
			my $portStartElement = $groupMember->first_child('port1_num');
			my $portEndElement = $groupMember->first_child('port2_num');
			if (defined($portStartElement) && defined($portEndElement))
			{
				my $portStart = $portStartElement->text();
				my $portEnd = $portEndElement->text();

				if (!defined($serviceGroup->{portRange}))
				{
					$serviceGroup->{portRange} = [];
				}
				push(@{$serviceGroup->{portRange}}, { portEnd => $portEnd, portStart => $portStart });
			}
			else
			{
				$LOGGER->debug("parse_object_groups: Faild to parse port1_num and port2_num for service object group: " . $serviceGroup->{id});
			}
		}
	}
	if (!defined($serviceGroup->{portExpression}) && !defined($serviceGroup->{portRange}))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for service object group: " . $serviceGroup->{id});
		return;
	}

	push(@{$objectGroups->{serviceGroups}}, $serviceGroup);
}


# This takes one argument, a reference to a scalar.
# It returns reference to a closure that can be used
# as a twig_handler to simply set the scalar referenced
# to the text context of the element the handler was
# triggered to handle.
#
# Note: If there is more than one instance of each element
#       the referenced scalar will end up being set to the
#       value of the last element.  createSimpleTwigSetter
#       probably will not be of any use if you expect
#       more than one instance of the element.
sub _createSimpleTwigSetter
{
	my ($varRef) = @_;
	return sub {
		my ($twig, $element) = @_;
		$$varRef = $element->text();
		return 1;
	};
}

# Parses the output of a dir command and populates a deviceStorage element,
# printing it directly to the SED.
sub _parseDirOutput
{
	my ($in, $out, $dir) = @_;

    my %months = (
        JAN => '01',
        FEB => '02',
        MAR => '03',
        APR => '04',
        MAY => '05',
        JUN => '06',
        JUL => '07',
        AUG => '08',
        SEP => '09',
        OCT => '10',
        NOV => '11',
        DEC => '12'
    );

	my $name = undef;
	my $size = undef;
	my $freeSpace = undef;

	if ($in->{$dir} =~ /dir (\S+):/)
	{
		$name = $1;
	}

	if ($in->{$dir} =~ /^\s*(\d+) bytes total/m)
	{
		$size = $1;
	}

	if ($in->{$dir} =~ /^\s*(\d+) bytes free/m)
	{
		$freeSpace = $1;
	}

	# The following two elements are required by the schema.
	if (!defined($name))
	{
		$LOGGER->debug('Failed to parse required element name for directory: ' . $dir);
		return;
	}
	if (!defined($size))
	{
		$LOGGER->debug('Failed to parse required element size for directory: ' . $dir);
		return;
	}

	$out->open_element('deviceStorage');
	$out->print_element('name', $name);
	$out->print_element('storageType', 'unknown');
	$out->print_element('size', $size);
	$out->print_element('freeSpace', $freeSpace) if (defined($freeSpace));

	my $rootDirOpen = 0;
	# Example string being parsed:
	# 29568000    Jan 07 07:01:42 2012  n7000-s1-kickstart.6.0.2.bin
	while ($in->{$dir} =~ /^\s*(\d+)\s+(\S+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)\s+(\S+)$/mg)
	{
		my $size = $1;

		my $month = $months{uc($2)};
		my $day = $3;
		my $time = $4;
		my $year = $5;

		my $fileName = $6;

		my $mtime = "$year-$month-$day" . "T$time";

		if (!$rootDirOpen)
		{
			$out->open_element('rootDir');
			$rootDirOpen = 1;
		}

		$out->open_element('file');
		$out->print_element('mtime', $mtime);
		$out->print_element('name', $fileName);
		$out->print_element('size', $size);
		$out->close_element('file');
	}
	if ($rootDirOpen)
	{
		# Schema says the name element can be left blank for a root directory.
		$out->print_element('name');

		$out->close_element('rootDir');
	}

	$out->close_element('deviceStorage');
}

sub _printInterface
{
	my ($zInterface, $out) = @_;
	my @orderedInterfaceZedFields = (
			"adminStatus", "description", "egressFilter",
			"failover", "ifIndex", "ingressFilter", 
			"interfaceEthernet", "interfaceGrouping",
			"interfaceIp", "interfaceBehavior", "interfaceType",
			"interfaceVlanTrunks", "mtu", "name", "physical",
			"speed"
	);	
	
	# start with an interface entry here...
	$out->open_element("cisco:interface");
	
	for my $nextIfField (@orderedInterfaceZedFields)
	{
		if (defined $zInterface->{$nextIfField})
		{
			$out->print_element($nextIfField, $zInterface->{$nextIfField});
		}	
	}
	
	$out->close_element("cisco:interface");	
}

#
#  This will scan the "RowSeqno"  term scoped here for the term data
#  to comprise a filter spec.
# 
#  $in - entire response data from context
#  $twigElement - element corresponding to ROW_Seqno in an access list
#  $filterEntry - term to "fill in" with ZED filterEntry conformant element names/values
#
#  RETURNS 'sequence number' of term (required element in Nexus access list spec)
sub _parse_alSpecTerm
{
	my ($in, $alName, $twigElement, $filterEntry) = @_;
	my $seqno = $twigElement->first_child_text('seqno');
	
	defined $seqno ||$LOGGER->debug("Cisco::Nexus::parse_access_lists; cannot find term sequence number in $alName");

	$filterEntry->{name} = $seqno;
	
	my $fEltText;
	my $nlPortHi;
	my $nlPortLo;
	my $nlPortOp;
	my $feDescription;
	my $primaryProtocol;
	
		# Assign mandatory elements with default values
	$filterEntry->{log} = "false";
	$filterEntry->{primaryAction} = "next"; # if nothing but metadata is in this filter, then simply pass to the next one

	foreach ($twigElement->children())
	{
		# Throw out structural metadata stuff
		next if (not ($_->is_elt && $_->string_value));
		$fEltText = $_->string_value;
		my $fEltTag = $_->name;		
		
		given($fEltTag)
		{
			when (/^dest_(\S+)/) { 
			    
			    # we have to special case the port numbers, so that we can locally
			    # assign them for fixup once the filter spec has been put together
			    if ($1 =~ /^port(\d)_num/)
			    {
					given ($1)
					{
					    when (1) { $nlPortLo->{destination} = $fEltText; }
					    when (2) { $nlPortHi->{destination} = $fEltText; }
					}
			    }
			    elsif ($1 =~ /^port(\d)_str/)
			    {
					given ($1)
					{
						# +++wayne this will return a single port number.  Do we need to worry about
						# multiple noted port numbers logically associated with one port (e.g., netbios)?
					    when (1) { $nlPortLo->{destination} = get_port_number($fEltText, $primaryProtocol ); }
					    when (2) { $nlPortHi->{destination} = get_port_number($fEltText, $primaryProtocol); }
					}			    	
			    }
			    elsif ($1 =~ /^port_op/)
			    {
					$nlPortOp->{destination} = $fEltText;
			    }
			    else
			    {
				# let the common dest/src argument handler deal with it
					_parse_access_list_entry_direction_options($fEltTag, $fEltText, 'dest', 'destination',
							       $filterEntry );

			    }
				
			}
			 
			# ""established" ... shows up as ""<established>enable</established>"
			when (/^established/) { $filterEntry->{established} = ($fEltText =~ /^enable/i) ? "true" : "false";} 
			# "log" is similarly encoded with an "enable" 
			when (/^log/) { $filterEntry->{log} = ($fEltText =~ /^enable/i) ? "true" : "false";} 
			when ("precedence_str") { $filterEntry->{precedence} = $fEltText; }  # nonenumerated anyways
			# dscp <-> tos : specify numerically
			when ("dscp") { $filterEntry->{tos} = $fEltText; } 
			
			when ("proto_str") 
			{
				 push(@{$filterEntry->{protocol}},  $fEltText);   # All cases I've seen match enumeration
				 $primaryProtocol = $fEltText;
			}
			when ("permitdeny") { $filterEntry->{primaryAction} = $fEltText; }
			when ("remark") { $feDescription = $fEltText unless ($feDescription); }
			# ICMP message filters go into "protocol filter"
			when ("icmp_str") { $filterEntry->{protocolFilter} = $fEltText; }
			# "[permit|deny] ip ..." seems to turn into a field *called* ip.  *Don't* assign $primaryProtocol here.
			when ("ip")  { push(@{$filterEntry->{protocol}}, "ip") if ($fEltText =~ /^ip$/i); }  # "ip" seems to be the only value, too  			
			# All source fields
			when (/^src_(\S+)/) { 
			    
			    # we have to special case the port numbers, so that we can locally
			    # assign them for fixup once the filter spec has been put together
			    if ($1 =~ /^port(\d)_num/)
			    {
					given ($1)
					{
					    when (1) { $nlPortLo->{source} = $fEltText; }
					    when (2) { $nlPortHi->{source} = $fEltText; }
					}
			    }
			    elsif ($1 =~ /^port(\d)_str/)
			    {
					given ($1)
					{
						# +++wayne this will return a single port number.  Do we need to worry about
						# multiple noted port numbers logically associated with one port (e.g., netbios)?
					    when (1) { $nlPortLo->{source} = get_port_number($fEltText, $primaryProtocol ); }
					    when (2) { $nlPortHi->{source} = get_port_number($fEltText, $primaryProtocol); }
					}			    	
			    }
			    elsif ($1 =~ /^port_op/)
			    {
					$nlPortOp->{source} = $fEltText;
			    }
			    else
			    {
				# let the common dest/src argument handler deal with it
					_parse_access_list_entry_direction_options($fEltTag, $fEltText, 'src', 'source',
							       $filterEntry );

			    }
				
			}
		}
	}
	
	# Now, we have to see how to assign the port op plus arguments....
	foreach my $gDir ("source", "destination")
	{
		my $gServ;
		my $gOp = $nlPortOp->{$gDir};
		
		if (not defined $gOp)
		{
			; # nothing
		}
	    # There has to be an operator for value, and the one we special case is 'range'...
	    elsif ($gOp eq 'range')
	    {
			if (defined $nlPortLo->{$gDir} and defined $nlPortHi->{$gDir})
			{
			    $gServ->{portRange}->{portStart} =	$nlPortLo->{$gDir};
			    $gServ->{portRange}->{portEnd} =	$nlPortHi->{$gDir};
			}
			else
			{
			   $LOGGER->debug("Cisco::Nexus::parse_access_lists; could not parse port arguments " .
					  "for range  in spec $seqno of access list $alName");
			}
	    }
	    else  # if defined otherwise, 'tis unary operator
	    {
			if (defined $nlPortLo->{$gDir})
			{
				# Mapping for range
				$gOp =~ s/neq/ne/;
			    $gServ->{portExpression}->{operator} = $gOp;
			    $gServ->{portExpression}->{port} =	$nlPortLo->{$gDir};
			}
			else
			{
		    	$LOGGER->debug("Cisco::Nexus::parse_access_lists; could not parse port argument " .
					  "for operator \' $gOp \'  in spec $seqno of " .
				  	" access list $alName");
			}
		}
		
		if ($gServ)
		{
			push (@{$filterEntry->{$gDir . "Service"}}, $gServ);
		}
	}

	return ($seqno, $feDescription);
}

# This parses "source" or "dest" arguments for a filter spec.  NOTE that actual port numbers and port ops are 
# parsed by the parent, since they have to be "fixed up", depending on port operator.
# 
sub _parse_access_list_entry_direction_options
{
	my ($elementTag, $elementSData, $tagDirection, $generatedDirection, $filterEntry) = @_;
	my $cidrMask = get_crep('cidr');
	my $gAddr;
	
	# Reality check that we're parsing the intended directional element
	if ($elementTag !~ /^$tagDirection/i )
	{
		$LOGGER->debug("Cisco::Nexus::parse_access_lists; _parse_access_list_entry_direction (direction $tagDirection) called with element $elementTag");
		return;	
	}	
	
	given ($elementTag)
	{
		when (/^${tagDirection}_(\S+)/) 
		{
			given ($1)
			{
				when ("ip_prefix") {
					if ($elementSData =~ /^$cidrMask/)
					{
						my $cAddr = parseCIDR($elementSData);
							
						# see if this is a /32
						if ( $cAddr->{prefix} == 32)
						{
							$gAddr->{host} = $cAddr->{host};
						}		
						else
						{
							# a real network
							$gAddr->{network} = create_network_element($cAddr->{host}, $cAddr->{network});
						}
					}
					else
					{
						$LOGGER->debug("Cisco::Nexus::parse_access_lists; cannot parse CIDR dest mask: $elementSData");
					}
					
				} # end of "ip_prefix"
				when ("any") {$gAddr->{network} = get_any_address_encoding(0); }
				when ("addrgrp") {$gAddr->{objectGroupReference} = $elementSData; }
				when ("portgrp") {push (@{$filterEntry->{$generatedDirection . "Service"}}, 
								{ objectGroupReference => $elementSData}); }
			} # end of "given $1 "	
		} # end of "when $tagDirection"...
	} # end of "given element tag"

	if ($gAddr)
	{
		push(@{$filterEntry->{$generatedDirection . "IpAddr"}}, $gAddr);						
	}
}

# This simply allows the parser to check for valid XML before passing the string to Twig->parse() (where non
# conforming XML will result in an uncheckable error.).  Returns boolean.
# +++wayne should probably just move this to the ZAF.
sub _is_conforming_xml
{
	my $xmlBlock = shift;

	eval
	{
		my $dontcareRef = XMLin($xmlBlock);	
	};

	return (not $@);
}

sub _parse_routing_protocol_interfaces
{
	my ($in, $rProtName, $vrfName) = @_;
	my @rpInterfaces = ();
	
	if (defined $in->{"$rProtName"}->{interfaces})
	{
    	# +++wayne oof, this has individual "as"-level entries nested in the XML structure.  This will take redesign for
    	# the ability to scan through multiple AS'es when we do BGP and OSPF

		# This is XML encoded, hence we can twig our way through pulling up the 
		# interface name
		my $twig = XML::Twig->new(twig_roots => {
				ROW_vrf => sub { 
					my ($twig, $element) = @_;
					# Note that in the XML, the VRF *is* assigned.
					if ($element->first_child_text('vrf') !~ /^$vrfName/)
					{
						return;  # Not the right VRF
					}
					
					for ($element->first_child('TABLE_if'))
					{
					    foreach my $rIfElement ($_->children('ROW_if'))
					    {
					    	my $ifN = $rIfElement->first_child_text('ifname');
					    	if ($ifN)
					    	{
						    	push (@rpInterfaces, expand_nexus_interface_name($ifN));				    		
					    	}
					    }						
					}
				}
		});
		
		$twig->parse($in->{"$rProtName"}->{interfaces})		
	}
	
	return \@rpInterfaces;
}

sub clean_interfaces {
	my $in = shift;

	my $interfacesValid = $in->{interface};
	$interfacesValid =~ s/(<\/.*)nftlv\(output\):.*/$1>/mg;
	$interfacesValid =~ s/nftlv\(output\):.*|^\s+>$//mg;
	$interfacesValid =~ s/^\s*\n+//mg;
	return ($interfacesValid);
}

1;

__END__

=pod

=head1 NAME

    Parsers.pm

=head1 DESCRIPTION

    Parsing logic for Cisco Nexus configuration.

=head2 Methods

=over

=item parse_system() 

=item parse_interfaces() 

=item parse_access_ports()

=item parse_chassis()

=item parse_interfaces()

=item parse_object_groups()

=item parse_access_lists()

=item parse_routing()

=item parse_static_routing()

=item expand_nexus_interface_name()

=item open_vdc_section()

=item close_vdc_section()

=item create_config()

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2012, IBM Corp.

=head1 AUTHOR

Contributor(s): Wayne Tackabury (wtackabury@us.ibm.com) David Payne (david.payne@ca.ibm.com)
Date: November 15, 2012

=cut
