package ZipTie::Adapters::Cisco::Nexus::ResponseHandler;

use strict;
use warnings;
use v5.10;

use XML::Twig;
use ZipTie::Logger;
use ZipTie::CLIProtocol;

use Exporter 'import';
our @EXPORT_OK = qw(populate_virtual_device_context delete_instance_data netconf_elements_extract netconf_response_elements_extract netconf_text_response_elements_extract nexus_command_cli_response get_virtual_device_contexts cli_xml_command);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# These are manifest constants to direct how we interpret XML command issuance

# If this is XML over SSH, we issue commands with the " | xml" CLI designator
use constant NEXUS_XML_COMMANDS_XML_SSH => 1;
# if this is a real Netconf connection, then we issue Netconf requests to the port requested
use constant NEXUS_XML_COMMANDS_NETCONF => 2;
# The real kicker is that at least one device (the Nexus 4000) in fact does not support Netconf or
# XML command formatting *at all*.  We actually have to back off to a pure CLI text response mode of
# parsing, and the command as issued has no XML form or request designator (although we'll try the latter
# and glean if we got an error to see if we're in this predicament
use constant NEXUS_XML_NO_XML_TEXT_ONLY => 3;

# populate_virtual_device_context
#
# Set up the instance data for the context (or global context) as passed by issuing the
# sequence of commands to collect said data to the device.
#
# RETURNS: hash at level of context (or global context) populated
#
sub populate_virtual_device_context
{
	my ($args) = @_;

	# The following arguments are not optional
	if (!defined($args->{responses}))
	{
		$LOGGER->fatal('populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, responses not defined!');
	}
	if (!defined($args->{cliProtocolInst}))
	{
		$LOGGER->fatal('populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, cliProtocolInst not defined!');
	}
	if (!defined($args->{promptRegex}))
	{
		$LOGGER->fatal('populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, promptRegex not defined!');
	}

	my $responses = $args->{responses};
	my $cliProtocolInst = $args->{cliProtocolInst};
	my $promptRegex = $args->{promptRegex};
	my $locVdcPrompt = $promptRegex;

	my $altSBktName = {};
	if (defined($args->{altBucketName}))
	{
		$altSBktName = $args->{altBucketName};
	}
	
	my $dcResponses;
	my $showCmd;
	my $sBkt;
		
	if ($args->{vdcName})
	{
		$locVdcPrompt = _switchto_vdc($args->{vdcName}, $args->{cliProtocolInst});
		
		if ($locVdcPrompt)  
		{
			# Change to virtual device context
			my $vdcName = $args->{vdcName};
			
			$responses->{contexts}->{"$vdcName"} = {};
			# migrate down the xml form of the parent
			$dcResponses = $responses->{contexts}->{"$vdcName"};

			if ($responses->{xmlForm})
			{
				$dcResponses->{xmlForm} = $responses->{xmlForm}; 
			}
			else
			{
				$LOGGER->debug("populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, " ,
					"unable to find xmlForm of admin context, not copying through to context $vdcName");
			}
						
			# Only set terminal length if we're on a true terminal line--netconf doesn't issue
			# more prompts.
			if ($cliProtocolInst->get_protocol_name() !~ /Netconf/i)
			{
				$cliProtocolInst->send_and_wait_for( "terminal length 0", $locVdcPrompt );			
			}
		}
		else
		{
			return $dcResponses; # *not* an error if the switch didn't work
		}		
	}
	else
	{
		$dcResponses = $responses;
	}
	
	# Gather inputs for the model
	
	# Get the textual command responses
	
	for $showCmd (@{$args->{textShowCommands}})
	{
		# Trim off any command modifiers, and turn dashes into underscores since we're used
		# to that, unless there's an explicit mapping to response bucket.
		
		if (defined $altSBktName->{"$showCmd"})
		{
			$sBkt = $altSBktName->{"$showCmd"};
		}
		else
		{
			($sBkt) = $showCmd =~ /^(?:ip\s+)?(\S+)(\s*\S*)?/;
			$sBkt =~ s/\-/_/g;			
		}
		$dcResponses->{"$sBkt"} = _nexus_command_text_response($dcResponses, $cliProtocolInst, "show $showCmd", $locVdcPrompt );
	}
	
	# Populate the file systems
	
	for $showCmd  (@{$args->{filesystems}})
	{
		$dcResponses->{show_fs}->{"$showCmd"} = _nexus_command_text_response($dcResponses, $cliProtocolInst, "dir " . $showCmd . ":", 
											$locVdcPrompt);
	}

	# Get the XML command responses

	for $showCmd (@{$args->{xmlShowCommands}})
	{
		if (defined $altSBktName->{"$showCmd"})
		{
			$sBkt = $altSBktName->{"$showCmd"};
		}
		else
		{
			# Trim off any command modifiers, and any leading "ip" for the show, unless there's an explicity override
			($sBkt) = $showCmd =~ /^(?:ip\s+)?(\S+)(\s*\S*)?/;
			$sBkt =~ s/\-/_/g;			
		}
		$dcResponses->{"$sBkt"} = nexus_command_cli_response($dcResponses, $cliProtocolInst, "show $showCmd", $locVdcPrompt);				
	}
		
	# Populate interfaces.  We have the running config.  Use lookhead to scan to the first line after "interface"
	# which has a nonspace char in the first column after the line separator.
	if (defined($args->{additionalInterfaceCommands}) && defined($dcResponses->{"running_config"}))
	{
	    while ( $dcResponses->{"running_config"} =~ /^interface\s+(\S+)(.+?)\n(?=\S)/smig ) {
			my $ifName      = $1;
			my $ifBlob      = $2;
		
			# only interested in explicitly configured interfaces - some non-whitespace element within.
			if ($ifName and $ifBlob =~ /\S/)
			{
				$dcResponses->{interfaces}->{"$ifName"} = nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show interface $ifName",
													$locVdcPrompt); 			
			}		
	    }
	}

	# If we have feature commands to run, and we don't have the running config then someone made a mistake
	if (defined($args->{featureCommands}) && !defined($dcResponses->{"running_config"}))
	{
		$LOGGER->fatal('populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, ' .
				'trying to run feature commands without running_config!');
	}
	# Now, certain commands are only going to be meaningful if we see a "feature xxx"
	# atop the config.  Otherwise, "show"'ing them actually generates an error.  The 
	# parser will know whether to parse the command.
	for $showCmd (@{$args->{featureCommands}})
	{
		if ($dcResponses->{"running_config"} =~ /^feature $showCmd/smig)
		{
			$dcResponses->{"$showCmd"} = nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show $showCmd" , $locVdcPrompt  );
			
		}
	}	

	# If we have routing protocol commands to run, and we don't have the running config then someone made a mistake
	if (defined($args->{ipRoutingProtocols}) && !defined($dcResponses->{"running_config"}))
	{
		$LOGGER->fatal('populate_virtual_device_context in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, ' .
				'trying to run routing protocol commands without running_config!');
	}
	# issue routing commands
	for my $rpShow (@{$args->{ipRoutingProtocols}})
	{
		if ($dcResponses->{"running_config"} =~ /^feature $rpShow/smig)				
		{
			$dcResponses->{"$rpShow"}->{config} = 
				nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show ip $rpShow", $locVdcPrompt );
			$dcResponses->{"$rpShow"}->{routes} = 
					nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show ip route $rpShow",  $locVdcPrompt );		
			$dcResponses->{"$rpShow"}->{topology} = 
					nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show ip $rpShow topology", $locVdcPrompt );
			$dcResponses->{"$rpShow"}->{neighbors} = 
					nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show ip $rpShow neighbors",$locVdcPrompt );
			$dcResponses->{"$rpShow"}->{interfaces} = 					
				nexus_command_cli_response($dcResponses, $cliProtocolInst,  "show ip $rpShow interfaces brief", $locVdcPrompt );			
		}
		else
		{
			$LOGGER->debug("Cisco::Nexus::backup : Routing protocol \'" . $rpShow . "\' found disabled");			
		}
	}		

	# reset the device context if we changed it
	if ($args->{vdcName})
	{
		_nexus_command_text_response($dcResponses, $cliProtocolInst, "exit" ,	$promptRegex );	# wait for the global prompt	
	}
	

	return $dcResponses;
}

sub delete_instance_data
{
	my ($args) = @_;

	# The following argument is not optional
	if (!defined($args->{responses}))
	{
		$LOGGER->fatal('delete_instance_data in ZipTie::Adapters::Cisco::Nexus::ResponseHandler, responses not defined!');
	}

	my $in = $args->{responses};
	my $cBkt;
	
	if ($in->{contexts})
	{
		for my $ctxResponseLevel ( undef, keys %{$in->{contexts}})
		{
			my $ctxResponses = (defined $ctxResponseLevel) ? $in->{contexts}->{"$ctxResponseLevel"} :
				$in;
			for  (@{$args->{textShowCommands}}, @{$args->{xmlShowCommands}}, 
				@{$args->{featureCommands}}, @{$args->{ipRoutingProtocols}})
			{
				$cBkt = $_;
				# turn dashes into underscores
				$cBkt =~ s/\-/_/g;
				if (defined $ctxResponses->{"$cBkt"})
				{
					undef $ctxResponses->{"$cBkt"};
				}		
			} 
		}
	}
	
	# Now at global level...
	for (@{$args->{textShowCommands}}, @{$args->{xmlShowCommands}}, 
		@{$args->{featureCommands}}, @{$args->{ipRoutingProtocols}})
	{
		$cBkt = $_;
		# turn dashes into underscores
		$cBkt =~ s/\-/_/g;
		if (defined $in->{"$cBkt"})
		{
			undef $in->{"$cBkt"};
		}		
	} 	
}

sub get_virtual_device_contexts
{
	my ($dcResponses, $sysName, $cliProtocolInst, $closingPrompt) = @_;
	my $vdcList;
	my $vdcTwig;

	# There's not a great place to put this, really.  There is a bug in the Nexus, at least the version
	# I'm working with, where there is no ability to navigate to subordinate VDC's.  Investigating...
	if ($cliProtocolInst->get_protocol_name() =~ /Netconf/i)
	{
		return $vdcList;
	} 
	
	# issue show vdc command
	my $vdcOutput = nexus_command_cli_response($dcResponses, $cliProtocolInst, "show vdc", $closingPrompt );
	my ($sSysName) = ($sysName =~ /^([\w_\-]+)\.?/i);
	
	
	# A lack of output in the netconf-trimmed payload isn't an error at all.  This would indicate 
	# that, for example, the "show vdc" command has been administratively disabled.
	if ($vdcOutput)
	{
		my $vdcState;
		my $vdcName;
		# +++waynet should this be in parsers?  Putting it here for usability for telemetry

		# we have to force the issue here between ssh xml and ssh text only
		if ($dcResponses->{xmlForm} ~~ NEXUS_XML_NO_XML_TEXT_ONLY)
		{
			# These all have numbers in front of them, followed by: name, state, and mac addr
			while ($vdcOutput =~ /^\d+\s+(\S+)\s+(\S+)/mig)
			{
				$vdcName = $1;
				# don't need the global context
				if ($sSysName !~ /$vdcName$/i)
				{
					if ($2 eq 'active')
					{
						push (@{$vdcList}, $vdcName);
					}
				}
			}
		}
		else
		{
			$vdcTwig = XML::Twig->new(twig_roots => {
						ROW_vdc => sub {
								my ($twig, $element) = @_;
			
								# First VDC if present is the global context
								$vdcName = $element->first_child('vdc_name')->text();
								
								# This check is done this way deliberately.  $sysName may be a vdc path, and if so
								# the vdcName would be at the end of it in one error case
								if ($sSysName !~ /$vdcName$/i)
								{
									$vdcState = $element->first_child('state')->text();
									
									if ($vdcState =~ /active/i)
									{
										$LOGGER->debug("Cisco::Nexus::backup : adding $vdcName to scanned device contexts");
										push (@{$vdcList}, $vdcName);
									}
								}
								return 1;
							}
			});
			
			$vdcTwig->parse($vdcOutput);
		}
	}
	
	return $vdcList;
}
# Issue a command and strip off the beginning and ending echo, sync, etc. around the actual netconf payload

sub netconf_elements_extract
{
	my ($cmd, $closingPrompt, $cliPInst) = @_;
	
	my $rStr = $cliPInst->send_and_wait_for( $cmd, $closingPrompt);
	return (netconf_response_elements_extract($rStr));		
	
}

sub netconf_response_elements_extract
{
	my $response = shift;

	my $xStr = "";
	
	if ($response =~ /(<\?xml version\=.*)/si)
	{
		$xStr = $1;  # so, we've trimmed off the echo of the command, etc.
		my $cTag;
		
		# There is some tricky and obscure compatibility issue here between Telnet from the Nexus
		# and Telnet from the ZipTie.  This is where the  ".+?" between the two elements in the
		# XML stream comes in--match eol/sol malarkey until next tag..
		#
		if ($xStr =~ /^<\?xml version\=[\d\.\"]+(\s\S+)?\?>.+?<([\w:\-]+)/msi)
		{
			$cTag = $2;  # grab the tag on line 2 to get the right closing tag as well
			
			# Now, just sync to the final occurance of the closing tag.  We want greedy matching here.
			if ($xStr =~ /^(.*\n<\/($cTag)>)/msi)
			{
				$xStr = $1;  # extract the closing stuff
			}
			else { $xStr = ""; }
		}
		else { $xStr = ""; }
		
		if (not $xStr)
		{
			$LOGGER->debug("Cisco::Nexus::backup : no tag found to sync to in command response");
			$xStr = "";  # Zero out response						
		}
		
		return $xStr;		
	}
	else
	{
		$LOGGER->debug("Cisco::Nexus::backup : could not find opening XML header in response to command");
		return $xStr;								
	}
	
}

sub netconf_text_response_elements_extract
{
	my ($rsp, $tRespTag) = @_;
	
	# default tRespTag
	my $mtRespTag = $tRespTag // "nf:data";
	
	# Get the main overall response XML elements
	my $mTResp = netconf_response_elements_extract($rsp);
	
	# Now, extract the overall text between the tag that consitutes the raw response
	(my $mTTextRsp) = ($mTResp =~ /^\s*<$mtRespTag>(.+?)<\/$mtRespTag>/msi);
	
	if (not defined $mTTextRsp)
	{
		$LOGGER->debug("Cisco::Nexus::backup : could not match tag $tRespTag in command response");
	}

	$mTTextRsp = clean_response($mTTextRsp);
	return $mTTextRsp;
}

sub _switchto_vdc
{
	my ($vdcName, $cliProtocolInst) = @_;
	
	my $vdcSwitchResponse = ZipTie::Adapters::Cisco::Nexus::AutoLogin::connect_vdc("switchto vdc $vdcName", $cliProtocolInst);
	
	# see what we got back, and if it contains the vdc name, then return it
	return ($vdcSwitchResponse =~ /$vdcName/i) ? $vdcSwitchResponse : "";
}

# The following two routines are where the issue of "are we doing pure netconf or just doing SSH CLI with netconf
# responses where we can" come in.  Ultimately, the two are the same for pure netconf requests. 
#
sub _nexus_command_text_response
{
	my ($in, $cliProtocolInst, $cmd, $termPrompt) = @_;
	my $value;
	
	# issue this as an exec command if this is netconf, and sync on the closing of the response,
	# not the prompt
	if ($in->{xmlForm} ~~ NEXUS_XML_COMMANDS_NETCONF)
	{
		$value = (netconf_text_response_elements_extract($cliProtocolInst->send_netconf_exec_msg($cmd)));			
	}
	else
	{
		$value = $cliProtocolInst->send_and_wait_for($cmd , $termPrompt );		
	}
	
	$value = clean_response($value);
	return $value;	
}

sub nexus_command_cli_response
{
	my ($in, $cliProtocolInst, $cmd, $termPrompt) = @_;
	my $value;

	if ($in->{xmlForm} ~~ NEXUS_XML_COMMANDS_NETCONF)
	{
		$value = ($cliProtocolInst->strip_netconf_elemment_namespaces(
			netconf_text_response_elements_extract($cliProtocolInst->send_netconf_exec_msg($cmd))));
	}
	elsif ($in->{xmlForm} ~~ NEXUS_XML_NO_XML_TEXT_ONLY)
	{
		$value = _nexus_command_text_response($in, $cliProtocolInst, $cmd, $termPrompt);
	}
	else
	{
		$value = netconf_elements_extract( "$cmd | xml",	$termPrompt, $cliProtocolInst );
	}
	
	$value = clean_response($value);	
	return $value;	
}	

sub cli_xml_command
{
	my ($in, $command) = @_;
	return ($in->{xmlForm} ~~ NEXUS_XML_COMMANDS_NETCONF? "$command | xml" : $command);
}

sub clean_response
{
	my $response = shift;
	
	$response =~ s/\e\[\d*m*--More--\e\[\d*m*\s*\n*//mig;
	$response =~ s/\e\[K//mig;
	$response =~ s/(<\/\w*)\s*\n\s*(\w*>)/$1$2/;
	
	return $response;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::Nexus::ResponseHandler - Module for populating and deleting the response hash for the Cisco Nexus adapter.

=head1 DESCRIPTION

This module is designed to be used by Nexus.pm and Telemetry.pm files in the Cisco Nexus adapter.  Its subroutines takes in lists of commands
as arguments to act on.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<populate_virtual_device_context($args)>

Changes to a context and issues the commands to populate the responses for that context.
This allows a common set of items for the configuration for all contexts.

Returns a hash at level of context (or global context) populated.

The arguments are passed in via a hash reference.  The following are required elements for the hash reference.  If any of them
are not provided the adapter operation will fail with a fatal error message.

	key: responses, type: reference to hash

	key: cliProtocolInst, type: ZipTie::CLIProtocol

	key: promptRegex, type: scalar

The following are optional elements for the hash reference.

	key: vdcName, type: scalar

	key: filesystems, type: reference to array

	key: xmlShowCommands, type: reference to array

	key: textShowCommands, type: reference to array

	key: featureCommands, type: reference to array

	key: ipRoutingProtocols, type: reference to array

	key: altBucketName, type: reference to array

	key: additionalInterfaceCommands, type: scalar

=item C<delete_instance_data($args)>

Deletes instance data from a responces hash.

The arguments are passed in via a hash reference.  The following are required elements for the hash reference.  If any of them
are not provided the adapter operation will fail with a fatal error message.

	key: responses, type: reference to hash

The following are optional elements for the hash reference.

	key: textShowCommands, type: reference to array

	key: xmlShowCommands, type: reference to array

	key: featureCommands, type: reference to array

	key: ipRoutingProtocols, type: reference to array

=item C<netconf_elements_extract($cmd, $closingPrompt, $cliPInst)>

Issue a command and strip off the beginning and ending echo, sync, etc, around the actual payload.

The following arguments are required:

	arg: $cmd, the command to issue.

	arg: $closingPrompt, the prompt regex.

	arg: $cliPInst, the ZipTie::CLIProtocol object already connected to the Nexus device.

=item C<netconf_response_elements_extract($response)>

Based on passed command response, strip off the beginning and ending echo, sync, etc, around the actual payload.

The following arguments are required:

	arg: $response, the response payload to extract out of.

=item C<get_virtual_device_contexts($args)>

Extract the device contexts from the device.  Return a list of discovered context names.

	arg: cli_protocol_inst : the ZipTie::CLIProtocol object already connected to the Nexus device.
	arg: $closingPrompt, the prompt regex.
	 
=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHORS

Contributor(s): Wayne Tackabury (wtackabury@us.ibm.com), David Payne (david.payne@ca.ibm.com)
Date: February 20, 2013

=cut
