package ZipTie::Adapters::Cisco::SecurityAppliance;

use strict;
use warnings;
use v5.10;

use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin;
use ZipTie::Adapters::Cisco::SecurityAppliance::GetConfig qw(get_config);
use ZipTie::Adapters::Cisco::SecurityAppliance::Parsers qw(parse_names parse_routing create_config parse_local_accounts parse_chassis parse_object_groups parse_filters parse_snmp parse_system parse_interfaces parse_dynamic_routes parse_static_routes parse_vlans parse_stp parse_subelements parse_subelement_data);
use ZipTie::Adapters::Cisco::SecurityAppliance::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::SecurityAppliance::Restore;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Model::XmlPrint;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::Adapters::GenericAdapter;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

my $SA_MORE_PROMPT_REGEX = '<?\-{2,3} ?[mM]ore ?\-{2,3}>?\s*$';

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup
{
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ( $connection_path ) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );
	
	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the enable prompt that is returned as a result of a successful connection.
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );

	# Grab an output filehandle for the model.  This usually points to STDOUT
	my $filehandle = get_model_filehandle( 'SecurityAppliance', $connection_path->get_ip_address() );

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# Begin executing commands on the device.  The results of each command will
	# be stored in a single hashtable ($parsers) and fed into each parsing method
	my $responses = {};

	# Configure the terminal page to be unlimited in length
	my $tpResp = $cli_protocol->send_and_wait_for( "terminal pager 0", $enable_prompt_regex );

	$responses->{names} = $cli_protocol->send_and_wait_for( 'show names', $enable_prompt_regex );
	$responses->{names} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	parse_names($responses);

	# now output the configuration
	$responses->{running_config} = get_config( $cli_protocol, $connection_path, 'running-config' );
	if ($responses->{running_config} =~ /% Invalid input detected|for a list of available commands/)
	{
		$LOGGER->fatal_error_code($INSUFFICIENT_PRIVILEGE, $cli_protocol->get_ip_address(), "Unable to issue \"show running-config\"");
	}
	$responses->{running_config} =~ s/$enable_prompt_regex$//;
	$responses->{running_config} =~ s/^.*?(?=^(?:!|\S+\s+Version\s+\d))//mis;
	$responses->{running_config} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	$responses->{version}	     = $cli_protocol->send_and_wait_for( 'show version', $enable_prompt_regex );
	$responses->{version}        =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );

	# Detirmine if this is a transparent firewall
	my $firewallTransparent = 0;
	if ($responses->{running_config} =~ /^firewall transparent/m)
	{  
		$firewallTransparent = 1;
	}
	$responses->{firewallTransparent} = $firewallTransparent;

	# Crypto maps only exist on single context devices, only need to issue this command once.
	$responses->{crypto_ipsec_sa} = $cli_protocol->send_and_wait_for('show crypto ipsec sa', $enable_prompt_regex);

	# Gather information on eigrp routes
	$responses->{show_route} = $cli_protocol->send_and_wait_for('show route', $enable_prompt_regex);
	$responses->{show_eigrp_topology} = $cli_protocol->send_and_wait_for('show eigrp topology', $enable_prompt_regex);
	$responses->{show_eigrp_neighbors} = $cli_protocol->send_and_wait_for('show eigrp neighbors', $enable_prompt_regex);

	$responses->{show_shun} = $cli_protocol->send_and_wait_for('show shun', $enable_prompt_regex);
	$responses->{show_shun} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt

	# change to system context where available (FWSMs)
	my $changeSystem = $cli_protocol->send_and_wait_for( 'changeto system', $enable_prompt_regex.'|[#>]\s*$' );
	
	# If the prompt changed from what it was, then we made something happen.
	
	if ($changeSystem !~ /$enable_prompt_regex/)
	{
		my $systemPrompt = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::_get_prompt( $cli_protocol );
		$LOGGER->debug("SecurityAppliance::backup() : found system prompt \'$systemPrompt\'");			
		$responses->{na_running_config} = get_config( $cli_protocol, $connection_path, 'running-config', $systemPrompt );
		$responses->{na_running_config} =~ s/$systemPrompt//;
		$responses->{na_running_config} =~ s/^.*?(?=^(?:!|\S+\s+Version\s+\d))//mis;
		$responses->{na_running_config} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
		$responses->{na_startup_config} = get_config( $cli_protocol, $connection_path, 'startup-config', $systemPrompt );
		$responses->{na_startup_config} =~ s/$systemPrompt//;
		$responses->{na_startup_config} =~ s/^.*?(?=^(?:!|\S+\s+Version\s+\d))//mis;
		$responses->{na_startup_config} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
		
		my ($adminContext) = $responses->{na_running_config} =~ /^admin-context\s+(\S+)/mi;
		if ($adminContext) 
		{
			# The top-level parsing logic for user-level security contexts will simply create sufficient, er,
			# internal context in $responses to be able to allow us to scan for them
			parse_subelements($responses, $printer);
				
			foreach my $uContext ( @{$responses->{userContexts}} )
			{
				# pass on the admin context, we're have that as the top level config elements ane are or will be
				# good with that, it's user contexts we're after here
				if ($uContext->{name} eq $adminContext)
				{
					next;
				}
				
				# context prompt should be a slash followed by the context name
				my $uContextPrompt = "\\/$uContext->{name}";
				my $changeSystem = $cli_protocol->send_and_wait_for( "changeto context $uContext->{name}", $uContextPrompt  );
				
				if ( $changeSystem !~ /invalid/i)
				{
					# firewall transparent is set in the system context but is applied across the whole device.
					$uContext->{firewallTransparent} = $firewallTransparent;

					# Names are inherited from the admin context
					$uContext->{names} = $cli_protocol->send_and_wait_for( 'show names', $uContextPrompt );

					$uContext->{running_config} = get_config( $cli_protocol, $connection_path, 'running-config', $uContextPrompt );
					$uContext->{names} = $cli_protocol->send_and_wait_for( 'show names', $uContextPrompt );
					
					# Collect interfaces in either of two ways....
					$uContext->{interfaces} = $cli_protocol->send_and_wait_for( "show interface detail", $uContextPrompt );
					if ($uContext->{interfaces} =~ /[Ii]ncorrect|[Ii]nvalid input|Usage:|% Error/mi )
					{
						$uContext->{interfaces} = $cli_protocol->send_and_wait_for( "show interface", $uContextPrompt );						
					}
					$uContext->{show_shun} = $cli_protocol->send_and_wait_for('show shun', $uContextPrompt);

					for ($uContext->{running_config}, $uContext->{names}, $uContext->{interfaces}, $uContext->{show_shun})
					{
						$_ =~ s/$uContextPrompt//;
						$_ =~ s/<---\s+More\s+--->\s*//mig;			
					}						
					# Store off the running config to be able to create a composite return configuration
					# store element...
				$LOGGER->debug("SecurityAppliance::backup() : storing config for context $uContext->{name} ");			
					$responses->{context_configs}->{$uContext->{name}} = $uContext->{running_config};	
				}
				else
				{
					$LOGGER->debug("SecurityAppliance::backup() : ** unable to changeto context $uContext->{name} ; error: " .
						$changeSystem);
				}
				# Either way, back to the changeable point of system context +++waynet is it "changeto context admin"
				# I actually have to issue here?
				$cli_protocol->send_and_wait_for( 'changeto system', $enable_prompt_regex.'|[#>]\s*$' );				
			} 
			
			$LOGGER->debug("SecurityAppliance::backup() : changing to admin-context \'$adminContext\'");			
			$cli_protocol->send_and_wait_for( 'changeto context '.$adminContext, $enable_prompt_regex );			
		}
		else
		{
			$adminContext = $responses->{na_startup_config} =~ /admin-context(.*)/mi;
			$LOGGER->debug("SecurityAppliance::backup() : avoiding admin-context change!  Found in startup config: \'$adminContext\'.");
			$adminContext = $responses->{na_running_config} =~ /admin-context(.*)/mi;
			$LOGGER->debug("Found in running config: \'$adminContext\'.");
		}
	}

	$responses->{startup_config} = get_config( $cli_protocol, $connection_path, 'startup-config' );
	$responses->{startup_config} =~ s/$enable_prompt_regex$//;
	$responses->{startup_config} =~ s/^.*?(?=^(?:!|\S+\s+Version\s+\d))//mis;
	$responses->{startup_config} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt
	create_config( $responses, $printer );
	delete $responses->{version};
	delete $responses->{startup_config};
	delete $responses->{na_running_config};
	delete $responses->{na_startup_config};
	$LOGGER->debug("SecurityAppliance::backup() : deleting context configs");
	delete $responses->{context_configs};

	$responses->{interfaces} = $cli_protocol->send_and_wait_for( "show interface detail", $enable_prompt_regex );
	if ( $responses->{interfaces} =~ /[Ii]ncorrect|[Ii]nvalid input|Usage:|% Error/mi )
	{
		$responses->{interfaces} = $cli_protocol->send_and_wait_for( "show interface", $enable_prompt_regex );
	}
	$responses->{interfaces} =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt

	$LOGGER->debug("Parse Object Groups.");	
	parse_object_groups( $responses, $printer );
	$LOGGER->debug("Parse Dynamic Routes.");	
	parse_dynamic_routes( $responses, $printer );
	$LOGGER->debug("Parse Filters.");	
	parse_filters( $responses, $printer );

	delete $responses->{names};

	$LOGGER->debug("Parse Interfaces.");	
	parse_interfaces( $responses, $printer );
	delete $responses->{interfaces};
	$LOGGER->debug("Parse Local Accounts.");	
	parse_local_accounts( $responses, $printer );
	$LOGGER->debug("Parse SNMP.");	
	parse_snmp( $responses, $printer );
	$LOGGER->debug("Parse Static Routes.");	
	parse_static_routes( $responses, $printer );

	# Go back to parse the subelement data collected previously
	$LOGGER->debug("Parse Subelement.");
	parse_subelement_data($responses, $printer);
	delete $responses->{userContexts};
	delete $responses->{running_config};

	# Vlan information was parsed in parse_interafces, but is printed out here.
	$LOGGER->debug("Parse Vlans.");
	parse_vlans($responses, $printer);
	delete $responses->{_vlans};

	# close out the ZiptieElementDocument
	$printer->close_model();
	
	# Make sure to close the model file handle
	close_model_filehandle($filehandle);
	
	# Disconnect from the specified device
	disconnect($cli_protocol);
}

sub commands
{
	my $package_name = shift;
	my $command_doc = shift;
	my ( $connection_path, $commands ) = ZipTie::Typer::translate_document( $command_doc, 'connectionPath' );
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );
	my $termLen = $cli_protocol->send_and_wait_for( "terminal length 0", $enable_prompt_regex );
	
	my $result = ZipTie::Adapters::GenericAdapter::execute_cli_commands('SecurityAppliance', $cli_protocol, $commands, $enable_prompt_regex.'|(#|\$|>)\s*$');
	disconnect($cli_protocol);
	return $result;
}

# Invockes the restore module
sub restore
{
	my $package_name = shift;
	my $command_doc  = shift;

	# Get the connection path and the restore file
	my ( $connection_path, $restoreFile ) = ZipTie::Typer::translate_document( $command_doc, 'connectionPath' );

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the enable prompt that is returned as a result of a successful connection.
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );

	# Restore the configuration
	ZipTie::Adapters::Cisco::SecurityAppliance::Restore::execute( $connection_path, $cli_protocol, $enable_prompt_regex, $restoreFile );

	# Disconnect from the specified device
	disconnect($cli_protocol);
}

sub _connect
{
	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);
	#$cli_protocol->set_more_prompt( $SA_MORE_PROMPT_REGEX, '20');

	# Make a connection to and successfully authenticate with the Security Appliance device
	my $enable_prompt_regex = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::execute( $cli_protocol, $connection_path );

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name( "enablePrompt", $enable_prompt_regex );
	
	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $enable_prompt_regex );
}

sub _get_file_systems
{
	# give the output of "show file systems", this method returns
	# and array of the file systems
	my $show_file_systems_response = shift;
	my @fs;
	while ( $show_file_systems_response =~ /^\*?\s+\d+\s+\d+\s+\b(\S+)\b\s+[A-Za-z]+\s+(\S+):/mg )
	{
		my $type = $1;
		my $name = $2;
		if ( $type !~ /opaque|nvram/i )
		{
			push( @fs, $name );
		}
	}
	return @fs;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::SecurityAppliance - Adapter for performing various operations against Cisco Security Appliance devices.

=head1 SYNOPSIS

    use ZipTie::Adapters::Cisco::SecurityAppliance;
	ZipTie::Adapters::Cisco::SecurityAppliance::backup( $backup_document );

=head1 DESCRIPTION

This module represents an adapter that can be used to perform various operations against against Cisco Security Appliance devices.

Generally you would run this module through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document.  This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question.  This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retreived, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.

=item C<_disconnect($cli_protocol)>

Disconnects from the CLI of the device.

=item C<_get_file_systems($show_file_systems_response)>

Parses the response of the "show file systems" and returns an array representing the file system of the Cisco Security
Appliance device.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org)
Date: August 10, 2007

=cut
