package ZipTie::Adapters::Fortinet::FortiOS;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Fortinet::FortiOS::AutoLogin;
use ZipTie::Adapters::Fortinet::FortiOS::Parsers qw(parse_vlans get_hardware_nics parse_object_groups parse_routing create_config parse_chassis parse_filters parse_system parse_interfaces parse_static_routes parse_vlans parse_stp);

use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::SNMP;
use ZipTie::SnmpSessionFactory;
use ZipTie::Typer;

use Exporter 'import';
our @EXPORT_OK = qw(get_services_predefined_list clean_responses);

use Readonly;
Readonly my $SHOW_FULL_CONF_WAIT_SECONDS => 600;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup
{
	my $packageName = shift;
	my $backupDoc   = shift;    # how to backup this device
	my $responses    = {};       # will contain device responses to be handed to the Parsers module

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'Fortinet FortiOS', $connectionPath->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# Make a Telnet or SSH connection
	my ($cliProtocol, $promptRegex) = _connect($connectionPath);
	
	disable_pagination ($cliProtocol, $promptRegex);
	
	$responses = get_command_responses($cliProtocol, $promptRegex);
	clean_responses ($responses);

	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );
	create_config( $responses, $printer );
	parse_object_groups( $responses, $printer );
	parse_filters($responses, $printer);
	parse_interfaces( $responses, $printer );
	parse_static_routes( $responses, $printer );
	parse_vlans( $responses, $printer);
	
	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle

	return;
}

sub get_command_responses {
	my ($cliProtocol, $promptRegex) = @_;
	
	my $responses = {};
	
	# Get ALL responses
	$responses->{system_interfaces} = $cliProtocol->send_and_wait_for( 'show system interface', $promptRegex );
	$responses->{hardware_nics} = $cliProtocol->send_and_wait_for( 'get hardware nic', $promptRegex );
	# need to get responses for each nic here
	my @hardwareNicsArray = get_hardware_nics($responses);
	foreach my $nic (@hardwareNicsArray){
		$responses->{nic_info}{$nic} = $cliProtocol->send_and_wait_for( "get hardware nic $nic", $promptRegex );
	}
	$responses->{status} = $cliProtocol->send_and_wait_for( 'get system status', $promptRegex);
	$responses->{performance} = $cliProtocol->send_and_wait_for( 'get system performance status', $promptRegex);
	$responses->{config} = $cliProtocol->send_and_wait_for( 'show full-configuration', $promptRegex, $SHOW_FULL_CONF_WAIT_SECONDS ); 
	$responses->{staticRoutes} = $cliProtocol->send_and_wait_for( 'get router info routing-table static', $promptRegex);
	$responses->{addresses} = $cliProtocol->send_and_wait_for( 'show firewall address', $promptRegex );
	$responses->{FQDNs} = $cliProtocol->send_and_wait_for( 'get test dnsproxy 6', $promptRegex );
	$responses->{addressGroups} = $cliProtocol->send_and_wait_for( 'show firewall addrgrp', $promptRegex );
	$responses->{servicesPredefinedList} = $cliProtocol->send_and_wait_for( 'get firewall service predefined', $promptRegex );
	# need to get responses for each predefined service here
	my @servicesPrefinedList = get_services_predefined_list($responses);
	foreach my $service (@servicesPrefinedList){
		$responses->{servicesPredefined}{$service} = $cliProtocol->send_and_wait_for( "get firewall service predefined $service", $promptRegex );
	}
	$responses->{servicesCustom} = $cliProtocol->send_and_wait_for( 'show firewall service custom', $promptRegex );
	$responses->{serviceGroups} = $cliProtocol->send_and_wait_for( 'show firewall service group', $promptRegex );
	$responses->{policies} = $cliProtocol->send_and_wait_for( 'show firewall policy', $promptRegex );
	$responses->{zones} = $cliProtocol->send_and_wait_for( 'show system zone', $promptRegex );
	$responses->{vips} = $cliProtocol->send_and_wait_for( 'show firewall vip', $promptRegex );
	$responses->{vipGroups} = $cliProtocol->send_and_wait_for( 'show firewall vipgrp', $promptRegex );
	$responses->{virtual_ip_pools} = $cliProtocol->send_and_wait_for( 'show firewall ippool', $promptRegex);
		
	_disconnect($cliProtocol);
	return $responses;
}

sub get_services_predefined_list {
	my $in = shift;

	my @servicesPredefined;
	foreach my $line ( split( /\n/, $in->{servicesPredefinedList} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^name:\s(\S+)/ ) {
			push @servicesPredefined, $1;
		}
	}
	return @servicesPredefined;
}

sub commands
{
	my $packageName = shift;
	my $commandDoc  = shift;

	my ( $connectionPath, $commands ) = ZipTie::Typer::translate_document( $commandDoc, 'connectionPath' );
	my ( $cliProtocol, $devicePromptRegex ) = _connect($connectionPath);

	my $result = ZipTie::Adapters::GenericAdapter::execute_cli_commands( 'Fortinet FortiOS', $cliProtocol, $commands, $devicePromptRegex . '|(#|\$|>)\s*$' );
	_disconnect($cliProtocol);
	return $result;
}

sub _connect
{

	# Grab our arguments
	my $connectionPath = shift;

	# Create a new CLI protocol object by using the ZipTie::CLIProtocolFactory::create sub-routine
	# to examine the ZipTie::ConnectionPath argument for any command line interface (CLI) protocols
	# that may be specified.
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);

	# Make a connection to and successfully authenticate with the device
	my $devicePromptRegex = ZipTie::Adapters::Fortinet::FortiOS::AutoLogin::execute( $cliProtocol, $connectionPath );

	# Store the regular expression that matches the primary prompt of the device under the key "prompt"
	# on the ZipTie::CLIProtocol object
	$cliProtocol->set_prompt_by_name( 'prompt', $devicePromptRegex );

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ( $cliProtocol, $devicePromptRegex );
}

sub _disconnect
{

	# Grab the ZipTie::CLIProtocol object passed in
	my $cliProtocol = shift;

	# Close this session and exit
	$cliProtocol->send("exit");

	return;
}

sub disable_pagination {
	my $cliProtocol = shift;
	my $promptRegex = shift;

	my $consolePrompt = "\\(console\\)|$promptRegex";
	$cliProtocol->send_and_wait_for('config system console', $consolePrompt );
	$cliProtocol->send_and_wait_for( 'set output standard', $consolePrompt );
	$cliProtocol->send_and_wait_for( 'end', $promptRegex );

	return;
}

sub clean_responses {
	my $responses = shift;
	foreach my $command ( keys %$responses ) {
		unless ($command eq "policies" || $command eq "vips"){
			$responses->{$command} =~ s/^--More--(\s+)?//mg
		}else{
			# need to restore spacing for policy parsing to work
			my $commandResponse = $responses->{$command};
			my $updateNextLine = 0;
			my $newCommandResponse;
			foreach my $line (split ("\n", $commandResponse)){
				if ($line =~ /^--More--/){
					$line =~ s/^--More--(\s+)?//;
					# update next line spaces
					$updateNextLine = 1;
				}elsif ($updateNextLine){
					# remove first 9 spaces
					$line =~ s/^\s{9}//;
					$updateNextLine = 0;
				}
				unless ($updateNextLine){
					$newCommandResponse .= $line . "\n";
				}
			}
			$responses->{$command} = $newCommandResponse; # update the response
		}
	}
	return;
}
1;
