package ZipTie::Adapters::Fortinet::FortiOS::Parsers;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Utils qw(get_interface_type get_crep seconds_since_epoch get_mask get_port_number trim get_interface_type mask_to_bits get_port_expression);
use MIME::Base64 'encode_base64';

use Exporter 'import';
our @EXPORT_OK = qw(parse_vlans get_hardware_nics parse_routing create_config parse_chassis parse_filters parse_system parse_interfaces parse_static_routes parse_vlans parse_stp parse_object_groups);

my $unknown = "Unknown";

my $LOGGER = ZipTie::Logger::get_logger();

my $policyFoundCount = 0;
my $policyDroppedCount = 0;

sub parse_chassis
{
	my ( $in, $out ) = @_;
	my $chassis;
	my $versionRegex = qr/Version:\s(\S+)\s(\S.+)/;
	my $serialNumberRegex = qr/Serial-Number:\s(\S+)/;

	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Fortinet';

	if (defined $in->{status}) {
		if($in->{status} =~ $versionRegex) {
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $2;
		} else {
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $unknown;
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $unknown;
		}
		if ($in->{status} =~ $serialNumberRegex) {
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		}
	}
	#$chassis->{'core:description'} = 'TBD';
	$out->print_element( "chassis", $chassis );

	return;
}

sub parse_system {
	my ( $in, $out ) = @_;

	my $versionRegex = qr/Version:\s(\S+)\s(\S.+)/;
	my $uptimeRegex = qr/Uptime:\s(.*)/;

	if ($in->{status} =~ /Hostname:\s(\S+)/) {
		$out->print_element('core:systemName', $1);
	} else {
		$out->print_element( 'core:systemName', $unknown );
	}
	
	$out->open_element('core:osInfo');
		$out->print_element('core:make', 'Fortinet' );
		if($in->{status} =~ $versionRegex) {
			$out->print_element('core:version', $2);
		} else {
			$out->print_element('core:version', $unknown);		
		}
		$out->print_element('core:osType', 'FortiOS');
	
	_generate_packet_processing_declaration($out);	
		
	$out->close_element('core:osInfo');
	
	$out->print_element('core:deviceType', 'Firewall');

	return;
}

sub create_config {
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';

	# build the simple text configuration
	my $config;
	$config->{'core:context'}    = 'active';
	$config->{'core:mediaType'}  = 'text/plain';
	$config->{'core:name'}       = 'config';
	$config->{'core:promotable'} = 'true';
	$config->{'core:textBlob'}   = encode_base64( $in->{'config'} );

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $config );

	# print the repository
	$out->print_element( 'core:configRepository', $repository );
	
	# memory management
	delete ($in->{config});

	return;
}

sub parse_routing
{
	my ( $in, $out ) = @_;
	return;
}

sub parse_filters
{
	my ( $in, $out ) = @_;
	
	# parse zones to find interface groupings
	my $zone = parse_zones($in); 
	my @filters = parse_firewall_policies($in);
	
	my @staticNat = parse_static_nat($in);
	my $staticNatList = create_filter_lists(\@staticNat, 1);
	
	# parse SDNAT using VIP
	my @vipNatArray = parse_vip_nat($in, \@filters);
	my $vipNatList = create_filter_lists(\@vipNatArray, 1,);
	
	$out->open_element("filterLists");
	
	my $filterList = create_filter_lists(\@filters);
	# print only the implicit filter if no policies defined
	unless (defined $filterList){
		print_implicit_deny_filter($out);
		$out->close_element("filterLists");
		return;
	}
	$out->open_element("filterListsSection");
	
	foreach my $key ( sort {$filterList->{$a}->{order} <=> $filterList->{$b}->{order}} keys %{$filterList}){ # iterate filterlists in order
		print_filter_lists($out, $zone, $key, $filterList);
	}
	
	$LOGGER->debug("TOTAL policies dropped is $policyDroppedCount out of $policyFoundCount");
	# reset counters
	$policyDroppedCount = 0;
	$policyFoundCount = 0;
	
	print_implicit_deny_filter($out); # needs to be last list if no matches
	$out->close_element("filterListsSection");
	
	# nat section after packet filters
	if (defined $staticNatList){
		$out->open_element("filterListsSection");
		foreach my $key ( sort {$staticNatList->{$a}->{order} <=> $staticNatList->{$b}->{order}} keys %{$staticNatList}){ # iterate filterlists in order
			print_filter_lists($out, $zone, $key, $staticNatList, 1);
		}
		$out->print_element("phaseDiscriminator", {processingPhase => {'core:packetProcessingPhase' => "routing"}, processingPhasePosition => "after"});
		$out->close_element("filterListsSection");
	}
	# print vip nat filters
	if (defined $vipNatList){
		$out->open_element("filterListsSection");
		foreach my $key ( sort {$vipNatList->{$a}->{order} <=> $vipNatList->{$b}->{order}} keys %{$vipNatList}){ # iterate filterlists in order
			print_filter_lists($out, $zone, $key, $vipNatList, 1);
		}
		$out->print_element("phaseDiscriminator", {processingPhase => {'core:packetProcessingPhase' => "routing"}, processingPhasePosition => "before"});
		$out->close_element("filterListsSection");
	}
		
	$out->close_element("filterLists");

	# mm
	delete $in->{vipMappings} if defined $in->{vipMappings};
	return;
}

sub parse_static_routes {
	my ( $in, $out ) = @_;
	my $directlyConnected = 0;
	
	# Some static routes can be directly connected
	# Alternate regex to parse the different command structure
	my $indirectRouteRegex = qr/(S\S*)\s+(\S+)\s(\S+)\s\S+\s(\S+)\s(\w*)(.*)/;
	my $directRouteRegex = qr/(S\S*)\s+(\S+)\s(\S+)\s(\w+\s+\w+\s+\w+)(.*)/;
	my $routeRegex;
	my $staticRoutes;
	
	# Parse the values for each route
	foreach my $line ( split( "\n", $in->{staticRoutes})) {
		# Determine which regex to use
		if ($line =~ /is directly connected/){
			$directlyConnected = 1;
			$routeRegex = $directRouteRegex;
		} else {
			$routeRegex = $indirectRouteRegex;
		}
		if ($line =~ $routeRegex){
			my $route = {};
			$route->{'destinationAddress'} = get_address($2);
			$route->{'destinationMask'} = get_cidr($2);
			if (!$directlyConnected) {
				$route->{'gatewayAddress'} = clean_address($4);
			} 
			$route->{'interface'} = clean_interface($5);
			$route->{'routeMetric'} = get_distance($3);
			$route->{'routePreference'} = get_preference($3);
			
			my $defaultGateway = 0;
			if ($1 =~ qr/S\*/) {
				$defaultGateway = 1;
			}
			$route->{'defaultGateway'} = $defaultGateway ? 'true' : 'false';	

			# Push each route into a staticRoute element 
			push( @{ $staticRoutes->{staticRoute} }, $route );	
		}
	}
	
	# Print parent staticRoutes element
	$out->print_element('staticRoutes', $staticRoutes);
	
	# memory management
	delete $in->{staticRoutes};

	return;
}

sub get_address {
	my $addressWithCIDR = shift;
	if ($addressWithCIDR =~ qr/(\S+)\/(\S+)/) {
		return $1;
	}
}

sub get_cidr {
	my $addressWithCIDR = shift;
	if ($addressWithCIDR =~ qr/(\S+)\/(\S+)/) {
		return $2;
	}
}

sub clean_address {
	my $uncleanAddress = shift;
	if ($uncleanAddress =~ qr/(\d+\.\d+\.\d+.\d+)/) {
		return $1;
	}
}

sub clean_interface {
	my $uncleanInterface = shift;
	if ($uncleanInterface =~ qr/(\w+)/) {
		return $1;
	}
}

sub get_distance {
	my $unparsedMetric = shift;
	if ($unparsedMetric =~ qr/(\d+)\/(\d+)/) {
		return $1;
	}
}

sub get_preference {
	my $unparsedMetric = shift;
	if ($unparsedMetric =~ qr/(\d+)\/(\d+)/) {
		return $2;
	}
}

sub parse_interfaces
{
	my ( $in, $out ) = @_;
	
	# parse the system interfaces
	my @systemInterfaceArray = parse_system_interfaces($in);

	# get the names of all hardware nics
	my @hardwareNicNamesArray = get_hardware_nics($in);
	
	# parse the hardware nics (previously returned nic names)
	my @hardwareNicDetailsArray = parse_hardware_nics($in, @hardwareNicNamesArray);
	
	# merge hashes and store in final array for printing
	my @interfaceArray = merge_hardware_and_system_interfaces(\@systemInterfaceArray, \@hardwareNicDetailsArray);
	
	# print the interfaces SED
	$out = print_interfaces($in, $out, \@interfaceArray);
	
	# memory management
	delete $in->{interfaces};
	delete $in->{hardware_nics};

	return;
}

sub parse_system_interfaces{
	my $in = shift;
	
	unless (defined $in->{system_interfaces}){
		return;
	}
	
	my @interfaceArray;
	my %interface;
	foreach my $line ( split( "\n", $in->{system_interfaces} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /edit\s\"(.*)\"/ ) { # new interface
			$interface{name} = $1;
		}
		elsif ( $line =~ /set\sip\s(\S+)\s(\S+)/ ) {
			$interface{ip} = $1;
			my $mask = mask_to_bits($2);
			$interface{subnet} = $mask;
		}
		elsif ( $line =~ /set\stype\s(\S+)/ ) {
			$interface{type} = $1;
		}
		elsif ( $line =~ /set\sstatus\s(\S+)/ ) {
			$interface{status} = $1;
		}
		elsif ( $line =~ /set\sinterface\s\"(.*)\"/ ) {
			my $interface = $1;
			$interface{vlan}{interface} = $interface;
			# create primary interface list
			unless ($interface ~~ @{$in->{primaryInterfaces}}){
				push @{$in->{primaryInterfaces}}, $interface;
			}
		}
		elsif ( $line =~ /set\svlanid\s(\d+)/ ) {
			my $vlanId = $1;
			$interface{vlan}{id} = $1;
		}
		elsif ( $line =~ /set\sstatus\s(\S+)/ ) {
			$interface{status} = $1;
		}
		elsif ( $line =~ /set\smtu\s(\d+)/ ) {
			$interface{mtu} = $1;
			my $intfName = $interface{name};
			$in->{interface_mtu_mappings}{$intfName} = $interface{mtu};
		}
		elsif ( $line =~ /next/ ) { 
				if (defined $interface{vlan}{id}){
					my $vlanId = $interface{vlan}{id};
					push @{$in->{vlans}{$vlanId}{interfaces}}, $interface{name}; # store vlans for parse_vlans
				}
				push @interfaceArray, {%interface}; # hash clone
				%interface = (); # reset hash
				next;
		}
	}	
	return @interfaceArray;
}
	
sub get_hardware_nics
{
	my ( $in ) = @_;
	
	my @hardwareNicsArray;
	
	if (defined $in->{hardware_nics}){
		foreach my $line (split('\n', $in->{hardware_nics})){
			# we expect 8 spaces and a name
			if ($line =~ /^\s+\w+/){
				my $nicName = $line;
				# trim whitespace from either end of the string
				$nicName =~ s/^\s+|\s+$//g;
				push(@hardwareNicsArray, $nicName);
			}
		}
	}
	return @hardwareNicsArray;
}

sub parse_hardware_nics
{
	my ( $in, @nicsArray ) = @_;
	
	my %nic_info;
	my @nicInfoArray;
	my $name;
	
	foreach my $nic (@nicsArray){
		if (defined $in->{nic_info}{$nic}){
			foreach my $line (split('\n', $in->{nic_info}{$nic})){
				$line =~ s/(^\s+|\s+$)//g;
				$line =~ s/://g; # we can ignore all colons as not needed so would only add complexity to parsing
				$nic_info{name} = $nic;
				if ($line =~ /Current_HWaddr\s+(\S+)/){
					$nic_info{Current_HWaddr} = $1;
				}
				elsif ($line =~ /Speed\s+(\d+)/){
					$nic_info{Speed} = $1;
				}
			}
			push (@nicInfoArray, {%nic_info});
			%nic_info = ();
		}
	}
	return @nicInfoArray;
}

sub merge_hardware_and_system_interfaces
{
	my ( $systemHash, $hardwareHash ) = @_;
	
	my @systemHashArray = @$systemHash;
	my @hardwareHashArray = @$hardwareHash;
	my %merged_hash;
	my @mergedHashArray;
	
	# compare each element of systemHash array against hardwareHash array
	foreach my $system_nic (@systemHashArray){
		foreach my $hardware_nic (@hardwareHashArray){
			# if the nic names both match then merge the hash and push into array
			unless(%merged_hash){
				my $systemInterfaceName = $system_nic->{name};
				my $hardwareInterfaceName = $hardware_nic->{name};
				# regex to trim any whitespace from names
				$systemInterfaceName =~ s/^\s+|\s+$//g;
				$hardwareInterfaceName =~ s/^\s+|\s+$//g;
				# if names match then merge hashes
				if ($systemInterfaceName eq $hardwareInterfaceName){
					%merged_hash = (%$system_nic, %$hardware_nic);
					push (@mergedHashArray, {%merged_hash});
				}
			}
		}
		
		# if names did not match then add system interface hash
		if(!%merged_hash){
			%merged_hash = %$system_nic;
			push (@mergedHashArray, {%merged_hash});
		}
		# reset the hash
		%merged_hash = ();
	}
	
	# clean the array for SED formatting
	foreach my $merged_interface (@mergedHashArray){
		foreach my $element (%$merged_interface){
			$element =~ s/^\s+|\s+$//g;
			chomp($element);
		}
	}
	return @mergedHashArray;
}

sub print_interfaces {
	my ($in, $out, $interfaceArray) = @_;
	my $ifIndexCount = 0;
	
	$out->open_element('interfaces');
	
	foreach my $interface (@$interfaceArray){
		
		if (defined $interface->{name}){
			
			$out->open_element('interface');

			my $admin_status = $interface->{status} // 'up';
			$out->print_element('adminStatus', $admin_status);

			$out->print_element('ifIndex', $ifIndexCount++);

			if (defined $interface->{Current_HWaddr}){
				$out->open_element('interfaceEthernet');
				$out->print_element('macAddress', $interface->{Current_HWaddr});
				$out->close_element('interfaceEthernet');
			}
			
			print_interface_groupings ($out, $interface, $in);
			
			if (defined $interface->{ip} && $interface->{ip} ne "0.0.0.0"){
				$out->open_element('interfaceIp');
				$out->open_element('ipConfiguration');
				$out->print_element('ipAddress', $interface->{ip});

				if (defined $interface->{subnet} && $interface->{subnet} ne "0"){
					$out->print_element('mask', $interface->{subnet});
				}

				$out->close_element('ipConfiguration');
				$out->close_element('interfaceIp');
			}
			
			if (defined $interface->{type} && $interface->{type} eq 'switch'){
				$out->print_element('interfaceBehavior', 'switching');
			}
			
			my $interfaceType = "other";
			$interfaceType = $interface->{type} if defined $interface->{type};
			if (defined $interface->{vlan}{id}){
				$interfaceType = "virtual";
			}else{
				$interfaceType = get_interface_type($interface->{name});
				if ($interfaceType eq 'other'){
					$interfaceType = get_interface_type($interface->{type});
				}
			}
		
			if (defined $interface->{Current_HWaddr}){
				$interfaceType = "ethernet"; # has a MAC address so must be ethernet
			}
			if (defined $interface->{type} && $interface->{type} eq 'switch'){ # || defined $interface->{vlan}){
				$interfaceType = "virtual";
			}
			$out->print_element('interfaceType', $interfaceType);
			
			if (defined $interface->{vlan} && defined $interface->{vlan}->{id}){
				$out->open_element('interfaceVlanTrunks');
				$out->print_element('startVlan', $interface->{vlan}->{id});
				$out->close_element('interfaceVlanTrunks');
			}
			
			if (defined $interface->{mtu}){
				$out->print_element('mtu', $interface->{mtu});
			} elsif (defined $interface->{type} && ($interface->{type} eq "tunnel" || $interface->{type} eq "switch" || $interface->{type} eq "loopback")  || defined $interface->{Current_HWaddr}){
				$out->print_element('mtu', 1500); # default mtu 
			} elsif (defined $interface->{vlan} && defined $interface->{vlan}->{interface}){
				# get mtu from parent interface
				my $parentIntf = $interface->{vlan}->{interface};
				if (defined $in->{interface_mtu_mappings} && exists $in->{interface_mtu_mappings}->{$parentIntf}){
					$out->print_element('mtu',  $in->{interface_mtu_mappings}->{$parentIntf});
				} else{
					$out->print_element('mtu', 1500); # default mtu 
				}
			}
			
			$out->print_element('name', $interface->{name});

			my $physical = 'false';
			if ( (defined $interface->{type} && $interface->{type} eq "physical") || defined $interface->{Current_HWaddr}){
				 $physical = 'true';
			}
			$out->print_element('physical', $physical);

			if (defined $interface->{Speed}){
				$out->print_element('speed', ($interface->{Speed}*1_000_000)); # convert Mbps to bps
			}
			
			$out->close_element('interface');
		}
	}
	$out->close_element('interfaces');
	# return the SED for this hash
	return $out;
}

sub parse_vlans
{
	my ( $in, $out ) = @_;
	
	unless (defined $in->{vlans}){
		return;
	}
	
	$out->open_element('vlans');
	foreach my $vlan (keys %{$in->{vlans}}){
		$out->open_element('vlan');
		$out->print_element('enabled', 'true');
		foreach my $interface (@{$in->{vlans}->{$vlan}->{interfaces}}){
			$out->print_element('interfaceMember', $interface);
		}
		$out->print_element('number', $vlan);
		$out->close_element('vlan');
	}	
	$out->close_element('vlans');
	return;
}

sub parse_stp
{
	my ( $in, $out ) = @_;
	return;
}

sub parse_object_groups{
	my ( $in, $out ) = @_;
	
	# parse FQDNs first for to store domain IPs before parsing addresses
	my %domainIPs = parse_fqdns($in);
	# parse all address objects and generate network groups
	my @networkGroups = parse_addresses($in, \%domainIPs);
	parse_firewall_object_groups($in, "addressGroups", \@networkGroups);

	# parse all service objects and generate service groups
	my @serviceGroups = parse_predefined_services($in);
	parse_custom_services($in, \@serviceGroups);
	parse_firewall_object_groups($in, "serviceGroups", \@serviceGroups);
	
	# parse vip ips for NAT
	parse_vip_addresses ($in, \@networkGroups);
	# parse vip Groups
	parse_firewall_object_groups($in, "vipGroups", \@networkGroups, 1);
	# parse vip pools
	parse_vip_pools($in, \@networkGroups);
	
	# print all generated object groups
	$out->open_element('objectGroups');
	print_network_groups($out, \@networkGroups);
	print_service_groups ($out, \@serviceGroups);
	$out->close_element('objectGroups');

	return;
}

sub parse_fqdns {
	my $in = shift;

	my %domainIPs;
	my $currentDomainName;
	foreach my $line ( split( "\n", $in->{FQDNs} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^vfid=0 name=(\S+)/ ) {
			$currentDomainName = $1;
			$currentDomainName =~ s/:$//;
		}
		elsif ( $line =~ /^\d+.\d+.\d+.\d+/ && defined($currentDomainName) ) {
			my @currentIPs = split( /\s\(.\S+\)\s?/, $line );
			push @{ $domainIPs{$currentDomainName} }, @currentIPs;
		}
	}
	return %domainIPs;
}

sub print_network_groups {
	my $out               = shift;
	my $networkGroupArray = shift;

	foreach my $networkObject (@$networkGroupArray){
		if (defined $networkObject->{country}) {
			$LOGGER->debug("DROPPING address object $networkObject->{id}, geography-based addresses not currently supported by QRM.");
			next;
		}
		$out->open_element('networkGroup');
		$out->print_element('id', $networkObject->{id});
		if (defined $networkObject->{subnet}){
			$out->print_element('subnet', $networkObject->{subnet});
		}elsif (defined $networkObject->{range}){
			$out->print_element('range', $networkObject->{range});
		}elsif (defined $networkObject->{fqdn}){
			foreach my $IPAddress (@{$networkObject->{fqdn}}){
				$out->print_element('host', $IPAddress);
			}
		}elsif (defined $networkObject->{members}){
			foreach my $member (@{$networkObject->{members}}){
				$out->print_element('objectGroupReference', $member);
			}
		}
		$out->close_element('networkGroup');
	}
	return $out;
}

sub parse_addresses {
	my ($in, $domainIPs) = @_;

	my $addressType;
	my %networkGroup;
	my @networkGroupArray;
	my $linesAfterName;
	foreach my $line ( split( "\n", $in->{addresses} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^set\ssubnet\s(\S+)\s(\S+)/ ) {
			$networkGroup{subnet}{address} = $1;
			my $mask = mask_to_bits($2);
			$networkGroup{subnet}{mask} = $mask;
		}
		elsif ( defined $addressType && $line ne "next" ) {
			if ( $addressType eq "iprange" && $line =~ /^set\sstart-ip\s(\S+)/ )
			{
				$networkGroup{range}{startAddress} = $1;
			}
			elsif ($addressType eq "iprange"
				&& $line =~ /^set\send-ip\s(\S+)/ )
			{
				$networkGroup{range}{endAddress} = $1;
			}
			elsif ($addressType eq "geography"
				&& $line =~ /^set\scountry\s(\S+)/ )
			{
				$networkGroup{country} = $1;
			}
			elsif ( $addressType eq "fqdn" && $line =~ /^set\sfqdn\s\"(\S+)\"/ )
			{
				my $domain = $1;
				if ( exists $domainIPs->{$domain} ) {
					push @{ $networkGroup{fqdn} }, $domainIPs->{$domain};
				}
			}
		}
		elsif ( $line =~ /^edit\s+\"(.*)\"/ ) {
			$networkGroup{id} = $1;
		}
		elsif ( $line =~ /^set type\s(\S+)/ ) {
			$addressType = $1;
		}
		elsif ( $line eq "next" ) {
			unless ( defined $networkGroup{subnet} || defined $networkGroup{range} 
			|| defined $networkGroup{country} || $networkGroup{fqdn}) { # no address defined so must be 0.0.0.0
					$networkGroup{subnet}{address} = '0.0.0.0';
					$networkGroup{subnet}{mask} = '0';
			}
			push @networkGroupArray, {%networkGroup}; # hash clone
			%networkGroup = (); # reset hash
			$addressType = undef;
			next;
		}
	}
    return @networkGroupArray;
}

sub parse_predefined_services {
	my $in = shift;

	my @serviceGroupArray;
	foreach my $service ( keys %{ $in->{servicesPredefined} } ) {
		my %service;
		my $serviceResponse = $in->{servicesPredefined}{$service};
		foreach my $line ( split( "\n", $serviceResponse ) ) {
			if ( $line =~ /(\S+)\s+:\s(.*)/ ) {
				$service{$1} = $2;
				$service{$1} =~ s/\s+$//;
			}
		}
		unless ( defined $service{name} ) {
			next;
		}
		my %serviceGroup;
		$serviceGroup{id} = $service{name};
		if (
			$serviceGroup{id} eq 'ANY'){
			my $portRange = {portStart => '0', portEnd => '65535' };
			push @{$serviceGroup{portRanges}}, $portRange;
		}
		if ($service{'protocol'} eq 'ICMP'){
				my $portRange = {protocol => 'icmp', portStart => '0', portEnd => '65535' };
				push @{$serviceGroup{portRanges}}, $portRange;
			}
		if ($service{'protocol'} eq 'IP'){
			my $portExp = {protocol => 'ip', operator => 'eq', port => $service{'protocol-number'}};
			push @{$serviceGroup{portExpressions}}, $portExp;
		} elsif ($service{'name'} eq 'PPTP'){ # special case
			my $portExp = {protocol => 'ip', operator => 'eq', port => '47'};
			push @{$serviceGroup{portExpressions}}, $portExp;
		}
		if (defined $service{'tcpport-range'}){
			parse_predefined_service_range(\%service, \%serviceGroup, 'tcp');
		}
		if (defined $service{'udpport-range'}){
			parse_predefined_service_range(\%service, \%serviceGroup, 'udp');
		}
		if (defined $service{'sctpport-range'}){
			parse_predefined_service_range(\%service, \%serviceGroup, 'sctp');
		}
		push @serviceGroupArray, \%serviceGroup;
	}
    return @serviceGroupArray;
}

sub parse_predefined_service_range {
	my ($service, $serviceGroup, $rangeProtocol) = @_;

	my @rangeList = split (" ", $service->{$rangeProtocol.'port-range'});
	foreach my $range (@rangeList){
		my @splitRange = split(":",$range);
		my @firstRange = split("-", $splitRange[0]);
		if (scalar @firstRange == 1){
			my $portExp = {protocol => $rangeProtocol, port => $firstRange[0]};
			push @{$serviceGroup->{portExpressions}}, $portExp;
		}elsif (scalar @firstRange == 2){
			my $portRange = {protocol => $rangeProtocol, portStart => $firstRange[0], portEnd =>  $firstRange[1]};
			push @{$serviceGroup->{portRanges}}, $portRange;
		}
	}

	return;
}

sub print_service_groups {
	my $out               = shift;
	my $serviceGroupArray = shift;

	foreach my $serviceGroup (@$serviceGroupArray) {
		$out->open_element('serviceGroup');
		$out->print_element('id', $serviceGroup->{id});
		if (defined $serviceGroup->{members}){
			foreach my $member (@{$serviceGroup->{members}}){
				$out->print_element('objectGroupReference', $member);
			}
		}
		foreach my $portExp (@{$serviceGroup->{portExpressions}}){	
			$out->open_element('portExpression');
			$out->print_element('port', $portExp->{port});
			$out->print_element('operator', 'eq');
			if (defined $portExp->{protocol}){
				$out->print_element('protocol', $portExp->{protocol});
			}
			$out->close_element('portExpression');
		}
		foreach my $portRange (@{$serviceGroup->{portRanges}}){	
			$out->open_element('portRange');
			$out->print_element('portStart', $portRange->{portStart});
			$out->print_element('portEnd', $portRange->{portEnd});
			if (defined $portRange->{protocol}){
				$out->print_element('protocol', $portRange->{protocol});
			}
			$out->close_element('portRange');
		}
		$out->close_element('serviceGroup');
	}

	return;
}

sub parse_custom_service_range {
	my ($match, $rangeProtocol, $serviceGroup) = @_;

	my @splitLine = split( " ", $match );
	foreach my $range (@splitLine) {
		my @splitRange = split(":",$range);
		my @firstRange = split("-", $splitRange[0]);
		if ( scalar @firstRange == 1 ) {
			my $portExp = {protocol => $rangeProtocol, operator => 'eq', port => $firstRange[0]};
			push @{$serviceGroup->{portExpressions}}, $portExp;
		}elsif (scalar @firstRange == 2){
			my $portRange = {protocol => $rangeProtocol, portStart => $firstRange[0], portEnd =>  $firstRange[1]};
			push @{$serviceGroup->{portRanges}}, $portRange;
		}
	}

	return;
}

sub parse_custom_services {
	my $in = shift;
	my $serviceGroupsArray = shift;
	
	my %serviceGroup;
	foreach my $line ( split( "\n", $in->{servicesCustom} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^edit\s+\"(.*)\"/ ) {
			$serviceGroup{id} = $1;
		}
		elsif ( $line =~ /set\stcp-portrange\s(.*)/ ) {
			parse_custom_service_range($1, 'tcp', \%serviceGroup);
		}
		elsif ( $line =~ /set\sudp-portrange\s(.*)/ ){
			parse_custom_service_range($1, 'udp', \%serviceGroup);
		}
		elsif ( $line =~ /set\ssctp-portrange\s(.*)/ ){
			parse_custom_service_range($1, 'sctp', \%serviceGroup);
		}
		elsif ( $line =~ /set protocol-number\s(\d+)/){
			my $portExp = {protocol => 'ip', operator => 'eq',  port => $1};
			push @{$serviceGroup{portExpressions}}, $portExp;
		}
		elsif ( $line =~ /set protocol\sICMP/){
			my $portRange = {protocol => 'icmp', portStart => '0', portEnd => '65535' };
			push @{$serviceGroup{portRanges}}, $portRange;
		}
		elsif ( $line eq "next" ) {
				push @{$serviceGroupsArray}, {%serviceGroup}; # hash clone
				%serviceGroup = (); # reset hash
				next;
		}
	}
	return;
}

sub parse_firewall_object_groups {
	my $in = shift;
	my $response = shift;
	my $fwObjectGroupArray = shift;
	my $isVipGroups = shift;

	my %fwObjectGroup;
	my $fwObjectGroupId;
	foreach my $line ( split( "\n", $in->{$response}) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^edit\s\"(.*)\"/ ) {
			$fwObjectGroupId = $1;
			$fwObjectGroup{id} = $fwObjectGroupId;
		}
		elsif ( $line =~ /^set member\s\"(.*)\"/ ) {
			my @members = split /\"\s\"/, $1;
			push @{ $fwObjectGroup{members} }, @members;
			if ($isVipGroups){
				$in->{vipGroupMappings}{$fwObjectGroupId} = \@members;
			}
		}
		elsif ( $line eq "next" && defined $fwObjectGroup{id} ) {
			push @{$fwObjectGroupArray}, {%fwObjectGroup};
			%fwObjectGroup = ();
		}
	}

	return;
}

# Optional second parameter to filter policies based on NAT type:
# - staticNatOnly
sub parse_firewall_policies {
	my $in = shift;
	my $policyFilter = shift;

	unless (defined $in->{policies}){
		return;
	}

	# Determine if any policy filters are specified
	my $staticNatOnly = undef;
	if (defined $policyFilter && $policyFilter eq "staticNatOnly"){
			$staticNatOnly = 1;
	}

	my %filterEntry;
	my @filtersArray;
	foreach my $line ( split( "\n", $in->{policies} ) ) {
		if ( $line =~ /^\s{4}edit\s(\d+)/ ) { # new rule
			$filterEntry{name} = $1;
			$filterEntry{ztSignature} = $1;
		}
		elsif ( $line =~ /set\ssrcaddr\s"(.*)"/ ) {
			my @addresses = split /"\s"/, $1;
			push @{ $filterEntry{srcAddresses} }, @addresses;
		}
		elsif ( $line =~ /set\sdstaddr\s"(.*)"/ ) {
			my @addresses = split /\"\s\"/, $1;
			push @{ $filterEntry{dstAddresses} }, @addresses;
		}
		elsif ( $line =~ /set\saction\s(\S+)/ ) {
			my $action = $1;
			$action =~ s/^accept$/permit/;
			$filterEntry{primaryAction} = $action;
		}
		elsif ( $line =~ /set\sservice\s"(.*)"/ ) {
			my @services = split /\"\s\"/, $1;
			push @{ $filterEntry{services} }, @services;
		}
		elsif ( $line =~ /set\ssrcintf\s"(.*)"/ ) {
			my @interfaces = split /\"\s\"/, $1;
			push @{$filterEntry{sourceInt}},  @interfaces;
		}
		elsif ( $line =~ /set\sdstintf\s"(.*)"/ ) {
			# split line
			my @interfaces = split /\"\s\"/, $1;
			push @{$filterEntry{destinationInt}},  @interfaces;
		}
		elsif ( $line =~ /set\slogtraffic\senable/ ) {
			$filterEntry{log} = "true";
		}
		elsif ( $line =~ /set\sstatus\sdisable/ ) {
			$filterEntry{disabled} = 1;
		}
		elsif ( $line =~ /set\sidentity-based\senable/ ) {
			$filterEntry{identityBased} = 1;
		}
		elsif ( $line =~ /set\snat\senable/ ) {
			$filterEntry{staticNat} = 1;
		}
		elsif ( $line =~ /set\scentral-nat\senable/ ) {
			$filterEntry{centralNat} = 1;
		}
		elsif ( $line =~ /set\sippool\senable/ ) {
			$filterEntry{ippool} = 1;
		}
		elsif ( $line =~ /set\sutm-status\senable/ ) {
			$filterEntry{'utm-status'} = 1;
		}
		elsif ( $line =~ /set\sservice-negate\senable/ ) {
			$filterEntry{'service-negate'} = 1;
		}
		elsif ( $line =~ /set\ssrcaddr-negate\senable/ ) {
			$filterEntry{'srcaddr-negate'} = 1;
		}
		elsif ( $line =~ /set\sdstaddr-negate\senable/ ) {
			$filterEntry{'dstaddr-negate'} = 1;
		}
		elsif ( $line =~ /^\s{4}next/ ) { # end of rule

			if (defined $filterEntry{disabled} && $filterEntry{disabled} == 1){ # don't store disabled polices
				%filterEntry = ();                    
				next;
			}
			# set various default values if not parsed
			$filterEntry{log} = "false" unless defined $filterEntry{log};
			$filterEntry{primaryAction} = "deny" unless defined $filterEntry{primaryAction};
			$filterEntry{identityBased} = 0 unless defined $filterEntry{identityBased};
			$filterEntry{'utm-status'} = 0 unless defined $filterEntry{'utm-status'};
			
			# Determine policy NAT type: static or central
			if (defined $filterEntry{centralNat} || $filterEntry{ippool}) {
				$filterEntry{staticNat} = 0;
			} else {
				$filterEntry{staticNat} = 0 unless defined $filterEntry{staticNat};	
			}
			
			# If nat filter is enabled, push only those that match
			if (defined $staticNatOnly) {
				if ($filterEntry{staticNat} == 1) {
					push @filtersArray, {%filterEntry};
				}
			} else { # push all filterEntries
				push @filtersArray, {%filterEntry};	
			}
			
			%filterEntry = ();                     # reset hash
			next;
		}
	}

	return @filtersArray;
}


sub print_filter_entry {
	my $out = shift;
	my $filterEntry = shift;
	my $processOrder = shift;
	my $natPolicy = shift;

	if($filterEntry->{identityBased}
	|| $filterEntry->{primaryAction} eq "ipsec"
	|| $filterEntry->{primaryAction} eq "ssl-vpn"){
			if ($natPolicy){
				return;
			}
			my $logMessage = "DROPPING policy id $filterEntry->{name}";
			if ($filterEntry->{identityBased}){
				$LOGGER->debug("$logMessage - identity-based policies not currently supported by QRM.");
			} else{
				$LOGGER->debug("$logMessage - VPN policies not currently supported by QRM.");
			}
			$policyDroppedCount++;
			return;
	}
	$out->open_element("filterEntry");
	if (defined $filterEntry->{dstAddrTranslation}){
		$out->print_element('destAddressTranslation', $filterEntry->{dstAddrTranslation});
	}
	foreach my $address (@{$filterEntry->{dstAddresses}}){
		if (defined $filterEntry->{vipDstFilter}){
			$out->print_element('destinationIpAddr', $filterEntry->{vipDstFilter});
		}elsif (defined $filterEntry->{natReverse}){
			$out->print_element('destinationIpAddr', $address);
		} else{
			$out->print_element('destinationIpAddr', {objectGroupReference => $address});
		}
	}
	foreach my $service (@{$filterEntry->{services}}){
			$out->print_element('destinationService', {objectGroupReference => $service});
	}
	foreach my $serviceVip (@{$filterEntry->{dstServicesVip}}){
		$out->print_element('destinationService',  $serviceVip);
	}
	$out->print_element("log", $filterEntry->{log});
	$out->print_element("name", $filterEntry->{name});
	$out->print_element("primaryAction", $filterEntry->{primaryAction});
	$out->print_element("processOrder", $processOrder);
	
	if (defined $filterEntry->{srcAddrTranslation}){
		$out->print_element('sourceAddressTranslation', $filterEntry->{srcAddrTranslation});
	}
	foreach my $address (@{$filterEntry->{srcAddresses}}){
		if (defined $filterEntry->{vipSrcFilter}){
			$out->print_element('sourceIpAddr', $filterEntry->{vipSrcFilter});
		}elsif (defined $filterEntry->{natReverse}){
			$out->print_element('sourceIpAddr', $address);
		}else {
			$out->print_element('sourceIpAddr', {objectGroupReference => $address});
		}
	}
	foreach my $srcService (@{$filterEntry->{srcServices}}){
		$out->print_element('sourceService', {objectGroupReference => $srcService});
	}
	foreach my $serviceVip (@{$filterEntry->{srcServicesVip}}){
		$out->print_element('sourceService',  $serviceVip);
	}
	
	$out->print_element("ztSignature", $filterEntry->{ztSignature});
	$out->close_element("filterEntry");

	return;
}

sub create_filter_lists {
	my $filtersArray = shift;
	my $natList = shift;

	my $filterList;
	my $orderCount;
	foreach my $filterEntry (@$filtersArray) {
		my $srcInterface   = $filterEntry->{sourceInt};
		my $dstInterface   = $filterEntry->{destinationInt};
		my $filterListName = $filterEntry->{name};
		if ($natList){
			if ($filterEntry->{natReverse}){
				$filterListName .= "_NATMapping_reverse";
			}else {
				$filterListName .= "_NATMapping_forward";
			}
			$srcInterface = $filterEntry->{srcIntfFilter} if defined $filterEntry->{srcIntfFilter};
			$dstInterface = $filterEntry->{dstIntfFilter} if defined $filterEntry->{dstIntfFilter};
		}
		unless ( exists $filterList->{$filterListName} ) {
			$filterList->{$filterListName}->{order}        = $orderCount++;
			$filterList->{$filterListName}->{srcInterface} = $srcInterface;
			$filterList->{$filterListName}->{dstInterface} = $dstInterface;
		}
		push @{ $filterList->{$filterListName}->{filters} }, $filterEntry;
	}
	return $filterList;
}

sub print_implicit_deny_filter {
	my $out = shift;

	# print implied deny all rule
	$out->open_element("filterList");
	$out->print_element( "defaultAction", "deny" );
	$out->print_element( "appliesTo", { source => "any" } );
	$out->print_element( "mode", "stateful" );
	$out->print_element( "name", "Implicit" );
	$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "packet-filtering"});
	$out->close_element("filterList");

	return;
}

sub parse_zones {
	my $in = shift;

	unless (defined  $in->{zones} ){
		return;
	}
	
	my $zone;
	my $zoneName;
	foreach my $line ( split( "\n", $in->{zones} ) ) {
		if ( $line =~ /^\s\s\s\sedit\s\"(.*)\"/ ) { 
			$zoneName = $1;
		}
		elsif ( $line =~ /set\sinterface\s\"(.*)\"/ ) {
			my @interfaces = split /\"\s\"/, $1;
			push @{ $zone->{$zoneName}{interfaces} }, @interfaces;
			push @{ $in->{parsed_zones}->{$zoneName}{interfaces} }, @interfaces; # store zones for later
		}
	}
	return $zone;
}

sub print_filter_lists {
	my $out        = shift;
	my $zone       = shift;
	my $key        = shift;
	my $filterList = shift;
	my $isNatList  = shift;

	$out->open_element("filterList");
	$out->print_element( "defaultAction", "none" );
	# so core moves on to next list if no match

	# check if src/dst interfaces are actually zones
	my $srcIntArray = $filterList->{$key}->{srcInterface};
	my $dstIntArray = $filterList->{$key}->{dstInterface};
	
	foreach my $srcInt (@{$srcIntArray}){
		foreach my $dstInt (@{$dstIntArray}){
			$out->open_element("appliesTo");
			if ( exists $zone->{$srcInt} ) {
				$out->print_element( "sourceGroup", $srcInt);
			}
			else {
				$out->print_element( "source", $srcInt);
			}
			if ( exists $zone->{$dstInt} ) {
				$out->print_element( "destinationGroup", $dstInt);
			}
			else {
				$out->print_element( "destination", $dstInt);
			}
			$out->close_element("appliesTo");
		}
	}
	
	my $processOrder = 1;
	foreach my $filterEntry ( @{ $filterList->{$key}->{filters} } ) {
		if (defined $filterEntry->{'service-negate'} || defined $filterEntry->{'srcaddr-negate'} 
		|| defined $filterEntry->{'dstaddr-negate'} && $filterEntry->{primaryAction} =~ /permit|deny/ && !$isNatList){
			process_filter_negation ($processOrder, $filterEntry, $out);
		}else{
			print_filter_entry( $out, $filterEntry, $processOrder++, $isNatList);
		}
		unless ($isNatList){
			$policyFoundCount++;
		}
	}
	$out->print_element( "mode", "stateful" );
	$out->print_element( "name", $key);
	if ($isNatList){
		$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "address-translation"});
	} else {
		$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "packet-filtering"});
	}
	$out->close_element("filterList");

	return;
}

sub print_interface_groupings {
	my $out       = shift;
	my $interface = shift;
	my $in        = shift;

	foreach my $zone ( keys %{ $in->{parsed_zones} } ) {
		my @zoneInterfaces = @{ $in->{parsed_zones}->{$zone}->{interfaces} };
		if ( $interface->{name} ~~ @zoneInterfaces ) {
			$out->open_element('interfaceGrouping');
			$out->print_element('ifGroupName', $zone);
			$out->print_element('ifGroupingCharacteristic', "securityZoneMember");
			$out->close_element('interfaceGrouping');
		}
	}
			
	if (defined $interface->{vlan} && defined $interface->{vlan}->{interface}){
		$out->open_element('interfaceGrouping');
		$out->print_element('ifGroupName', $interface->{vlan}->{interface});
		$out->print_element('ifGroupingCharacteristic', "subInterface");
		$out->close_element('interfaceGrouping');
	}
			
	if (defined $in->{primaryInterfaces} && $interface->{name} ~~ @{$in->{primaryInterfaces}}){
		$out->open_element('interfaceGrouping');
		$out->print_element('ifGroupName', $interface->{name});
		$out->print_element('ifGroupingCharacteristic', "primaryInterface");
		$out->close_element('interfaceGrouping');
	}

	return;
}

sub parse_static_nat {
	my $in = shift;

	my @staticNatArray = parse_firewall_policies( $in, "staticNatOnly" );
	my @systemInterfaceArray = parse_system_interfaces($in);
	my $interfaceIpMappings;
	# create interface ip mappings
	foreach my $interface (@systemInterfaceArray) {
		if ( defined $interface->{ip} && $interface->{subnet} ) {
			$interfaceIpMappings->{ $interface->{name} }->{ip} =
			  $interface->{ip};
			$interfaceIpMappings->{ $interface->{name} }->{subnet} =
			  $interface->{subnet};
		}
	}
	my @staticNatForwardArray;
	my @staticNatReverseArray;
	foreach my $filterEntry (@staticNatArray) {

		$filterEntry->{services} = [];    # remove services as already in associated packet-filter
		# find ip of dest intf
		my $dstinterfaceHasIP = 0;
		my $dstInterfaceArray = $filterEntry->{destinationInt};
		foreach my $dstInterface (@{$dstInterfaceArray}){
			if ( exists $interfaceIpMappings->{$dstInterface} ) {
				$dstinterfaceHasIP = 1;
			}
		}
		unless ($dstinterfaceHasIP){
			next;
		}
		
		my $dstInterface = $filterEntry->{destinationInt}[0];
		$filterEntry->{srcAddrTranslation} = {	translateAppliedAddressing => 'true',
												translationAddress => {translationIP => {network => {address => $interfaceIpMappings->{$dstInterface}->{ip} 
																				 , mask =>  32}}
																				, translationServicePort => {translationPort => {portRange =>{portStart => 1024 , portEnd => 65535}},
																				, translationProtocol => 'tcp' }}};
		# create reverse filter
		push @staticNatForwardArray, $filterEntry;
		my $reverseFilter = get_reverse_nat_filter($filterEntry, 1);
		if (defined $reverseFilter){
			push @{ $reverseFilter->{dstAddresses} }, {network => {address => $interfaceIpMappings->{$dstInterface}->{ip}, mask =>  32}};
			push @staticNatReverseArray, $reverseFilter;
		}
		
	}
	my @mergedStaticNat = (@staticNatForwardArray, @staticNatReverseArray);
	return @mergedStaticNat;
}

sub parse_vip_pools {
	my $in = shift;
	my $networkGroupArray = shift;
	
	my $SINGLE_IP_REGEX = get_crep("cipm");

	my %poolEntry;
	foreach my $line ( split( "\n", $in->{virtual_ip_pools} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ qr(^edit\s\"(.*)\") ) {
			$poolEntry{id} = $1;
		}
		elsif ( $line =~ qr(^set\sendip\s($SINGLE_IP_REGEX))) {
			$poolEntry{range}{endAddress} = $1;
		}
		elsif ($line =~ qr(^set\sstartip\s($SINGLE_IP_REGEX))) {
			$poolEntry{range}{startAddress} = $1;
		}
		elsif ( $line =~ qr(^next)) { 
			push @{$networkGroupArray}, {%poolEntry};
			%poolEntry = ();	
		}
	}
}

sub get_reverse_nat_filter {
	my $filterEntry = shift;
	my $includePAT = shift;
	my $vipMapping = shift;

	my $reverseFilter;
	if (defined $filterEntry->{srcAddrTranslation}){
		foreach my $address (@{$filterEntry->{dstAddresses}}){
			if ($address eq "all"){
				return undef; # don't create a reverse rule with a source address of 0.0.0.0 
			}
			push @{ $reverseFilter->{srcAddresses} }, {objectGroupReference => $address};
		}
		foreach my $srcAddress (@{$filterEntry->{srcAddresses}}){
			# create translation address
			my $translationAddress = {translationIP => {objectGroupReference => $srcAddress}};
			if ($includePAT){
				$translationAddress->{translationServicePort} = {translationPort => {portRange => {portStart => 1024, portEnd => 65535}}
																				, translationProtocol => 'tcp' };
		}	
	    	$reverseFilter->{dstAddrTranslation}->{translationAddress} = $translationAddress;
	    	$reverseFilter->{dstAddrTranslation}->{translateAppliedAddressing} = 'true';
		}	
	}elsif (defined $filterEntry->{srcAddrTranslationMultiple}){
		foreach my $address (@{$filterEntry->{dstAddresses}}){
			push @{ $reverseFilter->{srcAddresses} }, {objectGroupReference => $address};
		}
	} 
	elsif (defined $filterEntry->{dstAddrTranslation}){
		foreach my $address (@{$filterEntry->{srcAddresses}}){
			if ($address eq "all" && not defined $vipMapping->{srcFilter}){ # special case for VIP source filters
				return undef; # don't create a reverse rule for a destination address of 0.0.0.0 
			}
			push @{ $reverseFilter->{dstAddresses} }, {objectGroupReference => $address};
		}
		foreach my $dstAddress (@{$filterEntry->{dstAddresses}}){
			# create translation address
			my $translationAddress = {	translateAppliedAddressing => 'true',
									translationAddress => {translationIP => {objectGroupReference => $dstAddress}}};
			if ($includePAT){
				$translationAddress->{translationAddress}->{translationServicePort} = {translationPort => {portRange => {portStart => 1024, portEnd => 65535}}
																				, translationProtocol => 'tcp' };
		}	
	    	$reverseFilter->{srcAddrTranslation} = $translationAddress
	}
	}elsif (defined $filterEntry->{dstAddrTranslationMultiple}){
		foreach my $address (@{$filterEntry->{srcAddresses}}){
			push @{ $reverseFilter->{dstAddresses} }, {objectGroupReference => $address};
		}
	} 
	$reverseFilter->{sourceInt}             = $filterEntry->{destinationInt};
	$reverseFilter->{destinationInt}        = $filterEntry->{sourceInt};
	$reverseFilter->{natReverse}            = 1;
	$reverseFilter->{log}                   = $filterEntry->{log};
	$reverseFilter->{name}                  = $filterEntry->{name};
	$reverseFilter->{primaryAction}         = $filterEntry->{primaryAction};
	$reverseFilter->{processOrder}          = $filterEntry->{processOrder};
	$reverseFilter->{'utm-status'}          = $filterEntry->{'utm-status'};
	return $reverseFilter;
}

sub parse_vip_addresses {
	my $in                = shift;
	my $networkGroupArray = shift;
	
	my $SINGLE_IP_REGEX = get_crep("cipm");

	my $vipName;
	my %networkGroup;
	my $isDynamic = 0;
	foreach my $line ( split( "\n", $in->{vips} ) ) {
		if ( $line =~ /^\s\s\s\sedit\s\"(.*)\"/ ) {
			$vipName = $1;
			$networkGroup{id} = $1;
		}
		elsif ( $line =~ /set\sextip\s(?:\")?($SINGLE_IP_REGEX)(?:\")?(?:-($SINGLE_IP_REGEX))?/ ) {
			if (defined $2){
				$networkGroup{range}{startAddress} = $1;	
				$networkGroup{range}{endAddress} = $2;
			} else {
			$networkGroup{subnet}{address} = $1;
			$networkGroup{subnet}{mask} = 32;
			}
		}
		elsif ( $line =~ /set\smappedip\s(?:\")?($SINGLE_IP_REGEX)(?:\")?(?:-($SINGLE_IP_REGEX))?/ ) {
			if (defined $2){
				$in->{vipMappings}->{$vipName}->{mappedIp}{range}{startAddress} = $1;
				$in->{vipMappings}->{$vipName}->{mappedIp}{range}{endAddress} = $2;
			} else{
				$in->{vipMappings}->{$vipName}->{mappedIp}{host} = $1;
			}
		}
		elsif ( $line =~ /set\sextport\s(\d+)(?:-(\d+))?/ ) {
			if (defined $2){
				$in->{vipMappings}->{$vipName}->{extPort}{portRange}{portStart} = $1;
				$in->{vipMappings}->{$vipName}->{extPort}{portRange}{portEnd} = $2;
			} else{
				$in->{vipMappings}->{$vipName}->{extPort}{portExpression}{port} = $1;
				$in->{vipMappings}->{$vipName}->{extPort}{portExpression}{operator} = "eq";
			}
		}
		elsif ( $line =~ /set\smappedport\s(\d+)(?:-(\d+))?/ ) {
			if (defined $2){
				$in->{vipMappings}->{$vipName}->{mappedPort}{portRange}{portStart} = $1;
				$in->{vipMappings}->{$vipName}->{mappedPort}{portRange}{portEnd} = $2;
			} else{
				$in->{vipMappings}->{$vipName}->{mappedPort}{portExpression}{port} = $1;
				$in->{vipMappings}->{$vipName}->{mappedPort}{portExpression}{operator} = "eq";
			}
		}
		elsif ( $line =~ /set\ssrc-filter\s(?:\")?($SINGLE_IP_REGEX)(?:\")?(?:-($SINGLE_IP_REGEX))?/ ) {
			if (defined $2){
				$in->{vipMappings}->{$vipName}->{srcFilter}{range}{startAddress} = $1;
				$in->{vipMappings}->{$vipName}->{srcFilter}{range}{endAddress} = $2;
			} else{
				$in->{vipMappings}->{$vipName}->{srcFilter}{host} = $1;
			}
		}
		elsif ( $line =~ /set\ssrcintf-filter\s"(.*)"/ ) {
			my @interfaces = split /\"\s\"/, $1;
			push @{$in->{vipMappings}->{$vipName}->{srcIntfFilter}},  @interfaces;
		}
		elsif ( $line =~ /set\stype\s(load-balance|server-load-balance|dns-translation)/){
			$isDynamic = 1;	
		}
		elsif ( $line =~ /set\sprotocol\s(icmp)/){
			$in->{vipMappings}->{$vipName}->{protocol} = $1;
		}
		elsif ( $line =~ /next/ ) {
			push @$networkGroupArray, {%networkGroup};    # hash clone
			%networkGroup = ();
			if ($isDynamic){
				delete $in->{vipMappings}->{$vipName} if defined  $in->{vipMappings}->{$vipName}; # remove mappings for dynamic NAT
			}
			$isDynamic = 0;
			next;
		}
	}
	return;
}

sub parse_vip_nat {
	my $in      = shift;
	my $filters = shift;

	my $vipNatArray = get_vip_filters ($in, $filters);
	
	my @vipNatFilters;
	my @vipKeys = keys %{$in->{vipMappings}};
	my @vipGroupKeys = keys %{$in->{vipGroupMappings}};
	foreach my $vipFilter (@$vipNatArray) {
		my $count = 1;
		foreach my $address (@{$vipFilter->{dstAddresses}}){
			if ($address ~~ @vipKeys){
				my $vipFilterForward = create_vip_nat_filter ($in, $vipFilter, $address, $count);
				push @vipNatFilters, $vipFilterForward;
				my $vipFilterReverse = create_vip_nat_filter_reverse ($in, $vipFilterForward, $address, $count);
				push @vipNatFilters, $vipFilterReverse if defined $vipFilterReverse;
				$count++;
			} elsif ($address ~~ @vipGroupKeys){
				foreach my $vipName ( @{ $in->{vipGroupMappings}->{$address} } ) {
					my $vipFilterForward = create_vip_nat_filter ($in, $vipFilter, $vipName, $count);
					push @vipNatFilters, $vipFilterForward;
					my $vipFilterReverse = create_vip_nat_filter_reverse ($in, $vipFilterForward, $vipName, $count);
					push @vipNatFilters, $vipFilterReverse if defined $vipFilterReverse;
					$count++;
				}	
			}
		}
	}	
	return @vipNatFilters;
}

sub get_vip_filters {
	my $in      = shift;
	my $filters = shift;

	my @vipNatArray;
	FILTER: foreach my $filterEntry (@$filters) {
		# if any vips exist in src address or dst addresses return filter
		my @srcAddresses = $filterEntry->{srcAddresses};
		my @dstAddresses = $filterEntry->{dstAddresses};
		my @vips = keys %{ $in->{vipMappings}};
		my @vipGroups = keys %{ $in->{vipGroupMappings}};
		my @mergedVips = (@vips, @vipGroups);
		foreach my $vipName ( @mergedVips ) {
			if ( $vipName ~~ @srcAddresses || $vipName ~~ @dstAddresses ) {
				push @vipNatArray, {%$filterEntry};    # push hard copy of hash
				next FILTER;
			}
		}
	}
	return \@vipNatArray;
}

sub create_vip_nat_filter {
	my $in = shift;
	my $vipFilter = shift;
	my $address = shift;
	my $count = shift;
	my $returnReverse = shift;
	
	my $vipFilterCopy = {%$vipFilter};
	my $vipMapping = $in->{vipMappings}->{$address};
	$vipFilterCopy->{name} .= ".$count" if defined $vipFilterCopy->{name};
	$vipFilterCopy->{dstAddresses} = [];
	push @{$vipFilterCopy->{dstAddresses}}, $address;

	# get correct vip ext ips
	$vipFilterCopy->{services} = [];    # remove services as already in associated packet-filter
	$vipFilterCopy->{dstAddrTranslation}->{translationAddress} = { translationIP => $vipMapping->{mappedIp} };
	$vipFilterCopy->{dstAddrTranslation}->{translateAppliedAddressing} = "true";
	
	my $protocol = "tcp"; # default
	if (defined $vipMapping->{protocol}){
		$protocol = $vipMapping->{protocol};
	}
	
	# add source IP filter if defined
	$vipFilterCopy->{vipSrcFilter} = $vipMapping->{srcFilter} if defined $vipMapping->{srcFilter};
	# add source intf filter if defined
	$vipFilterCopy->{srcIntfFilter} = $vipMapping->{srcIntfFilter} if defined $vipMapping->{srcIntfFilter};
	
	if ( defined $vipMapping->{mappedPort}) {
		$vipFilterCopy->{dstAddrTranslation}->{translationAddress}->{translationServicePort} = {
			translationPort => $vipMapping->{mappedPort}, translationProtocol => $protocol };
		push @{$vipFilterCopy->{dstServicesVip}}, $vipMapping->{extPort};
	}elsif (defined $vipMapping->{protocol} && $vipMapping->{protocol} eq "icmp"){
		$vipFilterCopy->{dstAddrTranslation}->{translationAddress}->{translationServicePort} = {
			translationPort => {portRange => {portStart => 0, portEnd => 65535}}, translationProtocol => 'icmp' };
	}
	return $vipFilterCopy;
}

sub create_vip_nat_filter_reverse {
	my $in = shift;
	my $vipFilter = shift;
	my $address = shift;
	my $count = shift;
	
	my $vipMapping = $in->{vipMappings}->{$address};
	
	my $reverseFilter = get_reverse_nat_filter($vipFilter, 0, $vipMapping);
	unless (defined $reverseFilter){
		return undef;
	}
	if ( defined $vipMapping->{mappedPort}  ) {
		$reverseFilter->{srcAddrTranslation}->{translationAddress}->{translationServicePort} = {translationPort => $vipMapping->{extPort},
																	, translationProtocol => 'tcp' };
	}
	# add dst IP filter if defined
	$reverseFilter->{vipDstFilter} = $vipMapping->{srcFilter} if defined $vipMapping->{srcFilter};
	
	# add dst intf filter if defined
	$reverseFilter->{dstIntfFilter} = $vipMapping->{srcIntfFilter} if defined $vipMapping->{srcIntfFilter};
	
	push @{ $reverseFilter->{srcAddresses} }, $vipMapping->{mappedIp};
	push @{$reverseFilter->{srcServicesVip}}, $vipMapping->{mappedPort} if defined $vipFilter->{dstServicesVip};
	return $reverseFilter;
}

sub _generate_packet_processing_declaration
{
	my $out = shift;
	
	$out->open_element("core:packetProcessingOrdering");
	
	foreach my $pRPhase ('packet-filtering/source','routing','packet-filtering/destination','address-translation' )
	{
		my ($pPhase, $pPosition) = split ('/', $pRPhase);
		$out->open_element( 'core:packetProcessingElement');
		$out->print_element( 'core:packetProcessingPhase', $pPhase );
		if (defined $pPosition)
		{
			$out->print_element( 'core:processingPoint', $pPosition);
		}		
		$out->close_element( 'core:packetProcessingElement');
	}
	$out->close_element("core:packetProcessingOrdering");

	return;
}

1;

sub process_filter_negation {
	my $processOrder = shift;
	my $filterEntry  = shift;
	my $out          = shift;

	# create a copy of filterEntry and remove services
	my $negateFilter = { %{$filterEntry} };
	if ( $filterEntry->{primaryAction} eq "permit" ) {
		$negateFilter->{primaryAction} = "deny";
	}
	else {
		$negateFilter->{primaryAction} = "permit";
	}
	print_filter_entry( $out, $negateFilter, $processOrder++ );
	my $allowFilter = { %{$filterEntry} };
	if (defined $filterEntry->{'service-negate'}){
		$allowFilter->{services} = [];
	}
	if (defined $filterEntry->{'srcaddr-negate'}){
		$allowFilter->{srcAddresses} = [];
	}
	if (defined $filterEntry->{'dstaddr-negate'}){
		$allowFilter->{dstAddresses} = [];
	}
	print_filter_entry( $out, $allowFilter, $processOrder++ );
	return;
}
