#!/usr/bin/perl

package ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser;

use strict;
use warnings;
use v5.10;

use ZipTie::Logger;
use ZipTie::Model::AddressTranslation;
use ZipTie::Adapters::Utils qw(get_any_address_address get_crep get_port_number mask_to_bits ncname_from_string);

use Exporter qw(import);
our @EXPORT_OK = qw(formatArgs);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Common ip/mask regular expression
our $CIPM_RE = '\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}';

# Classless InterDomain Routing address regular expression
our $CIDR_RE = '\d{1,3}(?:\.\d{1,3}){0,3}\/\d+';

# MAC addresses regular expressions
our $MAC_RE1 = '[0-9a-f]{12}';
our $MAC_RE2 = '[0-9a-f]{1,2}(?:[:\.][0-9a-f]{1,2}){5}';
our $MAC_RE3 = '[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}';

sub new {
	
	my ( $proto, $in ) = @_;
    my $package = ref($proto) || $proto;
	my $self = {};
	
	$self = {
		in => $in,
		logger => ZipTie::Logger::get_logger(),
		zoneMapperHash => _setZoneMapperHash($in),
	};

	 bless($self, $package);
}


# This function returns a processingPhase element
sub getPacketProcessingPhase {
	
	my ($self) = @_;

	if (!$self->{in}->{NAT}){
		return undef;
	}
	
	my $packetProcessingPhase = '';
	
	open (my $packetProcessingPhaseFh, '>', \$packetProcessingPhase);
	
	if (defined $self->{in}->{NAT}->{settings}->{processingPhase}){
		my $packetProcessingPhasePrinter = ZipTie::Model::XmlPrint->new(\*$packetProcessingPhaseFh, 'core');
		$packetProcessingPhasePrinter->open_element('processingPhase');
		$packetProcessingPhasePrinter->print_element('core:packetProcessingPhase', $self->{in}->{NAT}->{settings}->{processingPhase});
		$packetProcessingPhasePrinter->close_element('processingPhase');	
	}
	
	close($packetProcessingPhaseFh);

	return ($packetProcessingPhase);
	
}

# This function returns a phaseDiscriminator element
sub getPhaseDiscriminator {
	
	my ($self) = @_;

	if (!$self->{in}->{NAT}){
		return undef;
	}
	
	my $phaseDiscriminator = '';
	
	open (my $phaseDiscriminatorFh, '>', \$phaseDiscriminator);
	
	if (defined $self->{in}->{NAT}->{settings}->{processingPhasePosition}){
		my $phaseDiscriminatorPrinter = ZipTie::Model::XmlPrint->new(\*$phaseDiscriminatorFh, 'core');
		$phaseDiscriminatorPrinter->open_element('phaseDiscriminator');
		
		my $packetProcessingPhase =  $self->getPacketProcessingPhase();
		$phaseDiscriminatorPrinter->print_string($packetProcessingPhase);
		$phaseDiscriminatorPrinter->print_element('processingPhasePosition', $self->{in}->{NAT}->{settings}->{processingPhasePosition});
		$phaseDiscriminatorPrinter->close_element('phaseDiscriminator');	
	}
	
	close($phaseDiscriminatorFh);

	return ($phaseDiscriminator);
	
}

# This function generates routing NAT
sub routingNAT {
	
	my ($self, $typeNAT, $section) = @_;
	
	if (not defined ($self->{in}->{NAT})){
		return undef;
	}
	my $RoutingNATFilterList = '';

	# open a printer to hold pre-routing NAT FilterList 
	open (my $RoutingNATFilterListFh, '>', \$RoutingNATFilterList);
	my $RoutingNATFilterListPrinter = ZipTie::Model::XmlPrint->new(\*$RoutingNATFilterListFh, 'core'); 
		
	if ($typeNAT eq 'interfaceNAT') {
		$self->_generateInterfaceNATXML($RoutingNATFilterListPrinter);
	}
	if ($typeNAT eq 'MIPNAT') {
		$self->_generateMIPNATXML($RoutingNATFilterListPrinter);
	}
	if ( (defined $section) and ($typeNAT eq 'policyNAT') ){
		$self->_generatePolicyNATXML($RoutingNATFilterListPrinter, $section);
	}
	if ($typeNAT eq 'VIPNAT'){
		$self->_generateVIPNATXML($RoutingNATFilterListPrinter);
	}
		
	close($RoutingNATFilterListFh);
	return $RoutingNATFilterList;

}

# This function generates XML for interface NAT
sub _generateInterfaceNATXML{
	
	my ($self, $routingNATFilterListPrinter) = @_;
	
	#interface NAT translation occurs after a route and policy lookup.
	$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
	$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'after';
	
	$self->_getInterfaceNATRules();
	
	# loop through rules
	foreach my $rule (@{$self->{in}->{NAT}->{interfaceNAT}->{rules}}) {
		my $srcInterface = $rule->{natIntName};
		my $dstInterface = $rule->{routeIntName};
		my @natSides = ($srcInterface, $dstInterface);
		my $addressTranslationObject = ZipTie::Model::AddressTranslation->new(@natSides);
		# create fake zone mappings for the src/dst interfaces
		$self->{zoneMapperHash} = {$srcInterface => [$srcInterface],
		$dstInterface => [$dstInterface]};
		$addressTranslationObject->setInterfaceZoneMapper(_getInterfaceZoneMapper($self));
		# Interface NAT always applies PAT
		$addressTranslationObject->addPortTranslationRange(1024, 65535);
		$self->_createNATRule ($addressTranslationObject, $rule);		
		# return unless there are available translated rules
		next unless (scalar(@{$addressTranslationObject->{_atTranslateRules}}));
		my @directions = ("source");
		# only create reverse list if srcAddr not 'any' (otherwise would create empty list)
		if ($rule->{realSrcAddr} !~ /0.0.0.0/){
			push @directions, "destination";
		}
		foreach my $direction (@directions){
			# generate translation list
			$rule = _generateTranslationList($addressTranslationObject, $direction);
			$routingNATFilterListPrinter->print_string($rule);	
		}
	}
}

# This function retrieves interface NAT rules  
sub _getInterfaceNATRules{
	
	my ($self) = @_;
	
	my @natInterface = @{$self->{in}->{NAT}->{interfaceNAT}->{interfaces}};
	foreach my $interface (@natInterface){
		if (($interface->{intMode} eq 'nat') and ($interface->{zone} !~ /untrust/i )){
			my $rule = {};
			$rule->{realSrcAddr} = $interface->{ipAddress};
			$rule->{isStaticNATMapping} = 1;
			$rule->{PAT}->{enabled} = 1;
			$rule->{service} = ("any");
			$rule->{protocol} = $self->_getServicesProtocol("any");
			$rule->{natIntName} = $interface->{intName};
			my @natInterface2 = @{$self->{in}->{NAT}->{interfaceNAT}->{interfaces}};
			foreach my $interface2 (@natInterface2){
				if (($interface2->{intMode} eq 'route') and ($interface2->{zone} =~ /untrust/i )){
					my %newRule = %$rule;
					my $newRuleRef = \%newRule;
					$newRuleRef->{mapSrcAddr} = $interface2->{ipAddress};
					$newRuleRef->{routeIntName} = $interface2->{intName};
					push (@{$self->{in}->{NAT}->{interfaceNAT}->{rules}}, $newRuleRef);
				}
			}
		}
	}
}

# This function collects policy NAT configuration and policies
# and generates XML for policy based NAT
sub _generatePolicyNATXML{
	
	my ($self, $routingNATFilterListPrinter, $section) = @_;
	my $printedTranslationList;
	my @NATRules;
		
	# retrieve interface settings and policies
	$self->_getPolicyNATFWPolicies();
	
	# set NAT rules
	$self->_setPolicyNATRules();
	
	if (defined ($self->{in}->{NAT}->{policyNAT}->{rules})){
		@NATRules = @{$self->{in}->{NAT}->{policyNAT}->{rules}};
	} else {
		return;
	}
	my %uniqueNatSides = _getUniqueNATSides (\@NATRules);
	foreach my $key ( keys %uniqueNatSides ){
    	my @natSides = @{$uniqueNatSides{$key}};
    	my $listZoneToZone = $self->_printTranslationList(\@natSides,\@NATRules, $routingNATFilterListPrinter, $section);
    	if (defined($listZoneToZone)){
    		$printedTranslationList = 1;
    	}
    }
	# set phaseDiscriminator if there is an available list
	if ($printedTranslationList) {
		$self->_setPhaseDiscriminator($routingNATFilterListPrinter, $section);	
	}
}

# This function joins a set of arguments into a scalar separated by :
# and joins back arguements separated by space
# that eventually were splitted during parsing  
sub formatArgs{
	my (@args) = @_;
	
	my $first;
	my $buf;
	
	foreach my $arg (@args) {
		# got an arguement between ""
		if ($arg =~ /^\"(\S+)\"$/){
			if (not (defined $buf)){
				$buf = $arg;	
			} else {
				$buf = join(':',$buf,$arg);
			}
			#go to the next arguement	
			next;
		}
		# keep the first arguement starting with "
		if ($arg =~ /^\"/){
			$first = $arg;
		}
		# join the first argument with a starting "
		if(defined $first) {
			if ($arg !~ /\"$/) {
				$first = join (' ', $first);	
			}	
		}
		# join the last arguement before closing "
		if (defined $first) {
			if ($arg =~ /\"$/) {
				$arg = join(' ', $first, $arg);
				$first = undef;
			}	
		}
		if ((not defined $first) and (defined $buf)){
			$buf = join(':',$buf,$arg);	
		} 
		if ((not defined $first) and (not defined $buf)){
			$buf = $arg;
		}		 
	} #end foreach
	
	return $buf;
}

# This function retrieves policies related to policy based NAT-Src and NAT-Dst
sub _getPolicyNATFWPolicies {
	
	my ($self) = @_;
	
	open FH,  '<', \$self->{in}->{config} or $LOGGER->fatal_error_code("ERROR couldn't open configuration policies") ;
	while (<FH>){
		# look for the policy statements
		my $set_policy;
		my $policyNATType;
		
		# store policy NAT type src or dst
		if ($_ =~ /nat\s+src/i){
			$policyNATType = 'src';
			# i.e set policy from trust to untrust group1 any http nat src dip-id 10
			if ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+nat\s+src\s+dip-id\s+(\d+)/i) {
				$set_policy = formatArgs($policyNATType, $2, $3, $4, $5, $6, $7);
			# set policy id 26 from "Trust" to "Untrust"  "10.193.4.50/32" "MIP(10.50.244.9)" "ICMP-ANY" nat src permit log
			}elsif ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+\"MIP\((\S+)\)\"\s+(\S+)\s+nat\s+src\s+/i) {
				$set_policy = formatArgs($policyNATType, $2, $3, $4, $5, $6);
			# set policy from Trust to Untrust 10.1.1.15 2.2.2.15 ssh nat src
			}elsif ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+nat\s+src\s+/i) {
				$set_policy = formatArgs($policyNATType, $2, $3, $4, $5, $6);
			}
			push (@{$self->{in}->{NAT}->{policyNAT}->{set_policy}}, $set_policy) if defined $set_policy;
		}
		if ($_ =~ /nat\s+(src.*)?dst/i){
			$policyNATType = 'dst';
			# in case of policy NAT destination
			# routing happens after policy NAT-Dst
			$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
			$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'before';
			
			# Case 1: Destination NAT to another IP with PAT to a different port
			# set policy from zone1 to zone1 srcAddr destAddr service nat [src [dip-id id]] dst ip ipAddr port portNo.
			if ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+nat\s+(?:src.*)?dst\s+ip\s+(\S+)\s+port\s+(\d+)/){
				$set_policy = formatArgs($policyNATType, $2, $3, $4, $5, $6, $7, $8);
			# Case 2: Destination NAT to another IP
			# set policy from zone1 to zone2 srcAddr destAddr service nat [src [dip-id id]] dst ip ipAddr
			}elsif ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+nat\s+(?:src.*)?dst\s+ip\s+(\S+) permit/){
				$set_policy = formatArgs($policyNATType, $2, $3, $4, $5, $6, $7);
			# Case 3: Destination NAT to an IP range
			# set policy from zone1 to zone2 srcAddr destAddr service nat [src [dip-id id]] dst ip ipAddr1 ipAddr2
			}elsif ($_ =~ /set\s+policy\s+(.*?)from\s+(\S+)\s+to\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+nat\s+(?:src.*)?dst\s+ip\s+(\S+)\s+(\S+)/){
				$set_policy = formatArgs($policyNATType."-case3", $2, $3, $4, $5, $6, $7, $8);
			}
			push (@{$self->{in}->{NAT}->{policyNAT}->{set_policy}}, $set_policy) if defined $set_policy;
		}
	} #end while
	close FH;
}

# this function sets the PhaseDiscriminator element
sub _setPhaseDiscriminator {
	
	my ($self,$routingNATFilterListPrinter, $section) = @_;
	
	if (not defined $routingNATFilterListPrinter){
		return undef;
	}
	
	my $phase;
	if ($section eq 'preRouting') {
		$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
		$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'before';
		
		$phase = $self->getPhaseDiscriminator(); 
		$routingNATFilterListPrinter->print_string($phase);
	}
	if ($section eq 'postRouting') {
		$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
		$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'after';
		
		$phase = $self->getPhaseDiscriminator(); 
		$routingNATFilterListPrinter->print_string($phase);
	}
}

# This functions sets policy based NAT-Src and NAT-Dst rules 
sub _setPolicyNATRules {
	
	my ($self) = @_;

	foreach my $policy (@{$self->{in}->{NAT}->{policyNAT}->{set_policy}}) {
		my $rule = {};
		my $service;
		my $policy_dip_ip;
		my $policyNATType;
		
		#clean policy from ""
		$policy =~ s/\"//g;
		if ($policy =~ /dst-case3.*/){
			($policyNATType, $rule->{zoneSrc}, $rule->{zoneDst}, $rule->{realSrcAddr}, $rule->{realDstAddr}, 
			$service, $rule->{mapDstAddr}->{startAddress}, $rule->{mapDstAddr}->{endAddress}) = split(':',$policy);
		} else {
			# check if the policy arguments are built based on the address configurations or groups settings
			($policyNATType, $rule->{zoneSrc}, $rule->{zoneDst}, $rule->{realSrcAddr}, $rule->{realDstAddr}, 
			$service, $policy_dip_ip, $rule->{destPortMapped}) = split(':',$policy);
		}
		# will store services with lower case for uniformity
		$rule->{service} = lc($service);
		$rule->{protocol} = $self->_getServicesProtocol(lc($service));
		
		# check for DIP id
		if ((defined $policy_dip_ip) and ($policyNATType =~ /src/) ) {
			$rule->{isStaticNATMapping} = 0;
		} else {
			# by default it is considered static
			$rule->{isStaticNATMapping} = 1;
		}
		
		if ($policyNATType eq 'src') {
			# store policy NAT type
			$rule->{section} = 'postRouting';
			# by default PAT is enabled
			if (not defined($rule->{PAT}->{enabled}) ) {
				$rule->{PAT}->{enabled} = 1;	
			}
		
			if (not defined($policy_dip_ip) ){
				$rule->{mapSrcAddr} = $self->_getIPofEgressIntf($rule->{zoneDst});
			}
							
			# lookup DIP id ($dip_id) between set interface configurations
			if( ( defined ($self->{in}->{NAT}->{policyNAT}->{set_interface}) ) and (scalar @{$self->{in}->{NAT}->{policyNAT}->{set_interface}}) ) {
				my $lastDipMatched;
				my $matchCount;
				foreach my $set_interface(@{$self->{in}->{NAT}->{policyNAT}->{set_interface}}) {
					if ($set_interface =~ /shift-from/i) {
						my ($IfZone, $interface_dip_id, $shift_from, $IPFrom, $IPStart, $IPEnd) = split (':',$set_interface);
						if ( (defined $interface_dip_id) and (defined $policy_dip_ip) and ($interface_dip_id eq $policy_dip_ip) ){	
							$rule = setPolicySourceAddress($rule, $IPStart, $IPEnd);
							# PAT is disabled for DIP shift 
							$rule->{PAT}->{enabled} = 0;
						}	
					}else {
						my ($IfZone, $interface_dip_id, $IPStart, $IPEnd, $fix_port) = split (':',$set_interface);
						if ( (defined $interface_dip_id) and (defined $policy_dip_ip) and ($interface_dip_id eq $policy_dip_ip) ){
							if ((defined $fix_port) and ($fix_port =~ /fix-port/i) ) {
								# PAT is disabled 
								$rule->{PAT}->{enabled} = 0;
							}
							# handle the case where we can have up to 3 ranges defined in a DIP
							if (defined $lastDipMatched and $lastDipMatched == $interface_dip_id){
								$matchCount++;
								if ($matchCount == 1){
									$rule = setPolicySourceAddress($rule, $IPStart, $IPEnd, "mapSrcAddr2");
								} elsif ($matchCount == 2){
									$rule = setPolicySourceAddress($rule, $IPStart, $IPEnd, "mapSrcAddr3");
								}
							} else {
								$rule = setPolicySourceAddress($rule, $IPStart, $IPEnd);
							}
							$lastDipMatched = $interface_dip_id;
						}
					} #else for fix-port
				}# foreach set_interface				
			}#end if set_interface	
		} # end if policyNATType = 'src'
		
		if ($policyNATType =~ /dst.*/) {
			# store policy NAT type
			$rule->{section} = 'preRouting';
			# in case of NAT-Dst the $policy_dip_ip argument is an IP
			# and need to check if the real source or real destination is a subnet 
			# in which case the mapped destination subnet will need to match the mask
			my $netobj;
			if (not defined ($netobj) and ($policyNATType ne 'dst-case3')){
				$rule->{mapDstAddr} = $policy_dip_ip;
				foreach my $refRealRule (\$rule->{realSrcAddr},\$rule->{realDstAddr} ) {
					$netobj = _checkIP($refRealRule); 
					if ( (defined $netobj) and (exists $netobj->{prefix}) and (defined($policy_dip_ip)) ) {	
						$rule->{mapDstAddr} = $policy_dip_ip."/".$netobj->{prefix};	
					}
				} #end foreach
			} #if netobj
		} # end if policyNATType = 'dst'
				
		if (scalar keys %$rule >3) {
			push (@{$self->{in}->{NAT}->{policyNAT}->{rules}}, $rule) if defined $rule;	
		}
		 		
	} # foreach policy	
	
}

# This function collects MIPNAT configuration and policies
# and generates XML for Mapped IP address (MIP) NAT
sub _generateMIPNATXML{
	
	my ($self, $routingNATFilterListPrinter) = @_;
	my @NATRules;
	
	#MIP NAT translation occurs before a route and policy lookup.
	$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
	$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'before';
		
	# retrieve interface settings and policies
	$self->_getMIPNATConfiguration();
	$self->_getMIPNATFWPolicies();
	
	if (defined ($self->{in}->{NAT}->{MIPNAT}->{rules})){
		@NATRules = @{$self->{in}->{NAT}->{MIPNAT}->{rules}};
	} else {
		return;
	}
	my %uniqueNatSides = _getUniqueNATSides (\@NATRules);
	foreach my $key ( keys %uniqueNatSides ){
    	my @natSides = @{$uniqueNatSides{$key}};
    	my $listZoneToZone = $self->_printTranslationList(\@natSides,\@NATRules, $routingNATFilterListPrinter, 'preRouting');
    }
}

# This function creates NAT rule per each transitioning zone and generates translation list
sub _printTranslationList {
	
	my ($self, $refNATSides, $refNATRules, $routingNATFilterListPrinter, $section, $sourceDirectionOnly) = @_;
	
	my @NATSides = @$refNATSides;
	my @NATRules = @$refNATRules;
	
	# generate addressTranslationObject for both NAT direction
	my $addressTranslationObject = $self->_createAddressTranslationObject(@NATSides);
	my @PortRange;
				
	# loop through rules
	foreach my $rule (@NATRules) {	
		if ( ($rule->{zoneSrc} =~ /$NATSides[0]/i) and ($rule->{zoneDst} =~ /$NATSides[1]/i) ){
			#if there is no PortRange yet defined will define it only once and add port translation range for this  
			if (not scalar(@PortRange)){
				if ( (defined $rule->{PAT}->{enabled}) and ($rule->{PAT}->{enabled} == 1) ){
					if (defined $rule->{PAT}->{PortRange}) {
						@PortRange = @{$rule->{PAT}->{PortRange}};
					} else {
						@PortRange = (1024, 65535);
					}
					$addressTranslationObject->addPortTranslationRange(@PortRange);
				} # if PAT enabled
			} # if not defined @PortRange
			if ( ( defined($rule->{section}) and ($rule->{section} eq $section)) or (not defined($rule->{section})) ){
				$self->_createNATRule ($addressTranslationObject, $rule);
			}
				
		} #end if zone		
	} #end foreach
	
	# return unless there are available translate rules
	return unless (scalar(@{$addressTranslationObject->{_atTranslateRules}}));
	foreach my $direction ("source", "destination"){
		# generate translation list
		my $rule = _generateTranslationList($addressTranslationObject, $direction);
		$routingNATFilterListPrinter->print_string($rule);
		if ($sourceDirectionOnly){
			return;
		}	
	}
}	

sub _createAddressTranslationObject{ 
	
	my ($self,@natSides) = @_;
	
	# create a new AddressTranslation object
	my $addressTranslationObject = ZipTie::Model::AddressTranslation->new(@natSides);
	
	# set the interfaceZoneMapper based on the on the gateway's interfaces characteristic
	$addressTranslationObject->setInterfaceZoneMapper(_getInterfaceZoneMapper($self));
	
	return $addressTranslationObject;
}

# This function collects interface settings
sub _getMIPNATConfiguration {

	my ($self) = @_;
	
	#collect MIP NAT configuration info
	#i.e set interface ethernet2 mip 1.1.1.5 host 10.1.1.5 netmask 255.255.255.255 vrouter trust-vr 
	while ($self->{in}->{config} =~  /set interface\s+\"?(.*?)\"?\s+mip\s+(\S+)\s+host\s+(\S+)\s+netmask\s+([\d\.]+)/mig){
		my $ifname = lc ($1);
		my $mip = $2;
		my $host = $3;
		my $netmask = $4;
		my $prefix;
		
		#clean MIP from possible ""
		$mip =~ s/\"//g;
		
		if (defined ($netmask) ){
			$prefix = mask_to_bits($netmask);	
		}
		
		# find zone interface belongs to
		my $zone = ZipTie::Adapters::Juniper::ScreenOS::Parsers::getIntfinZone($self->{in}, $ifname);
		$zone =~ s/\"//g if defined $zone;
		
		# store per MIP host and prefix	
		if ( defined($mip) && defined $zone){
			$self->{in}->{NAT}->{'if_alias_'.$zone}->{$mip}->{host} = $host if defined $host;
			$self->{in}->{NAT}->{'if_alias_'.$zone}->{$mip}->{prefix} = $prefix if defined $prefix;	
		} 	
		 
	}#end while 

}

# This function collects MIPNAT firewall's policies
sub _getMIPNATFWPolicies {
	
	my ($self) = @_;
	
	#collect MIP NAT policies infos
	
	#parse configuration 
	open FH,  '<', \$self->{in}->{config} or $LOGGER->fatal_error_code("ERROR couldn't open configuration policies") ;
	while (<FH>){
		# case 1 set policy from zone1 to zone2 any mip(mip_ipAddr) http permit
		if ($_ =~  /set policy\s+(.*?)from\s+\"?(.*?)\"?\s+to\s+\"?(.*?)\"?\s+\"?(.*?)\"?\s+\"?mip\s*?\(?\s*?($CIPM_RE|$CIDR_RE)\s*?\)?\"?\s+(\S+)/i){
			my $policyid = $1;
			my $zoneSrc = $2;
			my $zoneDst = $3;
			my $from = $4;
			#MIP IP
			my $to = $5;
			my $service = $6;
			
			foreach my $mip ( keys %{$self->{in}->{NAT}->{'if_alias_'.$zoneSrc}} ) {	
				# loop through available MIP
				my $rule = {};
				
				$rule->{policyid} = $policyid;
				$rule->{zoneSrc} = $zoneSrc;
				$rule->{zoneDst} = $zoneDst;	
				$rule->{protocol} = $self->_getServicesProtocol(lc($service));
				#MIP NAT is a static NAT
				$rule->{isStaticNATMapping} = 1;
				
				# set rule only if the destination address = MIP address
				if ($to =~ /$mip/i) {	
					$rule->{realSrcAddr} = $from;
					$rule->{realDstAddr} = $to;
					if (defined ($self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{prefix}) and $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{prefix} eq '32') {
						$rule->{mapDstAddr} = defined $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{host} ? $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{host} : undef ;
					}
					if (defined ($self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{prefix}) and $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{prefix} ne '32') {
						my $address = defined $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{host} ? $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{host} : undef ;
						my $prefix = $self->{in}->{NAT}->{'if_alias_'.$zoneSrc}->{$mip}->{prefix};
						$rule->{mapDstAddr} = $address.'/'.$prefix;
					}	
				}
				#store rule only if we have enough info
				if (scalar keys %$rule >5)	{
					push (@{$self->{in}->{NAT}->{MIPNAT}->{rules}}, $rule);	
				}
			} #end foreach value of MIP
				
		} #end if case 1
		
		# case 2 set policy id 19 from "Trust" to "Untrust" any any any permit log 
		# but not containing 'nat src' which would be interface or policy nat
		# or 'nat dst' which would be policy destination nat
		# and still having MIP specified through set interface
		# i.e set interface "untrust" mip 2.2.2.10 host 10.1.1.10 netmask 255.255.255.255 vr "trust-vr"
		elsif (($_ =~  /set policy\s+(.*?)from\s+\"?(.*?)\"?\s+to\s+\"?(.*?)\"?\s+\"?(.*?)\"?\s+\"?(.*?)\"?\s+\"?(.*?)\"?\s+/i) 
			and ($_ !~ /(.*?)nat src/) 
			and ($_ !~ /(.*?)nat dst/) 
			and (defined($self->{in}->{NAT}->{MIPNAT}->{enabled})) and ($self->{in}->{NAT}->{MIPNAT}->{enabled}==1) ){
			my $from,my $to,my $service;	
			# make a second parsing for the case when set policy's argument might contain ""
		 	if ($_	 =~ /set policy\s+(.*?)from\s+\"?(.*?)\"?\s+to\s+\"?(.*?)\"?\s+\"(.*?)\"\s+\"(.*?)\"\s+\"(.*?)\"\s+/i) {
		 		$from = $4;
				$to = $5;
				$service = $6;
		 	}else {
		 		$from = $4;
				$to = $5;
				$service = $6;
		 	}
		 	my $policyid = $1;
		 	my $zoneSrc = $2;
		 	my $zoneDst = $3;
		 	
		 	# for each MIP that matches the dest. zone
		 	foreach my $mip (keys %{$self->{in}->{NAT}->{'if_alias_'.$zoneDst}}) {
		 		# loop through available MIP	
			 	my $rule = {};				
			 	
				$rule->{policyid} = $policyid;
				$rule->{zoneSrc} = $zoneSrc;
				$rule->{zoneDst} = $zoneDst;
				$rule->{protocol} = $self->_getServicesProtocol(lc($service));
				#MIP NAT is a static NAT
				$rule->{isStaticNATMapping} = 1;
				
				if (($zoneDst) =~ /untrust/i) {
					#make sure mip host falls within the range of policy source address
					my $mipHost = $self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{host};
					my $isMipHostInRange = _checkMipHostIsInRange($mipHost, $from);
					if (!$isMipHostInRange){
						next;
					}
					$rule->{realSrcAddr} = $mipHost;
					$rule->{realDstAddr} = $to;
					if (defined ($self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{prefix}) and $self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{prefix} eq '32') {	
						$rule->{mapSrcAddr} = defined ($mip) ? $mip:undef;
					}
					if (defined ($self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{prefix}) and $self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{prefix} ne '32') {
						# for this case will map a subnet to a subnet
						my $address = defined ($mip) ? $mip:undef;
						my $prefix = $self->{in}->{NAT}->{'if_alias_'.$zoneDst}->{$mip}->{prefix};
						$rule->{mapSrcAddr} = $address.'/'.$prefix;
					}
				} 
				#store rule only if we have enough info
				if (scalar keys %$rule >5)	{
					push (@{$self->{in}->{NAT}->{MIPNAT}->{rules}}, $rule);	
				}
			} #end foreach mip		
		} #end if case2
		
	} # end while 
	close FH or $LOGGER->fatal_error_code("ERROR couldn't close configuration policies") ;
	
}

# This function collects policy NAT configuration and policies
# and generates XML for policy based NAT
sub _generateVIPNATXML{
	
	my ($self, $routingNATFilterListPrinter) = @_;
	my @NATRules;
	
	#VIP NAT translation occurs before a route and policy lookup.
	$self->{in}->{NAT}->{settings}->{processingPhase} = 'routing';
	$self->{in}->{NAT}->{settings}->{processingPhasePosition} = 'before';
		
	# retrieve interface settings and policies
	# and set NAT rules
	$self->_setVIPNATRules();
	if (defined ($self->{in}->{NAT}->{VIPNAT}->{rules})){
		@NATRules = @{$self->{in}->{NAT}->{VIPNAT}->{rules}};
	} else {
		return;
	}
	my %uniqueNatSides = _getUniqueNATSides (\@NATRules);
	foreach my $key ( keys %uniqueNatSides ){
    	my @natSides = @{$uniqueNatSides{$key}};
    	$self->_printTranslationList(\@natSides,\@NATRules, $routingNATFilterListPrinter, 'preRouting', 1);
    }
}

# This function collects VIPNAT firewall's policies and sets SED's NAT rules
sub _setVIPNATRules {
	
	my ($self) = @_;
	
	#collect VIP NAT policies infos
	
	#parse configuration 
	open FH,  '<', \$self->{in}->{config} or $LOGGER->fatal_error_code("ERROR couldn't open configuration policies") ;
	while (<FH>){
		# case 1 
		# set policy from untrust to trust any vip(1.1.1.5) http permit
		# or set policy from "Untrust" to "Trust" "Any" "VIP(210.1.1.10)" "HTTP" permit
		if ($_ =~  /set policy\s+(.*?)from\s+\"?(.*?)\"?\s+to\s+\"?(.*?)\"\s+(\S+)\s+\"?vip\(?($CIPM_RE|$CIDR_RE)\)?\"?\s+(\S+)/i){
			
			my $rule = {};
			$rule->{policyid} = $1;
			$rule->{zoneSrc} = $2;
			$rule->{zoneDst} = $3;
			my $from = $4;
			my $to = $5;
			my $service = $6;
			
			#clean service from possible ""
			$service =~ s/\"//g;
			$rule->{protocol} = $self->_getServicesProtocol(lc($service));
			$rule->{realSrcAddr} = $from;
			$rule->{realDstAddr} = $to;
			#VIP is a static NAT 
			$rule->{isStaticNATMapping} = 1;
			
			# lookup the vip address from set interfaces
			
			foreach my $set_interface (@{$self->{in}->{NAT}->{VIPNAT}->{set_interface}}) {
				# i.e. 'untrust:210.1.1.10:+:80:"HTTP":192.168.1.10'
				my ($IfZone,$VIPAddr,$destServicePort,$destPortNameMapped,$mapDstAddr,$destServicePortEnd,$destPortMapped,$destPortEndMapped,$protocol);
				my @setInterfaceArray = split(':',$set_interface);
				my $lengthOfSetInterfaceArray = @setInterfaceArray;
				if ($lengthOfSetInterfaceArray == 5){
					($IfZone,$VIPAddr,$destServicePort,$destPortNameMapped,$mapDstAddr)= split (':',$set_interface);
				# or untrust:172.18.10.57:101:120:172.18.1.3:141:160
				} elsif ($lengthOfSetInterfaceArray == 7) {
					($IfZone,$VIPAddr,$destServicePort,$destServicePortEnd,$mapDstAddr,$destPortMapped,$destPortEndMapped)= split (':',$set_interface);
				# or untrust:172.18.10.57:101:120:172.18.1.3:141:160:UDP
				} elsif ($lengthOfSetInterfaceArray == 8){
					($IfZone,$VIPAddr,$destServicePort,$destServicePortEnd,$mapDstAddr,$destPortMapped,$destPortEndMapped,$protocol)= split (':',$set_interface);
				} else{
					next;
				}
				if  ( ( defined($VIPAddr) ) and ($rule->{realDstAddr} eq $VIPAddr) ){
				 	# check if the last arguement is available
					 if ( defined($mapDstAddr) ) {
					 	$rule->{mapDstAddr} = $mapDstAddr;
					 }
					 if ( defined($destServicePort) and defined($mapDstAddr) ) {
					 	$rule->{destServicePort} = $destServicePort;
					 }
					 if ( defined ($destPortNameMapped) and defined($mapDstAddr) ){
					 	$destPortNameMapped =~ s/\"//g;
					 	$rule->{destPortNameMapped} = $destPortNameMapped;
					 }	
					 if (defined($destServicePort and $destServicePortEnd and $destPortMapped and $destPortEndMapped)){
					 	$rule->{destServicePort} = $destServicePort;
					 	$rule->{destServicePortEnd} = $destServicePortEnd;
					 	$rule->{destPortMapped} = $destPortMapped;
					 	$rule->{destPortEndMapped} = $destPortEndMapped;
					 }
					 if (defined $protocol){
					 	$rule->{protocol} = $protocol;
					 }
				}	
			} # foreach
			#store rule only if we have enough info
			if (scalar keys %$rule >7)	{
				push (@{$self->{in}->{NAT}->{VIPNAT}->{rules}}, $rule);	
			}
		} #if for set policy
	} #end while
	close FH;		
			
}

# This function lookup a given service and returns the associated protocol. 
# By default it returns an 'ip' protocol if it couldn't find an associated service 
sub _getServicesProtocol{
	my ($self, $service) = @_;
	my $protocol = defined($self->{in}->{portServiceGroups}->{$service}->{'protocol'}) ? $self->{in}->{portServiceGroups}->{$service}->{'protocol'} : 'ip';
	return $protocol;
}

# This function returns the IP address of the egress interface attached to a destination zone
sub _getIPofEgressIntf{
	my ($self, $zone) = @_;
	my $IPofIntf = undef;
	my $thisconfig = my $config = $self->{in}->{config};
	
	while ($config =~  /set interface\s+(\"?\S+\"?)\s+zone\s+(\"?\S+\"?)/mig){
		my $alias = $1;
		my $intfZone = $2;
		my $ip;
		
		if ($intfZone =~ /$zone/i and defined ($alias)){
			#clean alias from possible ""
			$alias =~ s/\"//g;
			while ($thisconfig =~ /set interface\s+(\"?$alias\"?)\s+ip\s+(\S+)/mig) {
				$ip = $2;
				if ($ip =~ /($CIPM_RE) ($CIPM_RE)/){
					$IPofIntf = $1;
					my $mask = $2;
				}
				if ($ip =~ /$CIDR_RE/) {
					($IPofIntf) = split('/',$ip);
				}
			}#end while
		}#end if
	}#end while	

	return $IPofIntf;
	
}

# This function assembles the required NAT translation arguments and calls _createAddressTranslationRule 
# to add the rule and the address rule criterion to an address translation object
sub _createNATRule {
	#my ($self,$addressTranslationObject, $rule, $direction) = @_;
	my ($self,$addressTranslationObject, $rule) = @_;
	
	my $addRuleArguments = {};
	my $criterionArguments = {};
	
	# set to static or dynamic
	if ( (not defined ($addRuleArguments->{isStaticNATMapping}) ) and (defined $rule->{isStaticNATMapping}) ){
		$addRuleArguments->{isStaticNATMapping} = $rule->{isStaticNATMapping};		
	}	
	
	# set protocol
	if ( (not defined($addRuleArguments->{protocol}) ) and (defined $rule->{protocol}) ) {
		$addRuleArguments->{protocol} = $rule->{protocol};
	}
	
	# set PAT 
	if ( (not defined($addRuleArguments->{usePATOverload}) ) and (defined $rule->{PAT}->{enabled}) and ($rule->{PAT}->{enabled} == 1) ) {
		$addRuleArguments->{usePATOverload} = $rule->{PAT}->{enabled};
	}
	
	if ( defined($rule->{realSrcAddr})) {
		#i.e $criterionArguments->{sourceIp} = $rule->{realAddr};
		# must be an objectgroupreference with a special name
		if( (($rule->{realSrcAddr} !~ /($CIDR_RE|$CIPM_RE)/) and ($rule->{realSrcAddr} =~ /\w+/)) or ($rule->{realSrcAddr} =~ /(.*?)\s+/) or ( ($rule->{realSrcAddr} !~ /($CIDR_RE|$CIPM_RE)/) and ($rule->{realSrcAddr} =~ /\w\./i) ) ){
			my $ruleSrcAddr = $rule->{realSrcAddr};
			$rule->{realSrcAddr} = _setNameAddress($self, $rule->{zoneSrc}, $rule->{realSrcAddr});
			unless (defined $rule->{realSrcAddr}){
				$LOGGER->debug("WARNING: Suppressing NAT filter, unable to find address object $ruleSrcAddr");
				return;
			}
		}
		_setCriterion($criterionArguments,$rule->{realSrcAddr}, "source");		
	}# realSrcAddr
	
	if ( defined($rule->{realDstAddr})) {
		#i.e $criterionArguments->{destIp} = $rule->{realAddr};
		# must be an objectgroupreference with a special name
		if( (($rule->{realDstAddr} !~ /($CIDR_RE|$CIPM_RE)/) and ($rule->{realDstAddr} =~ /\w+/)) or ($rule->{realDstAddr} =~ /(.*?)\s+/) or ( ($rule->{realDstAddr} !~ /($CIDR_RE|$CIPM_RE)/) and ($rule->{realDstAddr} =~ /\w\./i) ) ){
			my $ruleDstAddr = $rule->{realDstAddr};
			$rule->{realDstAddr} = _setNameAddress($self, $rule->{zoneDst}, $rule->{realDstAddr});
			unless (defined $rule->{realDstAddr}){
				$LOGGER->debug("WARNING: Suppressing NAT filter, unable to find address object $ruleDstAddr");
				return;
			}
		}
		_setCriterion($criterionArguments,$rule->{realDstAddr}, "dest");		
	}# realDstAddr
	
	if (defined ($rule->{destServicePort})) {
		$criterionArguments->{destServicePort} = $rule->{destServicePort};
	}
	if (defined ($rule->{destServicePortEnd})) {
		$criterionArguments->{destServicePortEnd} = $rule->{destServicePortEnd};
	}
	if(defined ($rule->{destPortNameMapped})) {
		$addRuleArguments->{destPortNameMapped} = $rule->{destPortNameMapped};
	}
	if (defined ($rule->{destPortMapped})) {
		$addRuleArguments->{destPortMapped} = $rule->{destPortMapped};
	} 
	if (defined ($rule->{destPortEndMapped})) {
		$addRuleArguments->{destPortEndMapped} = $rule->{destPortEndMapped};
	} 
	if ( defined($rule->{mapSrcAddr}) ) {
		#i.e $addRuleArguments->{sourceAddrMapped} = $rule->{mapSrcAddr};
		_setRule($addRuleArguments,$rule->{mapSrcAddr}, "source");
	} #mapSrcAddr
	my $addRuleArguments2 = {};
	if ( defined($rule->{mapSrcAddr2}) ) {
		_setRule($addRuleArguments2,$rule->{mapSrcAddr2}, "source");
	}
	my $addRuleArguments3 = {};
	if ( defined($rule->{mapSrcAddr3}) ) {
		_setRule($addRuleArguments3,$rule->{mapSrcAddr3}, "source");
	}
	if ( defined($rule->{mapDstAddr}) ) {
		#i.e $addRuleArguments->{destAddrMapped} = $rule->{mapDstAddr};
		_setRule($addRuleArguments, $rule->{mapDstAddr}, "dest");		
	} #mapDstAddr	
	
	_createAddressTranslationRule ($addressTranslationObject, $addRuleArguments, $criterionArguments,$addRuleArguments2, $addRuleArguments3);
}

# This function changes the name for a source/destination address to a possible available objectGroup name
# i.e. "OneGroup" is transformed into Untrust.OneGroup or Trust.OneGroup based on the passed zone (Trust/Untrust)
 
sub _setNameAddress{
	
	my ($self, $zone, $Addr) = @_ ;
	
	if ( (not defined($zone)) or (not defined($Addr)) ) {
		return undef;
	}  
	
	my $restore = $Addr;
	my $newAddr = ncname_from_string($zone . "." . $Addr);
	my $tmp = _checkObjectGroup($self, $newAddr);
	
	if ( defined($tmp)) {
		if ($tmp =~ /\bany\b/i){
			$newAddr = $restore;
		} else {
			$newAddr = $tmp;
		}
	}elsif ($restore =~ /\bany\b/i){
		$newAddr = $restore;
	} else {
		$newAddr = undef;
	}
	return $newAddr;
} 


# This function checks the existence of a given objectGroup 
sub _checkObjectGroup {
	
	my ($self, $obj) = @_;
	
	if (defined($self->{in}->{objectGroupLookup}->{$obj}) ){
		return $obj;
	} else {return undef};
}

#This function assigns criterion values to the criterion arguments
sub _setCriterion{
	
	my ($criterionArguments, $Addr, $addrDirection) = @_;
	
	my $netobj = _checkIP(\$Addr);
	my $criterionArg = _setCriterionArg(\$Addr, $addrDirection);
	
	if( (defined $netobj) and (exists $netobj->{prefix}) ){
		$criterionArguments->{$criterionArg->{address}} = $netobj->{address};
		$criterionArguments->{$criterionArg->{prefix}} = $netobj->{prefix};	
	}elsif( (defined $netobj) and (exists $netobj->{startAddress}) ) {
		# for IP Range
		$criterionArguments->{$criterionArg->{startAddress}} = $netobj->{startAddress};
		$criterionArguments->{$criterionArg->{endAddress}} = $netobj->{endAddress};
	}else {
		# must be a single address or network object
		# clean ""
		$netobj->{address} =~ s/\"//g;
		if ($netobj->{address} =~/\bany\b/i){
			$criterionArguments->{$criterionArg->{address}} = lc($netobj->{address}) 
		}else {
			$criterionArguments->{$criterionArg->{address}} = $netobj->{address};	
		}		
	}	
	 
}

#This function assigns rule values to the rule arguments
sub _setRule {
	
	my ($addRuleArguments, $Addr, $addrDirection) = @_;
	
	my $netobj = _checkIP(\$Addr);
	my $ruleArg = _setRuleArg($Addr, $addrDirection);
	
	#check the format of the netobj content
	if ( (defined $netobj) and (exists $netobj->{prefix}) ) {
		$addRuleArguments->{$ruleArg->{address}} = $netobj->{address};
		$addRuleArguments->{$ruleArg->{prefix}} = $netobj->{prefix};	
	}elsif( (defined $netobj) and (exists $netobj->{startAddress}) ) {
		# for IP Range
		$addRuleArguments->{$ruleArg->{startAddress}} = $netobj->{startAddress};
		$addRuleArguments->{$ruleArg->{endAddress}} = $netobj->{endAddress};
	}else {
		# must be a single address or network object
		# clean ""
			$netobj->{address} =~ s/\"//g;
		if ($netobj->{address} =~/\bany\b/i){
			 
			$addRuleArguments->{$ruleArg->{address}} = lc($netobj->{address});
		}else {
			$addRuleArguments->{$ruleArg->{address}} = $netobj->{address};	
		}	
	} 
}		

# this function returns the arguments name for Rules based on the passed Addr type
sub _setRuleArg {
	my ($Addr, $addrDirection) = @_;
	
	my $ruleArg = {};
	
	if (defined($Addr) and ($Addr =~ /$CIDR_RE/) ) {
		$ruleArg->{address} = $addrDirection.'AddrMapped';
		$ruleArg->{prefix} = $addrDirection.'AddrMaskMapped';
	}elsif ($Addr =~ /$CIPM_RE/){
		$ruleArg->{address} = $addrDirection.'AddrMapped';	
	}elsif($Addr =~ /\bany\b/i){
		$ruleArg->{address} = $addrDirection.'AddrMapped';
	}elsif( defined($Addr->{startAddress}) ){
		$ruleArg->{startAddress} = $addrDirection.'AddrMapped';
		$ruleArg->{endAddress} = $addrDirection.'AddrEndMapped';	
	}else {
		# must be an objectgroup reference
		$ruleArg->{address} = $addrDirection.'AddrNameMapped';		
	}
	
	return $ruleArg;	
	
}

# this function returns the arguments name for Criterion based on the passed Addr type
sub _setCriterionArg {
	my ($Addr, $addrDirection) = @_;
	
	my $criterionArg = {};
	
	if (not defined($Addr)){
		return undef;
	}
	
	if ( (ref($Addr) eq 'SCALAR') and ($$Addr =~ /$CIDR_RE/) ) {
		$criterionArg->{address} = $addrDirection.'Ip'; 
		$criterionArg->{prefix} = $addrDirection.'Prefix';
	}elsif ( (ref($Addr) eq 'SCALAR') and ( $$Addr =~ /$CIPM_RE/) ) {
		$criterionArg->{address} = $addrDirection.'Ip';	
	}elsif ( (ref($Addr) eq 'SCALAR') and ($$Addr =~ /\bany\b/i) ){
		$criterionArg->{address} = $addrDirection.'Ip';
	}elsif ( ref($Addr) eq 'REF') {
		if ( (ref($Addr) eq 'HASH') and ( exists($$Addr->{startAddress}) ) ){
			$criterionArg->{startAddress} = $addrDirection.'Ip';
			$criterionArg->{endAddress} = $addrDirection.'IpEnd';	
		}		
	}else {
		# must be an objectgroup reference
		$criterionArg->{address} = $addrDirection.'IpObjectGroup';
	}	
	
	
	return $criterionArg;
	
}

# This function checks the IP format (CIPM, CIDR) of a given IP address
# and returns a hash containing the address, prefix of the address
sub _checkIP {
	
	my ($obj) = @_;
	my $netobj= {};
	
	if (not defined($obj)) {
		return undef;
	}
	
	if ( (ref($obj) eq 'SCALAR') and ($$obj =~ /$CIDR_RE/) ){
		($netobj->{address},$netobj->{prefix}) = split('/',$$obj);	
	}elsif ( (ref($obj) eq 'SCALAR') and ($$obj =~ /$CIPM_RE/) ){
		$netobj->{address} = $$obj;
	}elsif ( ref($obj) eq 'REF'){  
		if ( (ref($$obj) eq 'HASH') and  ( exists($$obj->{startAddress}) ) ){
			$netobj->{startAddress} = $$obj->{startAddress};
			$netobj->{endAddress} = $$obj->{endAddress};
		}		
	}else {
		#it is just a network object element
		$netobj->{address} = $$obj;
	}
	
	
	return $netobj;
	
}

# This function creates a new address translation rule in $addressTranslation.  
# $addRuleArguments and $criterionArguments are both references to a hash of arguments to be passed into addRule and
# addRuleAddressCriterion methods.  
#
# For the passed in argument sets for addRuleAddressCriterion the rule argument
# does not need to be set.  This is set automatically after the call to addRule is completed.
sub _createAddressTranslationRule
{
	my ($addressTranslationObject, $addRuleArguments, $criterionArguments, $addRuleArguments2, $addRuleArguments3) = @_;

	my $rule = $addressTranslationObject->addRule($addRuleArguments);

	$criterionArguments->{rule} = $rule;
	$addressTranslationObject->addRuleAddressCriterion($criterionArguments);
	if (defined $addRuleArguments2){
		$addRuleArguments2->{rule} = $rule;
		$addressTranslationObject->addRuleAddressTranslationMapping($addRuleArguments2);
	}
	if (defined $addRuleArguments3){
		$addRuleArguments3->{rule} = $rule;
		$addressTranslationObject->addRuleAddressTranslationMapping($addRuleArguments3);
	}
}

# This function generate translation list for an address translation object
sub _generateTranslationList
{
    my ( $addressTranslationObject, $direction, $type ) = @_;
	
	my $nonDefaultListName = defined $type ? $type : undef;
	my $translationList = '';
	
	open (my $translationListFh, '>', \$translationList);
	
	my $translationListPrinter = ZipTie::Model::XmlPrint->new(\*$translationListFh, 'core'); 
	$addressTranslationObject->generateTranslationList($translationListPrinter, $direction, $nonDefaultListName, 1);

	close($translationListFh);

	return ($translationList);

}

# addressTranslation's setInterfaceZoneMapper takes in a function pointer to a function that returns a hash
# containing the zone mappings.  This function returns that function pointer.
sub _getInterfaceZoneMapper
{
	my ($self) = @_;
	
	return sub { return $self->{zoneMapperHash}; };
}

# This function returns a hash (zoneMapperHash) with keys containing the mapped interfaces 
sub _setZoneMapperHash
{
	my ($in) = @_;
	
	my $zoneMapperHash;
	my @interfaces;
	
	# start a scan of the interface sections, looking for the "nat" and "route" indicator
	while ( $in->{interfaces} =~ /^(\S+)\s+($CIPM_RE)\/(\d+)\s+(\S+)?\s+($MAC_RE3|N\/A)\s+(\S+)\s+(A|U|D|I|R)\s+\S+\s*$/mig ){
	    my ($ifname, $ipAddress, $mask, $zone, $mac, $status) = ($1, $2, $3, $4, $5, $6);
	    #for interface based nat the ingress interface is set to nat mode while the egress interface to route mode 
	    if (defined $in->{ 'if_' . $ifname } and $in->{ 'if_' . $ifname } =~ /mode\s+(route|nat)/){
			push (@{$zoneMapperHash->{"$zone"}}, $ifname);
			push(@interfaces, $ifname);
	    }	
		
	}
	
	@{$zoneMapperHash->{"any"}} = @interfaces;
	
	return $zoneMapperHash;
}

sub setPolicySourceAddress {
	my ($rule, $IPStart, $IPEnd, $mapSrcAddr) = @_;
	if (not defined $mapSrcAddr){
		$mapSrcAddr = "mapSrcAddr";
	}
	if ( ( defined $IPStart ) and ( defined $IPEnd ) ) {
		if ( $IPStart ne $IPEnd && $IPEnd ne "") {
			$rule->{$mapSrcAddr}->{startAddress} = $IPStart;
			$rule->{$mapSrcAddr}->{endAddress}   = $IPEnd;
		}
		else {
			$rule->{$mapSrcAddr}->{startAddress} = $IPStart;
		}
	}
	return $rule;
}

sub _getUniqueNATSides {
	my $NATRules = shift;
	my %uniqueNatSides;
	my @natZonePairs;
	foreach my $natRule (@$NATRules) {
		my @zonePair = ( $natRule->{zoneSrc}, $natRule->{zoneDst} );
		my $pairName = $natRule->{zoneSrc} . "||" . $natRule->{zoneDst};
		$uniqueNatSides{$pairName} = \@zonePair;
	}
	return %uniqueNatSides;
}

sub _checkMipHostIsInRange {
	my $mipHost = shift;
	my $from    = shift;
	my $subnet;
	if ($from =~ /any/i){
		return 1;
	}
	if ($from !~ /($CIDR_RE|$CIPM_RE)/) {
		return 0;
	}
	if ($from =~ /(\d{1,3}(?:\.\d{1,3}){0,3})\/(\d+)/){
		my $mask = $2;
		$subnet = ZipTie::Addressing::Subnet->new($from, $mask);
	} else{
		$subnet = ZipTie::Addressing::Subnet->new($from);
	}
	if ($subnet->contains($mipHost) ) {
		return 1;
	} else {
		return 0;
	}
}

=head1 NAME

	ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser
	Parse Juniper ScreenOS NAT configuration

=head1 DESCRIPTION

	This module implements the NAT parsers for Juniper Screen OS adapter.  
	It uses ZipTie::Model::AddressTranslation for the low level XML generation.

=head1 AUTHOR(S)

	Toni Barabas
	toni.barabas@ca.ibm.com

=head1 COPYRIGHT

	Copyright (c) 2013, IBM Corporation. All Rights Reserved.
	This module is free software. It may be used, redistributed
	and/or modified under the same terms as Perl itself.

=cut

1;
