package ZipTie::Adapters::Juniper::ScreenOS::Parsers;

use 5.010_000;
use strict;
use warnings;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser;
use ZipTie::Adapters::Utils qw(mask_to_bits seconds_since_epoch get_mask get_port_number get_port_expression trim get_interface_type strip_mac parseCIDR get_crep ncname_from_string get_protocol_number get_protocol_name);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_object_groups create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_services);

use constant MS_RPC_LISTENER_PROT => 135;
use constant SUN_RPC_LISTENER_PROT => 111;

my $CIPM_RE = get_crep('cipm');
my $CIDR_RE = get_crep('cidr');
my $MAC_RE3 = get_crep('mac3');

my $LOGGER = ZipTie::Logger::get_logger();

my $serviceLookup;
my $serviceGroupLookup;
my $serviceWarningLookup;

sub parse_chassis {
	my ( $in, $out ) = @_;

	my $chassisAsset = { 'core:assetType' => 'Chassis' };
	if ( $in->{system} =~ /^\s*Serial Number:\s+([^\,\s]+),/mi ) {
		$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		$chassisAsset->{'core:factoryinfo'}->{'core:make'} = 'Juniper';
	}
	if ( $in->{system} =~ /^\s*Hardware\s+Version:\s+([^\,\s]+)/mi ) {
		$chassisAsset->{'core:factoryinfo'}->{'core:partNumber'} = $1;
	}
	if ( $in->{system} =~ /^Product Name:\s*(\S+)/mi ) {
		$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
	} elsif ( $in->{system} =~ /File\s*Name:\s*\/?(\S+?)\./mi ) {
		$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
	}

	$out->open_element('chassis');
	$out->print_element( 'core:asset', $chassisAsset );

	_parse_file_storage( $in, $out );

	$out->close_element('chassis');
}

sub _parse_file_storage {
	my ( $in, $out ) = @_;

	if ( $in->{file_info} =~ /^There\s+are\s+(\d+)\s+bytes\s+free\s+\((\d+)\s+total\)\s+on\s+disk\s+\"(\S+)\"/mi ) {
		my $storage = {};
		$storage->{name}        = $3;
		$storage->{size}        = $2;
		$storage->{freeSpace}   = $1;
		$storage->{storageType} = 'flash' if ( $storage->{name} =~ /flash/i );
		$storage->{storageType} = 'disk' if ( $storage->{name} =~ /disk/i );
		$storage->{storageType} = 'NVRAM' if ( $storage->{name} =~ /nvram/i );
		$storage->{storageType} = 'other' if ( !defined $storage->{storageType} );
		$storage->{rootDir}     = { name => 'root' };
		while ( $in->{files} =~ /^\s*(\S+)\s+(\d+)\s*$/mig ) {
			push( @{ $storage->{rootDir}->{file} }, { name => $1, size => $2 } );
		}
		$out->print_element( 'deviceStorage', $storage );
	}
}

sub parse_system {
	my ( $in, $out ) = @_;

	my ($systemName) = $in->{config} =~ /^set\s+hostname\s+(\S+)/mi;
	if ( !defined $systemName ) {
		if ( $in->{snmp} =~ /^Sysname\s+:(\S+)/mi ) {
			$systemName = $1;
		}
	}
	$out->print_element( 'core:systemName', $systemName );

	$out->open_element('core:osInfo');
	if ( $in->{system} =~ /^File\s+Name:\s+([^\s\,]+)/mi ) {
		$out->print_element( 'core:fileName', $1 );
	}
	$out->print_element( 'core:make', 'Juniper' );
	if ( $in->{system} =~ /^Product\s+Name:\s+(\S+)/mi ) {
		$out->print_element( 'core:name', $1 );
	}
	if ( $in->{system} =~ /^Software\s+Version:\s+([^\s\,]+),\s+Type:\s+(\S+)/mi ) {
		$out->print_element( 'core:version', $1 );
	}
	$out->print_element( 'core:osType', 'ScreenOS' );

	# Generate the manifest of phases that the device uses sequentially on a given packet.
	# This gives the calling application a hint as to how to simulate device operation.
	_generate_packet_processing_declaration($out);

	$out->close_element('core:osInfo');

	if ( $in->{system} =~ /\bFirewall/mi ) {
		$out->print_element( 'core:deviceType', 'Firewall' );
	}

	my ($contact) = $in->{'config'} =~ /^set\s+snmp\s+contact\s+"([^\s\"]+)"/mi;
	$out->print_element( 'core:contact', $contact ) if defined($contact);

	# Up 8791 hours 15 minutes 9 seconds Since 30 Sept 2006 11:36:01
	if ( $in->{system} =~ /^Up (.+)$/mi ) {
		$_ = $1;
		my ($years)   = /(\d+)\s*years?/;
		my ($weeks)   = /(\d+)\s*weeks?/;
		my ($days)    = /(\d+)\s*days?/;
		my ($hours)   = /(\d+)\s*hours?/;
		my ($minutes) = /(\d+)\s*minutes?/;

		# subract the last reboot from the current time
		my $lastReboot = time();
		$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
		$lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
		$lastReboot -= $days * 24 * 60 * 60           if ($days);
		$lastReboot -= $hours * 60 * 60               if ($hours);
		$lastReboot -= $minutes * 60                  if ($minutes);
		$out->print_element( 'core:lastReboot', $lastReboot );
	}
}

sub _generate_packet_processing_declaration {
	my ($out) = @_;

	$out->open_element('core:packetProcessingOrdering');

	foreach my $pRPhase ('packet-filtering/source','routing','packet-filtering/destination','address-translation' ) {
		my ($pPhase, $pPosition) = split ('/', $pRPhase);
		$out->open_element('core:packetProcessingElement');
		$out->print_element( 'core:packetProcessingPhase', $pPhase );
		if (defined $pPosition) {
			$out->print_element( 'core:processingPoint', $pPosition );
		}
		$out->close_element('core:packetProcessingElement');
	}

	$out->close_element('core:packetProcessingOrdering');
}

sub create_config {
	my ( $in, $out ) = @_;

	# The name of the repository
	my $repository = { 'core:name' => '/' };

	if ( $in->{config} =~ /^get\s+config/i ) {
		$in->{config} =~ s/^get\s+config.*$//mi;
	}
	$in->{config}	=~ s/\s+[^\s>]+>\s*$//i;
	$in->{config}	=~ s/^Total\s+Config\s+size\s+\S+\s*$//mi;
	$in->{config}	= trim ( $in->{config} );

	my $startup = {
		'core:context' => 'boot',
		'core:mediaType' => 'text/plain',
		'core:name' => 'config',
		'core:textBlob' => encode_base64( $in->{config} ),
		'core:promotable' => 'true'
	};

	push( @{ $repository->{'core:config'} }, $startup );

	$out->print_element( 'core:configRepository', $repository );
}

sub parse_local_accounts {
	my ( $in, $out ) = @_;
	my $local_accounts;

	while ( $in->{users} =~ /^(\S+)\s+(\S+)\s+?$/mig ) {
		my $accountName = $1;
		my $account     = { accountName => $accountName };
		if ( ($accountName ne 'Name') && (substr($accountName, 0, 1) ne '-') ) {
			push( @{ $local_accounts->{localAccount} }, $account );
		}
	}

	if ( $local_accounts ) {
		$out->print_element( 'localAccounts', $local_accounts);
	}
	delete $in->{users};
}

sub parse_object_groups {
	my ($in, $out) = @_;

	$out->open_element('objectGroups');

	# We check we collect as many addresses as the device says it has, also how many we drop and log
	my $expected_address_count = 0;
	my $actual_address_count = 0;
	my $total_drop_count = 0;
	if ( $in->{addresses} =~ /Total\s+(\d+)\s+addresses/ ) {
		$expected_address_count = $1;
		$LOGGER->debug("Expecting to collect $expected_address_count address objects");
	}

	# We start with declared entries by zone
	foreach my $zone (split(/^addr zone/m, $in->{addresses})) {
		next unless ($zone =~ /^\s*name\s+(.+?)\s*$/m);
		my $name = $1;

		my $network_group = { id => "zone_$name" };
		my @collection_count = (0,0,0);
		my @drop_count = (0,0);

		foreach my $line (split(/^/m, $zone)) {
			my $addr_name;
			my $ip;
			my $mask;
			my $list_of_addrs;
			my @ip_list;

			if ($line =~ /^(.+?)\s+($CIPM_RE)\/($CIPM_RE|\d+)\s+\d{4}/) {
				($addr_name, $ip, $mask) = (ncname_from_string($1), $2, $3);
				unless ($mask =~ /^\d+$/) {
					$mask = mask_to_bits($mask);
					unless ( $mask =~ /^\d+$/ ) {
						$LOGGER->debug("WARNING: ScreenOS object group parsing: Unable to resolve mask: \'$mask\'");
						++$drop_count[0];
						next;
					}
				}
				++$collection_count[0];
			} elsif ($line =~ /^(.+?)\s+?($CIPM_RE)\s+\d{4}/) {
				# This is a host ref by address without a mask--presumed reference to a single
				# host in IP dot-delimited form.  No lookup.
				($addr_name, $ip, $mask) = ($1, $2, 32);
				++$collection_count[1];
			} elsif ($line =~ /^(.+?)\s+(\S+)\s+\d{4}/) {
				# Just a host NAME ref?  This would be a string, with possible embedded blanks,
				# followed by a lot of white space, then the host ref
				my $host_name;
				($addr_name, $host_name, $mask) = (ncname_from_string($1), $2, 32);

				# Resolve dns name to ipv4 addresses
				if ( defined $in->{'dns_resolver'}->{$addr_name} && $in->{'dns_resolver'}->{$addr_name} =~ /IPv4 Addresses:(.*)/mi
					&& $in->{'dns_resolver'}->{$addr_name} !~ /\s+None\s+/mi){
					$list_of_addrs = $';
					@ip_list = _get_ip_list($list_of_addrs);
				}else {
					# Drop this object ref. term
					$LOGGER->debug("WARNING: ScreenOS object group parsing: Unable to resolve presumed host address: \'$host_name\'");
					$LOGGER->debug("DROPPING address $host_name");
					++$drop_count[1];
					next;
				}
				++$collection_count[2];
			} elsif ($line =~ /^.+?\s+Group\s+Addresses:\s*$/) {
				# We have reached the end of the single address objects for this zone
				last;
			} else {
				# We won't be generating anything from this--no mapping on this line
				next;
			}

			# If we got here, we have something to generate as an obj. grp. elt.
			my $group_name = ncname_from_string("$name.$addr_name");
			push(@{$network_group->{objectGroupReference}}, $group_name);

			if ($mask == 32) {
				$out->open_element('networkGroup');
				$out->print_element('id', $group_name);
				if (defined $ip_list[0] ){
					for $ip (@ip_list){
						$out->print_element('host', $ip);
					}
				}else{
					$out->print_element('host', $ip);
				}
				$out->close_element('networkGroup');
			} else {
				$out->print_element('networkGroup', {
					id => $group_name,
					subnet => {
						address=> $ip,
						mask => $mask
					}
				});
			}

			# Dock a global hash pointer to this to be able to resolve it in policy...
			$in->{objectGroupLookup}->{$group_name} = $addr_name;
		}
		
		$out->print_element('networkGroup', $network_group);

		# Keep track of the total amount of collected address/network groups
		$actual_address_count += $collection_count[0] + $collection_count[1] + $collection_count[2];
		$total_drop_count += $drop_count[0] + $drop_count[1];
	}

	if ($expected_address_count) {
		$LOGGER->debug("Actually collected $actual_address_count addresses");
		$LOGGER->debug("Dropped $total_drop_count addresses");
		my $difference = $expected_address_count - ($actual_address_count + $total_drop_count);
		if ($difference) {
			$LOGGER->debug("WARNING: Sum of dropped addresses and addresses collected differs from expected total");
			$LOGGER->debug("Expected: $expected_address_count\nActual: $actual_address_count\nDropped: $total_drop_count\nDifference: $difference");
		}
	}

	parse_address_groups( $in, $out );
	parse_service_groups( $in, $out );

	$out->close_element('objectGroups');
}

sub parse_address_groups {
	my ( $in, $out ) = @_;

	while ( $in->{config} =~ /set group address \"(.*?)\" \"(.*?)\"( add)?/mig ) {
		my $address_zone = ncname_from_string($1);
		my $address_group = ncname_from_string($2);
		if ( ! defined($3) ) {
			if ( defined $in->{"address_group_$address_group"} ) {

				# What we're looking for here is a pretty hard line to parse. Embedded spaces, and easily references
				# to what would otherwise be the terminating strings we're looking for here abound.  Examples are (it's the
				# "Group Name:" line we're after):
				#
				#   get group address "V1-Trust" "445"
				#   Group Name: 445		IP: v4		Comment:
				#   Group Items:  7		Type: User-defined
				#   Members: "10.0.12.120" "10.0.12.121" "10.0.22.242" "10.0.22.246" "10.40.1.48" "10.6.32.111" "10.6.7.1"
				#
				#   get group address "V1-Untrust" "BGP Peer-PSINet"
				#   Group Name: BGP Peer-PSINet		IP: v4		Comment:
				#   Group Items:  2		Type: User-defined
				#   Members: "BGP Peer1-PSINet" "BGP Peer2-PSINet"
				#
				#  get group address "V1-Trust" "PROD IPADMINS"
				#  Group Name: PROD IPADMINS		IP: v4		Comment:
				#  Group Items:  2		Type: User-defined
				#  Members: "10.152.132.46" "10.152.32.46"
				if ( $in->{"address_group_$address_group"} =~ /^\s+?Group Name: (.+?)\s{2,}IP.*Comment:(.*)/mi) {
					my $cvtGroupName = ncname_from_string("$address_zone.$1");
					my $groupName = $cvtGroupName;
					my $groupComment = $2;
					my $network_group = { id => $cvtGroupName };

					if (defined $groupComment && $groupComment =~ /\S/ && $groupComment ne ""){
						$network_group->{description} = $groupComment;
					}

					# Furthermore, if this is a global group, then mark this in a separate
					# section of the lookup state.  Since the notion of a "global" group is
					# independent of zone, a global group can be qualified by any nonglobal
					# zone name (or by "global" itself...got that?)
					if ($address_zone =~ /^global/i) {
						$in->{objectGroupLookup}->{"Global.$address_group"} = $cvtGroupName;
					}

					if ( $in->{"address_group_$address_group"} =~ /^\s+?Members: (.*)/mi ) {
						my @groupsWithinGroups = split(' "', $1);
						my $droppedGroupsCount = 0;
						foreach my $group (@groupsWithinGroups) {
							my $thisgroup = substr($group, 0, index($group, '"', 1) );
							if ($thisgroup)	{
								$thisgroup =~ s/[\'\"]//g;
								$thisgroup = ncname_from_string("$address_zone.$thisgroup");
								# if this group is a domain name then check to see if modelled
								if (defined $in->{objectGroupLookup}->{$thisgroup}){
									push( @{$network_group->{objectGroupReference}}, $thisgroup );
								} else {
									$droppedGroupsCount++;
								}
							}
						}
						if ($droppedGroupsCount == scalar @groupsWithinGroups){
							$LOGGER->debug("WARNING: ScreenOS object group parsing: No nested objects found for address group $network_group->{id}");
							$LOGGER->debug("WARNING: Dropping address group $network_group->{id}");
							next;
						}
					}
					# Dock a pointer to this group so as to be able to resolve it at policy parse time.
					$in->{objectGroupLookup}->{$cvtGroupName} = $address_group;
					$out->print_element('networkGroup', $network_group);
				}
			}
		}
	}
}

sub parse_service_groups {
	my ($in, $out) = @_;
	return if not defined $in->{service_groups};

	while($in->{service_groups} =~ /^(\S+)\s+\d+/mg) {
		my $service_group = $1;

		if( defined $in->{"service_group_$service_group"} ) {
			my $service_group_ref;

			if( $in->{"service_group_$service_group"} =~ /^\s+Group\s+Name:\s+(\S+)\s+Comment:\s+(.*?)\s*$/mi ) {
				my ($name, $description) = ($1, $2);
				my $ncname = ncname_from_string($name);

				$service_group_ref->{id} = $name;

				if(defined $description && $description =~ /\S/ && $description ne ""){
					$service_group_ref->{description} = $description;
				}

				$serviceGroupLookup->{$ncname} = $name;

				if ($in->{"service_group_$service_group"} =~ /^\s+Members:\s+(.*?)\s*$/mi) {
					my @groupsWithinGroups = split(' ', $1);
					foreach my $group(@groupsWithinGroups) {
						$group =~ s/^"?(.*?)"?$/$1/;
						my $group_ncname = ncname_from_string($group);
						my $service_detail = $in->{"service_$group"};

						if (defined($serviceLookup->{$group_ncname})
						or defined($serviceGroupLookup->{$group_ncname})) {
							push(@{$service_group_ref->{objectGroupReference}}, $group);
						} elsif (defined($service_detail)) {
							if( $service_detail =~ /Transport.*Application\s*\n/) {
								my $service = { id => $group };

								foreach my $line(split(/^/, ${^POSTMATCH})) {
									if ($line =~ /^([udtcimp]+)\s{2,}[\d\/]+\s+(\d+)\/(\d+)/) {
										my ($protocol, $port_lo, $port_hi) = ($1, $2, $3);
										if ($port_lo == $port_hi) {
											push(@{$service->{portExpression}}, {
												port => $port_lo,
												operator => 'eq',
												protocol => $protocol
											});
										} else {
											push(@{$service->{portRange}}, {
												portStart => $port_lo,
												portEnd => $port_hi,
												protocol => $protocol
											});
										}
									}
								}

								push(@{$service_group_ref->{objectGroupReference}}, $group);
								$serviceLookup->{$group_ncname} = $service;
							}
						} else {
							$LOGGER->debug("WARNING: Unable to create service '$group', leaving it out of the group '$name'");
						}
					}
				}
			}

			if(defined $service_group_ref) {
				_print_service_group($out, $service_group_ref);
			}
		}
	}

	foreach my $service_key(keys(%{$serviceLookup})){
		my $service = $serviceLookup->{$service_key};
		_print_service_group($out, $service);
	}
}

sub parse_services {
	my ($in, $out) = @_;

	foreach my $line(split(/^/, $in->{services})) {
		if ($line =~ /^(.+?)\s{2,}([\dRPC]+)\s+([\d\/\-]+)\s+(.+?)\s+(default|\d+).*efined/gi) {
			my ($name, $protocol, $port_term, $desc) = ($1, $2, $3, $4);

			# Need this to assign as a hash key
			my $ncname = ncname_from_string($name);

			if(!$serviceLookup->{$ncname}){
				my $service = {
					id => $name,
					description => $desc
				};

				if($protocol eq 'RPC') {
					my $service_detail = $in->{"service_$name"};
					if($service_detail){
						if($service_detail =~ /Transport.*Application\s*\n([A-Z]+)\-RPC/) {
							my $rpc_flavour = $1;
							foreach my $ip_protocol('tcp', 'udp') {
								push(@{$service->{portExpression}}, {
									port => ($rpc_flavour eq 'MS') ? MS_RPC_LISTENER_PROT : SUN_RPC_LISTENER_PROT,
									operator => 'eq',
									protocol => $ip_protocol
								});
							}
						}
					}
				}else{
					my $port_expression = {protocol => get_protocol_name($protocol)};

					if($port_term =~ /^(\d+)\/(\d+)/)
					{
						$port_expression->{portStart} = $1;
						$port_expression->{portEnd} = $2;
						push(@{$service->{portRange}}, $port_expression);
					}
					elsif($port_term =~ /^(\d+)/)
					{
						$port_expression->{port} = $1;
						$port_expression->{operator} = 'eq';
						push(@{$service->{portExpression}}, $port_expression);
					}
				}

				# Store for later cross reference the protocol associated to the service
				unless ($name =~ /^any$/i){
					$in->{portServiceGroups}->{lc($name)}->{protocol} = get_protocol_name($protocol);
				}
				$serviceLookup->{$ncname} = $service;
			}
		}
		last if ($line =~ /^Total number of services/);
	}
	delete $in->{services};
}

sub parse_filters {
	my ($in, $out) = @_;

	$out->open_element('filterLists');

	# Call parse_interfaces without printing to $out
	parse_interfaces($in);

	# Check NAT's existence
	_checkNAT($in);

	# Check interface NAT's existence
	if ( defined($in->{isNAT}) ){
		# Print out packetFiltering
		$out->open_element('filterListsSection');
		_packetFiltering($in,$out);
		$out->close_element('filterListsSection');

		# Print out address-translation
		_addressTranslation ($in, $out);
	} else {
		_packetFiltering($in,$out);
	}

	$out->close_element('filterLists');
}

sub _print_service_group {
	my ($out, $service_group) = @_;

	$out->open_element('serviceGroup');

	foreach my $service_group_element('description', 'id', 'objectGroupReference'){
		if(defined($service_group->{$service_group_element})){
			$out->print_element($service_group_element, $service_group->{$service_group_element});
		}
	}

	if($service_group->{portExpression}){
		foreach my $port_expression(@{$service_group->{portExpression}}){
			$out->open_element('portExpression');
			$out->print_element('port', $port_expression->{port});
			$out->print_element('operator', $port_expression->{operator});
			if(defined($port_expression->{protocol}) && $port_expression->{protocol} ne ""){
				$out->print_element('protocol', $port_expression->{protocol});
			}
			$out->close_element('portExpression');
		}
	}

	if($service_group->{portRange}){
		foreach my $port_range(@{$service_group->{portRange}}){
			$out->open_element('portRange');
			$out->print_element('portStart', $port_range->{portStart});
			$out->print_element('portEnd', $port_range->{portEnd});
 			if(defined($port_range->{protocol}) && $port_range->{protocol} ne ""){
				$out->print_element('protocol', $port_range->{protocol});
			}
			$out->close_element('portRange');
		}
	}

	$out->close_element('serviceGroup');
}

# Checks NAT's existence
sub _checkNAT {
	my ($in, $out) = @_;

	# check for interface based NAT existence
	if ( not defined($in->{NAT}->{interfaceNAT}->{enabled}) ) {
		# get interface mode (i.e. route or nat)
		# and check if there is interface NAT defined
		_checkInterfaceNATAndStoreModes($in);
	}
	if (not defined($in->{NAT}->{MIPNAT}->{enabled}) ) {
		# check MIP NAT by parsing set interface and policies
		_checkMIPNAT($in);
	}
	# check for policy NAT-Src or NAT-Dst
	if (not defined ($in->{NAT}->{policyNAT}->{enabled}) ){
		# check is policy NAT
		_checkPolicyNAT($in);
	}
	# check for VIP NAT
	if (not defined ($in->{NAT}->{VIPNAT}->{enabled}) ){
		# check is VIP NAT
		_checkVIPNAT($in);
	}
}

# This function checks for interface NAT and stores interfaces' modes (i.e route or nat), IP address and zone.
sub _checkInterfaceNATAndStoreModes {
	my ($in) = @_;

	while ( $in->{interfaces} =~ /^(\S+)\s+($CIPM_RE|$CIDR_RE)\s+(\S.+?)\s+($MAC_RE3|N\/A)\s+(\S+)\s+(A|U|D|I|R)\s+\S+\s*$/mig ) {
		my ($ifname, $ipAddress, $zone, $mac, $status) = ($1, $2, $3, $4, $6);
		my $key_name = "if_$ifname";

		# For interface based nat the ingress interface is set to nat mode while the egress interface to route mode
		if (defined $in->{$key_name} and $in->{$key_name} =~ /mode\s+route/) {
			$in->{NAT}->{$key_name}->{interfaceMode} = 'route';
			my %interface = (
				intName => $ifname,
				intMode => 'route',
				zone => $zone,
				ipAddress => $ipAddress
			);
			push (@{$in->{NAT}->{interfaceNAT}->{interfaces}}, \%interface);
		} elsif (defined $in->{$key_name} and $in->{$key_name} =~ /mode\s+nat/) {
			$in->{NAT}->{$key_name}->{interfaceMode} = 'nat';
			$in->{NAT}->{$key_name}->{ipAddress} = $ipAddress;
			my %interface = (
				intName => $ifname,
				intMode => 'nat',
				zone => $zone,
				ipAddress => $ipAddress
			);
			push (@{$in->{NAT}->{interfaceNAT}->{interfaces}}, \%interface);

			# Record that interface NAT is available
			$in->{NAT}->{interfaceNAT}->{enabled} = 1;
			$in->{isNAT} = 1;
		}
	}
}

# This function checks the MIP NAT existence
sub _checkMIPNAT {
	my ($in) = @_;

	# set interface ethernet2 mip 1.1.1.5 host 10.1.1.5 netmask 255.255.255.255 vrouter trust-vr
	while ($in->{config} =~  /set interface\s+\"?(.*?)\"?\s+mip\s+(\S+)\s+host\s+(\S+)\s+netmask\s+([\d\.]+)/mig) {
		if (not defined($in->{NAT}->{MIPNAT}->{enabled}) ) {
			$in->{NAT}->{MIPNAT}->{enabled} = 1;
			$in->{isNAT} = 1;
		}
	}
}

# This function checks for VIP NAT existence
sub _checkVIPNAT {
	my ($in) = @_;

	open FH, '<', \$in->{config} or $LOGGER->fatal_error_code('ERROR: Couldn\'t open configuration settings');
	while (<FH>){
		# Get set interfaces
		my $set_interface;
		my $protocol = "";

		# Case 1 i.e set interface ethernet1/4 vip 210.1.1.10 + 80 "HTTP" 192.168.1.10
		# or set interface ethernet3 vip 1.1.1.10 80 http 10.1.1.10
		if ($_ =~ /set interface\s+(\S+)\s+vip\s+($CIPM_RE|$CIDR_RE)\s*\+?\s*(\d+)\s+(\S+)\s+($CIPM_RE|$CIDR_RE)/i){
			my $intf;
			if (defined $1){
				$intf = getIntfinZone($in,$1);
			}
			if (defined $intf) {
				$set_interface = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser::formatArgs($intf, $2, $3, $4, $5);
			} else {
				$set_interface = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser::formatArgs($1, $2, $3, $4, $5);
			}

		# Case 2: set interface <interface> vip <ip_addr1> [ + ] port-range <port1  port2> server-ip <ip-addr2> port-range <portx  porty> [ protocol TCP | UDP ] [ manual ]
		}elsif ($_ =~ /set interface\s+(\S+)\s+vip\s+($CIPM_RE|$CIDR_RE)\s+port-range\s+(\d+)-(\d+)\s+server-ip\s+($CIPM_RE|$CIDR_RE)\s+port-range\s+(\d+)-(\d+)\s?(?:protocol)?\s?(TCP|UDP)?/){
			if (defined $8){
				$protocol = $8;
			}
			$set_interface = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser::formatArgs($1, $2, $3, $4, $5, $6, $7, $protocol);
		}

		if (defined($set_interface)){
			push ( @{$in->{NAT}->{VIPNAT}->{set_interface}}, $set_interface);
		}
	}
	if (defined($in->{NAT}->{VIPNAT}->{set_interface}) and scalar(@{$in->{NAT}->{VIPNAT}->{set_interface}})){
		$in->{NAT}->{VIPNAT}->{enabled} = 1;
		$in->{isNAT} = 1;
	}
	close FH;
}

# This function gets the zone name if there is an interface defined into a zone
sub getIntfinZone {
	my ($in, $intf) = @_;

	if (!defined $intf){
		return undef;
	}

	open FHConfig, '<', \$in->{config} or $LOGGER->fatal_error_code('ERROR couldn\'t open configuration settings');
	my $zone;
	while (<FHConfig>){
		# set interface ethernet3 zone untrust
		if ($_ =~ /set interface\s+\"?$intf\"?\s+zone\s+(\S+)/i) {
			$zone = $1;
		}
	}
	close FHConfig;

	return $zone;
}

# This functions checks the policy based NAT-Src definition
sub _checkPolicyNAT {
	my $in = shift;

	open FH,  '<', \$in->{config} or $LOGGER->fatal_error_code("ERROR couldn't open configuration settings") ;
	while (<FH>){
		#get set interfaces
		#commands syntax 1: set interface <intName> [ ext ip <ip/mask> ] dip <id> ipAdr1 [ ipAdr2 ] [random-port | fix-port | incoming] [ scale-size <#> ]
		my $set_interface;
		my $match4 = "", my $match5 = "", my $match6 = "";
		if ($_ =~ /set interface\s+(\S+)\s?(?:ext ip)?\s?(?:$CIDR_RE)?\s+dip\s+(\d+)\s+($CIPM_RE)\s?($CIPM_RE)?\s?(fix-port)?/i ){
			if (defined $4){
				$match4 = $4;
			}
			if (defined $5){
				$match5 = $5;
			}
			$set_interface = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser::formatArgs($1, $2, $3, $match4, $match5);
			push (@{$in->{NAT}->{policyNAT}->{set_interface}}, $set_interface) if defined $set_interface;
		}
		# command syntax 2: set interface <intName> [ ext ip <ip/mask> ] dip <id> shift-from ipAdr3 to ipAdr4 ipAdr5
		if ($_ =~ /set interface\s+(\S+)\s?(?:ext ip)?\s?(?:$CIDR_RE)?\s+dip\s+(\d+)\s+(shift-from)\s+($CIPM_RE)\s+to\s+($CIPM_RE)\s+($CIPM_RE)/i){
			$set_interface = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser::formatArgs($1, $2, $3, $4, $5, $6);
			push (@{$in->{NAT}->{policyNAT}->{set_interface}}, $set_interface) if defined $set_interface;
		}

		if ( ($_ =~ /nat\s+src/i) or (defined($in->{NAT}->{policyNAT}->{set_interface}) ) or ($_ =~ /nat\s+dst/i) or ($_ =~ /nat\s+src\s+dst/i)  ){
			# must be defined at least policy based NAT
			$in->{NAT}->{policyNAT}->{enabled} = 1;
			$in->{isNAT} = 1;
		}

	}
	close FH;
}

sub _addressTranslation {
	my ($in, $out) = @_;
	my $addressTranslationParser = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser->new($in);

	# Set processingPhase
	$in->{NAT}->{settings}->{processingPhase} = "address-translation";
	my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
	# MIP NAT
	if ( defined($in->{NAT}->{MIPNAT}->{enabled}) and defined($in->{NAT}->{MIPNAT}->{enabled}) ) {
		$out->open_element('filterListsSection');
		my $routingNATFilterList = $addressTranslationParser->routingNAT('MIPNAT');
		$out->print_string($routingNATFilterList);
		# Add the phase discriminator
		my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
		$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	# VIP NAT
	if ( defined($in->{NAT}->{VIPNAT}->{enabled}) and defined($in->{NAT}->{VIPNAT}->{enabled}) ) {
		$out->open_element('filterListsSection');
		my $routingNATFilterList = $addressTranslationParser->routingNAT('VIPNAT');
		$out->print_string($routingNATFilterList);
		# Add the phase discriminator
		my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
		$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	# Interface NAT
	if ( defined($in->{NAT}->{interfaceNAT}) and defined($in->{NAT}->{interfaceNAT}->{enabled}) and not defined($in->{NAT}->{policyNAT}->{enabled}) ){
		$out->open_element('filterListsSection');
		my $routingNATFilterList = $addressTranslationParser->routingNAT('interfaceNAT');
		$out->print_string($routingNATFilterList);
		# Add the phase discriminator
		my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
		$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	# Policy based NAT
	if ( defined($in->{NAT}->{policyNAT}->{enabled}) and defined($in->{NAT}->{policyNAT}->{enabled}) ) {
		# Print the pre-routing section
		my $routingNATFilterList = $addressTranslationParser->routingNAT('policyNAT','preRouting');
		if ($routingNATFilterList ne ''){
			$out->open_element('filterListsSection');
			$out->print_string($routingNATFilterList);
			$out->close_element('filterListsSection');
		}
		# Reset values
		$in->{NAT}->{policyNAT}->{set_policy} = [];
		$in->{NAT}->{policyNAT}->{rules} = [];
		# Print the post-routing section
		$routingNATFilterList = $addressTranslationParser->routingNAT('policyNAT','postRouting');
		if ($routingNATFilterList ne ''){
			$out->open_element('filterListsSection');
			$out->print_string($routingNATFilterList);
			$out->close_element('filterListsSection');
		}
	}

}

sub _packetFiltering {
	my ( $in, $out ) = @_;

	my $ftr_from_add = '';
	my $ftr_to_add = '';

	# if NAT configuration is enabled then set processingPhase
	my $addressTranslationParser;
	my $packetProcessingPhase;
	if ($in->{isNAT}) {
		$in->{NAT}->{settings}->{processingPhase} = "packet-filtering";
	}

	# let's do a quick loop here to create a hash of interfaces to zones so we can populate appliesTo elements
	my %interfacesHash = ();
	while ( $in->{interfaces} =~ /^\s*(\S+)\s+(?:[\da-f\.\:]+)\/(?:\d+)\s+(\S+)\s+(?:[\da-f\.\:\-]+)\s+\S+\s+(?:\S+)\s+\S+\s*$/mig )
	{
		my $interface = $1;
		my $zone = ncname_from_string($2);
		if(exists $interfacesHash{$zone}){
			$interfacesHash{$zone} .= ",$interface";
		}else{
			$interfacesHash{$zone} = $interface;
		}
	};
	$in->{interface_groupings} = \%interfacesHash;

	# The keys for this hash will be a combo of "sourceZone"~"DestZone"
	my %zoneSourceDestHash = ();

	#my $openedFilterLists = 0;
	# Reference notes in the following:
	#   * I've not seen any BNF or the like as to the allowable zone name from CLI or HTTP config, but
	#           I've not seen any examples, of hundreds, that contain embedded blanks or non-NCName chars
	#   * As source/dest can refer to address refs in zone defs (HTTP config refs for Netscreen call these
	#             "address book entries"(), they can *definitely* contain embedded blanks and the like.  You
	#            can only parse these with a presumption of multiple blanks following the tabularly formatted
	#            policy display ("get policy all") display.
	#   * "Action" should be either "Permit" or "Deny", but we will allow it to be any alphabetized word.
	#   * "State" is "enabled" or "disabled", or "disabl~" (the '~' character is always used as an abbreviation),
	#              but we'll take any lowercase word with tilde
	#   * Each line ends with a series of '-' and 'X' characters, as a legend of flags set or not.
	#

	# The combination of the following are the key for each actual <filterList>.  Each filter ends up being a separate
	# <filterEntry>, evaluated sequentially.  All we do in the 'while' loop to follow is to put together that filter entry
	# and loop it into %zoneSourceDestHash, keyed by source and dest zones.

	my $ftr_from_zone;
	my $ftr_to_zone;
	my $zoneSrcDestKey;
	my $from_int;
	my $to_int;
	my $termFilterEntry;

	POLICY:	foreach my $line ( split( "\n", $in->{policy} ) ) {
		unless ($line =~ /^\s*(\d+)\s+(\S+)\s+(\S+)\s+(.+?)\s+(.+?)\s+(\S+)\s+([A-Z][a-z~]+)\s+([a-z~]+)\s[-X]+/){
			next;
		}
		# Need to map the zone names (object groups) in the ScreenOS represntation to canonical XML NCName parlance
		$ftr_from_zone = ncname_from_string($2);
		$ftr_to_zone   = ncname_from_string($3);

		my $ftr_id        = $1;
		my $ftr_from_add;
		my $ftr_to_add;
		my $ftr_service   = $6;
		my $ftr_action    = $7;
		my $ftr_state     = $8;
		my $operator      = 'eq';
		my $ftr_name      = '';
		my $ftr_log       = '';

		next unless ($ftr_state eq 'enabled' and defined $in->{ "acl_$ftr_id" });

		if ( $in->{ "acl_$ftr_id" } =~ /(?:^|, |^global policy: )name(?::"|\s)([^"]+)(?:"|,)/mi ) {
			$ftr_name = $1;
		}

		if ($ftr_action =~ /~$/) {
			if ($in->{ "acl_$ftr_id" } =~ /,action (\S+),/ ){
				$ftr_action = $1;
			}
		}
		$ftr_action = getSupportedFilterAction($ftr_id, $ftr_action);
		unless ($ftr_action) {
				next POLICY; # skip this policy 
		}
		
		if ($ftr_from_zone =~ /~$/) {
			($ftr_from_zone) = ncname_from_string($in->{ 'acl_' . $ftr_id } =~ /, zone (\S+) ->/);
		}
		if ($ftr_to_zone =~ /~$/) {
			($ftr_to_zone) = ncname_from_string($in->{ 'acl_' . $ftr_id } =~ /, zone (?:\S+) -> (\S+),/);
		}

		# See if we have a source/dest hash key for this combo.
		$zoneSrcDestKey = "$ftr_from_zone~$ftr_to_zone";
		if (not defined $zoneSourceDestHash{$zoneSrcDestKey}) {
			$zoneSourceDestHash{$zoneSrcDestKey} = ();
		}

		$ftr_log = ( $in->{ "acl_$ftr_id" } =~ /^log no/mi ) ? 'false' : 'true';
		my $filterEntry = {
			log		=> $ftr_log,
			primaryAction	=> lc($ftr_action),
			processOrder	=> $ftr_id
		};
		
		# set the ztSignature for rule counting
		$filterEntry->{ztSignature} = $ftr_id;

		# Can pull up source one of two ways
		my ($ftr_src_ref) = $in->{ "acl_$ftr_id" } =~ /^src "(.+?)",/m;
		if (not defined $ftr_src_ref) {
			($ftr_src_ref) =  $in->{ "acl_$ftr_id" } =~ /source: "(.*)"/m;
		}

		if ($ftr_src_ref) {
			$ftr_from_add = ncname_from_string($ftr_src_ref);
		}

		if ($ftr_from_add) {
			# Need to make sure this reference exists...
			my $cvtFromAdd = _generate_ip_object_group_reference($in, $ftr_from_zone, $ftr_from_add, 'source', $filterEntry, $ftr_id);

			if (not defined $cvtFromAdd) {
				# Object group not found, log message already printed
				next;
			}
		} else {
			if ( $in->{ "acl_$ftr_id" } =~ /^\d+ sources: (.*)/mi ) {
				$_ = $1;
				s/"//g;
				my @ftr_sources = split(/,\s+/);
				foreach my $source (@ftr_sources) {
					$source =~ s/\r|\n//g;
					my $cvtFromAdd = _generate_ip_object_group_reference($in, $ftr_from_zone, $source, 'source', $filterEntry, $ftr_id);
					if (not defined $cvtFromAdd) {
						# Object group not found, log message already printed
						next POLICY;
					}
				}
			}
		}

		# There are two forms of dest spec here
		my ($ftr_dest_spec) = $in->{ "acl_$ftr_id" } =~ /, dst "(.+?)"/m;

		if (not defined $ftr_dest_spec) {
			($ftr_dest_spec) =  $in->{ "acl_$ftr_id" } =~ /destination: "(.*)"/m;
		}

		if ($ftr_dest_spec) {
			$ftr_to_add = ncname_from_string($ftr_dest_spec);
		}

		if ($ftr_to_add) {
			my $ftrToRef = _generate_ip_object_group_reference($in, $ftr_to_zone, $ftr_to_add, 'destination', $filterEntry, $ftr_id);

			if (not defined $ftrToRef) {
				# Undefined object group
				next;
			}
		} else {
			if ( $in->{ "acl_$ftr_id" } =~ /^\d+ destinations: (.*)/mi ) {
				$_ = $1;
				s/"//g;
				my @ftr_destinations = split(/,\s+/);
				foreach my $destination (@ftr_destinations) {
					$destination =~ s/\r|\n//g;
					my $ftrToRef = _generate_ip_object_group_reference($in, $ftr_to_zone, $destination, 'destination', $filterEntry, $ftr_id);
					if (not defined $ftrToRef) {
						# Object group not found, log message already printed
						next POLICY;
					}
				}
			}
		}

		if ( $ftr_name ne 'none' ) {
		    $filterEntry->{name} = $ftr_name;
		}

		# First look for services ports in acl_[id] entry of responses. There are two ways these can show up in v.5.0 and v.6.0 of the system.
		my @ftr_ports;

		# if ( $in->{ 'acl_' . $ftr_id } =~ /^\d+ services:\s+("(?:[^\s"]+)"(?:,\s+"(?:[^\s"]+)")*)\s*$/mi )
		if ($in->{ "acl_$ftr_id" } =~ /^\d+ services?:\s+(.*)/mi) {
			$_ = $1;
			s/"//g;
			@ftr_ports = split(/,\s+/);
		} elsif ($in->{ 'acl_' . $ftr_id } =~ /,\s+serv\s+"(.+)\"/i) {
			my $sStr = $1;
			$sStr =~ s/[",]//g;
			@ftr_ports = split(/,\s+/, $sStr);
		}

		if (scalar @ftr_ports) {
			foreach my $service (@ftr_ports) {
				$service =~ s/\r|\n//g;
				if ($service =~ /^any/i) {
					# We can assume (!) that this is the only one in list, and hence we can
					# bypass generating a service entry, since "any" is the conceptual default
					last;
				}
				my $portExpressions = _service_port_expression(ncname_from_string($service));
	   			if( $portExpressions ) {
	   				foreach my $expression (@$portExpressions) {
	   					push(@{$filterEntry->{destinationService}},$expression);
	   				}
	   			}
			}
		# Then in general policy table
		} elsif ( $ftr_service =~ /any/i ) {
			# No destinationService means any
		} else {
			my $portExpressions = _service_port_expression(ncname_from_string($ftr_service));
			if( $portExpressions ) {
				foreach my $expression (@$portExpressions) {
					push(@{$filterEntry->{destinationService}},$expression);
				}
			}
		}

		# Store this filter entry as being pertinent to the scope of *this* source and dest.
		# We need to make sure we do this in order of sequence in the config.
		unshift @{$zoneSourceDestHash{$zoneSrcDestKey}}, $filterEntry;
	}
	
	# Now, we walk the list of source and dest filter elements.  Each unique combination in this hash is a zone set.
	foreach $zoneSrcDestKey ( keys(%zoneSourceDestHash) ) {
		($ftr_from_zone, $ftr_to_zone) = split(/~/, $zoneSrcDestKey);

		# Start recording the element...
		$out->open_element('filterList');

		# Set default action for each list to "none" so that if there is no match QRM core simply moves on to the next list.
		$out->print_element('defaultAction', 'none');

		# appliesTo elements for regular policies (the first policies listed)
		if ( ( exists $interfacesHash{$ftr_from_zone} ) || ( exists $interfacesHash{$ftr_to_zone} ) ) {
			$out->open_element('appliesTo');
			$out->print_element('sourceGroup', $ftr_from_zone) if  exists $interfacesHash{$ftr_from_zone};
			$out->print_element('destinationGroup', $ftr_to_zone) if exists $interfacesHash{$ftr_to_zone};
			$out->close_element('appliesTo');
		}

		# appliesTo elements for global policies which follow in the listing after regular policies
		if (( $ftr_from_zone eq 'Global' ) || ( $ftr_to_zone eq 'Global' )) {
			$out->open_element('appliesTo');
			$out->print_element('sourceGroup', 'Global') if $ftr_from_zone eq 'Global';
			$out->print_element('destinationGroup', 'Global') if $ftr_to_zone eq 'Global';
			$out->close_element('appliesTo');
		}

		# By using 'pop', here, we're honoring use as a queue since it was populated with
		# unshift.
		my $feLastIndex = 0;

		while (my $sdFilterEntry = pop @{$zoneSourceDestHash{$zoneSrcDestKey}}) {
			# Need to save this for terminating default
			$feLastIndex = $sdFilterEntry->{processOrder};
			$out->print_element( 'filterEntry', $sdFilterEntry );
		}

		$out->print_element( 'mode', 'stateful');

		# This predates the tilde delimiter.
		$out->print_element( 'name', $ftr_from_zone . "_" . $ftr_to_zone);
		if (defined $in->{isNAT}) {
			# Get an addressTranslationParser object
			if (not defined($addressTranslationParser)) {
				$addressTranslationParser = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser->new($in);
			}
			if (not defined($packetProcessingPhase)) {
				$packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
			}

			$out->print_string($packetProcessingPhase);
		}
		$out->close_element('filterList');
	}

	# So, to match the final-forwarding-logic-of-last-resort on the NetScreen, we need a per-zone
	# rule which says if flows are going from this zone to any OTHER zone, and nothing has applied until
	# this point to create an authoritative action disposition, then deny.  If the zone is going TO someplace
	# *in this zone*, then set the disposition to whether a feature called "intrazone blo0cking" has been enabled:
	#
	#     final-rule-1:
    #        <Applies to>:
    #           <source>: zone-1
    #           <filter-entry>
    #              <destination IP addr>
    #                 object-group-reference: zone-1.any
    #               </destination IP addr>
    #               <primary-action>:  (if intrazone blocking enabled in zone config then deny, else permit).
    #           </filter-entry>
    #           <Filter-entry>
    #                 source: all
    #                 dest: all
    #                <primary-action>: deny
    #           </filter-entry>
	#
	#   Repeat for all configured non-global zones.

	# Set up an entry which will be usable for all
	$termFilterEntry = {
		log => 'false',
		primaryAction => 'permit',
		processOrder => 1
	};

	foreach $ftr_from_zone (keys %interfacesHash) {
		# Since we can't limit intrazone traffic via any known destinationIPAddr or the like, need
		# to generate two filter lists -- one with a dest of same from_zone, one with a dest applies to of any.
		# The first, then, is the intrazone one.  Only need to generate it for zones which do not have intrazone
		# blocking enabled.
		if (($ftr_from_zone !~ /^Global/) and (not _intrazone_blocking_enabled($in, $ftr_from_zone))) {
			$out->open_element("filterList");

			foreach my $direction('source', 'destination'){
				$out->print_element('appliesTo', {$direction.Group => $ftr_from_zone});
			}

			$out->print_element ( 'filterEntry', $termFilterEntry );
			$out->print_element ( 'mode', 'stateful' );
			$out->print_element ( 'name', $ftr_from_zone . "_intrazone_default" );
			if (defined $in->{isNAT}) {
				# Get an addressTranslationParser object
				if (not defined($addressTranslationParser) ){
					$addressTranslationParser = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser->new ($in);
				}
				if (not defined($packetProcessingPhase)) {
					$packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
				}

				$out->print_string($packetProcessingPhase);
			}
			$out->close_element('filterList');
		}

	}

	# Now, for the final default
	$termFilterEntry->{primaryAction} = 'deny';

	foreach $ftr_from_zone (keys %interfacesHash) {
		if ($ftr_from_zone !~ /^Global/) {
			$out->open_element ( 'filterList' );
			$out->print_element ( 'appliesTo', {sourceGroup => $ftr_from_zone} );
			$out->print_element ( 'filterEntry', $termFilterEntry );
			$out->print_element ( 'mode', 'stateful' );
			$out->print_element ( 'name', $ftr_from_zone . '_interzone_default' );
			if (defined $in->{isNAT}) {
				# Get an addressTranslationParser object
				if (not defined($addressTranslationParser) ){
					$addressTranslationParser = ZipTie::Adapters::Juniper::ScreenOS::AddressTranslationParser->new ($in);
				}
				if (not defined($packetProcessingPhase)) {
					$packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
				}

				$out->print_string($packetProcessingPhase);
			}
			$out->close_element( 'filterList' );
		}
	}
}

sub parse_snmp {
	my ( $in, $out ) = @_;

	$out->open_element('snmp');

	my @traps = ();
	my ( $sysContact, $sysLocation, $sysName );
	while ( $in->{config} =~ /^set\s+snmp\s+(\S.+)$/mig ) {
		my $snmp_command = $1;
		if ( $snmp_command =~ /^community/i ) {

			#community "public" Read-Write Trap-on  traffic version v1
			my $accessType = ( $snmp_command =~ /\bRead-Write/i ) ? 'RW' : 'RO';
			my ($commString) = $snmp_command =~ /community\s+"([^\s\"]+)"/i;
			$out->print_element( 'community', { communityString => $commString, accessType => $accessType } );
		} elsif ( $snmp_command =~ /^host\s+"([^\s\"]+)"\s+($CIPM_RE)\s+($CIPM_RE)/i ) {
			push( @traps, { communityString => $1, ipAddress => $2, mask => $3 } );
		} elsif ( $snmp_command =~ /^(contact|location|name)\s+"(([^\s\"]+))"/i ) {
			$sysContact  = $2 if ( lc($1) eq 'contact' );
			$sysLocation = $2 if ( lc($1) eq 'location' );
			$sysName     = $2 if ( lc($1) eq 'name' );
		}
	}

	if ( !defined $sysName ) {
		if ( $in->{snmp} =~ /^Sysname\s+:(\S+)/mi ) {
			$sysName = $1;
		} elsif ( $in->{config} =~ /^set\s+hostname\s+(\S+)/mi ) {
			$sysName = $1;
		}
	}

	$out->print_element( 'sysContact', $sysContact ) if ( defined $sysContact );
	$out->print_element( 'sysLocation', $sysLocation ) if ( defined $sysLocation );
	$out->print_element( 'sysName', $sysName ) if ( defined $sysName );

	if ( defined $sysName ) {
		foreach my $trap (@traps) {
			$out->print_element( 'trapHosts', $trap );
		}
	}

	$out->close_element('snmp');
	delete $in->{snmp};
}

sub parse_static_routes {
	my ( $in, $out ) = @_;
	my $staticRoutes;

	while ( $in->{routes} =~ /^\*\s+\d+\s+($CIDR_RE)\s+(\S+)\s+($CIPM_RE)\s+(\S+)\s+(\d+)\s+(\d+)\s+\S+\s*$/mig ) {
		my $ip_address = parseCIDR($1);
		my $interface  = $2;
		my $gateway    = $3;
		my $type       = $4;
		my $preference = $5;
		my $metric     = $6;

		# Only include "Static" routes.
		# +++this might want to include "connected" routes as well
		next unless ($type =~ /S/);
		
		my $route = {
			destinationAddress => $ip_address->{host},
			destinationMask    => mask_to_bits ( $ip_address->{network} ),

			defaultGateway     => ( $ip_address->{host} ne '0.0.0.0' ? 'false' : 'true' ),
			interface          => $interface,
			routeMetric        => $metric,
			routePreference    => $preference
		};
		
		unless ($gateway eq "0.0.0.0"){ # if gateway address is 0.0.0.0 then no next hop has been decared so don't model
			$route->{gatewayAddress} = $gateway;
		}

		push( @{ $staticRoutes->{staticRoute} }, $route );
	}

	$out->print_element( 'staticRoutes', $staticRoutes ) if ( defined $staticRoutes );
	delete $in->{route};
}

sub parse_interfaces {
	my ( $in, $out ) = @_;

	$out->open_element('interfaces') if (defined $out);

	while ( $in->{interfaces} =~ /^(\S+)\s+(unnumbered|(($CIPM_RE)\/(\d+)))\s+\S.+?\s+($MAC_RE3|\S+)\s+(\S+)\s+(A|U|D|I|R)\s+\S+\s*$/mig ) {
		my ($ifname, $ipAddress, $mask, $mac, $status) = ($1, $4, $5, $6, $8);

		my $interface = {
			name          => $ifname,
			adminStatus   => ($status =~ /^[AUI]$/i) ? 'up' : 'down',
			interfaceType => ($ifname =~ /^(un)?trust/i) ? 'ethernet' : get_interface_type($ifname),
			physical      => _is_physical($ifname)
		};

		if ($mac ne 'N/A') {
			if ($mac =~ /^$MAC_RE3$/) {
				$interface->{interfaceEthernet}->{macAddress} = strip_mac($mac);
			} else {
				if ( $in->{ "if_$mac" } and $in->{"if_$mac"} =~ /\s+mac\s+($MAC_RE3)\s*$/im) {
					$interface->{interfaceEthernet}->{macAddress} = strip_mac($1);
				}
			}
		}

		if (defined($ipAddress) and $ipAddress ne '0.0.0.0') {
			push @{ $interface->{interfaceIp}->{ipConfiguration} }, { ipAddress => $ipAddress, mask => $mask, precedence => 1 };
		}

		if ( defined $in->{ "if_$ifname" } ) {
			if ( $in->{ "if_$ifname" } =~ /\b(auto|half|full)-duplex/mi ) {
				$interface->{interfaceEthernet}->{autoDuplex} = ( lc($1) eq 'auto' ? 'true' : 'false' );
				$interface->{interfaceEthernet}->{operationalDuplex} = lc($1);
			}
			if ( $in->{ "if_$ifname" } =~ /^\s*bandwidth:\s+physical\s+(\d+)([kmg])bps(?:,\s+configured\s+(\d+)([kmg])bps,\s+current\s+(\d+)([kmg])bps)?/mi ) {
				if ($5 and $6 and $5 ne '0') {
					$interface->{speed} = getUnitFreeNumber( $5, $6, 'bit' )
				} elsif ($3 and $4 and $3 ne '0') {
					$interface->{speed} = getUnitFreeNumber( $3, $4, 'bit' )
				} elsif ($1 and $2 and $1 ne '0') {
					$interface->{speed} = getUnitFreeNumber( $1, $2, 'bit' )
				}
			}
			if ( $in->{ "if_$ifname" } =~ /^\s*admin mtu (\d+)/mi ) {
				$interface->{mtu} = lc($1);
			}
		}

		push(@{$interface->{interfaceGrouping}}, {
			ifGroupName => 'Global',
			ifGroupingCharacteristic => 'securityZoneMember'
		});

		if($in->{interface_groupings}){
			my %interface_groupings = %{$in->{interface_groupings}};

			foreach my $key(keys(%interface_groupings)){
				foreach my $if(split(/,/, $interface_groupings{$key})){
					if($if eq $ifname){
						push(@{$interface->{interfaceGrouping}}, {
							ifGroupName => $key,
							ifGroupingCharacteristic => 'securityZoneMember'
						});
					}
				}
			}
		}
		$out->print_element( 'interface', $interface ) if (defined $out);
	}
	$out->close_element('interfaces') if (defined $out);
}

sub getUnitFreeNumber {
	my ($number, $unit, $base) = @_;

	my $m = 1024;
	if ( defined $base ) {
		# Memory size is measured in bytes
		if ( $base =~ /byte/i ) {
			$m = 1024;
		# Network speed is measured in bits
		} elsif ( $base =~ /bit/i ) {
			$m = 1000;
		}
	}

	if ( $unit =~ /K/i ) { $number * $m; }
	elsif ( $unit =~ /M/i ) { $number * $m * $m; }
	elsif ( $unit =~ /G/i ) { $number * $m * $m * $m; }
}

sub _generate_ip_object_group_reference {
	my ($in, $ogZone, $ogRef, $direction, $fe, $policy_id) = @_;
	my $ogPathLoc;
	my $generateFEnt = 0;

	my $lOgZone = ncname_from_string($ogZone);
	my $lOgRef = ncname_from_string($ogRef);

   	# Need to make sure this reference exists
   	if ($lOgRef =~ /^.*?\.?any/i) {
   		# "any" possibly qualified by a zone....which can be suppressed since it's the default
   		# but *don't* generate the filter entry....it's excess ("any" is implied)
   		$ogPathLoc = $lOgRef;
   	} elsif (defined $in->{objectGroupLookup}->{"$lOgZone.$lOgRef"}) {
   		$ogPathLoc = "$lOgZone.$lOgRef";
   		$generateFEnt = 1;
   	} elsif (defined $in->{objectGroupLookup}->{"Global.$lOgRef"}) {
   		# In this case, we need to actively change the name to the global one
   		$ogPathLoc = "Global.$lOgRef";
   		$generateFEnt = 1;
   	} else {
		$LOGGER->debug("WARNING: Suppressing policy $policy_id, unable to find $direction network/address group reference for $ogRef in zone $ogZone or Global");
   	}

   	if ($generateFEnt) {
		push(@{$fe->{$direction . 'IpAddr'}}, {objectGroupReference => $ogPathLoc});
   	}

   	return $ogPathLoc;
}

sub _is_physical {
	my $name = shift;
	my $physical = 'false';

	# Have to take into account v5.0 interfaces, which were just "trust", "untrust"
	if ( $name !~ /\.\d+$/ && $name =~ /seri|eth|gig|^fe|^fa|token|[bp]ri|trust/i ) {
		$physical = 'true';
	}

	return $physical;
}

# Removes unnecessary data from list_of_addrs
sub _get_ip_list {
	my ($list_of_addrs) = @_;
	my @ip_list = split(/\n/,$list_of_addrs);
	my @queue;

	foreach my $ip (@ip_list){
		# Remove possible tabs or carriage returns
		$ip =~ s/(\r|\t)//g;
		if ($ip =~ /$CIPM_RE/) {
			push(@queue, $ip);
		}
	}
	return @queue;
}

# This returns a fully initialized 'objectGroupReference', 'portExpression' or 'portRange' element,
# for service identification.
sub _service_port_expression {
	my ($port) = @_;
	my $returnService = ();

	if ( $port =~ /^\d+$/ )	{
		push @{$returnService}, {portExpression => {port => $port, operator => "eq"}};
	} else {
		if(defined($serviceGroupLookup->{$port})) {
			 push @{$returnService}, {objectGroupReference => $serviceGroupLookup->{$port}};
		} elsif(defined($serviceLookup->{$port})) {
			push @{$returnService}, {objectGroupReference => $serviceLookup->{$port}->{id}};
		} elsif ( defined (get_port_expression($port))) {
			$LOGGER->debug("Using hash in ZipTie Adapter Utils for port expression lookup for service $port");
			push @{$returnService}, {portExpression => {port => get_port_expression($port), operator => 'eq'}};
		} else {
			# This service is hitting the floor, suppress multiple warnings by keeping track of what has been logged
			if ( ! defined($serviceWarningLookup->{$port}) ) {
				$serviceWarningLookup->{$port} = 1;
				$LOGGER->debug("WARNING: No port, protocol, or object group reference information available for service $port");
			}
		}
	}
	return $returnService;
}

sub _intrazone_blocking_enabled {
	my ($in, $zoneIZ) = @_;
	if (defined($in->{"zone_$zoneIZ"})){
		return ($in->{"zone_$zoneIZ"} =~ /\nIntra\-zone block: On,/sig);
	} else {
		$LOGGER->debug("WARNING: zone $zoneIZ not found");
		return 0;
	}
}

1;


sub getSupportedFilterAction {
	my $ftr_id             = shift;
	my $ftr_action         = shift;

	my @validFilterActions = qw(deny drop forward next none permit);
	my $validAction;
	foreach my $candidateAction (@validFilterActions) {
		if ( $ftr_action =~ /^$candidateAction$/i ) {
			$validAction = lc($ftr_action);
		}
	}
	# Take some of the mappings which must be mapped
	unless ($validAction) {
		for ($ftr_action) {
			# Per group decision
			if (/permit-infranet-auth/i) { $validAction = "permit-authenticated"; }
			elsif (/reject/i) { $validAction = "drop"; }
		}
	}
	unless ($validAction) {
		$LOGGER->debug("WARNING: Suppressing filter entry $ftr_id, policy action not supported by QRM");
	}
	return $validAction;
}

__END__

=head1 Parsers

ZipTie::Adapters::Juniper::ScreenOS::Parsers

=head1 SYNOPSIS

	use ZipTie::Adapters::Juniper::ScreenOS::Parsers qw(create_config parse_local_accounts parse_chassis parse_object_groups parse_services parse_filters parse_snmp parse_system parse_interfaces parse_static_routes);
	parse_filters( $cliResponses, $xmlPrinter);

=head1 DESCRIPTION

Module with many static methods that take in a hash of
device responses and print out ZipTie model elements.

=head2 Methods

=over 12

=item C<create_config( $in, $out )>

Uses the device configs to put together a ZipTie configurationRepository element.

=item C<parse_local_accounts( $in, $out )>

Parse the local accounts.

=item C<parse_chassis( $in, $out )>

Parse out the chassis level elements.

=item C<parse_object_groups ( $in, $out )>

Parse and set state on network and service-based object gruops on the system, orienting them as grouped reference
sets bound to zones.  This will call parse_network_groups and parse_service_groups.

=item C<parse_services ( $in, $out )>

Parse and set state reflecting the individual services in the configuration.  This does not process service groups.

=item C<parse_filters( $in, $out )>

Parse out firewall rules.

=item C<parse_snmp( $in, $out )>

Parse out the SNMP elements.

=item C<parse_system( $in, $out )>

Parses out some top level core attributes of the ZipTie model ZiptieElementDocument

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.

  The Original Code is Ziptie Client Framework.

  The Initial Developer of the Original Code is AlterPoint.
  Portions created by AlterPoint are Copyright (C) 2006,
  AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): rkruse
  Date: Apr 23, 2007

=cut
