package ZipTie::Adapters::PaloAlto::PANOS;

use strict;
use warnings;

use 5.0100000;    # use perl version 5.10

use Data::Dumper;
use XML::Twig::XPath;
use English qw( -no_match_vars );    #Friendly names for punctuation variables

use ZipTie::ConnectionPath;
use ZipTie::Model::XmlPrint;
use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin qw(issue_command);
use ZipTie::Adapters::PaloAlto::PANOS::Parsers qw(parse_system print_system create_config print_config parse_chassis print_chassis
  parse_interfaces print_interfaces parse_vsystems print_sub_elements parse_vlans print_vlans parse_virtual_routers
  parse_snmp print_snmp parse_applications parse_predefined_and_shared_applications parse_shared_elements print_shared_object_groups);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle trim);
use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
use ZipTie::Typer;
use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK = qw(get_user_group_list);

#use ZipTie::Adapters::PaloAlto::PANOS::GetApplications;

our $VERSION = 1.0;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my $package_name = shift;
	my $backup_doc   = shift;

	$LOGGER->debug('Starting backup operation.');
	
########Create Initial Connection################
	my ($connection_path) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );
	my $connection = ZipTie::Adapters::PaloAlto::PANOS::AutoLogin->new($connection_path);

	$connection->establish_connection();
########End Create Initial Connection################

########Gather Device Responses################
	my $responses    = {};
	my $eval_success = 1;

	$eval_success = eval {
		$responses = _commands($connection);
		1;    #Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};
	if ( !$eval_success ) {
		$LOGGER->fatal("Parsing FAILED while issuing commands: $EVAL_ERROR");
	}
	
	# get QRadar version - (to determine if we model userFilters)
	$responses->{'qradarVersion'} = get_qradar_version();

########End Gather Device Responses################

	#Instantiate the LookupTable object that will be used to pass information between parsers
	$responses->{'lookup_tables'} = ZipTie::Adapters::PaloAlto::PANOS::LookupTable->new();

########Parse Inputs################
	# Grab an output filehandle for the model
	$LOGGER->debug('Opening Filehandle...');
	my $filehandle = get_model_filehandle( 'PANOS', $connection_path->get_ip_address() );

	# initialize the model printer
	$LOGGER->debug('Opening Printer...');
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	$LOGGER->debug('Parsing System Information...');
	parse_system( $responses, $printer );
	print_system( $responses, $printer );

	$LOGGER->debug('Parsing Chassis Information...');
	parse_chassis( $responses, $printer );
	print_chassis( $responses, $printer );

	$LOGGER->debug('Parsing Configuration Information...');
	create_config( $responses, $printer );
	print_config( $responses, $printer );

	$LOGGER->debug('Parsing VLAN Information...');
	parse_vlans( $responses, $printer );
	#Vlans are printed later

	# The "admin virtual system" shared area definitions need to be parsed somewhere here.  It shows
	# up variously elsewhere in the parsing (e.g., SNMP), but we're focusing here on things which will
	# need reference/resolution later in the parsing process.  The parsing here may be moved up in 
	# the event other definitions here need to be referenced earlier in the "print" process.  This 
	# is simply something which needs to go into 'objectGroups' and therefore needs to be printed
	# at this point as of this writing.
	$LOGGER->debug('Parsing shared "global" scope information...');
	parse_shared_elements($responses, $printer);
	print_shared_object_groups($responses, $printer);

	$LOGGER->debug('Parsing Interface Information...');
	parse_interfaces( $responses, $printer );
	print_interfaces( $responses, $printer );

	$LOGGER->debug('Parsing SNMP Information...');
	parse_snmp( $responses, $printer );
	print_snmp( $responses, $printer );

	$LOGGER->debug('Parsing Predefined and Shared Application Information...');
	parse_predefined_and_shared_applications($responses);

	$LOGGER->debug('Parsing Sub-Element Information...');
	parse_vsystems( $responses, $printer );
    #parse_virtual_routers( $responses, $printer );
	print_sub_elements( $responses, $printer);

	#parse vlan information before interface information so that Vlan associatings are available to include
	#in interfaceVlanTrunks elements
	print_vlans( $responses, $printer );

	# Close the model output file handle
	$LOGGER->debug('Closing Printer...');
	$printer->close_model;
	$LOGGER->debug('Printer closed');

	close_model_filehandle($filehandle);
	$LOGGER->debug('Filehandle closed');

	foreach my $key ( keys %{$responses} ) {
		delete $responses->{$key};
	}

########End Parse Inputs################

	$LOGGER->debug('Backup operation complete.');

#	$LOGGER->debug('Getting Applications');
#	ZipTie::Adapters::PaloAlto::PANOS::GetApplications::invoke($package_name, $backup_doc);
#	$LOGGER->debug('GetApplications Done');

	return;
}

sub _commands {
	my ($connection) = @_;

	my $responses;

### REQUIRED COMMANDS ###
	$responses->{'show_system_info'} = $connection->issue_command( '<show><system><info></info></system></show>', 'show system information' );

	$responses->{'show_config_running'} = $connection->issue_command( '<show><config><running></running></config></show>', 'show config running' );

	$responses->{'show_interface_all'} = $connection->issue_command( '<show><interface>all</interface></show>', 'show interfaces' );

	$responses->{'show_system_resources'} = $connection->issue_command( '<show><system><resources></resources></system></show>', 'show system resources' );

	$responses->{'show_predefined_applications'} = $connection->get_custom_url('api/?type=config&action=get&xpath=/config/predefined/application', 'show predefined applications');

	$responses->{'request_dbl_callback'} =
		sub{
			my ($name) = @_;

			my $response = $connection->issue_command(
				"<request><system><external-list><show><name>$name</name></show></external-list></system></request>",
				"request system external-list show name $name"
			);

			return($response);
		};
	
### OPTIONAL COMMANDS ###
	my $eval_success = eval {
		$responses->{'predefined_service_groups'} =
		  $connection->get_custom_url( 'api/?type=config&action=get&xpath=/config/predefined/service', 'show predefined service groups' );
		1;    #Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};
	if ( !$eval_success ) {
		$LOGGER->debug('Retrieval of predefined service groups failed. This is not a required command so the operation can proceed.');
		$LOGGER->debug($EVAL_ERROR);
	}
	return $responses;
}

=begin to be used in future user group modelling
sub get_user_group_list {
	my $responses = shift;

	my $show_user_group_list = $responses->{
		'show_user_group_list'};
	my $userGroupListResponse = $show_user_group_list->findvalue('/response/result');
	my @userGroups = split (/\s(?=cn)/, $userGroupListResponse);
	my @cleanUserGroups = grep { $_ ne ''} @userGroups; # remove any empty values
	$cleanUserGroups[-1] =~ s/\s+Total:\s\d+\s+//; # remove total from last value
	foreach my $userGroup (@cleanUserGroups) { # remove unwanted whitespace
		$userGroup =~ s/^\s+|\s+$//;
	}
    return @cleanUserGroups;
}
=cut


sub get_qradar_version {

	my $version = `/opt/qradar/bin/myver`;
	$version =~ s/\.//g;
	$version =~ /(\d\d\d)/;
	return $1;
}


__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS - ZipTie adapter for Palo Alto PANOS Firewalls and the Panorama management server

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS version 1.0

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS;
    backup ($backup_doc);

=head1 DESCRIPTION

This module is an adapter that allows the ZipTie framework to connect to Palo Alto firewalls and management servers.
It retrieves configuration information from a specified device and transforms it into a Standard Element Document.

=over 4

=item Supported Versions

This adapter supports versions 4.1.0 and above of the Palo Alto PAN-OS operating system.

=item Neighbor Data

Neighbor data is supported via the ZipTie::Adapters::PaloAlto::PANOS::Telemetry module.

=item SNMP Discovery

Matches B<Insert pattern here> in sysDescr.

=item Required Credentials

This adapter requires REST API access privilages to issue the listed commands. In PAN-OS 4.0.0 and earlier SuperUser access is required.
In PAN-OS 4.1.0 either SuperUser or SuperReader can be used. In PAN-OS 5.0.0 and later all role types can be used, however limited roles
may not result in a complete view of the device. SuperReader access is the recommended level of access.

Credentials for the selected level of access should be provided in the Username and Password field in CSM.

=item Connection Protocols

This adapter uses HTTPS to connect to the device.

=back

=head2 Commands Issued

Commands are issued to the device by sending HTTPS get requests to specific URLs on the device.
An initail connection is made and an authentication token is obtained by sending the following:

C<https://Device_IP_Address/api/?type=keygen&user=USERNAME&password=PASSWORD>

Subsequent commands are always sent to C<https://Device_IP_Address/api/> with a query string specifying the command.
The authentication token is appended to the end of the query string using C<&key=AUTHENTICATION_TOKEN>.

All commands issued and responses received are encrypted using the SSL certificate provided by the device.

=over 4

=item Commands issued during a B<Backup> operation

The following REST API commands are prefixed by 'C<?type=op&cmd=>'

'<show><system><info></info></system></show>' - B<REQUIRED>

'<show><config><running></running></config></show>' - B<REQUIRED>

'<show><interface>all</interface></show>' - B<REQUIRED>

'<show><system><resources></resources></system></show>'

When performing a backup of a Panorama management server the above commands are also issued to each device connected
to the management server. Commands issued to sub-devices have '&target=<serial number>' added to the query string.


It also issues:

C<?type=config&action=get&xpath=/config/predefined/service>

C<?type=config&action=get&xpath=/config/predefined/application>

=item Commands issued during a B<Neighbor Data> operation

The following REST API commands are prefixed by 'C<?type=op&cmd=>'

'<show><system><info></info></system></show>' - B<REQUIRED>

'<show><interface>all</interface></show>' - B<REQUIRED>

'<show><mac>all</mac></show>'

'<show><arp>all</arp></show>'

'<show><routing><route></route></routing></show>'

'<show><routing><interface></interface></routing></show>'

'<show><counter><interface>all</interface></counter></show>'

See Also: ZipTie::Adapters::PaloAlto::PANOS::Telemetry

=back


=head1 SUBROUTINES/METHODS

=over 12


=item C<backup($backup_doc)>

This function takes a backup doccument, uses the information that it contains to connect to the specified device, and reads configuration
information from that device. It then transforms that information into Standard Element Document format and prints it into an xml document.
This document provides information about the device's configuration in a standard format to the caller.

=back

=head1 PRIVATE SUB-ROUTINES

N/A

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

This module uses XML::Twig and LWP::UserAgent. These modules must be present on a system for this module to run.
In particular the version of XML::Twig must be the custom version available from IBM.

=head1 DEPENDENCIES

This module requires a version of adapters-common that includes RESTAPI.pm

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

N/A

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: October 2013

=cut

