package ZipTie::Adapters::PaloAlto::PANOS::LookupTable;

use strict;
use warnings;

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();

sub new {
	my ($class) = @_;

	my $self = {
		interface_by_name			=> {},
		vlans_by_interface			=> {},
		service_groups				=> {},
		security_zones				=> {},
		applications				=> {},
		application_filters			=> {},
		application_groups			=> {},
		default_port_service_groups	=> {},
		ipsec_tunnels				=> {},
		management_profiles			=> {},
		management_ip_addr			=> {},
		network_groups					=> {}
	};

	bless $self, $class;
	return $self;
}

sub add_interface {
	my ( $self, $name, $interface ) = @_;
	$self->{interface_by_name}->{$name} = $interface;
	return;
}

sub get_interface {
	my ( $self, $name ) = @_;

	return $self->{interface_by_name}->{$name};
}

sub add_vlan_member {
	my ( $self, $name, $interface ) = @_;

	push @{$self->{'vlans_by_interface'}->{$interface}}, $name;
	return;
}

sub get_interface_vlan_membership {
	my ( $self, $interface ) = @_;

	if (defined $self->{'vlans_by_interface'}->{$interface}) {
		return $self->{'vlans_by_interface'}->{$interface};
	}
	else {
		return;
	}
}

sub add_virtual_wire {
	my ( $self, $name, $first_interface, $second_interface ) = @_;

	$self->{'virtual_wires'}->{$first_interface} = $name;
	$self->{'virtual_wires'}->{$second_interface} = $name;

	return;
}

sub get_virtual_wire {
	my ( $self, $interface_name ) = @_;
	
	if (defined $self->{'virtual_wires'}->{$interface_name}) {
		return $self->{'virtual_wires'}->{$interface_name};
	}
	else {
		return;
	}
}


sub add_service_group {
	my ( $self, $name, $gObject ) = @_;

	my $base_name;
	#determine if the service group to be added is a source half or a destination half.
	#add the group to a hash indexed by the base name of the service group so that the two
	#halves can be related to each other easily
	if ( $name =~ /_source$/ ) {
		($base_name) = $name =~ /(.*)_source/;
		$self->{service_groups}->{$base_name}->{source} = $name;
	} elsif ( $name =~ /_destination$/ ) {
		($base_name) = $name =~ /(.*)_destination/;
		$self->{service_groups}->{$base_name}->{destination} = $name;
	} else {
		$LOGGER->debug("WARNING: Parsing of service group '$name' failed. Source/destination components not setup correctly.");
	}
	
	$self->{service_groups}->{$base_name}->{object_exp} = $gObject if ($gObject);
	
	return;
}

sub get_service_group {
	my ( $self, $base_name ) = @_;
	if ( !defined $self->{service_groups}->{$base_name} ) {
		$LOGGER->debug("WARNING: Service group '$base_name' not found in the lookup table.");
		return;
	}
	return $self->{service_groups}->{$base_name};
}

sub get_service_group_object {
	my ( $self, $base_name ) = @_;
	if ( !defined $self->{service_groups}->{$base_name} ) {
		# try to strip 'source' or 'destination' off of the base name
		if ($base_name =~ /^(.+?)_(source|destination)/)
		{
			$base_name = $1;
		}
		else
		{
			$LOGGER->debug("WARNING: Service group '$base_name' not found in the lookup table.");
			return {};
		}
	}
	# return either object group expansion or a degenerate shell.
	return $self->{service_groups}->{$base_name}->{object_exp} // { id => $self->{service_groups}->{$base_name} };
}

sub get_service_group_direction_set
{
	my ( $self, $base_name) = @_;	
	my $sgSet;
	
	if ($self->get_service_group($base_name))
	{
		for (values (%{$self->{service_groups}->{$base_name}}))
		{
			push (@{$sgSet}, $_);
		}
	}
	return $sgSet;
}

sub add_network_group {
	my ($self, $name, $group) = @_;

	if(!$self->{network_groups}->{$name}){
		$self->{network_groups}->{$name} = $group;
	}else{
		$LOGGER->debug("WARNING: Network group '$name' already exists, not overwriting");
	}
	return;
}

sub is_network_group_defined {
	my ($self, $name) = @_;

	if (defined $self->{network_groups}->{$name}) {
		return 1;
	}
	return;
}

sub add_security_zone {
	my ( $self, $name, $interface_list ) = @_;
	$self->{security_zones}->{$name} = $interface_list;
	return;
}

sub get_security_zone {
	my ( $self, $name ) = @_;
	if ( $name eq 'any' ) {
		return keys %{ $self->{interface_by_name} };
	} elsif ( !defined $self->{security_zones}->{$name} ) {
		$LOGGER->debug("WARNING: Security zone '$name' not found in the lookup table.");
		return;
	}
	return @{ $self->{security_zones}->{$name} };
}

sub applies_to_across_zones {
	my $self	   = shift;
	my @applies_to = ();

	foreach my $source_zone ( keys %{ $self->{security_zones} } ) {
		foreach my $destination_zone ( keys %{ $self->{security_zones} } ) {
			if ( $source_zone eq $destination_zone ) {
				next;
			}
			else {
				push @applies_to, { 'sourceGroup' => $source_zone, 'destinationGroup' => $destination_zone };
			}
		}
	}

	return @applies_to;
}

sub applies_to_within_zones {
	my $self	   = shift;
	my @applies_to = ();

	foreach my $zone ( keys %{ $self->{security_zones} } ) {
		push @applies_to, { 'sourceGroup' => $zone, 'destinationGroup' => $zone };
	}
	return @applies_to;
}

sub add_application{
	my ($self, $name, $application) = @_;

	if(!$self->{applications}->{$name}){
		$self->{applications}->{$name} = $application;
	}else{
		$LOGGER->debug("WARNING: Application '$name' already exists, not overwriting");
	}
	return;
}

sub get_applications{
	my ($self) = @_;
	return($self->{applications});
}

sub add_application_group {
	my ($self, $name, $group) = @_;

	if(!$self->{application_groups}->{$name}){
		$self->{application_groups}->{$name} = $group;
	}else{
		$LOGGER->debug("WARNING: Application group '$name' already exists, not overwriting");
	}
	return;
}

sub is_application_group_defined {
	my ($self, $name) = @_;

	if (defined $self->{application_groups}->{$name}) {
		return 1;
	}
	return;
}

sub add_application_filter{
	my ($self, $name, $filter) = @_;

	if(!$self->{application_filters}->{$name}){
		$self->{application_filters}->{$name} = $filter;
	}else{
		$LOGGER->debug("WARNING: Application filter '$name' already exists, not overwriting");
	}
	return;
}

sub get_application_filter{
	my ($self, $name) = @_;

	if($self->{application_filters}->{$name}){
		return($self->{application_filters}->{$name});
	}
	return;
}

=begin to be used for future user group modelling
sub add_user_group {
	my ($self, $name, $group, $shortName) = @_;

	if(!$self->{user_groups}->{$name}){
		$self->{user_groups}->{$name} = $group;
		$self->{user_groups}->{$name}->{shortName} = $shortName if defined $shortName;
	}else{
		$LOGGER->debug("WARNING: User group '$name' already exists, not overwriting");
	}
	return;
}

sub is_user_group_defined {
	my ($self, $name) = @_;

	if (defined $self->{user_groups}->{$name}) {
		return 1;
	}
	return;
}
sub get_user_shortname {
	my ($self, $name) = @_;
	if (defined $self->{user_groups}->{$name}->{shortName}) {
		return $self->{user_groups}->{$name}->{shortName};
	} else {
		return "";
	}
}
=cut

sub add_default_port_service_group{
	my ($self, $name) = @_;
	my $new_service_group = {
		id => $name,
		description => "The default service(s) for $name"
	};

	if($self->{applications}->{$name}){
		my $application = $self->{applications}->{$name};
		my @keys = keys(%{$application->{service_group}});
		foreach my $key(@keys){
			foreach my $service_group_value(@{$application->{service_group}->{$key}}){
				push(@{$new_service_group->{$key}}, $service_group_value);
			}
		}

	}elsif($self->{application_groups}->{$name}){
		my $application_group = $self->{application_groups}->{$name};

		foreach my $application_name(@{$application_group->{application}}){
			my $application = $self->{applications}->{$application_name};

			my @keys = keys(%{$application->{service_group}});
			foreach my $key(@keys){
				foreach my $service_group_value(@{$application->{service_group}->{$key}}){
					push(@{$new_service_group->{$key}}, $service_group_value);
				}
			}
		}

		foreach my $group_name(@{$application_group->{objectGroupReference}}){
			$self->add_default_port_service_group($group_name);
			push(@{$new_service_group->{objectGroupReference}}, $group_name);
		}

	}elsif($self->{application_filters}->{$name}){
		my $application_filter = $self->{application_filters}->{$name};

		foreach my $application_name(@{$application_filter->{application}}){
			my $application = $self->{applications}->{$application_name};

			my @keys = keys(%{$application->{service_group}});
			foreach my $key(@keys){
				foreach my $service_group_value(@{$application->{service_group}->{$key}}){
					push(@{$new_service_group->{$key}}, $service_group_value);
				}
			}
		}
	}

	$self->{default_port_service_groups}->{$name} = $new_service_group;

	return;
}

sub get_default_port_service_groups{
	my ($self) = @_;
	my @service_groups;
	
	foreach my $key(keys(%{$self->{default_port_service_groups}})){
		push(@service_groups, $self->{default_port_service_groups}->{$key});
	}

	return(\@service_groups);
}

sub add_ipsec_tunnel {
	my ($self, $name, $entry) = @_;
	$self->{ipsec_tunnels}->{$name} = $entry;
	return;
}

sub get_ipsec_tunnel {
	my ($self, $name) = @_;

	if (!defined $self->{ipsec_tunnels}->{$name}) {
		$LOGGER->debug("WARNING: ipSec Tunnel '$name' not found in the lookup table.");
		return;
	}
	return $self->{ipsec_tunnels}->{$name};
}

sub add_interface_management_profile{
	my ($self, $name, $network_group, $service_group) = @_;

	$self->{management_profiles}->{$name} = {
		network_group => $network_group,
		service_group => $service_group
	};
}

sub add_interface_to_management_profile{
	my ($self, $profile_name, $interface_name) = @_;

	if(!defined $self->{management_profiles}->{$profile_name}){
		$LOGGER->debug("WARNING: Interface Management Profile '$profile_name' not found in the lookup table, cannot add '$interface_name'");
	}else{
		push(@{$self->{management_profiles}->{$profile_name}->{interfaces}}, $interface_name);
	}
}

sub get_interface_management_profiles{
	my ($self) = @_;

	my @profiles = ();
	foreach my $profile_key(keys(%{$self->{management_profiles}})){
		$self->{management_profiles}->{$profile_key}->{name} = $profile_key;
		push(@profiles, $self->{management_profiles}->{$profile_key});
	}
	return(\@profiles);
}

sub add_management_ip_addr{
	my ($self, $ip, $mask) = @_;

	if(!defined($mask) || $mask == 32){
		$self->{management_ip_addr} = {host => $ip};
	}else{
		$self->{management_ip_addr} = {
			network => {
				address => $ip,
				mask => $mask
			}
		};
	}
}

sub get_management_ip_addr{
	my ($self) = @_;

	return($self->{management_ip_addr});
}

1;

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::LookupTable - Utility class to pass information between parsers.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::LookupTable version 1.1

=head1 SYNOPSIS

	use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
	my $lookup = ZipTie::Adapters::PaloAlto::PANOS::LookupTable->new();
	$lookup->add_application($name, $application);
	$lookup->add_application_filter($name, $filter);
	$lookup->add_application_group($name, $group);
	$lookup->add_interface($name, $interface);
	$lookup->add_security_zone($name, $interface_list);
	$lookup->add_service_group($name);
	$lookup->add_virtual_wire($name, $first_interface, $second_interface);
	$lookup->add_vlan_member($vlan_name, $interface_name);
	$lookup->get_interface($base_name);
	$lookup->get_interface_vlan_membership($interface_name);
	$lookup->get_security_zone($name);
	$lookup->get_security_zone_to_interface_hash();
	$lookup->get_service_group($name);
	$lookup->get_virtual_wire($interface_name);
	$lookup->get_service_group_object($name);

=head1 DESCRIPTION

This class encapsulated the idea of a set of lookup tables that store parsed information so that it can be passed between parsers.
It prevents the parsers from having to maintain a compicated hash structure for lookup tables, thus reducing the number of lines of
code needed in Parsers.pm. It exposes only the functions needed by parses and maintains its own data integrity in order to make the
use of lookup tables easy.

This class currently implements fonctions for the following lookup tables:

	interface_by_name - which stores interface information indexed by interface name.
	
	security_zones - which stores lists of interface names indexed by the name of the security zone those interfaces belong to.
	
	service_groups - which stores the source and destination components of each service group indexed by the service group name.
	
	application_groups - which stores a list of all application groups defined on the device.
	 
	vlans_by_interface - which stores lists of VLAN names indexed by the name of the interface that the VLAN contains.
	   This is in effect a reverse VLAN lookup table that stores interface-vlan association.
	
	ipsec_tunnels - which stores a list of parsed ipsec tunnels configured on the device indexed by name.
	
	applications - which stores the useful application data for applications, application groups and application filters
	
	default_port_service_groups - which stores the application-default service groups needed for a configuration

	management_profiles - which stores a list of parsed interface management profiles configured on the device, indexed by name.

	management_ip_addr - which stores the management address for later use by interface management profiles.

=head1 SUBROUTINES/METHODS

=over 12


=item C<new()>

Constructor to make a new instance of a LookupTable object.


=item C<add_interface($name, $interface)>

This method takes the name of an interface and a hash representing an interface. It adds the interface to the table of interfaces.


=item C<add_security_zone($name, $interface_list)>

This method takes the name of a security zone and a reference to a list of interface names in that zone. It adds the zone to the table of zones.


=item C<add_service_group($name, $object)>

This method takes the name of a decomposed service group (ie ending in _source or _destination) and stores the servie group in the table of 
service groups. It is indexed by the base name of the service group, that is to say the given name with the _source/_destination portion removed.
$object is optional, if present, a hash reference of groupObj will be created for subsequent lookup.

=item C<get_service_group($name)>

This method takes the name of a service group and returns a hash sontaining the source and destination elements of that group. If
Either the source or destination portion of the gropup is not defined then the cooresponding hash element will not exist. If neither is
defined then a warning is logged and undef (or the empty set) is returned.

=item C<get_service_group_object($name)>

This method takes the name of a service group and returns a hash sontaining the source and destination elements of that group. If
Either the source or destination portion of the gropup is not defined then the cooresponding hash element will not exist. If neither is
defined then a warning is logged and undef (or the empty set) is returned.  If either is defined and an object group expansion was not
registered in add_service group, then a degenerate service group with only the 'id' element defined will be returned.

=item C<get_interface($name)>

This method takes the name of an interface and returns the interface hash associated with it. If the given interface does not exist it returns undef.


=item C<add_virtual_wire($name, $first_interface, $second_interface)>

This method takes the name of a virtual wire and the interfaces that are part of it and stores this information so it can be retrieved.


=item C<add_vlan_member($vlan_name, $interface)>

This method takes the name of a VLAN and the name of an interface that is part of that VLAN and stores that association in the lookup table.


=item C<get_interface_vlan_membership($interface)>

This method takes the name of an interface and returns a list of all on the VLANs that the interface is part of.


=item C<get_security_zone($name)>

This method takes the name of a security zone and returns the list of interfaces associated with it. If the name of the zone is 'any' then ()
is returned to indicate that this zone means all interfaces. If the zone name does not exist a warning is logged.


=item C<get_security_zone_to_interface_hash()>

This method returns a hash, indexed by security zone name, where the value for a particular zone name is a list of
the interfaces that are part of that zone. This is used by NAT.
=item C<get_service_group($base_name)>

This method takes the name of a service group and returns the has containing the source and destination components of that group. If no service
group with the given name exists then a warning is logged and undef is returned.


=item C<get_service_group_direction_set($base_name)>

Given the name of a service group, return a reference to an array with the name of source, dest, and for that matter any other kind
(OK, source and dest) variant names of actual encoded service names. 


=item C<get_virtual_wire($interface_name)>

This method takes the name of an interface and returns the name of the virtual wire that the interface is part of.
If the interface is not part of a virtual wire than undef or an empty list is returned.


=item C<applies_to_across_zones()>

This method returns a set of appliesTo elements. The set contains all combinations of source and destination zones except zones with themselves.
This set of appliesTo elements is used to represent trafic whose source and destination are in different zones.
Its primary use is to describe Palo Alto's default transit policy which allows all trafic within security zones but blocks all
trafic between security zones. 


=item C<applies_to_within_zones()>

This method returns a set of appliesTo elements. The set contains one element for each zone with the source and destination set to that zone.
This set of appliesTo elements is used to represent trafic whose source and destination is within the same zone.
Its primary use is to describe Palo Alto's default transit policy which allows all trafic within security zones but blocks all
trafic between security zones. 


=item C<get_all_zone_names()>

This method returns the names of all security zones. Its primary use is to be able to convert the keyword 'any' into a list of all zones.


=item C<add_application_group($name)>

This method takes the name of an application group and adds it to the table of application groups.


=item C<is_application_group_defined($name)>

This method takes the name of an application group and returns true if that group is defined in the lookup table.
It is used to determin if a given name is a group reference or an application.


=item C<add_ipsec_tunnel($name, $value)>

This method adds a value to the ipsec tunnel lookup table. the value is expected to ba a hash containing parsed ipsec tunnel data.


=item C<get_ipsec_tunnel($name)>

This method returns a the value stored in the ipsec tunnel lookup table that matches the given name of undef if the name is not found.


=item C<add_interface_management_profile($name, $network_group, $service_group)>

This method creates an interface management profile that contains a network group and service group.


=item C<add_interface_to_management_profile($profile_name, $interface_name)>

This method adds an interface to a previously created interface management profile.


=item C<get_interface_management_profiles()>

This method returns an array of the interface management profile hashes and adds each profiles name to the hash.


=item C<add_management_ip_addr($ip, $mask)>

This method holds the management ipAddr element for the system.


=item C<get_management_ip_addr()>

This method gets the management ipAddr element for the system.


=back

=head1 DIAGNOSTICS

These warnings may appear in the adapter log file in /usr/share/ziptie-server/osgi-config/scratch/logs

=over 12

=item C<WARNING: Parsing of service group '$name' failed. Source/destination components not setup correctly.>

This warning indicates that something has gone wrong with service group parsing. It occurs when the adapter tries to save a
service group without either _source or _destination as a postfix. All service groups are broken down into source and destination
components by default. This message will only appear if this has failed.

=item C<WARNING: Service group '$base_name' not found in the lookup table.>

This warning indicates that the adapter attempted to lookup a service group that does not exist. It will appear in conjunction with the
above warning if parsing of a group has failed. If it appears alone it may indicate that one or more service groups were defined in
an unexpected place in the XML.

=item C<WARNING: Security zone '$name' not found in the lookup table.>

This warning indicates that the adapter attempted to lookup a security zone that does not exist. It may indicate a parse failure or that
one or more security zones were defined in an unexpected place in the XML.


=back

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

N/A

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

N/A

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com, Jak Devaney john.devaney@uk.ibm.com

Date: October 2013

=cut
