package ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils;

use strict;
use warnings;

use 5.0100000;    # use perl version 5.10
use Data::Dumper;
use ZipTie::Adapters::Utils qw(parseCIDR);
use ZipTie::Logger;
use XML::Twig::XPath;

our @EXPORT_OK = qw(print_if_defined get_value_from_xml unwind_port_expression parse_cidr_panos get_interface_type_panos);
use Exporter 'import';

our $VERSION = 1.0;

my $LOGGER = ZipTie::Logger::get_logger();

# This function takes a hash, a key and an output printer and prints the SED element identified by the key
# if that element is defined in the hash.
# Inputs:
# $element is a hash of parsed data where the keys are names of SED elements
# $key is the name of a SED element of be printed
# $out is the output printer where the SED element will printed

sub print_if_defined {
	my ( $element, $key, $out ) = @_;

	if ( defined( $element->{$key} ) ) {
		$out->print_element( $key, $element->{$key} );
	}
	return;
}

sub get_value_from_xml {
	my $path = shift;
	my $xml  = shift;

	my $value = $xml->findvalue($path);

	# On my local machine the findvalue function returns a string, while on my QRM VM it returns an XML::XPath::Literal
	# object. I'm not sure what is causing this difference, but this if statement deals with it.
	if ( ref($value) eq 'XML::XPath::Literal' ) {
		$value = $value->value();
	}

	if ( $value eq q{} ) {
		$value = undef;
	}

	return $value;
}

#given a string containing a port expression, produce a list of ports/portRanges
sub unwind_port_expression {
	my $expression = shift;

	my @ports;
	my @port_ranges;

	#if there is more that one port or range defined they will be separated by commas
	my @list_entries = split ',', $expression;

	foreach my $entry (@list_entries) {

		#if this item is a range it will be separated by a dash
		if ( $entry =~ /-/ ) {
			push @port_ranges, $entry;

			#otherwise it is a singel port
		} else {
			push @ports, $entry;
		}
	}

	return \@ports, \@port_ranges;

}    # end _unwind_port_expression

sub parse_cidr_panos
{
	my $input = shift;
	my $parsedCIDR;
	
	# If we only have an IP address, just assign it and code a /32.
	if ($input =~ /^([\d\.]+)$/)
	{
		$parsedCIDR = { 'host' => $1, 
					'prefix' => 32 };
	}
	else
	{
		$parsedCIDR = parseCIDR($input);
	}	
	return $parsedCIDR;		
}

sub get_interface_type_panos{
	my ($name) = @_;
	my $type = 'unknown';

	if($name =~ /^ethernet/){
		$type = 'ethernet';
	}elsif($name =~ /^ae|vlan/){
		$type = 'virtual';
	}elsif($name =~ /^tunnel/){
		$type = 'tunnel';
	}elsif($name =~ /^loopback/){
		$type = 'softwareLoopback';
	}elsif($name =~ /^ha/){
		$type = 'other';
	}else{
		$LOGGER->debug("Cannot deduce interface type from name '$name', setting to '$type'");
	}

	return($type);
}
1;
__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils - Subroutines to facilitate
parsing of PANOS configurations.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils version 1.0

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils;
    get_value_from_xml($path, $xml);
    print_if_defined($element, $key, $out );    
    unwind_port_expression($expression);


=head1 DESCRIPTION  

These are routines at global namespace scope which transcend any
specific element being parsed or generated into the Standard Element Document (SED).
Or, the routines are common across multiple similar SED elements.


=head1 SUBROUTINES/METHODS

=over 12

=item C<get_value_from_xml($path, $xml)>

This function takes an xpath string and an XML Element. It finds the value at the given xpath and returns it. If the value is blank, it
returns undef. Depending on the configuration environment the function 'findvalue' may return a 'XML::XPath::Literal' or a string.
if it returns a 'XML::XPath::Literal' then it is converted to a string before it is returned.


=item C<print_if_defined($element, $key, $out)>

This function takes a hash $element and checks to see if the key value $key is defined. If it is it prints that sub element as
a SED element with the name $key. This is used where SED elements are out of order and thus can not be printed as a block.


=item C<unwind_port_expression($expression)>

This function takes a string representing a set of ports and converts it into two lists. The first list contains all of the individual
ports in the list and the second list contains all the port ranges in the list.

The set is expected to contain elements separated by commas.

 Elements that represent single ports are single numbers.
 Elements that represent port ranges are pairs of numbers separated by a dash.


=item C<parse_cidr_panos($input)>

This function is an interlude to the ZAF parseCIDR, but takes into account that a /32 can be an 
optional thing for a panos address expression.

=back

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

N/A

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.


=head1 LICENSE AND COPYRIGHT

    Copyright (c) 2013, IBM Corporation. All Rights Reserved.
    This module is free software. It may be used, redistributed
    and/or modified under the same terms as Perl itself.


=head1 AUTHOR

    Wayne Tackabury (wtackabury@us.ibm.com), David Tickner (dtickner@ca.ibm.com)


=cut
