package ZipTie::Adapters::Utils;

use strict;
use warnings;
use v5.10;

use Exporter 'import';
use Archive::Tar;
use Compress::Zlib;
use File::Temp qw(tmpnam);
use File::Temp;
use Net::Ping;
use Time::Local;

use ZipTie::Logger;

our @EXPORT_OK =
  qw(create_network_element invert_mask is_contiguous_mask choose_admin_ip mask_to_bits net_addr_to_mask seconds_since_epoch strip_mac get_interface_type get_mask get_protocol_name get_port_number int_port trim get_model_filehandle close_model_filehandle get_cli_commands_filehandle parseCIDR bin2dec parse_targz_data get_crep create_empty_file getUnitFreeNumber merge_hashes escape_filename create_unique_filename get_port_expression get_any_address_encoding get_any_address_address get_protocol_number get_any ncname_from_string open_telemetry_neighbors close_telemetry_neighbors);

our %SeenMerged = ();    # this hash is used in merge_hashes sub and doesn't have to be exported

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

our $OS = $Config::Config{'osname'};

our %months = (
	JAN => 0,
	FEB => 1,
	MAR => 2,
	APR => 3,
	MAY => 4,
	JUN => 5,
	JUL => 6,
	AUG => 7,
	SEP => 8,
	OCT => 9,
	NOV => 10,
	DEC => 11,
);

our %timezones = (
	AEST     => +10,
	UTC      => 0,
	UTZ      => 0,
	EDT      => -4,
	EST      => -5,
	MICHIGAN => -5,
	CDT      => -5,
	CENTRAL  => -6,
	CST      => -6,
	MDT      => -6,
	MST      => -7,
	MOUNTAIN => -7,
	ARIZONA  => -7,
	PDT      => -7,
	PST      => -8,
	PACIFIC  => -8,
	AKDT     => -8,
	AKST     => -9,
	ALASKA   => -9,
	HAST     => -10,
	HAWAII   => -10,
	SAMOA    => -11,
);

# This maps *raw IP protocol numbers* to the corresponding named IP protocol.  This cannot
# extend beyond the enumeration within the ZipTie schema *core* protocol type enums, since
# this is the scoping contract between ZipTie and QRM.
#
#  N.B.!!! This only covers IPv4!  IPv6 has overlapping definitions.
# 
my $ipProtocolNames = {
	1		=>  "icmp",
	2       =>  "igmp",
	3		=>  "ggp",
	4       =>  "ip",  # "ip in IP" encapsulation, actually
	5       =>  "st",
	6		=>  "tcp",
	8       =>  "egp",
	9       =>  "igrp",
	17		=>  "udp",
	33		=>  "dccp",
	41		=>  "ipv6",  # encapsulation, RFC2473
	46      =>  "rsvp",
	47      =>  "gre",
	50		=>  "esp",
	51      =>  "ah",
	57      =>  "skip",
	88		=>  "eigrp",
	89		=>  "ospf",
	103		=>  "pim",
	112		=>  "vrrp",
	115     =>  "l2tp",
	132 	=>  "sctp"	
};

my $portNumbers = {
	"aol"         => 5190,
	"aol1"        => 5191,
	"aol2"        => 5192,
	"aol3"        => 5193,
	"bgp"         => 179,
	"biff"        => 512,
	"bootpc"      => 68,
	"bootps"      => 67,
	"destination-unreachable" => 3,
	"dhcp"        => 547,
	"dhcp-relay"  => 547,
	"chargen"     => 19,
	"cmd"         => 514,
	"cvspserver"  => 2401,
	"daytime"     => 13,
	"discard"     => 9,
	"dnsix"       => 195,
	"domain"      => 53,
	"dns"         => 53,
	"echo"        => 8,    # was 7 -- if this is ICMP echo, see RFC 792 - Wayne
	"echo-reply"  => 0,
	"eklogin"     => 2105,
	"ekshell"     => 2106,
	"exec"        => 512,
	"finger"      => 79,
	"ftp-data"    => 20,
	"ftp"         => 21,
	"gopher"      => 70,
	"h323"        => 1720,
	"hostname"    => 101,
	"ident"       => 113,
	"imap"        => 143,
	"imap4"       => 143,
	"irc"         => 194,
	"isakmp"      => 500,
	"non500-isakmp"      => 4500,
	"klogin"      => 543,
	"kshell"      => 544,
	"ldap"        => 389,
	"ldp"         => 646,
	"login"       => 513,
	"lpd"         => 515,
	"lotusnotes"  => 1352,
	"lotus notes" => 1352,
	"mobile-redirect" => 32,
	"mobile-registration" => 35,
	"mobile-registration-reply" => 36,
	"mobile-ip"   => 434,
	"msdp"        => 639,
	"nameserver"  => 42,
	"netbios-dgm" => 138,
	"netbios-ns"  => 137,
	"netbios-ss"  => 139,
	"netbios-ssn" => 139,
	"nntp"        => 119,
	"ntalk"       => 518,
	"ntp"         => 123,
	"pim-auto-rp" => 496,
	"pop2"        => 109,
	"pop3"        => 110,
	"radius"      => 1645,
	"radius-acct" => 1646,
	"redirect"    => 5,
	"rip"         => 520,
	"rkinit"      => 2108,
	"smtp"        => 25,
	"snmp"        => 161,
	"snmptrap"    => 162,
	"source-quench" => 4,
	"sqlnet"      => 1521,
	"ssh"         => 22,
	"ssl"         => 443,
	"sunrpc"      => 111,
	"sunrpc"      => 111,
	"syslog"      => 514,
	"tacacs"      => 49,
	"talk"        => 517,
	"telnet"      => 23,
	"tftp"        => 69,
	"time"        => 37,
	"time-exceeded" => 11,
	"timestamp"   => 13,  # ICMP
        "timestamp-reply" => 14,
	"traceroute"  => 30,  # ICMP tracerouter, per RFC1393
	"uucp"        => 540,
	"who"         => 513,
	"whois"       => 43,
	"www"         => 80,
	"http"        => 80,
	"https"       => 443,
	"xdmcp"       => 177,
	"x-windows"   => 6000,
	"voip"		  => 5060,
	"krb-prop"    => 754,
	"kpasswd"     => 761,
	"klogin"      => 543,
	"krbupdate"   => 760,
	"pcanywhere-data"   => 5631,
	"pcanywhere-status" => 5632,
};

# mostly taken from IANA port assignments, which means it's a little over-zealous with protocol coverage
my $portProtocols = {
	"aol"         => [{portRange=>{portStart=>5190,portEnd=>5193,protocol=> 'tcp',},},],
	"bgp"         => [{portExpression=>{operator=>'eq',port=>179,protocol=> 'tcp',},},],
	"biff"        => [{portExpression=>{operator=>'eq',port=>512,protocol=> 'udp',},},],
	"bootpc"      => [{portExpression=>{operator=>'eq',port=>68,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>68,protocol=> 'tcp',},},],
	"bootps"      => [{portExpression=>{operator=>'eq',port=>67,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>67,protocol=> 'tcp',},},],
	"chargen"     => [{portExpression=>{operator=>'eq',port=>19,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>19,protocol=> 'tcp',},},],
	"citrix-ica"   => [{portExpression=>{operator=>'eq',port=>1494,protocol=> 'tcp',},},],
	"ctiqbe"      => [{portExpression=>{operator=>'eq',port=>2748,protocol=> 'tcp',},},],
	"cmd"         => [{portExpression=>{operator=>'eq',port=>514,protocol=> 'tcp',},},],
	"cvspserver"  => [{portExpression=>{operator=>'eq',port=>2401,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2401,protocol=> 'tcp',},},],
	"daytime"     => [{portExpression=>{operator=>'eq',port=>13,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>13,protocol=> 'tcp',},},],
	"dhcp"        => [{portExpression=>{operator=>'eq',port=>547,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>547,protocol=> 'tcp',},},],
	"dhcp-relay"  => [{portExpression=>{operator=>'eq',port=>67,protocol=> 'udp',},},],
	"dhcp-client"  => [{portExpression=>{operator=>'eq',port=>67,protocol=> 'udp',},},],
	"dhcp-server"  => [{portExpression=>{operator=>'eq',port=>67,protocol=> 'udp',},},],
	"discard"     => [{portExpression=>{operator=>'eq',port=>9,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>9,protocol=> 'tcp',},},],
	"dnsix"       => [{portExpression=>{operator=>'eq',port=>195,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>195,protocol=> 'tcp',},},],
	"domain"      => [{portExpression=>{operator=>'eq',port=>53,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>53,protocol=> 'tcp',},},],
	"dns"         => [{portExpression=>{operator=>'eq',port=>53,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>53,protocol=> 'tcp',},},],
	"dns-tcp"         => [{portExpression=>{operator=>'eq',port=>53,protocol=> 'tcp',},},],
	"dns-udp"         => [{portExpression=>{operator=>'eq',port=>53,protocol=> 'udp',},},],
	"echo"        => [{portExpression=>{operator=>'eq',port=>7,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>7,protocol=> 'tcp',},},],
	"eklogin"     => [{portExpression=>{operator=>'eq',port=>2105,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2105,protocol=> 'tcp',},},],
	"ekshell"     => [{portExpression=>{operator=>'eq',port=>2106,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2106,protocol=> 'tcp',},},],
	"exec"        => [{portExpression=>{operator=>'eq',port=>512,protocol=> 'tcp',},},],
	"finger"      => [{portExpression=>{operator=>'eq',port=>79,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>79,protocol=> 'tcp',},},],
	"ftp-data"    => [{portExpression=>{operator=>'eq',port=>20,protocol=> 'tcp',},},],
	"ftp"         => [{portExpression=>{operator=>'eq',port=>21,protocol=> 'tcp',},},],
	"gopher"      => [{portExpression=>{operator=>'eq',port=>70,protocol=> 'tcp',},},],
	"gnutella"      => [{portExpression=>{operator=>'eq',port=>6346,protocol=> 'tcp',},},],
	"gprs-gtp"      => [{portExpression=>{operator=>'eq',port=>2123,protocol=> 'udp',},},],
	"gprs-gtp-c"      => [{portExpression=>{operator=>'eq',port=>2123,protocol=> 'udp',},},],
	"gprs-gtp-c-udp"      => [{portExpression=>{operator=>'eq',port=>2123,protocol=> 'udp',},},],
	"gprs-gtp-c-tcp"      => [{portExpression=>{operator=>'eq',port=>2123,protocol=> 'tcp',},},],
	"gprs-gtp-u"      => [{portExpression=>{operator=>'eq',port=>2152,protocol=> 'udp',},},],
	"gprs-gtp-u-udp"      => [{portExpression=>{operator=>'eq',port=>2152,protocol=> 'udp',},},],
	"gprs-gtp-u-tcp"      => [{portExpression=>{operator=>'eq',port=>2152,protocol=> 'tcp',},},],
	"gprs-gtp-v0"      => [{portExpression=>{operator=>'eq',port=>3386,protocol=> 'udp',},},],
	"gprs-gtp-v0-udp"      => [{portExpression=>{operator=>'eq',port=>3386,protocol=> 'udp',},},],
	"gprs-gtp-v0-tcp"      => [{portExpression=>{operator=>'eq',port=>3386,protocol=> 'tcp',},},],
	"gre"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'gre',},},],
	"h323"      => [{portRange=>{portStart=>1719,portEnd=>1720,protocol=> 'tcp',},},],
	"hostname"    => [{portExpression=>{operator=>'eq',port=>101,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>101,protocol=> 'tcp',},},],
	"http"        => [{portExpression=>{operator=>'eq',port=>80,protocol=> 'tcp',},},],
	"http-ext"        => [{portExpression=>{operator=>'eq',port=>8080,protocol=> 'tcp',},},],
	"https"       => [{portExpression=>{operator=>'eq',port=>443,protocol=> 'tcp',},},],
	"icmp"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'icmp',},},],
	"icmp-all"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'icmp',},},],
	"icmp-ping"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'icmp',},},],
	"ident"       => [{portExpression=>{operator=>'eq',port=>113,protocol=> 'tcp',},},],
	"ike"       => [{portExpression=>{operator=>'eq',port=>500,protocol=> 'udp',},},],
	"ike-nat"       => [{portExpression=>{operator=>'eq',port=>4500,protocol=> 'udp',},},],
	"imap"        => [{portExpression=>{operator=>'eq',port=>143,protocol=> 'tcp',},},],
	"imaps"       => [{portExpression=>{operator=>'eq',port=>193,protocol=> 'tcp',},},],
	"imap4"        => [{portExpression=>{operator=>'eq',port=>143,protocol=> 'tcp',},},],
	"internet-locator-service"       => [{portExpression=>{operator=>'eq',port=>389,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>1002,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>1503,protocol=> 'tcp',},},],
	"irc"         => [{portExpression=>{operator=>'eq',port=>194,protocol=> 'tcp',},},],
	"isakmp"      => [{portExpression=>{operator=>'eq',port=>500,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>500,protocol=> 'tcp',},},],
	"non500-isakmp"      => [{portExpression=>{operator=>'eq',port=>4500,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>4500,protocol=> 'tcp',},},],
	"kerberos"         => [{portExpression=>{operator=>'eq',port=>750,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>750,protocol=> 'udp',},}],
	"klogin"      => [{portExpression=>{operator=>'eq',port=>543,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>543,protocol=> 'tcp',},},],
	"kpasswd"     => [{portExpression=>{operator=>'eq',port=>761,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>761,protocol=> 'tcp',},},],
	"krb-prop"    => [{portExpression=>{operator=>'eq',port=>754,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>754,protocol=> 'tcp',},},],
	"krbupdate"   => [{portExpression=>{operator=>'eq',port=>760,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>760,protocol=> 'tcp',},},],
	"kshell"      => [{portExpression=>{operator=>'eq',port=>544,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>544,protocol=> 'tcp',},},],
	"l2tp"       => [{portExpression=>{operator=>'eq',port=>1701,protocol=> 'tcp',},},],
	"ldap"        => [{portExpression=>{operator=>'eq',port=>389,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>389,protocol=> 'tcp',},},],
	"ldaps"       => [{portExpression=>{operator=>'eq',port=>636,protocol=> 'udp',},},],
	"ldp"         => [{portExpression=>{operator=>'eq',port=>646,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>646,protocol=> 'tcp',},},],
	"login"       => [{portExpression=>{operator=>'eq',port=>513,protocol=> 'tcp',},},],
	"lpd"         => [{portExpression=>{operator=>'eq',port=>515,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>515,protocol=> 'tcp',},},],
	"lpr"         => [{portExpression=>{operator=>'eq',port=>515,protocol=> 'tcp',},},],
	"lpd-udp"         => [{portExpression=>{operator=>'eq',port=>515,protocol=> 'udp',},},],
	"lpd-tcp"         => [{portExpression=>{operator=>'eq',port=>515,protocol=> 'tcp',},},],
	"ldp-udp"         => [{portExpression=>{operator=>'eq',port=>646,protocol=> 'udp',},},],
	"ldp-tcp"         => [{portExpression=>{operator=>'eq',port=>646,protocol=> 'tcp',},},],
	"lotusnotes"  => [{portExpression=>{operator=>'eq',port=>1352,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>1352,protocol=> 'tcp',},},],
	"lotus notes" => [{portExpression=>{operator=>'eq',port=>1352,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>1352,protocol=> 'tcp',},},],
	"mobile-ip"   => [{portExpression=>{operator=>'eq',port=>434,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>434,protocol=> 'tcp',},},],
	"mgcp" => [{portExpression=>{operator=>'eq',port=>2427,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>2427,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2727,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>2727,protocol=> 'udp',},},],
	"mgcp-ua"       => [{portExpression=>{operator=>'eq',port=>2427,protocol=> 'udp',},},],
	"mgcp-ca" => [{portExpression=>{operator=>'eq',port=>2727,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>2727,protocol=> 'udp',},},],
	"msdp"        => [{portExpression=>{operator=>'eq',port=>434,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>434,protocol=> 'tcp',},},],
	"ms-rpc"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-epm"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-msexchange"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-msexchange-directory-nsp"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-msexchange-directory-rfr"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-msexchange-info-store"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-tcp"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'tcp',},},],
	"ms-rpc-udp"       => [{portExpression=>{operator=>'eq',port=>135,protocol=> 'udp',},},],
	"ms-sql"       => [{portExpression=>{operator=>'eq',port=>1433,protocol=> 'tcp',},},],
	"msn"       => [{portExpression=>{operator=>'eq',port=>1863,protocol=> 'tcp',},},],
	"nameserver"  => [{portExpression=>{operator=>'eq',port=>42,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>42,protocol=> 'tcp',},},],
	"nbds"       => [{portExpression=>{operator=>'eq',port=>138,protocol=> 'udp',},},],
	"nbname"       => [{portExpression=>{operator=>'eq',port=>137,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>137,protocol=> 'udp',},},],
	"netbios-dgm" => [{portExpression=>{operator=>'eq',port=>138,protocol=> 'udp',},},],
	"netbios-ns"  => [{portExpression=>{operator=>'eq',port=>137,protocol=> 'udp',},},],
	"netbios-ss"  => [{portExpression=>{operator=>'eq',port=>139,protocol=> 'tcp',},},],
	"netbios-ssn" => [{portExpression=>{operator=>'eq',port=>139,protocol=> 'tcp',},},],
	"netbios-session"       => [{portExpression=>{operator=>'eq',port=>139,protocol=> 'tcp',},},],
	"netbios"      => [{portRange=>{portStart=>137,portEnd=>139,protocol=> 'udp',},},],
	"nfs"       => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},],
	"nfsd-tcp"       => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},],
	"nfsd-udp"       => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},],
	"nntp"        => [{portExpression=>{operator=>'eq',port=>119,protocol=> 'tcp',},},],
	"nsm"       => [{portExpression=>{operator=>'eq',port=>8443,protocol=> 'udp',},},],
	"ntalk"       => [{portExpression=>{operator=>'eq',port=>518,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>518,protocol=> 'tcp',},},],
	"ntp"         => [{portExpression=>{operator=>'eq',port=>123,protocol=> 'udp',},},],
	"ospf"       => [{portExpression=>{operator=>'eq',port=>89,protocol=> 'ospf',},},],
	"pc-anywhere"       => [{portExpression=>{operator=>'eq',port=>5632,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>5631,protocol=> 'udp',},},],
	"pcanywhere-status" => [{portExpression=>{operator=>'eq',port=>5632,protocol=> 'udp',},},],
	"pcanywhere-data"        => [{portExpression=>{operator=>'eq',port=>5631,protocol=> 'tcp',},},],
	"persistent-nat"		  => [{portExpression=>{operator=>'eq',port=>3478,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>3478,protocol=> 'tcp',},},],
	"pim-auto-rp" => [{portExpression=>{operator=>'eq',port=>496,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>496,protocol=> 'tcp',},},],
	"ping"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'icmp',},},],
	"pingv6"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'icmp6',},},],
	"pop2"        => [{portExpression=>{operator=>'eq',port=>102,protocol=> 'tcp',},},],
	"pop3"        => [{portExpression=>{operator=>'eq',port=>110,protocol=> 'tcp',},},],
	"pptp"       => [{portExpression=>{operator=>'eq',port=>1723,protocol=> 'tcp',},},],
	"printer"         => [{portExpression=>{operator=>'eq',port=>646,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>646,protocol=> 'tcp',},},],
	"radius"      => [{portExpression=>{operator=>'eq',port=>1812,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>1812,protocol=> 'tcp',},},],
	"radius-acct" => [{portExpression=>{operator=>'eq',port=>1813,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>1813,protocol=> 'tcp',},},],
	"radacct" => [{portExpression=>{operator=>'eq',port=>1813,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>1813,protocol=> 'tcp',},},],
	"rcp"        => [{portExpression=>{operator=>'eq',port=>514,protocol=> 'tcp',},},],  # Shares port 514 w/rsh
	"realaudio"       => [{portExpression=>{operator=>'eq',port=>7070,protocol=> 'tcp',},},{portExpression=>{operator=>'eq',port=>554,protocol=> 'tcp',},},],
	"rexec"        => [{portExpression=>{operator=>'eq',port=>512,protocol=> 'tcp',},},],
	"rip"         => [{portExpression=>{operator=>'eq',port=>520,protocol=> 'udp',},},],
	"rkinit"      => [{portExpression=>{operator=>'eq',port=>2108,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2108,protocol=> 'tcp',},},],
	"rlogin"	=> [{portExpression=>{operator=>'eq',port=>513,protocol=> 'tcp',},},],
	"rtsp"       => [{portExpression=>{operator=>'eq',port=>554,protocol=> 'tcp',},},],
	"rsh"        => [{portExpression=>{operator=>'eq',port=>514,protocol=> 'tcp',},},],
	"sccp"       => [{portExpression=>{operator=>'eq',port=>2000,protocol=> 'tcp',},},],
	"secureid-udp" => [{portExpression=>{operator=>'eq',port=>5510,protocol=> 'udp',},},], 
	"sip"		  => [{portExpression=>{operator=>'eq',port=>5060,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>5060,protocol=> 'tcp',},},],
	"smb"       => [{portExpression=>{operator=>'eq',port=>445,protocol=> 'tcp',},},],
	"smtp"        => [{portExpression=>{operator=>'eq',port=>25,protocol=> 'tcp',},},],
	"snmp"        => [{portExpression=>{operator=>'eq',port=>161,protocol=> 'udp',},},],
	"snmptrap"    => [{portExpression=>{operator=>'eq',port=>162,protocol=> 'udp',},},],
	"snmp-agentx"  => [{portExpression=>{operator=>'eq',port=>705,protocol=> 'udp',},},], 
	"snpp"        => [{portExpression=>{operator=>'eq',port=>444,protocol=> 'tcp',},},],
	"snmptrap"    => [{portExpression=>{operator=>'eq',port=>162,protocol=> 'udp',},},],
	"sqlnet"       => [{portExpression=>{operator=>'eq',port=>1521,protocol=> 'tcp',},},],
	"sqlnet-v1"       => [{portExpression=>{operator=>'eq',port=>1521,protocol=> 'tcp',},},],
	"sqlnet-v2"       => [{portExpression=>{operator=>'eq',port=>1521,protocol=> 'tcp',},},],
	"sql-monitor"       => [{portExpression=>{operator=>'eq',port=>1434,protocol=> 'tcp',},},],
	"ssh"         => [{portExpression=>{operator=>'eq',port=>22,protocol=> 'tcp',},},],
	"ssl"         => [{portExpression=>{operator=>'eq',port=>443,protocol=> 'tcp',},},],
	"stun"		  => [{portExpression=>{operator=>'eq',port=>3478,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>3478,protocol=> 'tcp',},},],
	"sunrpc"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>111,protocol=> 'tcp',},},],
	"sun-rpc"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>111,protocol=> 'tcp',},},],
	"sun-rpc-tcp"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'tcp',},},],
	"sun-rpc-udp"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'udp',},},],
	"sun-rpc-nfs"      => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>2049,protocol=> 'tcp',},},],
	"sun-rpc-nfs-tcp"      => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'tcp',},},],
	"sun-rpc-nfs-udp"      => [{portExpression=>{operator=>'eq',port=>2049,protocol=> 'udp',},},],
	"sun-rpc-portmap"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>111,protocol=> 'tcp',},},],
	"sun-rpc-portmap-tcp"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'tcp',},},],
	"sun-rpc-portmap-udp"      => [{portExpression=>{operator=>'eq',port=>111,protocol=> 'udp',},},],
	"syslog"      => [{portExpression=>{operator=>'eq',port=>514,protocol=> 'udp',},},],
	"tacacs"      => [{portExpression=>{operator=>'eq',port=>49,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>49,protocol=> 'tcp',},},],
	"tacacs-ds"        => [{portExpression=>{operator=>'eq',port=>65,protocol=> 'tcp',},},],
	"talk"        => [{portExpression=>{operator=>'eq',port=>517,protocol=> 'udp',},},],
	"telnet"      => [{portExpression=>{operator=>'eq',port=>23,protocol=> 'tcp',},},],
	"tftp"        => [{portExpression=>{operator=>'eq',port=>69,protocol=> 'udp',},},],
	"time"        => [{portExpression=>{operator=>'eq',port=>37,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>37,protocol=> 'tcp',},},],
	"uucp"        => [{portExpression=>{operator=>'eq',port=>540,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>540,protocol=> 'tcp',},},],
	"vdo-live"      => [{portExpression=>{operator=>'eq',port=>7000,protocol=> 'tcp',},},],
	"vnc"      => [{portExpression=>{operator=>'eq',port=>5900,protocol=> 'tcp',},},],
	"voip"		  => [{portExpression=>{operator=>'eq',port=>5060,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>5060,protocol=> 'tcp',},},],
	"wais"      => [{portExpression=>{operator=>'eq',port=>210,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>210,protocol=> 'tcp',},},],
	"who"         => [{portExpression=>{operator=>'eq',port=>513,protocol=> 'udp',},},],
	"whois"       => [{portExpression=>{operator=>'eq',port=>43,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>43,protocol=> 'tcp',},},],
	"www"         => [{portExpression=>{operator=>'eq',port=>80,protocol=> 'tcp',},},],
	"xdmcp"       => [{portExpression=>{operator=>'eq',port=>177,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>177,protocol=> 'tcp',},},],
	"x-windows"   => [{portExpression=>{operator=>'eq',port=>6000,protocol=> 'udp',},},{portExpression=>{operator=>'eq',port=>6000,protocol=> 'tcp',},},],
	"xnm-clear-text"      => [{portExpression=>{operator=>'eq',port=>3221,protocol=> 'tcp',},},], 
	"xnm-ssl"      => [{portExpression=>{operator=>'eq',port=>3220,protocol=> 'tcp',},},],
	"ymsg"      => [{portExpression=>{operator=>'eq',port=>5050,protocol=> 'tcp',},},],

	"tcp-any"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'tcp',},},],
	"udp-any"      => [{portRange=>{portStart=>0,portEnd=>65535,protocol=> 'udp',},},]
};

my $masks = {
	0  => "0.0.0.0",
	1  => "128.0.0.0",
	2  => "192.0.0.0",
	3  => "224.0.0.0",
	4  => "240.0.0.0",
	5  => "248.0.0.0",
	6  => "252.0.0.0",
	7  => "254.0.0.0",
	8  => "255.0.0.0",
	9  => "255.128.0.0",
	10 => "255.192.0.0",
	11 => "255.224.0.0",
	12 => "255.240.0.0",
	13 => "255.248.0.0",
	14 => "255.252.0.0",
	15 => "255.254.0.0",
	16 => "255.255.0.0",
	17 => "255.255.128.0",
	18 => "255.255.192.0",
	19 => "255.255.224.0",
	20 => "255.255.240.0",
	21 => "255.255.248.0",
	22 => "255.255.252.0",
	23 => "255.255.254.0",
	24 => "255.255.255.0",
	25 => "255.255.255.128",
	26 => "255.255.255.192",
	27 => "255.255.255.224",
	28 => "255.255.255.240",
	29 => "255.255.255.248",
	30 => "255.255.255.252",
	31 => "255.255.255.254",
	32 => "255.255.255.255",
};

my $reverse_masks = {
    "0.0.0.0" => 0,
    "128.0.0.0" => 1,
    "192.0.0.0" => 2,
    "224.0.0.0" => 3,
    "240.0.0.0" => 4,
    "248.0.0.0" => 5,
    "252.0.0.0" => 6,
    "254.0.0.0" => 7,
    "255.0.0.0" => 8,
    "255.128.0.0" => 9,
    "255.192.0.0" => 10,
    "255.224.0.0" => 11,
    "255.240.0.0" => 12,
    "255.248.0.0" => 13,
    "255.252.0.0" => 14,
    "255.254.0.0" => 15,
    "255.255.0.0" => 16,
    "255.255.128.0" => 17,
    "255.255.192.0" => 18,
    "255.255.224.0" => 19,
    "255.255.240.0" => 20,
    "255.255.248.0" => 21,
    "255.255.252.0" => 22,
    "255.255.254.0" => 23,
    "255.255.255.0" => 24,
    "255.255.255.128" => 25,
    "255.255.255.192" => 26,
    "255.255.255.224" => 27,
    "255.255.255.240" => 28,
    "255.255.255.248" => 29,
    "255.255.255.252" => 30,
    "255.255.255.254" => 31,
    "255.255.255.255" => 32,
};


# common regex patterns
our $creps = {
	'cipm'  => '\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}',         # common ip/mask regular expression
	'ecipm' => '\d{1,3}(?:\.\d{1,3}){0,3}',                  # extended ip/mask regular expression
	'cidr'  => '\d{1,3}(?:\.\d{1,3}){0,3}\/\d+',             # classless interdomain routing address regular expression
	'ipv6'  => '[\da-f]{0,4}(?:\:{1,2}[\da-f]{0,4}){1,7}',   # ipv6 regular expression
	'mac1'  => '[0-9a-f]{12}',                               # mac address format 1
	'mac2'  => '[0-9a-f]{1,2}(?:[-:\.][0-9a-f]{1,2}){5}',    # mac address format 2
	'mac3'  => '[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}',      # mac address format 3
	'mac4'  => '[0-9a-f]{6}\-[0-9a-f]{6}',                   # mac address format 4
	'time'  => '\d{1,2}:\d{1,2}(?::\d{1,2})?',               # time
	'hex'   => '[a-fA-F0-9]',                                # hexadecimal digit
	'date1' => '\d{1,2}[-\/]\d{1,2}[-\/]\d{1,4}',            # numerical date 1
	'date2' => '\d{1,4}[-\/]\d{1,2}[-\/]\d{1,2}',            # numerical date 2
};

sub get_crep
{
	my $key = shift;
	if ( defined $creps->{$key} )
	{
		return $creps->{$key};
	}

	return 0;
}

sub escape_regex_metas
{
	$_[0] =~ s/([\(\)\[\]\\\/\+\*\?\.\^\$]{1})/\\$1/mg;
	$_[0];
}

sub unescape_regex_metas
{
	$_[0] =~ s/\\{1}([\(\)\[\]\\\/\+\*\?\.\^\$])/$1/mg;
	$_[0];
}

sub seconds_since_epoch
{
	my ( $sec, $min, $hour, $mday, $mon, $year, $timezone ) = @_;

	if ( $mon =~ /\d+/ )
	{

		# month array starts at 0
		$mon -= 1;
	}
	else
	{
		$mon = $months{ uc($mon) };
	}

	my $tz = $timezones{ uc($timezone) };
	if ( !defined $tz )
	{
		$tz = $timezones{'UTC'};    # default timezone if undefined
	}

	my $time = timegm( $sec, $min, $hour, $mday, $mon, $year );

	# handle the timezone difference if one was listed
	$time -= ( $tz * 60 * 60 );
	return $time;
}

sub is_contiguous_mask
{
    return exists($reverse_masks->{$_[0]});
}

sub get_mask
{

	# translates '24' into '255.255.255.0'
	my $cidr = shift;
	return $masks->{$cidr};
}

sub invert_mask
{
    # i.e. turn 0.0.255.255 into 255.255.0.0
    my $mask_wild_dotted = shift;
    my $mask_wild_packed = pack('C4', split(/\./, $mask_wild_dotted));
    my $mask_norm_packed = ~$mask_wild_packed;
    my $mask_norm_dotted = join('.', unpack('C4', $mask_norm_packed));
    return $mask_norm_dotted;
}

sub mask_to_bits
{
    my $mask = shift;

    if (exists($reverse_masks->{$mask}))
    {
        return $reverse_masks->{$mask};
    }

    return $mask;
}

=cut
=item C<create_ip_element($ip, $mask, $invert)>

Given an IP and a IP mask returns an network object suitable for inclusion in the ZED.

C<create_ip_element('10.10.1.1', '0.0.255.255', 1)> will return { address => '10.10.1.1', mask => 16 }.

=cut
sub create_network_element
{
    my ($ip, $mask, $invert) = @_;

    if ($invert)
    {
        $mask = invert_mask($mask);
    }

    my $bits = $reverse_masks->{$mask};
    if ($bits)
    {
        return {
            address => $ip,
            mask => $bits,
        };
    }

	# Let's not get tripped up on our own address encoding.
	if ($ip ~~ get_any_address_address())
	{
        return {
            address => $ip,
            mask => 0,
        };				
	}
	# See if this is specified as a numeric mask, in which case it is 
        # a generic mask, as opposed to an IPv4 mask.
	# +++wayne consider making a separate "parseIPv6SubnetMask" vector, 
	# limiting to <= 128
	
	if (($mask =~ /^\d{1,3}$/) && ($mask + 0))  # in range and nonzero numeric
	{
        return {
            address => $ip,
            mask => $mask,
        };		
	}
		 
    return {
        address => $ip,
        ipV4Mask => $mask,
    };
}

# This assumes an argument with the zeroes portion of the address trailing to
# indicate network mask up to a trailing final byte.  Hence,
#  NetAddrToMask("10.0.0.0") == 8
#  NetAddrToMask("10.10.0.0") == 16, but
#  NetAddrToMask("10.0.10.0") == 24 ,and
#  NetAddrToMask("10.0.0.10") == 32
#
#  This is useful to infer a mask in a context (such as a route network declaration)
#  where one isn't provided, just an address.
#
sub net_addr_to_mask
{
	my $netAddr = shift;
	my $mask = "";  # guarantee that mask is blank, not undefined.
	my $zeroBytes = 0;  # have to count zero bytes in case we have legit zero subbnets
	my @address = split( /\./, $netAddr );
	foreach my $piece (@address) {
		if ( $piece eq '0' ) {
			$zeroBytes++;
		}
		else {
			# if we've been counting zero bytes and we now see nonzero, we have
			# significant bytes to fill in
			$mask .= (length($mask)) ? ".255" : "255";  # there's at least one, and it may be leading
			$mask .= ".255" x $zeroBytes;
			$zeroBytes = 0;
		}
	}
	
	# Now, if we have zero bytes to fill
	if ($zeroBytes--)
	{
		# There's at least one--is it leading?
		$mask .= (length($mask)) ? ".0" : "0";  # there's at least one, and it may be leading
		$mask .= ".0" x $zeroBytes;
	}
	
	return $mask;	
}

sub parseCIDR
{
	my ( $cidr ) = @_;
	my $parsed_cidr = {
		host => '',
		network => '',
		prefix => ''
	};

	if ( $cidr ) {
		my ( $host_address, $first_seted_bits ) = split( /\//, $cidr );
		if ( $host_address =~ /^\d{1,3}$/ )
		{
			$host_address .= ".0.0.0";
		}
		elsif ( $host_address =~ /^\d{1,3}\.\d{1,3}$/ )
		{
			$host_address .= ".0.0";
		}
		elsif ( $host_address =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}$/ )
		{
			$host_address .= ".0";
		}
		my $network_address;
		my $octet;
		for ( my $x = 1 ; $x <= 32 ; $x++ )
		{
			$octet .= ( $x <= $first_seted_bits ) ? '1' : '0';
			if ( $x % 8 == 0 )
			{
				$network_address .= ( defined $network_address ) ? '.' . bin2dec($octet) : bin2dec($octet);
				$octet = '';
			}
		}

		$parsed_cidr->{host} = $host_address;
		$parsed_cidr->{network} = $network_address;
		$parsed_cidr->{prefix} = $first_seted_bits;
	}

	return $parsed_cidr;
}

sub getUnitFreeNumber
{
	my $number = shift;
	my $unit   = shift;
	my $base   = shift;

	my $m = 1024;
	if ( defined $base )
	{
		if ( $base =~ /byte/i )    # memory size is measured in bytes
		{
			$m = 1024;
		}
		elsif ( $base =~ /bit/i )    # network speed is measured in bits
		{
			$m = 1000;
		}
	}

	if ( $unit =~ /K/i ) { $number * $m; }

	elsif ( $unit =~ /M/i ) { $number * $m * $m; }

	elsif ( $unit =~ /G/i ) { $number * $m * $m * $m; }
}

sub bin2dec
{
	return unpack( "N", pack( "B32", substr( "0" x 32 . shift, -32 ) ) );
}

# this function gunzips and untars a file
sub parse_targz_data
{
	my $tarFilename    = shift;
	my $filenameFilter = shift;

	if ( !$filenameFilter )
	{
		$filenameFilter = ".*";
	}

	# uncompress the using Zlib interface functions
	my $buffer;
	my $gz = undef;

	my $temp_file     = new File::Temp();
	my $temp_filename = $temp_file->filename();
	binmode $temp_file;

	if ( $gz = gzopen( $tarFilename, "rb" ) )
	{
		while ( $gz->gzread($buffer) > 0 )
		{
			print $temp_file $buffer;
		}
		
		# Z_BUF_ERROR can be omitted
		if ( $gzerrno != Z_STREAM_END && $gzerrno != Z_BUF_ERROR )
		{
			die "Couldn't retrieve uncompressed data\n";
		}
		$gz->gzclose;
	}
	else
	{
		die "Couldn't open temporary gz file for reading\n";
	}

	# create file pointer for tar file
	my $tar = Archive::Tar->new($temp_filename);

	# get file list from temp tar file and sort them
	my @filenames = sort { lc($a) cmp lc($b) } $tar->list_files;

	# this variable will store the configuration of each file in tar archive
	my $configRepository = {};

	# run through the file list
	for my $filename (@filenames)
	{
		if ( $filename =~ /$filenameFilter/ )
		{
			my $fileContents = $tar->get_content($filename);    # get the content of this file
			if ($fileContents)                                  # directories don't need to be specifically created
			{
				my @pieces        = split( /\//, $filename );
				my $size          = @pieces;
				my $lastFolder    = $configRepository;
				my $hashStatement = "\$configRepository";
				for ( my $i = 0 ; $i < $size ; $i++ )
				{
					my $name = $pieces[$i];
					$hashStatement .= "->{\'$name\'}";
				}
				eval( $hashStatement .= " = \$fileContents;" );
			}
		}
	}

	# close temp tar file
	$temp_file->close();
	return $configRepository;
}

# this is for *IP* protocols
sub get_protocol_name
{
	my $protoNum = shift;
	return (defined $ipProtocolNames->{$protoNum}) ? $ipProtocolNames->{$protoNum} : ""; 	
}

# ..as is this
sub get_protocol_number
{
	my $protoIdent = shift;
	my ($matchingProtoNum) = grep { $ipProtocolNames->{$_} eq lc($protoIdent) } keys %{$ipProtocolNames};
	
	# this may be blank if the user is looking for whether it matches our schema val
	return  $matchingProtoNum; 
}


sub get_port_expression
{
	my $name = shift;
	if( $portProtocols->{$name} )
	{
		return $portProtocols->{$name};
	}	
	return undef;
}

sub get_port_number
{
	# Translates 'telnet' into '23'
	my $port = shift;
	my $proto = shift || 'tcp';

	if (defined $portNumbers->{$port})
	{
	    return $portNumbers->{$port}; 
	}
	
	# Use the mapping to port expression if we can
	my $pExp = get_port_expression($port);
	
	if (defined $pExp)
	{
		foreach my $pExpTerm (@{$pExp})
		{
			if ($pExpTerm->{portExpression}->{protocol} ~~ $proto)
			{
				return $pExpTerm->{portExpression}->{port};
			}
		}
	}

	# If we don't have it defined in our constant, use perl's built in lookup for it.
	my $number = getservbyname($port, $proto);
	if( defined $number )
	{
		return $number;
	}
	return $port;
}

# checks to see if the incoming port is an integer.  If not
# it will do a lookup of the service name to convert it to an integer
sub int_port {
	my ($port) = @_;
	if ( $port =~ /^\d+$/ ) {
		return $port;
	}
	else {
		return get_port_number($port);
	}
}


sub trim
{

	# remove leading and trailing whitespace
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

sub get_interface_type
{

	# given an interface type, figures out a valid ZipTie type that will
	# pass the zds:interface schema.  If there isn't a valid match, returns "unknown"
	my $type = shift;

	if ( !$type )
	{
		return "unknown";
	}
	if ( $type =~ /^lo|softwareLoopback/i )
	{
		return "softwareLoopback";
	}
	elsif ( $type =~ /frame/i )
	{
		return "frameRelay";
	}
	elsif ( $type =~ /tun/i)
	{
		return "tunnel"
	}
	elsif ( $type =~ /eth|^ge-|^fe-/i )
	{
		return "ethernet";
	}
	elsif ( $type =~ /token/i )
	{
		return "tokenRing";
	} 
	elsif ( $type =~ /BRI|PRI/i )
	{
		return "isdn";
	}
	elsif ( $type =~ /(gre|isdn|other|modem|ppp|serial|atm|sonet)/i )
	{
		return lc($1);
	}
	else
	{
		return "other";
	}
}

sub strip_mac
{

	# strips out everything but hex chars
	my $mac = shift;
	$mac =~ s/[^0-9a-f]//ig;
	return uc($mac);
}

sub get_model_filehandle
{
	my $adapter_name   = shift;
	my $ip_address     = shift;
	my $operation      = shift; # optional argument
	my $file_extension = "_model.xml";
	return _get_output_filehandle( $adapter_name, $ip_address, $file_extension, $operation );
}

sub get_cli_commands_filehandle
{
	my $adapter_name   = shift;
	my $ip_address     = shift;
	my $file_extension = "_cli_cmds.xml";
	return _get_output_filehandle( $adapter_name, $ip_address, $file_extension );
}

sub _get_output_filehandle
{
	my $adapter_name   = shift;
	my $ip_address     = shift;
	my $file_extension = shift;
	my $operation      = shift; # optional argument

	# Create a file handle to STDOUT unless it has been specified to output to a file
	my $filehandle = \*STDOUT;

	# Check to see if the "OUTPUT_MODEL_TO_FILE" exists
	my $output_model_to_file_flag = $ENV{OUTPUT_MODEL_TO_FILE};
	if ($output_model_to_file_flag)
	{

		# Cleanup the adapter name for any characters that are illegal to a file name
		$adapter_name =~ s/:://g;
		$adapter_name =~ s/ /_/g;

		# Add the operation name to the filename if provided
		if (!defined($operation))
		{
			$operation = '';
		}
		else
		{
			$operation = "_$operation";
		}

		# Calculate the time of the operation
		my ($sec, $min, $hour, $day, $month, $year) = localtime();
		my $current_time = sprintf('_%s-%02s-%02s_%02s-%02s-%02s', ($year + 1900), ($month + 1), $day, $hour, $min, $sec);

		# Create the file name of the model output file
		my $output_model_file = escape_filename( $adapter_name . $operation . "_" . $ip_address . $current_time . $file_extension );

		# Check to see if a directory has been specified
		my $output_model_dir = $ENV{OUTPUT_MODEL_DIR};
		if ($output_model_dir)
		{
			$output_model_file = $output_model_dir . "/" . $output_model_file;
		}

		# Open the output model file for writing to
		# The tee command below can be replaced by IO::Tee when convenient: http://search.cpan.org/~kenshan/IO-Tee-0.64/Tee.pm
		open( MODEL_OUTPUT_FH, "| tee $output_model_file" ) || $LOGGER->fatal("Unable to open '$output_model_file' for printing adapter output!");
		$filehandle = \*MODEL_OUTPUT_FH;
		$LOGGER->debug("Logging adapter output to '$output_model_file'");
	}
	else
	{
		$LOGGER->debug("Logging adapter output to STDOUT");
	}
	return $filehandle;
}

sub close_model_filehandle
{
	my $filehandle = shift;

	# Check to see if the "OUTPUT_MODEL_TO_FILE" exists.
	my $output_model_to_file_flag = $ENV{OUTPUT_MODEL_TO_FILE};

	# If it does exist, then the filehandle is expect to be a file, so we should close it.
	# If not, then do nothing since STDOUT should not be closed.
	if ($output_model_to_file_flag)
	{
		close($filehandle);
	}
}

sub create_empty_file
{
	my $filename = shift;
	open( FH, ">$filename" );
	close(FH);
}

sub escape_filename
{

	# Attempt to get any specified filename
	my $filename = shift;

	# If no file name was specified, leverage the File::Temp module to give us a unique filename
	if ( defined($filename) )
	{

		# Replace any non-compliant characters for filenames with a hyphen
		$filename =~ s/[\\\/:*\?"<>\|]/-/g;
		return $filename;
	}
	else
	{
		return "";
	}
}

sub create_unique_filename
{
	return tmpnam();
}

# Taken from perlmonks.org
sub merge_hashes    # returns 0 on circular references
{

	# function merges 2 or more hashes
	# circular/duplicate links make function return 0
	my @hashrefs = @_;

	return 0 if grep { ref $_ ne 'HASH' } @hashrefs;    # original function died here

	my %merged = ();

	my @seen = grep { ref $_ eq 'HASH' } @SeenMerged{@hashrefs};    # Break circular links..

	if (@seen)
	{
		return 0;                                                   # original function died here
	}

	@SeenMerged{@hashrefs} = @hashrefs;

	foreach my $h (@hashrefs)
	{
		while ( my ( $k, $v ) = each %$h )
		{
			push @{ $merged{$k} }, $v;
		}
	}

	while ( my ( $k, $v ) = each %merged )
	{
		my @hashes = grep { ref $_ eq 'HASH' } @$v;
		$merged{$k} = $v->[0] if ( @$v == 1 && !ref $v->[0] );
		$merged{$k} = merge_hashes(@hashes) if @hashes;
	}

	delete @SeenMerged{@hashrefs};
	return \%merged;
}

sub choose_admin_ip
{

	# select the best IP for management
	my ( $originalIp, $interfaces ) = @_;
	
	return $originalIp if !defined $interfaces->{interface};
	
	my @ifArray = @{ $interfaces->{interface} };

	# check if this is a NAT'd device.  If so, don't pick an IP off of an interface
	my $natted = 1;
	foreach my $interface (@ifArray)
	{
		if ( defined $interface->{ipEntry} )
		{
			my @ips = @{ $interface->{ipEntry} };
			foreach my $ip (@ips)
			{
				if ( $ip->{ipAddress} eq $originalIp )
				{
					$natted = 0;
					$LOGGER->debug( "Orignal IP address $originalIp found on interface " . $interface->{name} );
					next;
				}
			}
		}
	}
	if ($natted)
	{
		$LOGGER->debug("Original IP address $originalIp was not found on an interface, assuming a NAT.");
		return $originalIp;
	}

	# now analyze any loopbacks
	foreach my $interface (@ifArray)
	{
		if ( $interface->{operStatus} !~ /Down/i && $interface->{type} =~ /softwareLoopback/i )
		{
			$LOGGER->debug("Testing interface ".$interface->{name}." for an admin IP.");
			my $passingIp = _test_interface_ips($interface);
			if ($passingIp)
			{
				return $passingIp;
			}
		}
	}

	# now look at the rest if there hasn't been a match yet
	foreach my $interface (@ifArray)
	{
		if ( $interface->{operStatus} !~ /Down/i && !$interface->{-virtual} && $interface->{type} !~ /softwareLoopback/i )
		{
			$LOGGER->debug("Testing interface ".$interface->{name}." for an admin IP.");
			my $passingIp = _test_interface_ips( $interface );
			if ($passingIp)
			{
				return $passingIp;
			}
		}
	}

	return $originalIp;    # fall back choice
}

sub get_any_address_encoding {
    my $ipv6 = shift;
    return { address => get_any_address_address($ipv6), mask => '0'};
}

sub get_any_address_address
{
	my $ipv6 = shift;
	
	return ($ipv6) ?  '::' : '0.0.0.0';
	
}

# get_any()
# Returns the topology "canonical" address/mask encoding for "any" address
# PARAMS:
#  ipv6 (optional) - boolean, indicating whether to return in ipv6 address form.  FALSE if not
#   specified
# RETURNS:
#  ZED address/mask pair for assignment to src/dest, object group address entry, whatever.
sub get_any {
	my $ipv6 = shift;
	if ($ipv6) {
		return { address => '::', mask => '0' };
	}
	else {
		return { address => '0.0.0.0', mask => '0' };
	}
}

#  Key named identifiers in the group and filter areas are defined
#  by the Ziptie schema to be NCNames, which basically boil down
#  through the inheritance chain to be XMLSchema ID elements which
#  further prohibit the use of colons in their text.  This then, in
#  total, prohibits the characters:
#     : @ $ % & / + , ; { } { } [ ]  (and whitespace characters)
#
#  Wade through http://www.w3.org/TR/2001/REC-xmlschema-2-20010502/ and its
#  supporting documents for corroboration, or there's a good summary of this
#  point at
#  http://stackoverflow.com/questions/1631396/what-is-an-xsncname-type-and-when-should-it-be-used.
#  The Netscreen network information elements do not limit a number
#  of these however, so to map to the SED, we will map these to underscores.
sub ncname_from_string {
	my ($inIDStr) = @_;
	$inIDStr =~ s/[:@\$%&\/+\-\,;\{\}\{\}\[\]\s]/_/g;
	return $inIDStr;
}

sub _test_interface_ips
{
	my $interface = shift;
	if ( defined $interface->{ipEntry} )
	{
		my @ips = @{ $interface->{ipEntry} };
		foreach my $ip (@ips)
		{
			if ( $ip->{ipAddress} !~ /^127\./ )
			{
				my $pinger;
				if ( $OS =~ /(?<!dar)win|sun|solaris/i )
				{
					$pinger = Net::Ping->new('icmp');
				}
				else
				{
					$pinger = Net::Ping->new();
				}
				
				if ( $pinger->ping( $ip->{ipAddress} ) )
				{
					$LOGGER->debug( "Administrative IP address chosen: " . $ip->{ipAddress} );
					return $ip->{ipAddress};
				}
				else
				{
					$LOGGER->debug('Candidate admin IP '.$ip->{ipAddress}.' is not responding to ping.');
				}
			}
		}
	}
	return 0;
}

# Routines to allow external call to open and close the telmetry-level "neighbors" elements
# TODO: Even though $in is not used it is needed by unit tests, this should be fixed
# so we can drop it from the parameters
sub open_telemetry_neighbors {
	my ($in, $out) = @_;
	$out->attributes(1);
	$out->open_element('neighbors');	
}

sub close_telemetry_neighbors {
	my ($in, $out) = @_;
	$out->close_element('neighbors');
}

1;
__END__

=head1 NAME

ZipTie::Adapters::Utils - General Utility Methods for Adapters

=head1 SYNOPSIS

    use ZipTie::Adapters::Utils;
	my $port = get_port_number('ssh');
	my $mask = get_mask(16);
	my $date = w3c_datetime($commonDate);

=head1 Description

Utils provides several static utility methods to make it easier for
adapter parsing modules to produce output that will pass the ZipTie
model schema.

=head2 Methods

=over 12

=item C<get_mask>

Given a CIDR mask, returns the matching bit mask.  For example
C<get_mask('24')> will return "255.255.255.0".

=item C<mask_to_bits($ipMask)>

Given an IPv4 address mask such as 255.255.255.0, converts to its
bit mask (24).

=item C<get_port_number>

Given a service name, returns the matching port number.  For example
C<get_port_number('telnet')> will return "23".

=item C<seconds_since_epoch($sec, $min, $hour, $mday, $month, $year, $timezone)>

Converts dates into the ZED required long format in seconds since Unix epoch.

Inputs:		$sec - seconds
			$min - minutes
			$hour - hour in 24 hour format
			$mday - day of the month
			$month - the month integer (1-12) or the shorthand name, e.g. 'Jan'
			$year - the 4 digit year
			$timezone - the timezone name, e.g. CST.  This is optional, UTC is assumed if it is not provided.

=item C<trim>

removes leading and trailing whitespace from the input string.

=item C<get_interface_type>

given the type or name of an interface, it tries to make a best 
guess as the proper ZipTie interface type.

=item C<strip_mac>

Returns just the hex chars from the provided string

=item C<choose_admin_ip($originalIp, $interfaceHash)>

Chooses an IP address using criteria from the list of interfaces provided in the 
$interfacesHash.  

The $originalIp must also be passed through as it is used to determine if the device
is behind a NAT or not.  For example, if you found this device via an IP address
that it wasn't configured for locally, then it is likely that the device is behind some
sort or translation (NAT).  When that is the case, the $originalIp is returned as the preferred
administrative IP address.

This method will choose available loopback IP addresses first.
It will return a single scalar string that represents the IP address.

The incoming $interfacesHash should match the telemetry.xsd XML schema. Here 
is an example of the data structure of the incoming hash table.

$VAR1 = {
          'interface' => [
                           {
                             'operStatus' => 'Up',
                             'name' => 'Ethernet0/0',
                             'type' => 'ethernet',
                             'ipEntry' => [
                                            {
                                              'ipAddress' => '10.100.8.10',
                                              'mask' => '29'
                                            }
                                          ],
                             'inputBytes' => '873995991'
                           },
                           {
                             'operStatus' => 'Down',
                             'name' => 'Serial0/0',
                             'type' => 'serial',
                             'inputBytes' => '0'
                           },
                           {
                             'operStatus' => 'Up',
                             'name' => 'Ethernet0/1',
                             'type' => 'ethernet',
                             'ipEntry' => [
                                            {
                                              'ipAddress' => '10.100.8.25',
                                              'mask' => '29'
                                            }
                                          ],
                             'inputBytes' => '1194740983'
                           },
                           {
                             'operStatus' => 'Up',
                             'name' => 'Loopback0',
                             'type' => 'softwareLoopback',
                             'ipEntry' => [
                                            {
                                              'ipAddress' => '10.100.8.151',
                                              'mask' => '32'
                                            }
                                          ],
                             'inputBytes' => '0'
                           }
                         ]
        };


=item C<parse_targz_data($tgzFilename, $filenameFilter)>

$tgzFilename - the full path to the tar file to inspect
$filenameFilter - an optional setting, this is a regex that allows you to only untar
				  files that match this regular expression.  The regex is applied to the entire path.

Given a full path to a tar'd and zipped file (.tgz), unzip and untar the file
and place the contents of the file in a hash structure.

For example, if the tar contained these files.....
	/home/rkruse/filea
	/home/rkruse/fileb
	/etc/named.conf

The returned hash from this call would look like.....

	$VAR1 = {
          'home' => {
                      'rkruse' => {
                                    'fileb' => 'fileb contents',
                                    'filea' => 'filea contents'
                                  }
                    },
          'etc' => {
                     'namedconf' => 'named.conf contents'
                   }
	};


=item C<get_protocol_name($protoNum)>

Returns an IP protocol name of a passed IPv4 Protocol type.  Only (intentionally) handles 
legal enumerated values for simple type core:Protocol in the ZipTie schema.

$protoNum - ip protocol number to look up.

Returns "" if protocol not in those supported.

=item C<get_protocol_number($protoName)>

Returns an IP protocol number of a passed IPv4 Protocol type.  Only (intentionally) handles 
legal enumerated values for simple type core:Protocol in the ZipTie schema.

$protoName - ip protocol, by name to look up.
 
Returns "" if protocol not in those supported.

=item C<get_model_filehandle($adapter_name, $ip_address)>

Returns a file handle that an adapter can use to output model data to.  By default, I<STDOUT> will be used as the file handle;
however, the output can also be written to a file that uses the specified adapter name/ID and IP address to have a file name
that follows the syntax of: C<adapterName_ipAddress_model.xml>.

Output of the model to a file can be configured via two environment varaibles: B<OUTPUT_MODEL_TO_FILE> and B<OUTPUT_MODEL_DIR>.

B<OUTPUT_MODEL_TO_FILE> is a boolean environment variable to determine whether or not output of the model should be done to
a file at all.  If the value of it is false of empty, then the filehandle to use will be I<STDOUT>.  Otherwise, the model will
be written to a file with the syntax of: C<adapterName_ipAddress_model.xml>.

B<OUTPUT_MODEL_DIR> is only important if the B<OUTPUT_MODEL_TO_FILE> environment variable has been set.  If it has been
determined that the model should be output to a file, then B<OUTPUT_MODEL_DIR> will be checked to see which directory
the output model file will be written to.  If the environment variable exists, it will be prepended to the file name
generated from the adapter ID and IP address specified.  Otherwise, the output model file will be within the current
working directory of Perl.

=item C<get_cli_commands_filehandle($adapter_name, $ip_address)>

Very similar to the C<get_model_filehandle($adapter_name, $ip_address)> method, but delivers
an XML document name that is specific to an adapter that is running CLI commands only and
not performing a backup.

=item C<close_model_filehandle($filehandle)>

Closes the filehandle that has been used for output model information.  It is assumed that the filehandle specified is the
one retrieved from C<get_model_filehandle($adapter_name, $ip_address)>.  If the filehandle that is being used is STDOUT,
then it will B<NOT> be closed; otherwise it will be closed.

=item C<create_empty_file($filename)>

Creats an empty file using the 'open' command against the provided filename.
A filehandle is created but immediately closed.

=item C<get_crep('cipm')>

Returns the regex. pattern referenced by the key. All the (c)ommon (r)egular (e)xpression
(p)atterns are stored in the creps hash.

=item C<sub escape_regex_metas('(don't capture nor cluster this)')>

Escapes regex. metacharacters which are used in regular expressions. 

=item C<unescape_regex_metas('\(don't capture nor cluster this\)')>

Unescapes the regex. pattern escaped with escape_regex_metas.

=item C<parseCIDR('10.100/16')>

Convert Classless InterDomain Routing address to ip address and subnet mask.

=item C<bin2dec('1000')>

Converts binary number to decimal number.

=item C<merge_hashes($ref_hash1,$ref_hash2)>

Merges two or more hashes.

=item C<escape_filename($filename)>

Replaces any characters that are not suitable for use in a filename with a hyphen.  It is assumed that the filename
specified is a relative filename, and thus forward slashes and back slashes are not allowed.

=item C<create_unique_filename>

Simple wrapper for C<File::Name::tmpnam()> to grab fully-qualified name of an available temp file.

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.
 
  The Original Code is Ziptie Client Framework.
  
  The Initial Developer of the Original Code is AlterPoint.
  Portions created by AlterPoint are Copyright (C) 2006,
  AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): rkruse, bedwards, mkourbanov
  Date: Apr 25, 2007

=cut
