package ZipTie::RESTAPI;

use strict;
use warnings;

use 5.0100000;

our $VERSION = 1.0;

use LWP::UserAgent;
use ZipTie::ConnectionPath;

use ZipTie::Recording;
use ZipTie::Recording::Interaction;
my $RECORDING = ZipTie::Recording::get_recording();

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();

# Creates a new REST API object from a connectionPath Object
sub new {
	my ($class_name, $connection_path, $use_management_ip) = @_;

	$LOGGER->debug("Creating RESTAPI Object\n");

	my $this;

	$this->{client} = LWP::UserAgent->new();
	$this->{client}->timeout(30);

	my $protocol = $connection_path->get_protocol_by_name('HTTPS');
	if (!defined($protocol)) {
		$LOGGER->fatal("REST API Connections only support HTTPS at the present time.\n");
	}

	$this->{protocol_name} = $protocol->get_name();
	$LOGGER->debug( 'Using Protocol: ' . $this->{protocol_name} . "\n" );

	$this->{port} = $protocol->get_port();
	$LOGGER->debug( 'Using Port: ' . $this->{port} . "\n" );

	if ($use_management_ip) {
		$this->{ip_address} = $connection_path->get_management_server_ip_address();
	}else{
		$this->{ip_address} = $connection_path->get_ip_address();
	}

	$LOGGER->debug( 'Using IP Address: ' . $this->{ip_address} . "\n" );

	$this->{client}->protocols_allowed( [ lc( $this->{protocol_name} ) ] );

	$this->{connection_token} = q{};
	$this->{api_location}     = q{};

	# Turn $this into a ZipTie::RESTAPI object
	bless($this, $class_name);

	return($this);
}

# This function makes an initial connection to the specified device and sets up the required credential string
# to be used for future requests.

# It requires a device specific query string that will be used to retrieve an authentication token
# It also requires a function reference that will be used to parse the response into an appropriate authentication token
sub obtain_connection_token {
	my ($this, $query_string, $response_parser) = @_;

	my $my_url = lc( $this->{protocol_name} ) . q{://} . $this->{ip_address} . q{:} . $this->{port} . q{/} . $query_string;

	_start_recording( $query_string, 'obtain connection token' );

	$LOGGER->debug("Sending Authentication Request\n");
	my $response = ( $this->{client} )->get($my_url);

	_end_recording( $response->content() );

	$this->{connection_token} = &{$response_parser}( $response->content() );
	$LOGGER->debug("Authentication Successful\n");

	return;
}

# A Rest API URL is constructed from:
#	The protocol name + ://
#	the ip address:port/ combination
#	the api base location
#	the command to be issued
#	the connection token
#
# The content of the HTTP response is returned
sub get {
	my $this    = shift;
	my $command = shift;
	my $name    = shift;

	# A given adapter may wish to disable logging for specific commands for security reasons
	# If a command or response contains information that should not be logged, set this to some defined value
	my $disable_logging = shift;

	# https://IP_Address:Port/api_base (+) command (+) security_token
	my $my_url = lc( $this->{protocol_name} ) . q{://} . $this->{ip_address} . q{:} . $this->{port} . q{/} . $this->{api_location} . $command . $this->{connection_token};

	_start_recording( $this->{api_location} . $command . $this->{connection_token}, $name );

	my $start_time = time();

	if ( !( defined $disable_logging ) ) {
		$LOGGER->debug("Issuing Command: $command\n");
	}

	# Send the REST Request
	my $response = ( $this->{client} )->get($my_url);

	my $end_time = time();
	my $response_time = $end_time - $start_time;

	if ( !( defined $disable_logging ) ) {
		$LOGGER->debug( 'Response Received:' . $response->content() . "\n" );
		$LOGGER->debug("Response Time: $response_time seconds\n");
	}

	_end_recording( $response->content() );

	return $response->content();

}

# this function allows the caller to specify a URL in a custom manner if the resource they are interested
# does not follow the standard format used in get()
# the url is constructed from:
#	The protocol name + ://
#	the ip address:port/ combination
#	the specified custom location
sub get_custom_url {
	my $this = shift;
	my $url  = shift;
	my $name = shift;

	# A given adapter may wish to disable logging for specific commands for security reasons
	# If a command or response contains information that should not be logged, set this to some defined value
	my $disable_logging = shift;

	_start_recording( $url . $this->{connection_token}, $name );

	my $start_time = time();

	if ( !( defined $disable_logging ) ) {
		$LOGGER->debug("Issuing Custom URL: $url\n");
	}

	my $my_url = lc( $this->{protocol_name} ) . q{://} . $this->{ip_address} . q{:} . $this->{port} . q{/} . $url . $this->{connection_token};

	# Send the REST Request
	my $response = ( $this->{client} )->get($my_url);

	my $end_time = time();
	my $response_time = $end_time - $start_time;

	if ( !( defined $disable_logging ) ) {
		$LOGGER->debug( 'Response Received:' . $response->content() . "\n" );
		$LOGGER->debug("Response Time: $response_time seconds\n");
	}

	_end_recording( $response->content() );

	return $response->content();

}

# Returns the Base API Location
sub get_api_location {
	my $this = shift;
	return $this->{api_location};
}

# Sets the base API Location. Commands will be attached after this section of the URL
sub set_api_location {
	my $this         = shift;
	my $new_location = shift;
	$this->{api_location} = $new_location;
	return;
}

sub get_connection_token {
	my $this = shift;
	return $this->{connection_token};
}

sub set_connection_token {
	my $this  = shift;
	my $token = shift;
	$this->{connection_token} = $token;
	return;
}

# This function allows credentials to be set directly on the user agent object if
# that is required by a particular device.
sub set_credentials {
	my ($this, $username, $password) = @_;

	$this->{client}->set_username($username);
	$this->{client}->set_password($password);

	return;
}

sub set_timeout {
	my ($this, $new_timeout) = @_;

	if ($new_timeout > 0) {
		$this->{client}->timeout($new_timeout);
		$LOGGER->debug("Rest API timeout set to '$new_timeout'");
	}else{
		$LOGGER->debug("Rest API timeout not set due to invalid value '$new_timeout'");
	}

	return;
}

# This function sets up and invokes the start_current_interaction function of recording.
sub _start_recording {
	my ($command, $name) = @_;

	if ( $ENV{ENABLE_RECORDING} ) {

		# If the command contains xml tags, replace the angle braces with percent
		# representation so that the recording matches the url that gets sent
		$command =~ s/\</\%3C/msgx;
		$command =~ s/\>/\%3E/msgx;

		my $interaction = ZipTie::Recording::Interaction->new( richCommand => $command );
		if ( defined $name ) {
			$interaction->name($name);
		}
		$RECORDING->start_current_interaction($interaction);
	}

	return;
}

# This function sets up and invokes the finish_current_interaction function of recording.
sub _end_recording {
	my $response = shift;

	if ( $ENV{ENABLE_RECORDING} ) {
		my $interaction = $RECORDING->get_current_interaction();
		$interaction->cliResponse($response);
		$RECORDING->finish_current_interaction();
	}

	return;
}

1;

__END__

=head1 NAME

ZipTie::RESTAPI - Automates connection, authentication, and communication with https based REST api servers

=head1 VERSION

This document describes ZipTie::RESTAPI version 1.0

=head1 SYNOPSIS

    use ZipTie::RESTAPI;
    new($connection_path);
	obtain_connection_token ($query_string, $response_parser);
	get ($command);
	get_custom_url($url);
	get_api_location();
	set_api_location($new_location);
	set_credentials($username, $password);
	

=head1 DESCRIPTION

This Module provides functionallity to connect to a REST API server using the https protocol. It allows an initial authentication
request to be sent and then further requests can be made using the same credentias.
It uses the LWP::UserAgent module to handle https interaction.

=head1 SUBROUTINES/METHODS

=over 12

=item C<new($connection_path)>

This is the constructor for RESTAPI. It takes a connection path and sets up the required parameters for a REST API connection

=item C<obtain_connection_token($query_string, $response_parser)>

This method makes an initial connection to the REST server. It takes a query string specified by the caller and makes a call to that location
on the server that was configured when the RESTAPI object was created.

The caller provides a reference to a function that will be used to parse the response. That function must accept the response body from the
initial connection call and return a string that can be appended to future REST requests as an uathentication token. If credentials are
specified in some other manner this string can be empty. Thie function should also accept a fake authentication token and return a properly
formatted fake token similar to the real toke. This fake token will be output into any recording that is created in order to mask the real token.

=item C<get ($command)>

This method takes a command specified by the caller and constructs a url that will be used to make a REST API call.
It sends the request and returns the content of the response.

The URL is constructed in the following form:

https://IP_Address:Port/api_base+$command+security_token


=item C<get_custom_url($url)>

This method takes a custom query string specified by the caller and constructs a url that will be used to make a REST API call.
It sends the request and returns the content of the response.

The URL is constructed in the following form:

https://IP_Address:Port/$url

=item C<get_api_location()>

This method returns the api location string stored in this object. The api location string is the base location on the server where
commands will be issued.

=item C<set_api_location($new_location)>

This method sets the api location string in this object.

=item C<set_credentials($username, $password)>

This metho sets the username and password credentials on the LWP::UserAgent object that handles https communication.
This will be used if the REST API being accessed required standard http authentication rather than using a token.


=item C<get_connection_token()>

This metho returns the security token currently configured for this RestApi object.

=item C<set_connection_token($token)>

This method sets the connection token for this RestApi object to be the given string.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_start_recording($command)>

This function starts a new interaction in the current recording if recordings are enabled. It stores the command to be issued
in this interaction. Any angle braces ( < or > ) are converted to percentage representation (%3C and %3E respectively) before
the command is stored. This is to match the format that is actually received by the device simulator when a backup is run
against a simulation. A friendly name for the command can be optionally provided. If it is provided then this name is stored in
the interaction.

=item C<_end_recording($response)>

This function stores a response in the current interaction if recording is enabled and then ends the interaction.

=back

=head1 DIAGNOSTICS

=over 12

=item C<REST API Connections only support HTTPS at the present time.>

This message is displayed if the provided connection document does not include HTTPS as a supported protocol. At
present only HTTPS is supported for security reasons. This is a fatal error, its presence indicates that the 
requested operation did not complete.

=back

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

This module uses richCommands and provides names for each interaction when recordings are enabled. Recording.pm must be
a high enough version to support these operations.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.

=head1 LICENSE AND COPYRIGHT

Copyright IBM 2013


=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: October 2013

=cut

