=pod

=head1 NAME

    Parsers.pm

=head1 DESCRIPTION

    Parsing logic for Cisco IOS configuration.

=head2 Methods

=over

=item parse_arp()

=item parse_cdp() 

=item parse_routing_neighbors() 

=item parse_telemetry_interfaces() 

=item parse_mac_table() 

=item parse_vtp() 

=item parse_static_routes() 

=item parse_dynamic_routes()

=item parse_vlans() 

=item parse_stp parse_routing() 

=item parse_access_ports() 

=item create_config() 

=item parse_local_accounts() 

=item parse_chassis() 

=item parse_filters() 

=item C<parse_object_groups ($in, $out)>

    Parse the object groups in the configuration, referenced symbolically in access-lists (only 
    supported in IOS >= 12.4)

=item parse_snmp() 

=item parse_system() 

=item parse_interfaces() 

=item parse_services() 

=item parse_mpls() 

=item parse_qos()

=back

=head2 Internal Methods

=over

=item _process_if_failover_group()

    Process a passed failover group for an interface, building up a hash representing the ZED structure
    for a E<lt>failoverGroupE<gt> element.

=over 2

=item I<failoverConfig>: 
    The text element for a given failover group from the IOS config

=item I<groupNumber>: 
    The number of this failover group.  Do not parse for this here.

=item I<groupText>: 
    The textual representation for this failover group.  May differ from groupNumber.  You should parse for this here.

=item I<Returns>: reference to hash for E<lt>failoverGroupE<gt> element, or C<undef> if not correctly
    parseable/validatable.

=back 

=back

=cut

package ZipTie::Adapters::Cisco::IOS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::Cisco::IOS::AddressTranslationParser qw(parseAddressTranslation);
use ZipTie::Adapters::Utils qw(mask_to_bits strip_mac create_network_element seconds_since_epoch get_mask get_port_number trim get_interface_type get_crep net_addr_to_mask is_contiguous_mask get_any_address_encoding invert_mask);
use ZipTie::Addressing::Subnet;
use ZipTie::Addressing::RouteFilter;
use ZipTie::Logger;
use ZipTie::RemoteGatewayResolution;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

use Exporter qw(import);
our @EXPORT_OK = qw(parse_arp parse_cdp parse_dynamic_routes parse_routing_neighbors parse_telemetry_interfaces parse_mac_table parse_vtp parse_static_routes _parse_static_route_data parse_vlans parse_stp parse_routing parse_access_ports create_config parse_local_accounts parse_chassis parse_filters parse_object_groups parse_snmp parse_system parse_interfaces parse_services parse_mpls parse_qos);

# TODO: Do we need @VALID_MEMSIZES and $SIZEPATTERNS_STRING here, why not in common?
# Class constants.
# The reasonable memory sizes in MBs: 2 to the n-th power.
our @VALID_MEMSIZES = map { 2**$_ } qw(2 3 4 5 6 7 8 9 10 11 12 13 14 15 16);

# The string which collects the sizes from 4 (2**2) MBs to 65536 (2**16) MBs.
# Format the array into a space-padded string.
our $SIZEPATTERNS_STRING = " @VALID_MEMSIZES ";

# The number of KBs in 80 percent of 1 MB.
our $KBS_80PERCENTMB = int( my $volatile = 1024 * 0.8 );

# TODO: This was used for signature qualification which is no longer used, reduce scope?
our $systemName = "";

# parse_routing_neighbors()
#
# Parse the neighbor sections for supported routing protocols.  Provide a state of
# neighbor attributes, and if an output XMLPrinter context has been passed, fill
# that in with the 'neighbors' complex type instances (as found in telemetry.xsd).
#
# Note that in analysis of the Cisco 3825 Integrated Services Switch, I'm noticing at least
# two (wonderful :-\) variations on the "get <protocol> neighbor" output, at least for EIGRP.
#
# PARAMS: $in - input context, with (as of September 2011), sections for
#              {bgp}
#              {ospf}
#              {eigrp}
#
#           These in turn all have {neighbors} hash elements off of them if they're
#           to be processed here.
#
#          $out - output XML printer context.  *Optional*, can be undefined, in which
#              case no element generation will happen as a byproduct of the routine.
#
# RETURNS:  Output structure only defined currently for eigrp: for each neighbor
#              ->{key:neighborIP}  ->  {routeProcess: process ID}
#                                ->  {fwdInterface: interface name for forwarding to this neighbor}
#
sub parse_routing_neighbors {
	my ($in, $out) = @_;
	my $opened = 0;
	my $neighborData;

	if ($in->{ospf}->{neighbors}) {
		while ($in->{ospf}->{neighbors} =~ /^([\da-f.:]+)\s+\d+\s+\S+\s+\S+\s+([\da-f\.:]+)\s+(\S+)\s*$/gm) {
			if (!$opened) {
				$opened = 1;
				$out->open_element('routingNeighbors');
			}

			my $neighbor = {
				protocol  => 'OSPF',
				routerId  => $1,
				ipAddress => $2,
				interface => $3,
			};

			$out->print_element('routingNeighbor', $neighbor);
		}		
	}

	if ($in->{bgp}->{neighbors}) {
		while ($in->{bgp}->{neighbors} =~ /^BGP neighbor is ([\da-f:\.]+)(.+?)BGP state/gsm) {
			if (!$opened) {
				$opened = 1;
				$out->open_element('routingNeighbors');
			}

			my $neighbor = {
				protocol  => 'BGP',
				ipAddress => $1,
			};

			my $blob = $2;
			if ($blob =~ /router ID\s*([\da-f:.]+)/) {
				$neighbor->{routerId} = $1;
			}

			$out->print_element('routingNeighbor', $neighbor);
		}
	}

	if ($in->{eigrp}->{neighbors}) {
		while ($in->{eigrp}->{neighbors} =~ /^(IP-EIGRP neigh|EIGRP-IPv4 Neigh).*?(process|AS)[ \(](\d+).*?\n(.+)$/gsim) {
			# $1 and $2 are dontcares, provided as alternatives
			my ($eigrpProc, $eigrpSec) = ($3, $4);

			while ($eigrpSec =~ /^(?:\d+\s+)?([\da-f:\.]{3,})\s+(\S+)/gm) {
				# Fill in context about this route for return to caller
				$neighborData->{eigrp}->{$1} = { routeProcess => $eigrpProc, fwdInterface => _full_int_name($2) };

				if (defined $out) {
					if (!$opened) {
						$opened = 1;
						$out->open_element('routingNeighbors');
					}

					my $neighbor = {
						protocol  => 'EIGRP',
						ipAddress => $1,
						routerId  => $1,
						interface => _full_int_name($2),
					};

					$out->print_element('routingNeighbor', $neighbor);
				}
			}
		}
		if ($opened) {
			$out->close_element('routingNeighbors');
		}
	}
	return $neighborData;
}

sub parse_arp {
	my ($in, $out) = @_;
	my $IP = get_crep('cipm');
	my $MAC = get_crep('mac1') .'|' . get_crep('mac2') . '|' . get_crep('mac3') . '|' . get_crep('mac4');
	my $arps;

	while ($in->{arp} =~ /^\s*\S+\s+($IP)\s+[\d\-]+\s+($MAC)\s+\S+\s+(\S+)\s*$/mg) {
		my $arp = {
			ipAddress  => $1,
			macAddress => strip_mac($2),
			interface  => $3,
		};
		push(@{$arps->{arpEntry}}, $arp);
	}

	if ($arps) {
		$out->print_element('arpTable', $arps);
	}
}

sub parse_cdp {
	my ($in, $out) = @_;
	my $opened = 0;

	while ($in->{cdp} =~ /^-+\s*$(.+?advertisement.+?)(?=^\s)/msg) {
		if (!$opened) {
			$opened = 1;
			$out->open_element('discoveryProtocolNeighbors');
		}

		my $singleCdpBlob = $1;
		my $neighbor = { protocol => 'CDP', };
		if ($singleCdpBlob =~ /Device\s*ID:\s*(\S+)/) {
			$neighbor->{sysName} = $1;
		}
		if ($singleCdpBlob =~ /IP address:\s*(\S+)/i) {
			$neighbor->{ipAddress} = $1;
		}
		if ($singleCdpBlob =~ /Interface:\s*(\S+)/m) {
			$neighbor->{localInterface} = $1;
			$neighbor->{localInterface} =~ s/,$//;
		}
		if ($singleCdpBlob =~ /\(outgoing port\):\s*(\S+)/) {
			$neighbor->{remoteInterface} = $1;
		}
		if ($singleCdpBlob =~ /Platform:\s*(\b[^,]+)/) {
			$neighbor->{platform} = $1;
		}
		if ($singleCdpBlob =~ /Version\s*:\s*(.+?)(?=^\s+)/ms) {
			$neighbor->{sysDescr} = trim($1);
		}
		$out->print_element('discoveryProtocolNeighbor', $neighbor);
	}
	if ($opened) {
		$out->close_element('discoveryProtocolNeighbors');
	}
}

sub parse_dynamic_routes {
	my ($in, $out, $subnets) = @_;
	my $accessListAssignments;
	my $ipAddrRe = get_crep('cipm');
	my $dRSecStarted = 0;

	# For starters, we need to have the routing process context available
	# Do NOT pass a $out param here, intentionaally
	_parse_routing_processes($in);

	# Get some info on the routing protocol neighbors
	# Specifically do not pass a '$out' param
	my $routingNeighborContext = parse_routing_neighbors($in, undef);

	# Parse through the routing data associated with various protocols
	my @supportedRoutingProtocols = qw(ospf eigrp rip);
	foreach my $rProt (@supportedRoutingProtocols) {

		# We can save ourselves heaps of time here if we know there's no definition for this proc
		if (not keys %{$in->{routingProcesses}->{$rProt}}) {
			next;
		}

		# We need a subnets arg. If we can't find it, we'll have to dredge it up by (brrr!) 
		# calling parse_interfaces to get it. If so, we aren't interested in the output mappings.
		# $accessListAssignments is a dontcare here.
		if (not defined $subnets) {
			# Do not pass in output context
			($subnets, $accessListAssignments) = parse_interfaces($in);
		}

		my @routingProtocolSectionKeys = ({targetProtocol => $rProt});
		foreach my $rSrcProt (@routingProtocolSectionKeys) {
			my $dynRoute;
			my $dRNet;
			my $routeNH;
			my $lastResortRoute = undef;

			# We choose one of several sources for routes here. We have a parser for static routes
			# (for "redirect static") which we can convert into routes here as necessary...
			if ($rSrcProt->{targetProtocol} eq 'static') {
				# 'static' takes no secondary keywords
				my $srRoutes;
				($srRoutes, $lastResortRoute) = _parse_static_route_data($in, $subnets);

				if (defined $srRoutes) {					
					foreach my $rStaticRoute (@{$srRoutes->{staticRoute}}) {

						# Don't bother with default route
						if ($rStaticRoute->{destinationAddress} =~ /^[0.]+$/) {
							next;
						}

						# TODO 37808: check route against referenced route filter. "next" if it doesn't pass
						$dRNet = {
							address => $rStaticRoute->{destinationAddress},
							mask => $rStaticRoute->{destinationMask}						
						};

						# Handle next hop. If the interface came as part of the SR definition, then use it
						$routeNH = {
							adjacencyInterface => (defined $rStaticRoute->{interface})
								? $rStaticRoute->{interface}  
								: _pick_subnet($rStaticRoute, $subnets),
						};

						if (defined  $rStaticRoute->{gatewayAddress}) {
							$routeNH->{adjacencyNode} = $rStaticRoute->{gatewayAddress};
						}

						# If the static route didn't map to an adjacencyInterface--meaning that
						# it mapped to something off of any local net--then the default gateway should have been
						# routed to. Need to have recorded the next hop for the route 
						# of last resort and route to it, filling in $routeNH appropriately.

						# Did we end up resolving a next hop?
						if ( !$routeNH->{adjacencyInterface} ) {
							# Route of last resort....
							if ($lastResortRoute) {
								$routeNH = { 
									adjacencyInterface => $lastResortRoute->{interface}, 
									adjacencyNode => $lastResortRoute->{gatewayAddress} // $lastResortRoute->{destinationAddress}
								}
							}else{
								$LOGGER->debug("Cisco IOS/parse_dynamic_routes: " . 
								"No last resort route data for: " . $dRNet->{address} . "/" . $$dRNet->{mask}); 						
							}
						}

						push @{$dynRoute->{routeNextHop}}, $routeNH;

						# Some marker to indicate origin
						$dynRoute->{advertisingRoutingProcess} = $rSrcProt->{targetProtocol};
						$dynRoute->{routeNetwork} = $dRNet;

						if (not $dRSecStarted) {
							$dRSecStarted = 1;
							$out->open_element("dynamicRoutes");
						}

						$out->print_element('dynamicRoute', $dynRoute);
						undef $dynRoute;
					}
				}
			} elsif ($rSrcProt->{targetProtocol} eq "connected") {
				# 'connected' takes no secondary keywords
				my $crTopo = _parse_connected_route_topology($in, $rProt);

				if (defined $crTopo) {
					foreach my $rConnectedRoute (@{$crTopo}) {
						$dRNet = {
							address => $rConnectedRoute->{routeAddress},
							mask => $rConnectedRoute->{routeMask}
						};

						$routeNH = {
							adjacencyInterface => $rConnectedRoute->{routeInterface},
							adjacencyNode => $rConnectedRoute->{routeAddress}
						};

						$dynRoute->{advertisingRoutingProcess} = $rSrcProt->{targetProtocol};  # Some marker to indicate origin
						$dynRoute->{routeNetwork} = $dRNet;
						push @{$dynRoute->{routeNextHop}}, $routeNH;					

						if (not $dRSecStarted) {
							$dRSecStarted = 1;
							$out->open_element("dynamicRoutes");
						}
						
						$out->print_element('dynamicRoute', $dynRoute);
						undef $dynRoute;
					}
				}
			}else{
				my $rRoutesProt = $rSrcProt->{targetProtocol};
				if (exists $in->{$rRoutesProt}->{routes}) {
					if (not $dRSecStarted) {
						$dRSecStarted = 1;
						$out->open_element('dynamicRoutes');
					}

					# This is effectively a "scan for all text following a start of line with a capital letter
					# immediately following it, until you see another start of line with a capital following it..."
					# kinda regexp.  Route specifiers begin with any of a number of single upper case letters,
					# then potentially a multichar discriminant, then it spills the beans on the route itself.							
					# Subnet mask following IP addr may or may not exist, default it to 32 if not present
					# Look for [adminDistance/metric] as well.
					while ($in->{$rRoutesProt}->{routes} =~ /\n([A-Z] .+?) ($ipAddrRe)(\/([\d]{1,2}))?\s(\[.+?)(?=\n\S)/msg) {
						$dRNet = {
							address => $2,
							mask => ($3) ? $4 : 32,
						};

						$dynRoute = { routeNetwork => $dRNet };
						my $routeInfoBlob = $5;

						while ($routeInfoBlob =~ /\[\d+\/(\d+)\]\s+via ($ipAddrRe),\s+\S+? (\S+)\n?/mg) {
							$routeNH = {
								adjacencyInterface => $3,
								adjacencyMetric => $1,
								adjacencyNode => $2
							};

							# And to reiterate...
							$dynRoute->{routeSourceNeighbor} = $2;

							# This should only really be assigned once anyways
							if (defined $routingNeighborContext->{$rRoutesProt}->{$dynRoute->{routeSourceNeighbor}}) {
								$dynRoute->{advertisingRoutingProcess} = 
								$routingNeighborContext->{$rRoutesProt}->{$dynRoute->{routeSourceNeighbor}}->{routeProcess};
							}

							push @{$dynRoute->{routeNextHop}}, $routeNH;
						}

						# Check that we generated a next hop
						if (not defined $dynRoute->{routeNextHop}) {
							if ( not $lastResortRoute ) {
								my $srDontcare;
								($srDontcare, $lastResortRoute) = _parse_static_route_data($in, $subnets);
							}

							$LOGGER->debug("Last Resort Route:\n" . Dumper($lastResortRoute));
							if ( $lastResortRoute ) {
								$dynRoute->{routeNextHop} = {
									adjacencyInterface => $lastResortRoute->{interface},
									adjacencyNode => $lastResortRoute->{gatewayAddress}
								};
							} else {
								$LOGGER->debug("Cisco IOS/parse_dynamic_routes" . 
								"Unable to detect next hop data for: " . $dynRoute->{routeNetwork}->{address} . 
								"/" . $dynRoute->{routeNetwork}->{mask}); 					
							}
						}

						# And generate the route
						$out->print_element('dynamicRoute', $dynRoute);
						undef $dynRoute;
					}
				}
			}
		} 
	}

	if ($dRSecStarted) {
		$out->close_element('dynamicRoutes');
	}

	# Clear out context for next run
	if (defined $in->{routingProcesses}) {
		delete $in->{routingProcesses};
	}

	foreach my $rProt ('rip', 'bgp', 'eigrp', 'ospf', 'static', 'connected') {
		if (defined $in->{$rProt}) {
			delete $in->{$rProt};
		}
	}
}

sub parse_telemetry_interfaces {
	my ($in, $out) = @_;
	my $interfaces;

	while ($in->{interfaces} =~ /(\S+)\s+(is.+?)(?=^\S)/msg) {
		my $blob      = $2;
		my $name      = $1;
		my $interface = {
			name       => $name,
			type       => get_interface_type($name),
			inputBytes => 0,
		};

		if ($blob =~ /line protocol is (down|up)/i) {
			$interface->{operStatus} = ucfirst($1);
		}
		if ($blob =~ /packets input,.*?(\d+)\s*bytes/) {
			$interface->{inputBytes} = $1;
		}

		while ($blob =~ /Internet\s+address\s+is\s+([\d.:a-fA-F]+)\/(\d+)/gi) {
			my $ipEntry = {
				ipAddress => $1,
				mask      => $2,
				-virtual  => 0,
			};
			push(@{$interface->{ipEntry}}, $ipEntry);
		}

		# Now push any virtual IP addresses
		if ($in->{standby} =~ /$name\b.+?Virtual IP address is\s+(\S+).+?(?=^\S)/ms) {
			my $ipEntry = { ipAddress => $1, -virtual => 1, };
			push(@{$interface->{ipEntry}}, $ipEntry);
		}

		push(@{$interfaces->{interface}}, $interface);
	}

	# This needs to be assigned as attribute-level encoding
	$out->attributes(1);
	$out->print_element('interfaces', $interfaces);

	return $interfaces;
}

sub parse_mac_table {
	my ($in, $out) = @_;
	my $opened = 0;
	my $macEntry;

	# there are three different forms, one of which completely reorders the fields
	if ($in->{mac} =~ /Destination Address\s+Address Type\s+VLAN\s+Destination Port/i) {
		my $macAddrRe = get_crep('mac3');
		while ($in->{mac} =~ /($macAddrRe)\s+\w+\s+(\d+)\s+(\S+)/g) {
			if (!$opened) {
				$opened = 1;
				$out->open_element('macTable');
			}

			$macEntry = {
				vlan       => $2,
				macAddress => strip_mac($1),
				interface  => _full_int_name($3)
			};
			$out->print_element('macEntry', $macEntry);
		}
	}else{
		my $regex;
		if ($in->{mac} =~ /Vlan\s*Mac Address\s*Type\s*Ports/i) {
			$regex = '(\S+)\s*([\da-f.]{14})\s*\S+\s*(\S+)';
		} elsif ($in->{mac} =~ /Destination Address\s+Address Type\s+VLAN\s+Destination Port/i) {
			# TODO: What is the point in this?
			my $macAddrRe = get_crep('mac3');
		} else {
			$regex = '(\S+)\s*([\da-f.]{14})\s*\S+\s*(?:Yes|No)\s*(\S+)';
		}
		while ($in->{mac} =~ /$regex/g) {
			if (!$opened) {
				$out->open_element('macTable');
				$opened = 1;
			}
			$macEntry = {
				vlan       => $1,
				macAddress => strip_mac($2),
				interface  => _full_int_name($3),
			};
			$out->print_element('macEntry', $macEntry);
		}
	}
	if ($opened) {
		$out->close_element('macTable');
	}
}

sub parse_vtp {
	my ($in, $out) = @_;
	my $vtp;

	if ($in->{vtp_status} =~ /^VTP Version\s*:\s*(\d+)$/mi) {
		$vtp->{'cisco:version'} = $1;
	}
	if ($in->{vtp_status} =~ /^Configuration Revision\s*:\s*(\d+)$/mi) {
		$vtp->{'cisco:configVersion'} = $1;
	}
	if ($in->{vtp_status} =~ /^Maximum VLANs supported locally\s*:\s*(\d+)$/mi) {
		$vtp->{'cisco:maxVlanCount'} = $1;
	}
	if ($in->{vtp_status} =~ /^Number of existing VLANs\s*:\s*(\d+)$/mi) {
		$vtp->{'cisco:vlanCount'} = $1;
	}
	if ($in->{vtp_status} =~ /^VTP Operating Mode\s*:\s*(\S+)$/mi) {
		$vtp->{'cisco:localMode'} = $1;
	}
	if ($in->{vtp_status} =~ /^VTP Domain Name\s*:\s*(\S+)$/mi) {
		$vtp->{'cisco:domainName'} = $1;
	}
	if ($in->{vtp_status} =~ /^VTP Pruning Mode\s*:\s*(\S+)$/mi) {
		$vtp->{'cisco:vlanPruningEnabled'} = ($1 =~ /enabled/i) ? 'true' : 'false';
	}
	if ($in->{vtp_status} =~ /^VTP V2 Mode\s*:\s*(\S+)$/mi) {
		$vtp->{'cisco:v2Mode'} = $1;
	}
	if ($in->{vtp_status} =~ /^VTP Traps Generation\s*:\s*(\S+)$/mi) {
		$vtp->{'cisco:alarmNotificationEnabled'} = ($1 =~ /enabled/i) ? 'true' : 'false';
	}
	if ($in->{vtp_status} =~ /^MD5 digest\s*:\s*(.*[^\s])$/mi) {
		$vtp->{'cisco:password'} = $1;
	}
	if ($in->{vtp_status} =~ /^last modified by\s+(\S+)/mi) {
		$vtp->{'cisco:lastUpdater'} = $1;
	}
	if (defined $vtp->{localMode}) {
		$vtp->{'cisco:serviceType'} = 'vtp';
	}

	$out->print_element('cisco:vlanTrunking', $vtp);
}

sub parse_static_routes {
	my ($in, $out, $subnets) = @_;
	_parse_static_route_data($in, $subnets);
	_print_static_routes($in, $out);
}

# _parse_static_route_data
#
# This routine creates a ref to a hash which is compatible with a direct 
# print_element() into the output context of a caller (or, just for
# analysis as a hash). IF it finds no static route data, it returns 
# undef...caller should check return appropriately.  We also choose this
# as the place to record the route-of-last-resort.
#
# PARAMETERS: 
#    $in: response data.  To be treated as a READ-ONLY hash
#    $subnets: an output from parse_interfaces, a lookaside hash with which to
#          decide forwarding interfaces (via _pick_subnet)
#    
# RETURNS:
#    a 2-element tuple (duple?):
#     * Array of static routes, each element being an instance of the StaticRoute
#        complex type encoding
#     * A *single* instance of the StaticRoute complex type encoding which is the
#         discovered route of last resort (or undef if none is found).  If defined, this
#         will have the interface, and gateway address DEFINITELY assigned.
#	UPDATE - Populates $in->_remoteGatewayResolution with routes (including route of last resort)
#			 that have been resolved using RemoteGatewayResolution in adapters/common. 
#
sub _parse_static_route_data {
    my ($in, $subnets) = @_;
    my $staticRoutes = undef;
    my $defaultNetwork = "notAssigned";  # Intentionally initialized this way => not an IP
	if (!defined($in->{_remoteGatewayResolution}))
	{
		$in->{_remoteGatewayResolution} = ZipTie::RemoteGatewayResolution->new();
	}

   	# Now, record ip default-network... to identify the route of last resort.
	# We need this now to be able to identify the route which will represent the
	# forwarding data for this.	
	# 
	my $lastResortRoute = _parse_last_resort_route($in, $subnets);
	    
    while ( $in->{running_config} =~
	    /^ip route\s+(\d+\.\S+)\s+(\d+\.\S+)\s+(\S+)(.*)/mig )
    {
    	my $route = {};
    	my $destinationAddress = $1;
    	my $destinationMask = mask_to_bits($2);
		my $gatewayAddress = $3;
		my $remainder = $4;
		my $otherDetails = {};
		my $defaultGateway = 0;
		
		$route->{destinationAddress} = $destinationAddress;
		$route->{destinationMask} = $destinationMask;
		
		if ( $gatewayAddress =~ /\d+\.\d+\.\d+\.\d+/ ) 
		{
		    $route->{gatewayAddress} = $gatewayAddress;
		}
		elsif ($gatewayAddress =~ /Null\d+/i)
		{
		    # This is a route to the bit bucket
		    $route->{discard} = 'true';	

		    # Interface is required
		    $route->{interface} = $gatewayAddress;
		}
		else 
		{
		    $route->{interface} = $gatewayAddress;
		}
		
		if (!defined $route->{discard})
		{
			# Just for some legacy tests that have several means of setting default g/w simultaneously, we always
			# mark a 0.0.0.0 as a default gateway (why not)
		    if ( ( defined($lastResortRoute) && $route->{destinationAddress} eq $lastResortRoute->{destinationAddress} ) or
		    	($route->{destinationAddress} eq '0.0.0.0') )
		    {
				$defaultGateway = 1;
		    }

		    if (
			( defined $remainder )
			&& ( ( $remainder =~ /^\s*(\d+)\s*$/ )
			       || ( $remainder =~ /^\s*(\d+)\s+.*$/ )
			       || ( $remainder =~ /^\s*\S+\s+(\d+)\s+.*$/ ) )
			)
		    {
				$route->{routePreference} = $1;
		    }
		}

		$otherDetails->{defaultGateway} = $defaultGateway ? 'true' : 'false';
		
		# Pass in attributes to add_route for route object creation
		$in->{_remoteGatewayResolution}->add_route($route->{destinationAddress}, $route->{destinationMask}, $route->{gatewayAddress}, $route->{routePreference}, $defaultGateway, $otherDetails);
		
		# Add additional attributes to route object not set in add_route
		my @routes = @{$in->{_remoteGatewayResolution}->{_routes}};
		$routes[$#routes]->{interface} = $route->{interface};
		if ( defined $route->{discard} ) {
			$routes[$#routes]->{discard} = $route->{discard};
		}
	
		# Retain $staticRoutes to avoid breaking dependencies in parse_dynamic_routes	
		push( @{ $staticRoutes->{staticRoute} }, $route );
    }
    
    # Pass in the populated _remoteGatewayResolution object to resolve all next hop addresses
    $in->{_remoteGatewayResolution}->resolve_remote_gateways();

	# Static Routes to be output are within the _remoteGatewayResolution->resolvedRoutes object however    
    # Return these to avoid breaking dependencies in parse_dynamic_routes
    return ($staticRoutes, $lastResortRoute);
}

sub _print_static_routes {
	
	my ($in, $out) = @_;

	my $staticRoutesOpen = 0;
	foreach my $route (@{$in->{_remoteGatewayResolution}->get_routes()})
	{
		# Don't print out dynamic routes.
		if (defined($route->{otherDetails}->{dynamicRoute}))
		{
			next;
		}

		my $printableRoute = {};
		if (defined $route->{gatewayAddress}) { $printableRoute->{gatewayAddress} = $route->{gatewayAddress}; }
		$printableRoute->{destinationAddress} = $route->{destinationAddress};
		$printableRoute->{destinationMask} = $route->{destinationMask};
		$printableRoute->{interface} = $route->{interface};
		$printableRoute->{defaultGateway} = $route->{otherDetails}->{defaultGateway};
		$printableRoute->{routeMetric} = $route->{routeMetric} if (defined($route->{routeMetric}));
		if (defined $route->{discard}) { $printableRoute->{discard} = $route->{discard}; }

		if (!$staticRoutesOpen)
		{
			$staticRoutesOpen = 1;
			$out->open_element('staticRoutes');
		}

		$out->print_element('staticRoute', $printableRoute);
	}
	# Print the crypto map routes
	if (defined($in->{_crypto_map_routes}))
	{
		foreach my $route (@{$in->{_crypto_map_routes}})
		{
			if (!$staticRoutesOpen)
			{
				$staticRoutesOpen = 1;
				$out->open_element('staticRoutes');
			}

			$out->print_element('staticRoute', $route);
		}
	}

	if ($staticRoutesOpen)
	{
		$out->close_element('staticRoutes');
	}

	if (scalar(@{$in->{_remoteGatewayResolution}->get_problem_routes()}))
	{
		$LOGGER->debug('parse_static_routes: Couldn\'t resolve remote gateway addresses for ' . scalar(@{$in->{_remoteGatewayResolution}->get_problem_routes()}) . ' routes. Routes dropped.');
	}

	delete($in->{_remoteGatewayResolution});
}


# This returns *either* an array of connected routes or undef. Caller should check appropriately.
sub _parse_connected_route_topology {
	my ($in, $connectedFrom) = @_;
	my $ipAddrRe = get_crep('cipm');
	my @connectedRoutes;

	if (defined ($in->{$connectedFrom}->{topology})) {
		while ($in->{$connectedFrom}->{topology} =~ /\n[A-Z]\s+($ipAddrRe)\/([\d]{1,2}),\s\d+ successors?,\sFD\sis\s\d+\n\s+via Connected, (\S+)/msg) {
			my $thisConnectedRoute = {
				routeAddress => $1,
				routeMask => $2,
				routeInterface => $3
			};	

			push  @connectedRoutes, $thisConnectedRoute;
		}					
	}

	# Guarantee return of an empty array
	return \@connectedRoutes;
}

# Choose from a hash of subnets and return the matching value
sub _pick_subnet {
	my ($route, $subnets) = @_;

	if ($route->{interface}) {
		return $route->{interface};
	} elsif (defined $subnets) {
		foreach my $interfaceName ( sort ( keys(%$subnets) ) ) {
			my @subnetsArray = @{ $subnets->{$interfaceName} };
			foreach my $subnet (@subnetsArray) {
				if ( $subnet->contains( $route->{gatewayAddress} ) ) {
					return $interfaceName;
				}
			}
		}
	} else {
		return "";
	}
}

sub _parse_last_resort_route {
	my ($in, $subnets) = @_;
	my $lastResortRoute = undef;
	my $defaultNetwork;
	
	# For starters, scan for ip default-network...
	# Typically, this looks like: 
	#   
	# ip default-network 198.115.237.0
	# ....
	# ip route 198.115.237.0 255.255.255.0 158.123.2.1
	
    if ( $in->{running_config} =~
	 /^ip default-network\s+(\d+\.\d+\.\d+\.\d+)/im )
	{
		$defaultNetwork = $1;
		
		# Now, just out and look for a static route to the default net
		if ( $in->{running_config} =~ /^ip route\s+$defaultNetwork\s+(\d+\.\S+)\s+(\S+)/mi)
		{
			$lastResortRoute = 
			{
				destinationAddress => $defaultNetwork,
				destinationMask => mask_to_bits($1),
			};
			my $gateway = $2;
			if ($gateway =~ /^[\d\.]+/)
			{
				$lastResortRoute->{gatewayAddress} = $gateway;							
				$lastResortRoute->{interface} = _pick_subnet($lastResortRoute, $subnets);
			}
			else # interface identifier
			{
				$lastResortRoute->{interface} = $gateway;
			}
		}	
	}
	
	if (not $lastResortRoute)
	{
		# If not that, look for ip default-gateway, which should only be used in the absence of routing
	    if ( $in->{running_config} =~
		 /^ip default-gateway\s+(\d+\.\d+\.\d+\.\d+)\s*$/im )
	    {
			$lastResortRoute = 
			{
			    destinationAddress => '0.0.0.0',
			    destinationMask    => '0',
			    gatewayAddress     => $1,
			    defaultGateway     => 'true',
			};
			
			# Save off route of last resort....
			$lastResortRoute->{interface} = _pick_subnet( $lastResortRoute, $subnets );
	    }
		else
		{
			# Failing that, look for a route to 0.0.0.0.  This is the same semantically as ip default-network, but
			# has some limitations, such as the fact that IGRP/EIGRP will only route to or propagate it under certain
			# circumstances.
			$defaultNetwork = "0.0.0.0";
			
			if ( $in->{running_config} =~ /^ip route\s+$defaultNetwork\s+\d+\.\S+\s+(\d+\.\S+)/mi)
			{
				$lastResortRoute = 
				{
					destinationAddress => $defaultNetwork,
					destinationMask => 0,
					gatewayAddress => $1,
				};
				$lastResortRoute->{interface} = _pick_subnet($lastResortRoute, $subnets);
			}	
			
		}
	}		
		
	return $lastResortRoute;
}
sub parse_chassis {
    my ( $in, $out ) = @_;

    my $cpuref;
    my $cpuid           = -1;
    my $description     = "";
    my $systemimagefile = "";
    my $activeslot      = "";
    my $mac             = "";

    my ( $flashpbtype, $flashpbname, $flashpbsize ) = ( 0, 0, -1 );
    my ( $flashpc0type, $flashpc0name, $flashpc0size, $flashpc0num ) =
	( 0, 0, -1, 0 );
    my ( $flashpc1type, $flashpc1name, $flashpc1size, $flashpc1num ) =
	( 0, 0, -1, 0 );
    my ( $flashpu0type, $flashpu0name, $flashpu0size, $flashpu0num ) =
	( 0, 0, -1, 0 );
    my ( $flashpu1type, $flashpu1name, $flashpu1size, $flashpu1num ) =
	( 0, 0, -1, 0 );
    my ( $flashp1type, $flashp1name, $flashp1size, $flashp1num ) =
	( 0, 0, -1, 0 );
    my ( $flashp2type, $flashp2name, $flashp2size, $flashp2num ) =
	( 0, 0, -1, 0 );
    my ($dualFlashFlag) = 0;
    my ( $configMemory, $ramMemory, $packetMemory, $processorBoard );
    my ( $ciscoModel, $serialNumber ) = ( 0, 0 );

    foreach my $line ( split( /\n/, $in->{version} ) ) {
	if (
	    (
	     $line =~
	     /Cisco \S+\s+\((.*)\) processor\s+.*\s+with \S+ bytes of memory/i
	    )
	    or ( $line =~
		 /Cisco \S+\s+\((.*)\) processor with \S+ bytes of memory./i )
	    )
	{
	    $cpuid          = 0;
	    $processorBoard = $1;
	}
	elsif (
	    ( $line =~ /^(\S+\s+CPU\s+at\s+\S+,.*)/i )
	    or ( $line =~ /^(\S+)\s+CPU\s+at\s+\S+,.*/i )
	    or ( $line =~
		 /Cisco Catalyst ([19|28].*\s+processor) with \S+ bytes of memory/i
	    )
	    or ( $line =~ /CPU part number (\w+)/i )
	    )
	{
	    $cpuid       = 0;
	    $description = $1;
	}
	elsif ( $line =~ /System\s+image\s+file\s+is\s+"(\S+)"/i ) {
	    $systemimagefile = $1;
	    if ( $systemimagefile =~ /(\S+):(\S+)/i ) {
		$activeslot = $1;
	    }
	    elsif ( $line =~
		    /System\s+image\s+file\s+is\s+"\S+".*via\s+([A-Za-z]+)/i )
	    {
		$activeslot = $1;
	    }
	}

	#20480K bytes of Flash PCMCIA card at slot 0 (Sector size 128K).
	#20480K bytes of Flash PCMCIA card at slot 1 (Sector size 128K).
	elsif ( $line =~ /(\d+)K bytes of Flash PCMCIA card at (\w+) (\d) \(/i )
	{
	    $dualFlashFlag = 1;
	    if ( $3 == 0 ) {
		( $flashpc0type, $flashpc0name, $flashpc0size, $flashpc0num ) =
		    ( $2 . $3, "pcmciacard", $1, 0 );
	    }
	    elsif ( $3 == 1 ) {
		( $flashpc1type, $flashpc1name, $flashpc1size, $flashpc1num ) =
		    ( $2 . $3, "pcmciacard", $1, 1 );
	    }
	}

	#20480K bytes of processor board PCMCIA Slot0 flash (Read/Write)
	#20480K bytes of processor board PCMCIA Slot1 flash (Read/Write)
	elsif (
	    $line =~ /(\d+)K bytes of processor board PCMCIA (\w+) flash \(/i )
	{
	    $dualFlashFlag = 1;
	    my ( $flashsize, $slotnum ) = ( $1, $2 );
	    if ( $slotnum =~ /0/ ) {
		( $flashpu0type, $flashpu0name, $flashpu0size, $flashpu0num ) =
		    ( lcfirst($slotnum), "pcmcia", $flashsize, 0 );
	    }
	    elsif ( $slotnum =~ /1/ ) {
		( $flashpu1type, $flashpu1name, $flashpu1size, $flashpu1num ) =
		    ( lcfirst($slotnum), "pcmcia", $flashsize, 1 );
	    }
	}

	#8192K bytes of processor board System flash partition 1 (Read/Write)
	elsif ( $line =~
		/(\d+)K bytes of processor board System flash partition 1 \(/i )
	{
	    $dualFlashFlag = 1;
	    ( $flashp1type, $flashp1name, $flashp1size, $flashp1num ) =
	    { "flash", "partition", $1, 1 };
	}

	#8192K bytes of processor board System flash partition 2 (Read/Write)
	elsif ( $line =~
		/(\d+)K bytes of processor board System flash partition 2 \(/i )
	{
	    $dualFlashFlag = 1;
	    ( $flashp2type, $flashp2name, $flashp2size, $flashp2num ) =
		( "flash", "partition", $1, 2 );
	}

	#8192K bytes of processor board System flash (Read/Write)
	elsif ( $line =~ /(\d+)K bytes of processor board System flash \(/i ) {
	    $dualFlashFlag = 1;
	    ( $flashpbtype, $flashpbname, $flashpbsize ) =
		( "flash", "none", $1 );
	}

	# Configuration memory, RAM, Packet memory sizes.
	if (
	    ( $line =~ /Cisco\s+.*?with\s+(\d+)K\/(\d+)K\s+bytes of .*memory/i )
	    or
	    ( $line =~ /processor.*?with\s+(\d+)K\/(\d+)K\s+bytes of .*?memory/i )
	    )
	{

	    # See bug #7740 for more information on ramMemory and packetMemory
	    ( $ramMemory, $packetMemory ) = ( $1, $2 );

	    # Round up the ram size, if residue > 819 KBs (80% of 1 MB).
	    my $roundup = $ramMemory % 1024 > $KBS_80PERCENTMB ? 1 : 0;
	    my $ramMBs = int( my $volatile = $ramMemory / 1024 ) + $roundup;

	    # Use the ram size value, if it is reasonable (2**n);
	    # Otherwise, choose the (ram+packet) sum as the size value.
	    $ramMemory +=
		( $SIZEPATTERNS_STRING =~ /\s($ramMBs)\s/ )
		? 0
		: $packetMemory;
	}
	elsif ( $line =~ /Cisco\s+.*?with\s+(\d+)K\s+bytes of .*?memory/i ) {
	    $ramMemory = $1;
	}
	if (
	    ( $line    =~ /(\d+)K bytes of non-volatile configuration memory/i )
	    or ( $line =~
		 /(\d+)K bytes of flash-simulated non-volatile configuration memory./i
	    )
	    )
	{
	    $configMemory = $1;
	}
	elsif ( $line =~ /(\d+)K bytes of NVRAM/i ) {
	    $configMemory = $1;
	}

	if ( $line =~ /^IOS\s+\(tm\)\s+c(16|17|26|29)\d\d(XL)?\s+Software.*/i )
	{
	    $ciscoModel = $1;
	}

	if ( defined $ciscoModel && $ciscoModel =~ /^16$|^17$|^26$/ ) {
	    if ( $line =~ /^Processor\s+board\s+ID\s+(\w+)\s+(\(\d+\))/i ) {
			$serialNumber = $1;
	    }
	}
	if ( defined $ciscoModel && $ciscoModel == 29 ) {
	    if (   ( $line =~ /System\s+serial\s+number:\s+(\S+)/i )
		   or ( $line =~ /System\s+serial\s+number\s+:\s+(\S+)/i ) )
	 	{
			$serialNumber = $1;
	    }
	}
	else {
	    if ( $line =~ /^Processor\s+board\s+ID\s+([\S+ ]+).?/i ) {
			$serialNumber = $1;
	    }
	}
    }

    # only use the SNMP Chassis ID if it is longer than 6 chars
    if ( !defined($serialNumber) && $in->{snmp} =~ /^Chassis:\s*(\S{6,})/mi ) {
		$serialNumber = $1;
    }

    $out->open_element("chassis");

    my $chassisAsset = { "core:assetType" => "Chassis", };
    $chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $serialNumber
	if $serialNumber;
    $chassisAsset->{'core:factoryinfo'}->{'core:make'} = "Cisco";
    if ( $in->{version} =~ /^(?:Product\/)?Model number\s*:\s*(\S+)/mi ) {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~
	    /^cisco\s+((?:WS-C|Cat|AS|C|VG)?\d{3,4}\S*\b)\S*\b/mi )
    {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~ /^cisco\s+((?:SOHO)?\d{2}\S*\b)\S*\b/mi ) {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~ /^cisco\s+((?:uBR)?\d{3}\S*\b)\S*\b/mi ) {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~
	    /^Cisco\s+(\S+)(?:\s\(\S+\))?\s*processor.+?(?=with)/mi )
    {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~
	    /^Cisco\s+(\S+)\s.+?Voice\sLinecard.+?\sprocessor(?:\s\(.+?\))? with/mi
	)
    {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~ /^cisco\s+catalyst\s+(\d{3,4})\s/mi ) {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }
    elsif ( $in->{version} =~
	    /^cisco CISCO(\d+) [(]R\d+[)] processor [(]revision/mi )
    {
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
    }

    $out->print_element( "core:asset", $chassisAsset );

    # chassis/core:description
    if ( $in->{version} =~
	 /(^Cisco\s+(?:Internetwork\s+Operating\s+System|IOS)\s+Software.+^Compiled.+?$)/msi
	)
    {
	$out->print_element( "core:description", $1 );
    }
    elsif ( $in->{version} =~
	    /(Cisco Catalyst .*? Enterprise Edition Software.+Copyright.+?$)/msi )
    {
	$out->print_element( "core:description", $1 );
    }

    _parse_cards( $in, $out );
    
    # Not every chassis description details CPU attributes, but <cpu> is an optional element in the schema
    if ($cpuid != 1)
    {
	my $cpu;
	$cpu->{"core:description"} = $description if $description;
	$cpu->{cpuType} = $processorBoard if  $processorBoard;
	if ($cpu) 
	{
	    $out->print_element( "cpu", $cpu );					
	}
    }
    
    _parse_file_storage( $in, $out );

    # Chassis->MACAddress
    if ( $in->{version} =~ /Base\sEthernet\s(?:MAC |)Address:\s+([^\s]+)/msi ) {
	$mac = $1;
    }
    elsif ( $in->{version} =~ /Ethernet\sAddress:\s+([^\s]+)/msi ) {
	$mac = $1;
    }
    if ($mac) {
	$mac =~ s/[^\da-z]//gi;
	$out->print_element( "macAddress", $mac );
    }

    my @memories = ();
    push @memories, { kind => 'Flash', size => $flashpbsize * 1024 }
    if $flashpbsize != -1
	and $flashpbtype ne "";
    push @memories, { kind => 'Flash', size => $flashpc0size * 1024 }
    if $flashpc0size != -1
	and $flashpc0type ne "";
    push @memories, { kind => 'Flash', size => $flashpc1size * 1024 }
    if $flashpc1size != -1
	and $flashpc1type ne "";
    push @memories, { kind => 'Flash', size => $flashpu0size * 1024 }
    if $flashpu0size != -1
	and $flashpu0type ne "";
    push @memories, { kind => 'Flash', size => $flashpu1size * 1024 }
    if $flashpu1size != -1
	and $flashpu1type ne "";
    push @memories, { kind => 'Flash', size => $flashp1size * 1024 }
    if $flashp1size != -1
	and $flashp1type ne "";
    push @memories, { kind => 'Flash', size => $flashp2size * 1024 }
    if $flashp2size != -1
	and $flashp2type ne "";
    push @memories,
    { 'core:description' => 'RAM', kind => 'RAM', size => $ramMemory * 1024 }
    if $ramMemory;
    push @memories,
    {
	'core:description' => 'PacketMemory',
	kind               => 'PacketMemory',
	size               => $packetMemory * 1024
    }
    if $packetMemory;
    push @memories,
    { kind => 'ConfigurationMemory', size => $configMemory * 1024 }
    if $configMemory;

    foreach my $memory (@memories) {
	$out->print_element( "memory", $memory );
    }

    # Power supply parsing
    my $foundPowerSupplies;
    if ( defined $in->{'show_power'}
	 && $in->{'show_power'} =~ /(^PS.+?)(?=^\s)/ms )
    {
	my $psBlob = $1;
	if ( $in->{'show_power'} =~ /Supply\s+Model\s+No\s+Type\s+Status/ ) {

	    # 4000 chassis
	    while ( $psBlob =~ /^PS(\d+)\s+(\S+)\s+(\S+.+?W\b)\s+(\S+)/mg ) {
		my $number = $1;
		my $powerSupply = { number => $number, };
		$powerSupply->{'core:asset'}->{'core:assetType'} =
		    "PowerSupply";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:make'} = "Unknown";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:modelNumber'} = "Unknown";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:partNumber'} = $2;
		$powerSupply->{'core:description'} = $3;

		my $status = $4;
		if ( $status =~ /good/i ) {
		    $status = "OK";
		}
		else {
		    $status = "Fault";
		}
		$powerSupply->{'status'} = $status;

		$out->print_element( "powersupply", $powerSupply );
		$foundPowerSupplies = 1;
	    }
	}
	else {
	    while ( $psBlob =~
		    /(\d+)\s+(\S+)\s+(\d+.\d\d)\s+(\d+.\d\d)\s+(OK|\S+)\s+(OK|\S+)\s+(on|off)/mg
		)
	    {
		my $number      = $1;
		my $status      = $6;
		my $powerSupply = { number => $number, };
		$powerSupply->{'core:asset'}->{'core:assetType'} =
		    "PowerSupply";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:make'} = "Unknown";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:modelNumber'} = "Unknown";
		$powerSupply->{'core:asset'}->{'core:factoryinfo'}
		->{'core:partNumber'} = $2;
		if ( $in->{'show_inventory'} =~
		     /PS\s+$number\b.+?DESCR:\s"(.+?)".+?SN:\s+(\S+)/s )
		{
		    $powerSupply->{'core:description'} = $1;
		    $powerSupply->{'core:asset'}->{'core:factoryinfo'}
		    ->{'core:serialNumber'} = $2;
		}

		if ( $status =~ /OK/i ) {
		    $status = "OK";
		}
		else {
		    $status = "Fault";
		}
		$powerSupply->{'status'} = $status;

		$out->print_element( "powersupply", $powerSupply );
		$foundPowerSupplies = 1;
	    }
	}
    }

    # for Catalyst 3000 series switches
    if ( ( $in->{version} =~ /Power\s+supply\s+part\s+number\s*:\s+(\S+)/i )
	 && !$foundPowerSupplies )
    {
	my $powerSupply = { number => 1, };
	$powerSupply->{'core:asset'}->{'core:assetType'} = "PowerSupply";
	$powerSupply->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} =
	    "Unknown";
	$powerSupply->{'core:asset'}->{'core:factoryinfo'}
	->{'core:modelNumber'} = "Unknown";
	$powerSupply->{'core:asset'}->{'core:factoryinfo'}
	->{'core:partNumber'} = $1;
	if ( $in->{version} =~ /Power\s+supply\s+serial\s+number\s*:\s+(\S+)/i )
	{
	    $powerSupply->{'core:asset'}->{'core:factoryinfo'}
	    ->{'core:serialNumber'} = $1;
	}
	$out->print_element( "powersupply", $powerSupply );
    }
    $out->close_element("chassis");
}

sub parse_system {
    my ( $in, $out ) = @_;

    # This is latched for subsequent signature qualification
    ($systemName) = $in->{running_config} =~ /^hostname (\S+)/m;
    $out->print_element( 'core:systemName', $systemName );

    $out->open_element('core:osInfo');
    if ( $in->{version} =~ /^System image file is "([^"]+)/m ) {
	$out->print_element( 'core:fileName', $1 );
    }
    $out->print_element( 'core:make', 'Cisco' );
    if ( $in->{version} =~ /^(IOS.+?),/m ) {
	$out->print_element( 'core:name', $1 );
    }
    if ( $in->{version} =~ /^(?:Cisco )?IOS.+Version (\S[^\s,]+)/mi ) {
	$out->print_element( 'core:version', $1 );
    }
    elsif ( $in->{version} =~ /^Version\s+V(\d+\.\d+\.\d+)/mi ) {
	$out->print_element( 'core:version', $1 );
    }
    $out->print_element( 'core:osType', 'IOS' );
    
    # Call to generate the manifest of phases that the device uses sequentially on a given packet.
    # This gives the calling application a hint as to how to simulate device operation.
    _generate_packet_processing_declaration($out);

    if ( $in->{install_run} !~
	 /%\s+[Ii]nvalid input|%\s+[Ee]rror|%\s+[Uu]nknown/mi )
    {
	while ( $in->{install_run} =~ /^(.+)$/mig ) {
	    $_ = $1;
	    next if (/^\s*$/);
	    next if (/[Ss]oftware/i);

	    if (/^\s*(B|P|MP)\s+(\S+)\s+(\S+)(?:\s+-\s+Version\s+(\S+))?/i) {
		my $swMod;
		$swMod->{'core:make'}     = 'CISCO';
		$swMod->{'core:fileName'} = $3;
		$swMod->{'core:state'}    = $2;
		if ($4) {
		    $swMod->{'core:version'} = $4;
		}

		if ( uc($1) eq 'B' ) {
		    $swMod->{'core:moduleType'} = 'Base';
		}
		elsif ( uc($1) eq 'P' ) {
		    $swMod->{'core:moduleType'} = 'Patch';
		}
		elsif ( uc($1) eq 'MP' ) {
		    $swMod->{'core:moduleType'} = 'Maintenance';
		}
		if ( $3 =~ /([^\/]+)$/ ) {
		    $swMod->{'core:name'} = $1;
		}

		$out->print_element( 'core:softwareModule', $swMod );
	    }
	}
    }

    $out->close_element('core:osInfo');

    if (
	(
	 $in->{version} =~
	 /ROM:\s+System\s+Bootstrap[,]\s+Version\s+([^\s,]+)/i
	)
	or ( $in->{version} =~ /ROM:\s+Bootstrap\s+program\s+is\s+([\S ]+)/i )
	or ( $in->{version} =~
	     /BOOTLDR:\s+\S+\s+Boot\s+Loader.*?Version\s+([^\s,]+)/i )
	or ( $in->{version} =~ /^ROM:\s+TinyROM\s+version\s+([^\s]+)/mi )
	or ( $in->{version} =~ /^ROM:\s+([^\s]+)/mi )
	)
    {
	$out->print_element( 'core:biosVersion', $1 );
    }

    if ( $in->{version} =~ /\b(cat|(WS|ME)-C)\d{4}|catalyst|CIGESM/i ) {
	$out->print_element( 'core:deviceType', 'Switch' );
    }
    elsif ( $in->{version} =~ /\bC1200\b|\bAIR/ ) {
	$out->print_element( 'core:deviceType', 'Wireless Access Point' );
    }
    else {
	$out->print_element( 'core:deviceType', 'Router' );
    }

    my ($contact) = $in->{running_config} =~ /^snmp-server contact ([\S ]+)/m;
    
    if ($contact)
    {  
		$out->print_element( 'core:contact', $contact );		
    }
    
    # System restarted at 18:00:01 CST Sun Feb 28 1993
    if ( $in->{version} =~
	 /^System restarted (?:by \S+\s)?at\s+(\d{1,2}:\d{1,2}:\d{1,2})\s+(\S+)\s+\S+\s+(\S+)\s+(\d{1,2})\s+(\d{4})/mi
	)
    {
	my $year     = $5;
	my $month    = $3;
	my $day      = $4;
	my $time     = $1;
	my $timezone = $2;

	my ( $hour, $min, $sec ) = $time =~ /(\d+):(\d+):(\d+)/;

	$out->print_element(
	    "core:lastReboot",
	    seconds_since_epoch(
		$sec, $min, $hour, $day, $month, $year, $timezone
	    )
	    );
    }
    elsif ( $in->{version} =~ /uptime is\s+(.+)/i ) {
	$_ = $1;
	my ($years)   = /(\d+)\s*years?/;
	my ($weeks)   = /(\d+)\s*weeks?/;
	my ($days)    = /(\d+)\s*days?/;
	my ($hours)   = /(\d+)\s*hours?/;
	my ($minutes) = /(\d+)\s*minutes?/;

	# subract the last reboot from the current time
	my $lastReboot = time();
	$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
	$lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
	$lastReboot -= $days * 24 * 60 * 60           if ($days);
	$lastReboot -= $hours * 60 * 60               if ($hours);
	$lastReboot -= $minutes * 60                  if ($minutes);
	$out->print_element( "core:lastReboot", $lastReboot );
    }
}

sub _generate_packet_processing_declaration
{
	my $out = shift;
	
	$out->open_element("core:packetProcessingOrdering");
	foreach my $pRPhase ('layer2-bridging','packet-filtering/source', 'qos-evaluation/source', 'routing', 
		'address-translation', # addr-xlate overriden in sections anyways 
		'packet-filtering/destination', 'qos-evaluation/destination')
	{
		my ($pPhase, $pPosition) = split ('/', $pRPhase);
		$out->open_element( 'core:packetProcessingElement');
		$out->print_element( 'core:packetProcessingPhase', $pPhase );
		if (defined $pPosition)
		{
			$out->print_element( 'core:processingPoint', $pPosition);
		}		
		$out->close_element( 'core:packetProcessingElement');
	}
	$out->close_element("core:packetProcessingOrdering");
}

sub parse_vlans {

    # "show vlan" output parsing into the ZED Vlan model
    my ( $in, $out ) = @_;
    if ( $in->{vlans} =~ /^VLAN\s+Name\s+Status\s+Ports(.+?)^\s*$/msi ) {

	# IOS output

	# create a variable with only section 2 data
	my $section2;
	while ( $in->{vlans} =~
		/^VLAN\s+Type\s+SAID\s+MTU\s+Parent\s+RingNo\s+BridgeNo\s+Stp\s+BrdgMode\s+Trans1\s+Trans2(.+?)^\s*$/msig
	    )
	{
	    $section2 .= $1;
	}

	my $section1      = $1 . "\nEND";
	my $headerPrinted = 0;
	while ( $section1 =~ /^(\d+)\s+(\S+)\s+(\S+)(.*?)(?=^\b)/msg ) {
	    $out->open_element("vlans") if ( !$headerPrinted );
	    $headerPrinted = 1;

	    my $number = $1;
	    my $name   = $2;
	    my $status = $3;
	    my $ports  = $4;

	    my $vlan = { number => $number, name => $name, };
	    if ( $status =~ /^act/i ) {
		$vlan->{enabled} = "true";
	    }
	    else {
		$vlan->{enabled} = "false";
	    }

		# Add any vlan pseudo interfaces to the correct vlan as an interfaceMember.
		if (defined($in->{interfacesData}->{"Vlan$number"}))
		{
			push(@{$vlan->{interfaceMember}}, "Vlan$number");
		}

		if ($ports) {
			foreach my $line ( split /\n/, $ports ) {
				foreach my $port ( split /,/, $line ) {
					# Make sure there is at least one non white space character in the port name
					if ($port =~ /\S/) {
						push(
							@{ $vlan->{interfaceMember} },
							_full_int_name($port)
							);
					}
				}
			}
		}

	    if ( $section2 =~
		 /^$number+\s+(\S+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+(\d+)\s+(\d+)/m
		)
	    {
		$vlan->{implementationType} = $1;
		$vlan->{said}               = $2;
		$vlan->{mtu}                = $3;
		$vlan->{parent}             = $4 if ( $4 ne "-" );
		$vlan->{ringNumber}         = $5 if ( $5 ne "-" );
		$vlan->{bridgeNumber}       = $6 if ( $6 ne "-" );
		$vlan->{bridgeMode}         = $7 if ( $7 ne "-" );
		$vlan->{translationBridge1} = $8;
		$vlan->{translationBridge2} = $9;
	    }

	    $out->print_element( "vlan", $vlan );
	}
	$out->close_element("vlans") if ($headerPrinted);
    }
    else {

	# MSFC output
	my $headerPrinted = 0;
	while ( $in->{vlans} =~ /Virtual LAN ID:\s+(\d+)\s+(.+?)(?=^\S)/msig ) {
	    $out->open_element("vlans") if ( !$headerPrinted );
	    $headerPrinted = 1;
	    my $vlan = {
		number  => $1,
		name    => "vlan" . $1,
		enabled => "true",
	    };
	    my $blob = $2;
	    while ( $blob =~ /VLAN Trunk Interface:\s+(\S+)/gi ) {
		push( @{ $vlan->{interfaceMember} }, _full_int_name($1) );
	    }
	    $out->print_element( "vlan", $vlan );
	}
	$out->close_element("vlans") if ($headerPrinted);
    }

}

sub create_config {

    # Populates the configuration entity for the main IOS configs
    my ( $in, $out ) = @_;

    # the name of the repository
    my $repository;
    $repository->{"core:name"} = "/";

    # build the simple text configuration
    my $running;
    $running->{"core:name"}     = "running-config";
    $running->{"core:textBlob"} =
	encode_base64( _apply_masks( $in->{"running_config"} ) );
    $running->{"core:mediaType"}  = "text/plain";
    $running->{"core:context"}    = "active";
    $running->{"core:promotable"} = "false";

    # push the configuration into the repository
    push( @{ $repository->{'core:config'} }, $running );

    if ( defined $in->{"startup_config"} ) {
	my $startup;
	$startup->{"core:name"}     = "startup-config";
	$startup->{"core:textBlob"} =
	    encode_base64( _apply_masks( $in->{"startup_config"} ) );
	$startup->{"core:mediaType"}  = "text/plain";
	$startup->{"core:context"}    = "boot";
	$startup->{"core:promotable"} = "true";

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $startup );
    }

    # print the repository
    $out->print_element( "core:configRepository", $repository );
}

sub parse_local_accounts {

    # local accounts - local usernames
    my ( $in, $out ) = @_;
    my $c = $in->{running_config};
    $out->open_element("localAccounts");

    while ( $c =~ /^username\s+([^\s]+)\s+(.+)/mig ) {
	my $account = { accountName => $1, };
	my $userconfig = $2;
	if ( $userconfig =~ /privilege\s+([^\s]+)/ ) {
	    $account->{accessLevel} = $1;
	}
	if ( $userconfig =~ /password\s+([^\s]+)\s+([^\s]+)/ ) {
	    $account->{password} = $2;
	}
	$out->print_element( "localAccount", $account );
    }

    # local accounts - enable and enable secret
    while ( $c =~ /^enable\s+(\S+)\s+(level\s)?(\d*)\s?([^\s]+)/mig ) {
	my $enabletype = $1;
	my $username   = "enable";
	my $encryption = $3;
	if ( $encryption eq "" ) {
	    $encryption = "0";
	}

	if ( $enabletype eq "secret" ) { $username = "enablesecret"; }
	my $account = {
	    accountName => $username,
	    password    => $4,
	    accessLevel => 15,
	};
	$out->print_element( "localAccount", $account );
    }
    $out->close_element("localAccounts");

}

sub parse_filters {
 	my ( $in, $out ) = @_;
	my $timeRanges = parse_time_range($in);
	my $lineNumber   = 0;
	my $processOrder = 1;  # starts at 1
	my $listSignatureAssembly = undef;
	my $currentList;
	my $natFilterListInstances;

	# pull up a list of subnets with which to work.  We may need it for interface mapping and don't want to
	# disrupt our scan of $in->{running_config}.  We also want to find where named filters are assigned as 
	# ingress and egress from this interface as output.
	my ($ifSubnets, $filterAssignments) = parse_interfaces($in);  # no output param  

	$out->open_element('filterLists');
	
	# Parse NAT.  This involves using the AddressTranslationParser to drive population of an AddressTranslation
	# instance allocated here.  We need to do a pass first, however, to detect where the zones map
	
	# Every combination of source and dest interface provide a separate filter list for the 
	# NAT section of the FilterLists element.  They are split into those specifying rules corresponding strictly
	# to source-address-mapped packet flows, and into destination-mapped address packet flows.  One of those is
	# going to be an implied byproduct of the NAT rule as found in the running-config.
	#
	if (not defined $in->{interfaceZones})
	{
		$in->{interfaceZones} = _parse_nat_interfaces($in);
	}	
	
	# only define NAT if we see a hint it's present
	if ( defined ($in->{interfaceZones}))
	{		
		$natFilterListInstances = parseAddressTranslation( { in => $in, interfaces => $in->{interfaceZones},
						interfaceMapper => \&_mapIfNameToPrimaryAddr, interfaceMapperContext => $ifSubnets});
				
		if ($natFilterListInstances)
		{		
			# this maps
			#   phase position for enclosing filterListsSection to a set of groups of
			#     - directional orientation of enclosed rules
			#     - the argument that should be used to call AddressTranslation::generateTranslationList with them
			 
			# +++wayne FIX this damned interface to have the second arg to generateTranslationList
			# be canonically called "source" or "destination", since it's a completely overloaded
			# mess

			my $generateRulePatterns = { "before" => [ ["source", "destination"], ["dest", "source"]] ,
				"after" => [["source", "source"], ["dest", "destination"]]};
				
			for (keys(%{$generateRulePatterns}))
			{
				my $fListGenerated = 0;
				my $phasePos = $_;
				
				foreach my $rulesTuple (@{$generateRulePatterns->{$phasePos}})
				{
					(my $rulesDirection, my $genXlDirection) = @{$rulesTuple};
					
					foreach my $natFilterList (@{$natFilterListInstances})
					{
						if ($natFilterList->getTranslationName() =~ /$rulesDirection/)
						{
							if (not $fListGenerated)
							{
								$out->open_element('filterListsSection');
								$fListGenerated = 1;
							}
							$natFilterList->resetForNewFiltersSection();
							$natFilterList->generateTranslationList($out, $genXlDirection);																			
						}
					}
					
				}
				if ($fListGenerated)
				{
					$out->open_element('phaseDiscriminator');
					
					# phase discriminator generation
					$out->print_element('processingPhase', { 'core:packetProcessingPhase' => "routing"});
					$out->print_element('processingPhasePosition', $phasePos);
					$out->close_element('phaseDiscriminator');
					$out->close_element('filterListsSection');					
				}
			}	# end of iteration over %generatedRulePatterns keys
			
		} # end of check for $natFilterInstances
		# else, no 'ip nat' statements defined in running-config

		delete $in->{interfaceZones}				
	}

	$out->open_element('filterListsSection');
	
    foreach my $line ( split( /\n/, $in->{running_config} ) ) {
	$lineNumber++;
	if ( $line =~ /^ip access-list\s+(extended|dynamic|standard)\s+(.+\b)/ )
	{
	    if ($currentList) {
		$processOrder = 1;
		_close_list( $currentList, $out );
		undef $currentList;
	    }
	    $currentList = {
		type     => $1,
		protocol => 'ip',
		name     => $2,
	    };
	    $out->open_element('filterList configPath="running-config"');
	    # First element is always the default action.  For IOS, this is "deny".  Which is to say,
	    # if a filter is applied to an interface and it does not match a given packet (at the very)
	    # least, by ending with a "permit all" if that's the desired effect), then it is denied.
	    $out->print_element("defaultAction", "deny");
	    
	    # Filter list's are set as being tracked with signature assembly
	    $listSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");
	    # +++waynet we aren't doing signature qualification on filter lists by device name		
	    # $listSignatureAssembly->QualifySignature($systemName);
	} elsif ( $line =~ /^ipv6 access-list\s+(\b.+\b)/ ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping IPv6 access list: $1");
	} elsif ( $line =~ /^access-list\s+(\d+)(\s+.+)/ ) {
	    my $aclNumber = $1;
	    my $aceBody   = $2;
	    if ( $currentList && $currentList->{name} ne $aclNumber ) {
		_close_list( $currentList, $out );
		undef $currentList;
	    }

	    if ( !$currentList ) {
		$processOrder = 1;
		$currentList  = {
		    protocol => 'ip',
		    name     => $aclNumber,
		    type     => 'standard',
		};
		$out->open_element('filterList configPath="running-config"');
		# First element is always the default action.  For IOS, this is "deny".  Which is to say,
		# if a filter is applied to an interface and it does not match a given packet (at the very)
		# least, by ending with a "permit all" if that's the desired effect), then it is denied.
		$out->print_element("defaultAction", "deny");
		$listSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");						
		# +++waynet we aren't doing signature qualification on filter lists by device name		
		# $listSignatureAssembly->QualifySignature($systemName);	
	    }
	    if ( $line !~ /\s+remark/i ) {
		if ( ($aclNumber < 100) || (($aclNumber >= 1300) && ($aclNumber <= 1399)) ) {
		    _standard_term( $aceBody, $processOrder, $lineNumber,
				    $timeRanges, $out );
		}
		else { # technically, 100-199, 2600-2699
		    _extended_term( $aceBody, $processOrder, $lineNumber,
				    $timeRanges, $out );
		}
		$processOrder++;
	    }
	}
	elsif ( $line =~ /^\S/ ) {
	    if ($currentList) {
		$processOrder = 1;
		_close_list( $currentList, $out );
	    }
	    undef $currentList;
	}
	elsif ( $currentList && $line !~ /^\s*remark/ ) {
	    if ( $currentList->{type} =~ /standard/i ) {
		_standard_term( $line, $processOrder, $lineNumber, $timeRanges,
				$out );
	    }
	    elsif ( $currentList->{type} =~ /extended/i ) {
		_extended_term( $line, $processOrder, $lineNumber, $timeRanges,
				$out );
	    }
	    elsif ( $currentList->{type} =~ /ipv6/i ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping IPv6 rule: $line");
	    }
	    $processOrder++;
	}
    }
	$out->close_element('filterListsSection');
	
	# Now, we need to generate the filterLists section that is complete with the references for input
	# filter lists and output filter lists
	my $filterRefs; # indexed by ingress/egress hash ref
	my $filterRefInst;
	my $phaseDiscrim;
	
	foreach my $fAssignIf ( keys %{$filterAssignments})
	{
		foreach my $assignedFilter (@{$filterAssignments->{$fAssignIf}})
		{
			$filterRefInst = { name => $assignedFilter->{filterName}, direction => $assignedFilter->{'direction'}};
			
			if ($filterRefInst->{'direction'} eq 'ingress')
			{
				push ( @{$filterRefInst->{appliesTo}}, { source => $fAssignIf});
			}
			else
			{
				push ( @{$filterRefInst->{appliesTo}}, { destination => $fAssignIf});				
			}
			
			push ( @{$filterRefs->{$assignedFilter->{'direction'}}}, $filterRefInst);
		}	
	} 	
	
	# We should just be able to write the input/output filtering assignment sections in bulk now
	$out->open_element('filterListsSection');
	
	foreach $filterRefInst (@{$filterRefs->{'ingress'}})
	{
		$out->print_element('filterListReference', $filterRefInst);
	}
	
	$phaseDiscrim->{'processingPhase'} = { 'core:packetProcessingPhase' => 'packet-filtering', 
						'core:processingPoint' => 'source'};
	$out->print_element('phaseDiscriminator', $phaseDiscrim);
	
	$out->close_element('filterListsSection');

	$out->open_element('filterListsSection');
	
	foreach $filterRefInst (@{$filterRefs->{'egress'}})
	{
		$out->print_element('filterListReference', $filterRefInst);
	}
	
	$phaseDiscrim->{'processingPhase'} = { 'core:packetProcessingPhase' => 'packet-filtering', 
						'core:processingPoint' => 'destination'};
	$out->print_element('phaseDiscriminator', $phaseDiscrim);
	$out->close_element('filterListsSection');
	
    $out->close_element('filterLists');
}

sub _close_list {
    my ( $list, $out ) = @_;
    $out->print_element( 'mode', 'stateless' );
    # we don't generate the list name into the signature
    $out->print_element( 'name', $list->{name}, {suppressFromSignatureAssembly => 1});
    $out->close_element('filterList');
}

# process a standard term:
sub _standard_term {
    my ( $acl_line, $term_process_order, $lineNumber, $time_range, $out ) = @_;
    my $thisterm = { processOrder => $term_process_order, };
    $thisterm->{'-attributes'} =
    { lineStart => $lineNumber, lineEnd => $lineNumber, };
    if ( $acl_line =~ /^\s*(no permit|permit|deny)\s+(any)/i ) {
	my $primaryAction = $1;
	if ( $primaryAction eq "no permit" ) {
	    $primaryAction = "deny";
	}
	$thisterm->{"primaryAction"} = $primaryAction;
	push @{ $thisterm->{'sourceIpAddr'}->{'network'} }, get_any_address_encoding();
    }
    elsif ( $acl_line =~
	    /^\s*(no permit|permit|deny)\s+(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+)/i
	)
    {
	my $primaryAction = $1;
	if ( $primaryAction eq "no permit" ) {
	    $primaryAction = "deny";
	}
	$thisterm->{"primaryAction"} = $primaryAction;

	push @{ $thisterm->{"sourceIpAddr"}->{network} },
	create_network_element( $2, $3, 1 );
    }
    elsif ( $acl_line =~
	    /^\s*(no permit|permit|deny)\s+host\s+(\d+\.\d+\.\d+\.\d+)/i )
    {
	my $primaryAction = $1;
	if ( $primaryAction eq "no permit" ) {
	    $primaryAction = "deny";
	}
	$thisterm->{"primaryAction"} = $primaryAction;
	my $thisIPAddress = { "host" => $2, };
	push @{ $thisterm->{"sourceIpAddr"} }, $thisIPAddress;
    }
    elsif ( $acl_line =~
	    /^\s*(no permit|permit|deny)\s+(\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+)/i
	)
    {
	my $primaryAction = $1;
	if ( $primaryAction eq "no permit" ) {
	    $primaryAction = "deny";
	}
	$thisterm->{"primaryAction"} = $primaryAction;
	push @{ $thisterm->{"sourceIpAddr"}->{network} }, create_network_element( $2, $3, 1 );
    }
    elsif ( $acl_line =~ /^\s*(no permit|permit|deny)\s+(\d+\.\d+\.\d+\.\d+)/i )
    {
	my $primaryAction = $1;
	if ( $primaryAction eq "no permit" ) {
	    $primaryAction = "deny";
	}
	$thisterm->{"primaryAction"} = $primaryAction;
	my $thisIPAddress = { "host" => $2, };
	push @{ $thisterm->{"sourceIpAddr"} }, $thisIPAddress;
    }

    if ( $acl_line =~ /\s+log\s*$/ ) {
	$thisterm->{log} = "true";
    }
    else {
	$thisterm->{log} = "false";
    }

    # look for the time-range instruction
    if ( $acl_line =~ /\btime-range\s+(\S+)/i ) {
	$thisterm->{timeAllowed} = $time_range->{$1};
    }
    $out->print_element( 'filterEntry', $thisterm );
}

# Process an extended term
sub _extended_term {

	my ( $acl_line, $term_process_order, $lineNumber, $time_range, $out) = @_;

	my $thisterm = {
		processOrder => $term_process_order,
		'-attributes' => { lineStart => $lineNumber, lineEnd => $lineNumber, },
	};

	my $ipaddresses = [];

	# permit/deny and dynamic acl settings
	my $primaryAction;
	if ( $acl_line =~ /^\s*(no permit|deny|permit)/i ) {
		$primaryAction = $1;
	} elsif ( $acl_line =~ /^\s*dynamic\s+(\S+)\s+timeout\s+(\S+)\s+(no permit|deny|permit)/i ) {
		$primaryAction = $3;
	} elsif ( $acl_line =~ /^\s*dynamic\s+(\S+)\s+(no permit|deny|permit)/i ) {
		$primaryAction = $2;
	} else {
		return;
	}

	if ( $primaryAction eq 'no permit' ) {
		$primaryAction = 'deny';
	}
	$thisterm->{primaryAction} = $primaryAction;

	# Set term protocol. Keep in mind it can have been set as an object-group reference.
	my $termProtocol = "";
	my $ogSvcObjRef = undef;
	my $protoArgTxt = "";

	if ( $acl_line =~ /^\s*(no permit|deny|permit)\s+(object-group\s+(\S+))\s+/i ) {
		# Object group reference
		# This is the entire protocol term argument, including "object-group"
		$protoArgTxt = $2;
		$ogSvcObjRef = $3;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}\s+/i ) {
		$LOGGER->debug("WARNING: Layer 2 address matching not supported. Dropping rule at line $lineNumber in 'show startup-config'");
		return;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+0x\d{4}\s+/i ) {
		$LOGGER->debug("WARNING: Ethernet type matching not supported. Dropping rule at line $lineNumber in 'show startup-config'");
		return;
	} elsif ( $acl_line =~ /^\s*(no permit|deny|permit)\s+(\S+)\s+/i ) {
		$termProtocol = $2;
		$protoArgTxt = $termProtocol;
	}

	# Choose source address format
	my $srctext = "";
	my $termSrcAddress = ();

	if ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+any)(?:\s+any)?\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} }, get_any_address_encoding();
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+object-group\s+(\S+))/i )	{
		# This is a network object group reference
		$srctext = $1;
		push @{$termSrcAddress->{objectGroupReference}}, $3;
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(addrgroup\s+.+))/i ) {
		# This is a network object group reference but in "addrgroup" form..there could be multiple object groups after it.
		$srctext = $1;
		my $agBlob = $3;

		# Multiple groups could follow
		while ($agBlob =~ /addrgroup\s+(\S+)/gi) {
			push @{$termSrcAddress->{objectGroupReference}}, $1;
		}
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} },
		create_network_element( $3, $4, 1 );
	} elsif ($acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+([a-f\d:]+)\/(\d+))\s+/i ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping network element creation: $acl_line");
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{network} },
		create_network_element( $3, $4, 1 );
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+host\s+(\d+\.\d+\.\d+\.\d+|[a-f\d:]+))\s+/i ) {
		$srctext = $1;	
		push @{ $termSrcAddress->{host} }, $3;
	} elsif ( $acl_line =~ /^(\s*(no permit|deny|permit)\s+$protoArgTxt\s+(\d+\.\d+\.\d+\.\d+))\s+/i ) {
		$srctext = $1;
		push @{ $termSrcAddress->{host} }, $3;
	}

	# Now look for the src port
	my $srcport        = "";
	my $termSrcService = ();
	if ( $acl_line =~ /$srctext\s+((lt|gt|eq|neq)\s+(\S+))/i ) {
		$srcport = $1;

		my $operator = $2;
		if ( $operator =~ /neq/ ) {
			$operator = "ne";
		}

		# Note: this is the locus of port-map support, pulled in momentary short-term indecision
		# as to how to handle its support and testing (see PR 29487)
		my $snPort = $3;

		# If we didn't get a service mapping out of that, try it as a numeric or trivially "built in" port
		if ( _int_port($snPort) =~ /^\d+$/ ) {
			my $thisport = {
				port     => _int_port($snPort),
				operator => $operator,
			};
			push @{ $termSrcService->{portExpression} }, $thisport;
		} else {
			# We have to bypass mapping support for this port qualification
			$LOGGER->debug('Unable to parse or map presumed filter entry source port/service: ' . _int_port($snPort) );
		}

	} elsif ( $acl_line =~ /$srctext\s+(range\s+(\S+)[-\s](\S+))/i ) {
		$srcport = $1;
		my $thisport = {
			portStart => _int_port($2),
			portEnd   => _int_port($3),
		};
		push @{ $termSrcService->{portRange} }, $thisport;
	}

	# Destination address
	my $dsttext = "";
	my $termDestAddr = ();

	if( $acl_line =~ /$srctext\s*$srcport\s*(any(?:\s+any)?)/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, get_any_address_encoding();
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(object-group\s+(\S+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{objectGroupReference}}, $2;
	} elsif( $acl_line =~ /$srctext\s*$srcport\s+(addrgroup\s+.+)/i ) {
		# This is a network object group reference but in "addrgroup" form..there could be multiple object groups after it.
		$dsttext = $1;

		# Multiple groups could follow
		while ($dsttext =~ /addrgroup\s+(\S+)/gi) {
			push @{$termDestAddr->{objectGroupReference}}, $1;
		}
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*((\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(([a-f\d:]+)\/(\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*((\d+\.\d+\.\d+\.\d+),\s+wildcard\s+bits\s+(\d+\.\d+\.\d+\.\d+))/i ) {
		$dsttext = $1;
		push @{ $termDestAddr->{network} }, create_network_element( $2, $3, 1 );
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(host\s+(\d+\.\d+\.\d+\.\d+|[a-f\d:]+))/i ) {
		$dsttext = $1;
		push @{$termDestAddr->{host}}, $2;
	} elsif( $acl_line =~ /$srctext\s*$srcport\s*(\d+\.\d+\.\d+\.\d+|[a-f\d:]+)/i ) {
		$dsttext = $1;
		push @{$termDestAddr->{host}}, $1;
	}

	# now look for the dst port
	my $dstport            = "";
	my $termDestService    = ();
	my $termProtocolFilter = "";

	if ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+((lt|gt|eq|neq)\s+(\S+))/i ) {
		$dstport = $1;
		my $operator = $2;
		if ( $operator =~ /neq/ ) {
			$operator = "ne";
		}

		$ogSvcObjRef = undef;
		my $dnPort = $3;

		if ( _int_port($dnPort) =~ /^\d+$/) {
			my $thisport = {
				port     => _int_port($dnPort),
				operator => $operator,
			};
			push @{ $termDestService->{portExpression} }, $thisport;
		}

	} elsif ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+(range\s+(\S+)[-\s](\S+))/i ) {
		$dstport = $1;
		my $thisport = {
			portStart => _int_port($2),
			portEnd   => _int_port($3),
		};
		push @{ $termDestService->{portRange} }, $thisport;
	} elsif ( ( $termProtocol eq "icmp" ) && ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s+((?:icmp-type\s+)?([\w-]+)\s*)/i ) ) {

		# This is either a numeric or symbolic (e.g., "echo", "timestamp-reply")
		# ICMP message type reference.  Now, this is a bit of a conundrum...we could
		# shove this in as a "port type", and the table for protocol -> numeric mapping
		# in Utils.pm that is consulted by get_port_number() would have you believe that
		# this is what you should do (e.g., there is "echo" entry there that maps to
		# 7).  On the other hand, the ONM schema has an option for "protocol-filter",
		# whose descriptive text says its use case is ICMP message types.  It is also
		# generously typed as a simple string.  We will use the latter.
		$termProtocolFilter = $2;
	}

	# Assign term fields in order of definition in the Ziptie Adapter schema
	if ($termProtocol) {
		$thisterm->{protocol} = $termProtocol;
	}

	# Initialize a filter term
	if ($termProtocolFilter) {
		$thisterm->{protocolFilter} = $termProtocolFilter;
	}

	# These should be mutually exclusive
	if ( $ogSvcObjRef and ( not defined $thisterm->{protocol} ) ) {
		$thisterm->{protocolObjectGroupReference} = $ogSvcObjRef;
	}

	for my $addrTerm ('host', 'network', 'objectGroupReference') {
		for (@{$termSrcAddress->{$addrTerm}}) {
			# Push the type of addr term instance being referenced as a separate sourceIpAddr instance.
			push ( @{$thisterm->{sourceIpAddr}}, { $addrTerm => $_} );
		}
		
		# Now do the same thing for dest
		for (@{$termDestAddr->{$addrTerm}}) {
			# Push the type of addr term instance being referenced as a separate destinationIpAddr instance.
			push ( @{$thisterm->{destinationIpAddr}}, { $addrTerm => $_} );
		}
	}

	if ($termSrcService) {
		$thisterm->{sourceService} = $termSrcService;
	}

	if ($termDestService) {
		$thisterm->{destinationService} = $termDestService;
	}

	# Optional modifications and side-effects to filter:
	#	-- only filter TCP RST's ("established")
	#   -- cause matching packet to be logged ("l")
	#
	#  These terms come at the end of the source/dest specs we've
	#  parsed to this point.  They can come in any order, and be
	#  interspersed with a variety of other modifiers (tos, precedence, etc.)
	#  which can show up similarly trailing on the access-control spec.
	#
	#  +++wayne bug: 'tos' and 'precedence' should be processed.  See bug 29515.
	#

	# Initialize...may be overridden
	$thisterm->{log} = 'false';

	if ( $acl_line =~ /$srctext\s*$srcport\s*$dsttext\s*$dstport([ -_\w]+)/ ) {

		# Force a word start
		my $acEntry_modifiers = ' ' . $1;
		if ( ( $acEntry_modifiers =~ /\bestablished/ ) && ( $thisterm->{protocol} eq "tcp" ) ) {

			# Set optional ZED "established" filter element
			$thisterm->{established} = 'true';
		}

		if ( $acEntry_modifiers =~ /\bl/ ) {
			$thisterm->{log} = 'true';
		} else {
			$thisterm->{log} = 'false';
		}

		# look for the time-range instruction
		# +++wayne bug!: the "no" specifier is not taken into account here.
		#    see bug 29514.
		if ( $acl_line =~ /\btime-range\s+(\S+)/i ) {
			$thisterm->{timeAllowed} = $time_range->{$1};
		}
	}

	$out->print_element( 'filterEntry', $thisterm );
}

sub parse_time_range {
    my ($in) = @_;

    my $months = {
	JAN => '01',
	FEB => '02',
	MAR => '03',
	APR => '04',
	MAY => '05',
	JUN => '06',
	JUL => '07',
	AUG => '08',
	SEP => '09',
	OCT => '10',
	NOV => '12',
	DEC => '12',
    };
    my $time_range;
    while ( $in->{running_config} =~ /^time-range\s+(\S+)(.+?)(?:!|: end)/migs )
    {
	my $name = $1;
	$_ = $2;
	if (
	    /absolute\s+start\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)\s+end\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)/mi
	    )
	{
	    $time_range->{$name}->{startTime} = $1 . ':00';
	    $time_range->{$name}->{startDate} =
		$4 . '-' . $months->{ uc( substr( $3, 0, 4 ) ) } . '-' . $2;
	    $time_range->{$name}->{endTime} = $5 . ':00';
	    $time_range->{$name}->{endDate} =
		$8 . '-' . $months->{ uc( substr( $7, 0, 4 ) ) } . '-' . $6;
	    $time_range->{$name}->{startTime} =
		'0' . $time_range->{$name}->{startTime}
	    if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
	    $time_range->{$name}->{endTime} =
		'0' . $time_range->{$name}->{endTime}
	    if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
	}
	elsif (/periodic\s+(\D+)(\d+:\d+)\s+to\s+(\d+:\d+)/mi) {
	    my $days = $1;
	    $time_range->{$name}->{startTime} = $2 . ':00';
	    $time_range->{$name}->{endTime}   = $3 . ':00';
	    $time_range->{$name}->{startTime} =
		'0' . $time_range->{$name}->{startTime}
	    if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
	    $time_range->{$name}->{endTime} =
		'0' . $time_range->{$name}->{endTime}
	    if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
	    $days =~
		s{ (\w+) }{ ( lc ( $1 ) ne 'thursday' ? substr ( $1, 0, 3 ) : substr ( $1, 0, 4 ) ) }gex;
	    push @{ $time_range->{$name}->{days} }, split( /\s+/, $days );
	}
    }

    return $time_range;
}

sub print_filter_address {
    my ( $in, $out, $v ) = @_;

    $v ||= 6;

    my $any_network =
	$v == 6
	? { "address" => "::",      "mask" => "0" }
    : { "address" => "0.0.0.0", "mask" => "0" };
    if ( $in =~ /^any$/i ) {
	push @{ $out->{network} }, $any_network;
    }
    elsif ( $in =~ /^[\da-f\.\:\/]+$/i ) {
	my ( $address, $mask ) = split( /\//, $in );
	if ( $v == 6 ) {
	    push @{ $out->{network} }, { address => $address, mask => $mask };
	}
	else {
	    push @{ $out->{network} },
	    create_network_element( $address, $mask );
	}
    }
    elsif ( $in =~ /^([\da-f\.\:]+)\s+([\da-f\.\:]+)$/i ) {
	if ( $v == 6 ) {
	    push @{ $out->{network} }, { address => $1, mask => $2 };
	}
	else {
	    push @{ $out->{network} }, create_network_element( $1, $2 );
	}
    }
    elsif ( $in =~ /^host\s+(\S+)$/i ) {
	push @{ $out->{host} }, $1;
    }
    elsif ( $in =~ /^object-group\s+(\S+)$/i ) {
	push @{ $out->{objectGroupReference} }, $1;
    }
}

sub print_filter_service {
    my ( $in, $out ) = @_;

    if ( $in =~ /^(lt|gt|eq|neq)\s+(\S+)$/i ) {
	push @{ $out->{portExpression} },
	{
	    "port"     => _int_port($2),
	    "operator" => ( $1 ne 'neq' ? $1 : 'ne' )
	};
    }
    elsif ( $in =~ /^range\s+([^\s\-]+)[-\s]([^\s\-]+)$/i ) {
	push @{ $out->{portRange} },
	{ "portStart" => _int_port($1), "portEnd" => _int_port($2) };
    }
    elsif ( $in =~ /^object-group\s+(\S+)$/i ) {
	push @{ $out->{objectGroupReference} }, $1;
    }
}

# checks to see if the incoming port is an integer.  If not
# it will do a lookup of the service name to convert it to an integer
sub _int_port {
    my ($port) = @_;
    if ( $port =~ /^\d+$/ ) {
	return $port;
    }
    else {
	return get_port_number($port);
    }
}


=head

    =item See examples to be found at (as of this writing) http://www.cisco.com/en/US/docs/ios/sec_data_plane/configuration/guide/sec_object_group_acl.html.  See examples under "show object-group".
=cut

sub parse_object_groups {
    
    my ( $in, $out ) = @_;
    
    my $Blob;
    my $ogType;
    my $cipm = get_crep('cipm');
    my $currentGroup = undef;
    
    # parse actual object groups as blobs
    # Header line, e.g.,
    # "Network object group auth_proxy_acl_permit_dest
    
    foreach my $line ( split( /\n/, $in->{object_groups} ) ) {
    	# There is a type of reported object group in the 12.[23]SX release line (Catalyst 6509 MFSC) only, which is
    	# "IP address object group".  We will work diligently to normalize that to our notion of a 'Network' object group, 
    	# since the syntax appears identical otherwise.
	if ( $line =~ /^(\w+|IP address)\s+object group\s+(\S+)/ ) {
		
	    $currentGroup = {
		type => ( $1 eq 'IP address') ? 'Network' : $1,
		id => $2
	    };
	    
	    if ( $currentGroup->{type} eq 'Network' ) {
		    $in->{objectGroup_objects}->{networkGroup}->{ $currentGroup->{id} } = $currentGroup;
		}
	}
	elsif ( $line =~ /^\s+(\S+.+)$/ ) {
	    my $ogLine = $1;
	    
	    # as it's "non-blank" and doesn't open another access list, we'd better have
	    # an access list we're working on
	    if ( not defined $currentGroup ) {
		$LOGGER->debug(
		    "unexpected object-group definition outside of confines of parsed object group: "
		    . $line );
	    }
	    elsif ( $ogLine =~ /description\s+(.+)/ ) {
		
		# description can be for any type
		$currentGroup->{description} = $1;
	    }
	    else {
		
		# dispatch on object-type-specific configuration directive.
		if ( $currentGroup->{type} eq 'Network' ) {
		    
		    # Host.  This should be IP addressed,
		    if ( $ogLine =~ /host ($cipm)/ ) {
			push( @{ $currentGroup->{host} }, $1 );
		    }
		    # Currently, we're covering the case where this is "host (DNS name)" by just passing 
		    # the untranslated DNS name up to topology (where it will be flagged as an error, as least
		    # allowing visibility to the user).  Other alternatives (e.g., doing a DNS resolve here,
		    # at backup/parse time) don't seem significantly more appealing.	       
		    elsif ($ogLine =~ /host\s+(\S+)/) {
			my $dnsHostName = $1;
			$LOGGER->debug(
			    "found DNS object-group host name: \'$dnsHostName\' in object-group during backup.  Not currently suportted for translation, putting in SED anyways. ");
			push( @{ $currentGroup->{host} }, $dnsHostName );
		    }	       
		    # Network address
		    elsif ( $ogLine =~ /^\s*($cipm)\s+($cipm)/ ) {
			
			# (ipv4 addr) (ipv4 mask) form
			push(
			    @{ $currentGroup->{subnet} },
			    { address => $1, ipV4Mask => $2 }
			    );
		    }
		    elsif ( $ogLine =~ /^\s*($cipm)\/([\d]{1,2})/ ) {
			
			# (ipv4 addr/CIDR mask) variety
			push(
			    @{ $currentGroup->{subnet} },
			    { address => $1, mask => $2 }
			    );
		    }
		    
		    # range
		    elsif ( $ogLine =~ /range\s+($cipm)\s+($cipm)/ ) {
			push(
			    @{ $currentGroup->{range} },
			    { startAddress => $1, endAddress => $2 }
			    );
		    }
		    
		    # any
		    elsif ( $ogLine =~ /^\s+any/ ) {
			push( @{ $currentGroup->{subnet} }, get_any_address_encoding() );
		    }
		    
		    # other object ref.  This must be verified at the time this is serialized to the ZED element.
		    elsif ( $ogLine =~ /^\sgroup-object\s+(\S+)/ ) {
			push( @{ $currentGroup->{objectGroupReference} }, $1 );
		    }
		    else {
			$LOGGER->debug(
			    "unexpected object-group \'Network\' type detail: "
			    . $ogLine );
		    }
		    
		}
		elsif ( $currentGroup->{type} eq 'Service' ) {
		    my $svcDetail = undef;
		    
		    # well-knonwn primary IP protocols Keep in mind that the "service" protocol can end up covering an ICMP message type.
		    # oh, and while we're kvetching about it, a single spec can take on UDP and TCP.  This is why
		    # "protocol" is a secondary index-listed structure in the IOSNetworkGroups representation, and
		    # gets more suitably decomposed to ZED form at serialize time.
		    
		    # +++wayne we cannot deal with separate source/dest specs here in the ZED as IOS can.  We need to catch that here, until
		    # we can address this in the ZED and toopology.
		    if ( $ogLine =~ /\s+source\s.+/ ) {
			$LOGGER->debug(
			    "Ziptie backup cannot currently handle separate source/dest object-group network specs, discarding \'"
			    . $ogLine
			    . "\'" );
			next;
		    }
		    
		    if ( ( $ogLine =~ /\s*([udtcp-]{3,})\s+(.*)/ ) ) {
			my $prot       = $1;
			my $protDetail = $2;
			my $operator   = undef;
			my $port;
			my $rangeDef = 0;
			
			# Now, to dig into what that detail is...'eq' is a default, or it can be spec'd, along with eq, lt, or gt.
			if ( $protDetail =~ /^\s*(\d+)/ ) {
			    $operator = "eq";
			    $port     = $1;
			}
			elsif ( $protDetail =~ /^\s*([elg][qt])\s+(\S+)/ ) {
			    $operator = $1;
			    $port     = $2;
			}
			
			# did either of those parses work?
			if ( defined $operator ) {
			    
			    # Deal with the fact we may be looking atr both 
			    # tcp and udp if "tcp-udp" is the protocol.
			    # +++wayne I don't know if this actually shows up in 
			    # 'show object-group', or whether it gets
			    # split into two detail lines.  This should work 
			    # either way.
			    
			    foreach my $thisProt ( split( '-', $prot ) ) {
				$svcDetail =
				    $in->{objectGroup_objects}->{serviceGroup}->{$thisProt}
				->{ $currentGroup->{id} };
				
				if ( not defined $svcDetail ) {
				    $svcDetail =
				    { id => $currentGroup->{id} };
				    $in->{objectGroup_objects}->{serviceGroup}->{$thisProt}->{ $currentGroup->{id} } = $svcDetail;
				}
				push(@{$svcDetail->{ports}{ "ports_" . $operator }},
				     _int_port($port));
			    }		     
			}
			elsif ( $protDetail =~ /^\s*range\s+(\d+)\s+(\d+)/ ) {
			    $port = $1;
			    my $portEnd = $2;
			    
			    foreach my $thisProt ( split( '-', $prot ) ) {
				$svcDetail =
				    $in->{objectGroup_objects}->{serviceGroup}->{$thisProt}
				->{ $currentGroup->{id} };
				
				if ( not defined $svcDetail ) {
				    $svcDetail =
				    { id => $currentGroup->{id} };
				    push(
					@{
					    $in->{objectGroup_objects}->{serviceGroup}
					    ->{$thisProt}
					    ->{ $currentGroup->{id} }
					},
					$svcDetail
					);
				}
				push @{ $svcDetail->{ranges} }, [_int_port($port), _int_port($portEnd)];
			    }
			}
		    }
		    elsif ( ( $ogLine =~ /\s*icmp\s+(\S+)/ ) ) {
			
			# +++wayne I've not actually seen one of these in my samples, and I don't have access to a 12.4 IOS system
			# currently, so I'm using some pretty good inference as to how this will show up in a 'show object-groups'
			# output
			my $icmpMessageType = $1;
			$svcDetail =  $in->{objectGroup_objects}->{serviceGroup}->{icmp}->{ $currentGroup->{id} };
			
			if ( not defined $svcDetail ) {
			    $svcDetail = { id => $currentGroup->{id} };
			    $in->{objectGroup_objects}->{serviceGroup}->{icmp}->{ $currentGroup->{id} } = $svcDetail;
			}
			push(@{ $svcDetail->{messageTypes} }, $icmpMessageType);
		    }
		    elsif ( $ogLine =~ /\s*group-object\s+(\S+)/ ) {
		  	my $ogRef = $1;
			
		  	$svcDetail = $in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{ $currentGroup->{id} };
			
			if ( not defined $svcDetail ) {
			    $svcDetail = { id => $currentGroup->{id} };
			    $in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{ $currentGroup->{id} } = $svcDetail;
			}
			push( @{ $svcDetail->{groupObjects} }, $ogRef );
		    }
		    
		    # other IP protocol.  This is, apparently, just a single word in an entry for a pure IP
		    # protocol (e.g., ESP).
		    elsif ( $ogLine =~ /\s+(\S+)/ ) {
			my $ipProt = $1;
			
			$svcDetail =
			    $in->{objectGroup_objects}->{serviceGroup}->{ip}->{ $currentGroup->{id} };
			
			if ( not defined $svcDetail ) {
			    $svcDetail = { id => $currentGroup->{id} };
			    $in->{objectGroup_objects}->{serviceGroup}->{ip}->{ $currentGroup->{id} } = $svcDetail;
			}
			push( @{ $svcDetail->{ipProtocols} }, $ipProt );
		    }
		    else {
			$LOGGER->debug( "Unable to parse object-group detail entry \'"
					. $ogLine
					. "\'" );
		    }
		}
	    }
	}
    }
    
    # Even if there was no content from object_groups, call to generate object groups which have
    # been mapped into intermediate representation (in %IOSObjectGroups) from other native IOS
    # configuration elements.
    
    _generate_ziptie_object_groups($in, $out);
}

sub _generate_ziptie_object_groups {
    
    my ($in, $out) = @_;
    my $groupEnt;

# organization is
#  (IOSObjectGroups)
#	{network-group}	->
# 		{id} ->
# 			id = name
# 			description = description
# 			host ->
# 				w.x.y.z
# 			subnet ->
# 			 	{ address => w.x.y.z ipVrmask => a.b.c.d }
# 			 	{ address => w.x.y.z mask=nn }
# 			        ....
#	{service-group)->
#		{protocol}->     (this allows defs of a protocol separately by UDP and TCP ports)
#			{(id)}->
#				id = name
# 				description = (descrptive text)
#				ports ->
# 					eq = (port1, port2, port3)  (sorted)
# 					lt = port
# 					gt = port
# 				ranges ->
# 					(range1a, range1b)
# 					(range2a, range2b)....
#
#
#   Notes:
#    We have to do some odd checking and cross-referencing in service groups, but the idea is that network-group
#    keys have been faithfully preserved as ZED element keys, hence can have some loop-automation assigned to
#    their ZED serialization.

    my $opened = 0;

    # We have to do one scan of the service groups looking just for icmp entries, since they
    # get their own ZED object group element

    if (defined $in->{objectGroup_objects}->{serviceGroup}->{icmp})
    {
	
	foreach $groupEnt (values(%{$in->{objectGroup_objects}->{serviceGroup}->{icmp}})) 
	{
	    if ( not $opened ) {
		$out->open_element("objectGroups");
		$opened = 1;
	    }
	    
	    $out->open_element("icmpGroup");
	    if (defined $groupEnt->{description}) {
		$out->print_element("description", $groupEnt->{description});
	    }
	    
	    $out->print_element("id", $groupEnt->{id});
	    
	    foreach (@{$groupEnt->{messageTypes}}) 
	    {
		$out->print_element("icmpObject", $_);
	    }
	    $out->close_element("icmpGroup");
	}
    }
    
    if (defined $in->{objectGroup_objects}->{serviceGroup}->{ip})
    { 
	foreach $groupEnt (values(%{$in->{objectGroup_objects}->{serviceGroup}->{ip}}))
	{
	    # Same thing for protocol groups, from ip service elements (there are raw IP protocols
	    if ( not $opened ) {
		$out->open_element("objectGroups");
		$opened = 1;
	    }
	    
	    $out->open_element("protocolGroup");
	    if (defined $groupEnt->{description}) {
		$out->print_element("description", $groupEnt->{description});
	    }
	    
	    $out->print_element("id", $groupEnt->{id});
	    
	    foreach (@{$groupEnt->{ipProtocols}} )
	    {
		$out->print_element("protocol", $_);
	    }
	    $out->close_element("protocolGroup");
	}
    }

    # And now, network groups, which have their own analogous ZED tree structure
    foreach $groupEnt (values(%{$in->{objectGroup_objects}->{networkGroup}}))
    {
	if ( not $opened ) {
	    $out->open_element("objectGroups");
	    $opened = 1;
	}
	
	$out->open_element("networkGroup");
	
	if (defined $groupEnt->{description}) 
	{
	    $out->print_element("description", $groupEnt->{description});
	}
	$out->print_element("id", $groupEnt->{id});
	
	# If there are any object group references, generate them, but check here for whether
	# they reference actual network object groups, since if they don't, undefined refs will
	# KO the topology engine if it encounters them in the ZED.
	
	foreach my $indirectNwObjGrp (@{ $groupEnt->{objectGroupReference}}) {
	    if (defined $in->{objectGroup_objects}->{networkGroup}->{$indirectNwObjGrp}) {
		$out->print_element("objectGroupReference", $indirectNwObjGrp);
	    }
	    else
	    {
		$LOGGER->debug(  "Unable to find network object group \'" .
				 $indirectNwObjGrp . "\', referenced in \'" .
				 $groupEnt->{id} . "\', suppressing reference");
	    }
	}
	
	# Now, all other elements should have been set up with proper sturcture to support
	# direct print element to the ZED
	
	foreach my $nwSubelt ("host", "range", "subnet") 
	{
	    if (defined $groupEnt->{$nwSubelt})
	    {
		$out->print_element($nwSubelt, $groupEnt->{$nwSubelt});
	    }
	}
	$out->close_element("networkGroup");
    }
    
    
    if ( defined $in->{objectGroup_objects}->{serviceGroup} ) {
	my $protocol;
	my $protocolEnt;
	my $portsEnt;                
	my $portRangesEnt;
	
	if ( not $opened ) {
	    $out->open_element("objectGroups");
	    $opened = 1;
	}
	
	
	# The IOSObjectGroups are set up as 
	#  (hash root) -->
	#                 (udp) -->
	#                          (name) --> [fields, ports]
	#                          (name) --> [fields, ports]
	#                             ....
	#                 (tcp) -->
	#                          (name) --> [fields, ports]
	#                          (name) --> [fields, ports]
	#                             ....
	#
	#  HOWEVER, the ZED must have its serviceGroup names be unique, and in fact, wants multiple
	#  protocol-indexed elements within each named service group (in other words, the opposite of
	#  how the service groups have been set up.  What that means we have to do here is to do a "fixup"
	#  on the fly to combine elements where there are ports for BOTH UDP and TCP for the same named
	#  network group.
	my @nwProtocolKeys = ("udp", "tcp");
	do {
	    my $protocol = shift(@nwProtocolKeys);
	    $protocolEnt = $in->{objectGroup_objects}->{serviceGroup}->{$protocol};
	    
	    if (defined $protocolEnt)
	    {
		foreach my $nGroupEnt ($protocolEnt)  # this brings us to the name-indexed node...
		{      
		    foreach my $pGroupKey (keys %{$nGroupEnt}) {
		    	
		    my $pGroupEnt = $nGroupEnt->{$pGroupKey};                 	
			my @pGroupEvalSecs = ( {protocol => $protocol, nwObjectEnt => $pGroupEnt} );  # The starting section for <portExpression> generation
			
			# Look at remaining protocol entries to see if "fixup" is called for
			
			foreach my $altProtKey (@nwProtocolKeys)
			{
			    if (defined $in->{objectGroup_objects}->{serviceGroup}->{$altProtKey}->{$pGroupKey})
			    {
				push ( @pGroupEvalSecs, 
				       { protocol => $altProtKey, 
					 nwObjectEnt => $in->{objectGroup_objects}->{serviceGroup}->{$altProtKey}->{$pGroupKey}});
				delete $in->{objectGroup_objects}->{serviceGroup}->{$altProtKey}->{$pGroupKey};	
			    }
			}	
			# we have all we need to generate a serviceGroup
			$out->open_element("serviceGroup");
			
			# f there is an explicit description, slap it down
			$out->print_element( "description", $pGroupEnt->{description} )
			    if ( defined $pGroupEnt->{description} );
			$out->print_element( "id", $pGroupEnt->{id} );
			
			
			# Handle nested group refs.
			# +++how did I end up nesting a group ref under a protocol?
			if (defined $in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{$pGroupEnt->{id}})
			{
			    foreach my $ogIndirectRef (@{$in->{objectGroup_objects}->{serviceGroup}->{group_ref}->{$pGroupEnt->{id}}->{groupObjects}})
			    {
				$out->print_element("objectGroupReference", $ogIndirectRef);
			    }
			}
			
			# Now, work through the protocol-based subelements
			my @portExpressions;
			my @portRanges;
			while (@pGroupEvalSecs)
			{
			    my $pgSecEnt = shift @pGroupEvalSecs; 
			    
			    
			    # see if the discrete "ports" entries exists
			    $portsEnt = $pgSecEnt->{nwObjectEnt}->{ports};
			    
			    # this is in turn indexed by a number of well-known equivalence operators, with sets of ports
			    # as var-lenght arrays indexed by those operators for each time they show up in the
			    # group spec
			    
			    if ( defined $portsEnt ) {
				if ( defined $portsEnt->{ports_eq} ) {
				    
				    # What we want to do here is to determine the series of port scalars
				    # and ranges we can take advanage of.  Port sets here are guaranteed
				    # to be sorted on assignment.
				    my @portSet          = @{ $portsEnt->{ports_eq} };
				    my @mappedPortExps   = ();
				    my @mappedPortRanges = ();
				    
				    while (@portSet) {
					my $currentPort = shift @portSet;
					
					if ( @portSet && ( $currentPort + 1 ) == $portSet[0] ) {
					    
					    # we have a range!
					    # $out->open_element("portRange");
					    my $nextRange = {
						protocol  => $pgSecEnt->{protocol},
						portStart => _int_port($currentPort)
					    };
					    
					    # now just count contiguous ports
					    while ( @portSet
						    && ( ( $currentPort + 1 ) == $portSet[0] ) )
					    {
						$currentPort = shift @portSet;
					    }
					    
					    # we counted to the end
					    $nextRange->{portEnd} = _int_port($currentPort);
					    push( @mappedPortRanges, $nextRange );
					}
					else {
					    # just a port ent
					    my $nextExp = {
						port     => _int_port($currentPort),
						operator => "eq",
						protocol => $pgSecEnt->{protocol}
					    };
					    
					    push( @mappedPortExps, $nextExp );
					}
				    }
				    
				    # now, just walk the arrays of exps and ports.
				    if (@mappedPortExps)
				    {
					foreach my $portExp (@mappedPortExps) {
					    push @portExpressions, $portExp;
					}
				    }
				    
				    if (@mappedPortRanges)
				    {
					foreach my $portRange (@mappedPortRanges) {
					    push @portRanges, $portRange;
					}
				    }
				}
				
				# handle lt and gt expressions                           
				foreach my $ltPortVal (@{$portsEnt->{ports_lt}}) 
				{
				    my $ltElement = { port => $ltPortVal, operator => "lt", protocol => $pgSecEnt->{protocol} };
				    push @portExpressions, $ltElement;
				}
				
				foreach my $ltPortVal (@{$portsEnt->{ports_gt}}) 
				{
				    my $ltElement = { port => $ltPortVal, operator => "gt", protocol => $pgSecEnt->{protocol} };
				    push @portExpressions, $ltElement;
				}                                              
			    }
			    
			    foreach my $portRangeEnt (@{$pGroupEnt->{ranges}})
			    {
					my $rangeElement = { portStart => $portRangeEnt->[0], portEnd => $portRangeEnt->[1], protocol => $pgSecEnt->{protocol} };
					push @portRanges, $rangeElement;
			    }
			    
			    
			}
			# print ports
			foreach my $portExp (@portExpressions){
				$out->print_element("portExpression", $portExp);
			}
			foreach my $portRange (@portRanges){
				$out->print_element("portRange", $portRange);
			}
			
			$out->close_element("serviceGroup");
		    }      # end of service group handling
		}
	    }  # End of scoping on "defined $protocolGroup"             	             	
	} until (not @nwProtocolKeys);
	
    }    # end of service groups
    
    if ($opened) {
	$out->close_element("objectGroups");
    }
    
    # We need to clear out the objectGroup_objects from one run to the next.  This is probably the best place
    # to do it, since it would be regenerated from config again before this routine is called...
    if (defined $in->{objectGroup_objects})
    {
	delete $in->{objectGroup_objects};
    }
}

sub parse_access_ports {
    my ( $in, $out ) = @_;
    my $headerPrinted = 0;
    while ( $in->{running_config} =~
	    /\bline\s+(vty|aux|con)?\s*(\d+)\s*(\d*)(.+?)(?=^\S)/msig )
    {
	$out->open_element("accessPorts") if ( !$headerPrinted );
	$headerPrinted++;

	my $endInstance = $3;
	my $blob        = $4;
	my $port        = {
	    type => ( $1 or "unknown" ),
	    startInstance => $2,
	};
	$port->{endInstance} = $endInstance if ($endInstance);

	if ( $blob =~ /^\s+exec-timeout\s+(\d+)\s+(\d+)/mi ) {
	    $port->{inactivityTimeout} = ( $1 * 60 ) + $2;
	}
	if ( $blob =~ /^\s+session-timeout\s+(\d+)\s+(\d+)/mi ) {
	    $port->{sessionTimeout} = ( $1 * 60 ) + $2;
	}
	if ( $blob =~ /^\s+absolute-timeout\s+(\d+)\s+(\d+)/mi ) {
	    $port->{absoluteTimeout} = ( $1 * 60 ) + $2;
	}
	if ( $blob =~ /^\s+transport\s+input\s+(.*\S)\s*$/mi ) {
	    $port->{inboundProtocol} = $1;
	}
	if ( $blob =~ /^\s+transport\s+output\s+(.*\S)\s*$/mi ) {
	    $port->{outboundProtocol} = $1;
	}
	if ( $blob =~ /^\s+access-class\s+(\S+)\s+(in|out)\s*$/mi ) {
	    my $aclname   = $1;
	    my $direction = $2;
	    if ( $direction =~ /out/i ) {
		$port->{egressFilter} = $aclname;
	    }
	    else {
		$port->{ingressFilter} = $aclname;
	    }
	}

	$out->print_element( "accessPort", $port );
    }
    $out->close_element("accessPorts") if ($headerPrinted);
}

sub parse_snmp {
    my ( $in, $out ) = @_;
    my $c = $in->{running_config};
    $out->open_element("snmp");

    while ( $c =~ /^(snmp-server\s+community\s+([^\s]+).*)$/mig ) {
	my $line = $1;
	my $community = { communityString => $2, };
	if ( $line =~ /\s+(RO|RW)/i ) {
	    $community->{accessType} = uc($1);
	}
	if ( $line =~ /\s+view\s+([^\s]+)/ ) {
	    $community->{mibView} = $1;
	}
	if ( $line =~ /\s+(RO|RW)\s+(\d+)/i ) {
	    $community->{filter} = $2;
	}
	$out->print_element( "community", $community );
    }

    if ( $c =~ /^snmp-server\s+contact\s+(.*[^\s])\s*$/mi ) {
	$out->print_element( "sysContact", $1 );
    }
    if ( $c =~ /^snmp-server\s+location\s+(.*[^\s])\s*$/mi ) {
	$out->print_element( "sysLocation", $1 );
    }
    if ( $c =~ /^hostname\s+(\S+)/mi ) {
	$out->print_element( "sysName", $1 );
    }
    if ( $c =~ /^snmp-server\s+system-shutdown\s*$/mi ) {
	$out->print_element( "systemShutdownViaSNMP", "true" );
    }

    while ( $c =~ /^(snmp-server\s+host\s+\"?(\d+\.\d+\.\d+\.\d+)\"?.*$)/mig ) {
	my $trapHost = { ipAddress => $2, };
	my $line     = $1;

	if ( $line =~ /\"?\d+\.\d+\.\d+\.\d+\"?\s+([^\s]+.*)$/ ) {
	    my $extraInformation = $1;
	    my $communityString;

	    if ( $extraInformation =~ /^trap\s+version\s+\S+\s+(\S+)/i ) {
		$communityString = $1;
	    }
	    elsif ( $extraInformation =~ /^trap\s+(\S+)/i ) {
		$communityString = $1;
	    }
	    elsif ( $extraInformation =~ /^inform\s+version\s+\S+\s+(\S+)/i ) {
		$communityString = $1;
	    }
	    elsif ( $extraInformation =~ /^inform\s+(\S+)/i ) {
		$communityString = $1;
	    }
	    elsif ( $extraInformation =~ /^(\S+)/ ) {
		$communityString = $1;
	    }

	    if ($communityString) {
		$trapHost->{communityString} = $communityString;
	    }
	}
	$out->print_element( "trapHosts", $trapHost );
    }
    if ( $c =~ /^snmp-server\s+trap-source\s+(.*[^\s])\s*$/mi ) {
	$out->print_element( "trapSource", $1 );
    }
    if ( $c =~ /^snmp-server\s+trap-timeout\s+(\d+)/mi ) {
	$out->print_element( "trapTimeout", $1 );
    }
    $out->close_element("snmp");
}


# parse_interfaces
#
# Go through the interface section(s) parsing of the running config.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  CAN BE undef, in which case parsing and subnet formation will occur
#    without XML output.
#
# RETURNS:
#   a list:
#   	$subnets - a hash of Ziptie::Addressing::Subnets instances, indexed by interface name
#       $filterAssignments - a hash indexed by interface name, indicating assigned ingress and egress filters.
#
sub parse_interfaces {
    my ( $in, $out ) = @_;
    my $subnets = {};    # will be returned to the caller
    my $filterAssignments = {};  # so willl this.
	my $ipConfiguration = undef;
	my $order;    

    $out->open_element("interfaces") if (defined $out);
    my $cipm = get_crep('cipm');

    # Need to assure availability of routing process data
    _parse_routing_processes($in, $out);

    # Need to perform a secondary (well, asuming parse_interfaces is the primary) pass to
    # assess interface grouping characteristics.
    _parse_interface_groupings($in, $out);		

    while ( $in->{"running_config"} =~ /^interface\s+(\S+)(.+?)^!/smig ) {
	my $name      = $1;
	my $blob      = $2;
	my $interface = {
	    name          => $name,
	    interfaceType => get_interface_type($name),
	    physical      => _is_physical($name),
	};

	# Undef any accumulators used to collect per-interface parsing and transformation state
	undef $ipConfiguration;
	
	if ( $blob =~ /^\s*description\s+(.+?)\s*$/mi ) {
	    $interface->{description} = $1;
	}

	# get the ip addresses
	$order = 1;
	if ( $blob =~ /^\s*ip address\s+($cipm)\s+($cipm)\s*$/mi ) {
	    $ipConfiguration = {
			ipAddress  => $1,
			mask       => mask_to_bits($2),
			precedence => $order,
	    };
	    $order++;
	    push(
			@{ $interface->{"interfaceIp"}->{"ipConfiguration"} },
			$ipConfiguration
		);
	    my $subnet =
		new ZipTie::Addressing::Subnet( $ipConfiguration->{ipAddress},
						$ipConfiguration->{mask} );
	    push( @{ $subnets->{$name} }, $subnet );
	}
	while ( $blob =~ /^\s*ip address\s+($cipm)\s+($cipm)(\ssecondary)/mig )
	{
	    $ipConfiguration = {
			ipAddress  => $1,
			mask       => mask_to_bits($2),
			precedence => $order,
	    };
	    $order++;
	    push(
			@{ $interface->{"interfaceIp"}->{"ipConfiguration"} },
			$ipConfiguration
		);
	    my $subnet =
		new ZipTie::Addressing::Subnet( $ipConfiguration->{ipAddress},
						$ipConfiguration->{mask} );
	    push( @{ $subnets->{$name} }, $subnet );
	}

#		while ( $blob =~ /standby\s+\d+\s+ip\s+(\S+)/g )
#		{
#			my $ipConfiguration = {
#				ipAddress  => $1,
#				precedence => $order,
#			};
#			$order++;
#			push( @{ $interface->{"interfaceIp"}->{"ipConfiguration"} }, $ipConfiguration );
#		}

	$order = 1;
	while ( $blob =~ /^\s*ipv6 address\s+([A-Z\d:]+)\/(\d+)/mig ) {
		$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping interface IPv6 address: $1\/$2 on $name");
	}
	while ( $blob =~ /^\s*ip helper-address\s+(\S+)/mig ) {
	    push( @{ $interface->{"interfaceIp"}->{"udpForwarder"} }, $1 );
	}
	while ( $blob =~ /^\s*ip access-group\s*\b(.+?)\b\s+(in|out)\s*$/mig ) {
	    my $acl       = $1;
	    my $direction = $2;
	    if ( $direction eq "out" ) {
			$interface->{"egressFilter"} = $1;
			push (@{$filterAssignments->{$name}}, { filterName => $1, direction => 'egress'});
	    }
	    elsif ( $direction eq "in" ) {
			$interface->{"ingressFilter"} = $1;
			push (@{$filterAssignments->{$name}}, { filterName => $1, direction => 'ingress'});
	    }
	}
	
	# Process interface grouping characteristics.  We can optimize this by noting right away if this 
	# has been so marked by _parse_interface_groupings.
	for my $ifGrpIndex (@{$in->{interfacesData}->{"$name"}->{ifGroupings}}) 
	{
	    my $grpDescr = undef;
	    
	    # we can iterate across primary and secondary membership lists
	    for $grpDescr ( @{$in->{interfaceGroupings}->{"$ifGrpIndex"}->{secondaryInterfaces}}, 
			    ($in->{interfaceGroupings}->{"$ifGrpIndex"}->{primaryInterface}) )
	    {
		if ($grpDescr->{ifName} eq $name)
		{
		    push ( @{$interface->{interfaceGrouping}}, 
			   {ifGroupName => $ifGrpIndex,
			    ifGroupingCharacteristic => $grpDescr->{ifGroupCharacteristic}});
		}
	    }
	}
	# Independently, let's see if this is marked as primary on any interface grouping
	if (exists $in->{interfaceGroupings}->{"$name"})
	{
	    push ( @{$interface->{interfaceGrouping}},
		   { 	ifGroupName => $name,
			ifGroupingCharacteristic => $in->{interfaceGroupings}->{"$name"}->{"primaryInterface"}->{"ifGroupCharacteristic"}
		   } 
		);
	}

	# process ethernet properties
	if ( $name =~ /eth/i ) {
	    if ( $blob =~ /^(\s+(half|full|auto).duplex|\s+duplex\s+(\S+))/mi )
	    {
			my $duplex;
			$duplex = $2 if defined $2;    # left of the pipe
			$duplex = $3 if defined $3;    # right of the pipe
			if ( ( defined $duplex ) && ( $duplex =~ /auto/i ) ) {
		    	$interface->{interfaceEthernet}->{autoDuplex} = "true";
			}
			elsif ( defined $duplex ) {
		    	$interface->{interfaceEthernet}->{autoDuplex} = "false";
		    	$interface->{interfaceEthernet}->{operationalDuplex} =
				$duplex;
			}
		}
	    elsif ( $blob =~ /^\s+speed\s+\d+/mi ) {
		$interface->{interfaceEthernet}->{autoSpeed} = "false";
	    }
	    elsif ( $blob =~ /^\s+speed\s+auto/mi ) {
		$interface->{interfaceEthernet}->{autoSpeed} = "true";
	    }
	    elsif ( $blob =~ /^\s+no negotiation/mi ) {
		$interface->{interfaceEthernet}->{autoSpeed} = "false";
	    }
	}

	# see if this is a tunneling interface
	if ($blob =~ /^\s+tunnel\s+source\s+(\S+)/mi)
	{
	    my $tunnelLocalEndBlob = $1;
	    my $tunnelLocalEndProps = undef;
	    my $tunnelIfProps;
	    
	    
	    if ($tunnelLocalEndBlob =~ /\d+\.\d+\.\d+\.\d+/)
	    {
			$tunnelLocalEndProps->{tunnelLocalPeeringIPAddress} = $tunnelLocalEndBlob;
	    }
	    elsif (get_interface_type($tunnelLocalEndBlob) ne "other")  # if it returns as a well known i/f type of SOME sort
	    {
			$tunnelLocalEndProps->{tunnelLocalPeeringInterface} = $tunnelLocalEndBlob;				
	    }
	    else
	    {
			$LOGGER->debug(
			    "Cannot find local peer info in configuration for tunnel with parsed local endpoint $tunnelLocalEndBlob; bypassing interface");
			next; # to next interface					
	    }
	    
	    if (defined $tunnelLocalEndBlob)
	    {
			# Note, try not to rely on the interface being *called* tunnelxxx
			$tunnelIfProps = { tunnelLocalPeering => $tunnelLocalEndProps};
		
			if ($blob =~ /^\s+tunnel\s+destination\s+(\d+\.\d+\.\d+\.\d+)/mi)
			{
			    $tunnelIfProps->{tunnelRemotePeerIPAddress} = $1;
			}
			else
			{
			    $LOGGER->debug(
				"Cannot find remote peer address in configuration for tunnel with local address $tunnelIfProps->{tunnelLocalPeeringAddress}");				
			}
	    
		    
			# find tunnel interface encapsulation type, or rather, if it deviates from GRE
		
			if ($blob =~ /^\s+tunnel\s+mode\s+(\S+)/m)
			{
		    	for ($1)
		    	{
					if    (/ipsec/) { $tunnelIfProps->{encapsulationProtocol} = "ipSec"; }
					elsif (/ipip/)  { $tunnelIfProps->{encapsulationProtocol} = "ipv4-in-ipv4"; }
					elsif ( /ip6ip/ ) { $tunnelIfProps->{encapsulationProtocol} = "ipv6-in-ipv4" ;}
		    	}					
			}
    	} 
	    $tunnelIfProps->{encapsulationProtocol} = "gre" if not defined ( $tunnelIfProps->{encapsulationProtocol});
	    $interface->{interfaceTunnel} = $tunnelIfProps;
		    
	}
	

       while ( $blob =~
		/^\s+(no\s+ip|ip)\s+(directed.broadcast|local.proxy.arp|redirects|route.cache|mroute.cache)\s*$/mig
		  )
       {
			my $checkneg  = $1;
			my $ipsetting = $2;
			my $bool      = "true";
			$bool = "false" if ( $checkneg eq "no ip" );
			$interface->{interfaceIp}->{directedBroadcast} = $bool
			  if ( $ipsetting =~ /^directed.broadcast$/ );
			$interface->{interfaceIp}->{localProxyARP} = $bool
			  if ( $ipsetting =~ /^local.proxy.arp$/ );
			$interface->{interfaceIp}->{redirects} = $bool
			  if ( $ipsetting =~ /^redirects$/ );
			$interface->{interfaceIp}->{routeCache} = $bool
			  if ( $ipsetting =~ /^route.cache$/ );
			$interface->{interfaceIp}->{mRouteCache} = $bool
			  if ( $ipsetting =~ /^mroute.cache$/ );
	  }
	    
	    
	while ( $blob =~
		/^\s+(no\s+ip|ip)\s+(directed.broadcast|local.proxy.arp|redirects|route.cache|mroute.cache)\s*$/mig
	    )
	{
	    my $checkneg  = $1;
	    my $ipsetting = $2;
	    my $bool      = "true";
	    $bool = "false" if ( $checkneg eq "no ip" );
	    $interface->{interfaceIp}->{directedBroadcast} = $bool
		if ( $ipsetting =~ /^directed.broadcast$/ );
	    $interface->{interfaceIp}->{localProxyARP} = $bool
		if ( $ipsetting =~ /^local.proxy.arp$/ );
	    $interface->{interfaceIp}->{redirects} = $bool
		if ( $ipsetting =~ /^redirects$/ );
	    $interface->{interfaceIp}->{routeCache} = $bool
		if ( $ipsetting =~ /^route.cache$/ );
	    $interface->{interfaceIp}->{mRouteCache} = $bool
		if ( $ipsetting =~ /^mroute.cache$/ );
	}

	# SONET (as POS)
	if ( $interface->{name} =~ /^pos(\d|\s+\d)/i ) {
	    if ( $blob =~ /^\s+bandwidth\s+(\d+)/mi ) {
		$interface->{interfaceSonet}->{bandwidth} = $1;
	    }
	    if ( $blob =~ /^\s+clock\s*source\s+(\S+)/mi ) {
		$interface->{interfaceSonet}->{clockSource} = $1;
	    }
	    if ( $blob =~ /^\s+crc\s+(\d+)/mi ) {
		$interface->{interfaceSonet}->{crc} = $1;
	    }
	    if ( $blob =~ /^\s+encapsulation\s+(\S+)/mi ) {
		$interface->{interfaceSonet}->{encapsulation} = $1;
	    }
	    if ( $blob =~ /^\s+keepalive\s+(\d+)/mi ) {
		$interface->{interfaceSonet}->{keepAlive} = $1;
	    }
	    if ( $blob =~ /^\s+framing\s+(\S+)/mi ) {
		$interface->{interfaceSonet}->{framing} = $1;
	    }
	    if ( $blob =~ /^\s+flag[s]*\s+(.+)/mi ) {
		$interface->{interfaceSonet}->{flag} = $1;
	    }
	    if ( $blob =~ /^\s+loopback\s+(\S+)/mi ) {
		$interface->{interfaceSonet}->{loopBack} = $1;
	    }
	    if ( $blob =~ /^\s+down-when-looped/mi ) {
		$interface->{interfaceSonet}->{downWhenLooped} = "true";
	    }
	    else {
		$interface->{interfaceSonet}->{downWhenLooped} = "false";
	    }
	}

	# PPP
	if ( $blob =~ /^\s+ppp\s+authentication\s+(\S+)/mi ) {
	    $interface->{interfacePPP}->{authenticationType} = $1;
	}
	if ( $blob =~ /^\s+ppp\s+\S+\s+sent-username\s+(\S+)/mi ) {
	    $interface->{interfacePPP}->{username} = $1;
	}
	if ( $blob =~ /^\s+ppp\s+callback\s+(\S+)/mi ) {
	    $interface->{interfacePPP}->{callBack} = $1;
	}
	if ( $blob =~ /^\s+ppp\s+compression\s+(\S+)/mi ) {
	    $interface->{interfacePPP}->{compression} = $1;
	}
	if ( $blob =~ /^\s+password\s+(\S+)/mi ) {
	    $interface->{interfacePPP}->{password} = $1
		if ( $interface->{interfacePPP} );

	    # cmd is unspecific to ppp, check for other elements exist
	}

	# SERIAL
	if ( $interface->{name} =~ /^serial(\d|\s+\d)/i ) {
	    if ( $blob =~ /^\s+dsu.?bandwidth\s+(\d+)/mi ) {
		$interface->{interfaceSerial}->{dsuBandwidth} = $1;
	    }
	    if ( $blob =~ /^\s+keepalive\s+(\d+)/mi ) {
		$interface->{interfaceSerial}->{keepAlive} = $1;
	    }
	    if ( $blob =~ /^\s+cable.?length\s+(\d+)/mi ) {
		$interface->{interfaceSerial}->{cableLength} = $1;
	    }
	    if ( $blob =~ /^\s+encapsulation\s+(\S+)/mi ) {
		$interface->{interfaceSerial}->{encapsulation} = $1;
	    }
	    if ( $blob =~ /^\s+idle.?character\s+(\S+)/mi ) {
		$interface->{interfaceSerial}->{idleCharacter} = $1;
	    }
	    if ( $blob =~ /^\s+framing\s+(\S+)/mi ) {
		$interface->{interfaceSerial}->{framing} = $1;
	    }
	    if ( $blob =~ /^\s+scramble\s+(\S+)/mi ) {
		$interface->{interfaceSerial}->{scramble} = $1;
	    }
	    if ( $blob =~ /^\s+dsu.?mode\s+(\S+)/mi ) {
		$interface->{interfaceSerial}->{dsuMode} = $1;
	    }
	    if ( $blob =~ /^\s+crc\s+(\d+)/mi ) {
		$interface->{interfaceSerial}->{crc} = $1;
	    }
	    if ( $blob =~ /^\s+down-when-looped/mi ) {
		$interface->{interfaceSerial}->{downWhenLooped} = "true";
	    }
	    else {
		$interface->{interfaceSerial}->{downWhenLooped} = "false";
	    }
	}

	# ATM
	if ( $interface->{name} =~ /^atm(\d|\s+\d)/i ) {
	    if ( $blob =~ /^\s+loopback\s+(\S+)/mi ) {
		$interface->{interfaceATM}->{loopback} = $1;
	    }
	    if ( $blob =~ /^\s+atm\s+\S+?keepalive\s+(\d+)/mi ) {
		$interface->{interfaceATM}->{keepAlive} = $1;
	    }
	    if ( $blob =~ /^\s+atm\s+clock\s+(.+)/mi ) {
		my $clock = lc($1);
		$interface->{interfaceATM}->{clock} = $clock;
	    }
	    if ( $blob =~ /^\s+atm\s+\S+?pvc-discovery\s+(.+)/mi ) {
		$interface->{interfaceATM}->{ilmiPVCDiscovery} = $1;
	    }
	    if ( $blob =~ /^\s+atm\s+scrambling\s+(.+)/mi ) {
		$interface->{interfaceATM}->{scrambling} = $1;
	    }

	    # This is one way that IOS displays ATM PVCs, one per line
	    while ( $blob =~
		    /^(\s*atm\s+pvc\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+).*)$/mig )
	    {
		my $line          = $1;
		my $vcd           = $2;
		my $vpi           = $3;
		my $vci           = $4;
		my $encapsulation = $5;

		my $logicalAtm = {
		    vpi           => $vpi,
		    vci           => $vci,
		    vcd           => $vcd,
		    encapsulation => $encapsulation,
		};

		if ( $line =~ /$encapsulation\s+(\d+)\s+(\d+)\s+(\d+)/ ) {
		    $logicalAtm->{peak}    = $1;
		    $logicalAtm->{average} = $2;
		    $logicalAtm->{burst}   = $3;
		}
		elsif ( $line =~ /$encapsulation\s+(\d+)\s+(\d+)/ ) {
		    $logicalAtm->{peak}    = $1;
		    $logicalAtm->{average} = $2;
		}

		if ( $line =~ /oam\s+(\d+)/i ) {
		    $logicalAtm->{oam} = $1;
		}
		push(
		    @{ $interface->{interfaceATM}->{interfaceLogicalATM} },
		    $logicalAtm
		    );
	    }
	}

	# ISDN (as BRI)
	if ( $interface->{name} =~ /^bri(\d|\s+\d)/i ) {
	    if ( $blob =~ /^\s+encapsulation\s+(\S+)/mi ) {
		$interface->{interfaceISDN}->{encapsulation} = $1;
	    }
	    if ( $blob =~ /^\s+keepalive\s+(\d+)/mi ) {
		$interface->{interfaceISDN}->{keepAlive} = $1;
	    }
	    if ( $blob =~ /^\s+isdn\s+switch.?type\s+(\S+)/mi ) {
		$interface->{interfaceISDN}->{switchType} = $1;
	    }
	    if ( $blob =~ /^\s+isdn\s+spid1\s+(.+)/mi ) {
		$interface->{interfaceISDN}->{spid1} = $1;
	    }
	    if ( $blob =~ /^\s+isdn\s+spid2\s+(.+)/mi ) {
		$interface->{interfaceISDN}->{spid2} = $1;
	    }
	    if ( $blob =~ /^\s+isdn caller\s+(\S+)/mi ) {
		$interface->{interfaceISDN}->{caller} = $1;
	    }
	    if ( $blob =~ /^\s+loopback\s+(\S+)/mi ) {
		$interface->{interfaceISDN}->{loopback} = $1;
	    }
	}

	# IPX and IPX addresses <-- needs more testing
	if ( $blob =~ /^\s+ipx\s+link-delay\s+(\d+)/mi ) {
	    $interface->{interfaceIPX}->{linkDelay} = $1;
	}
	if ( $blob =~ /^\s+ipx\s+throughput\s+(\d+)/mi ) {
	    $interface->{interfaceIPX}->{throughPut} = $1;
	}
	if ( $blob =~
	     /^\s+ipx\s+network\s+(\S+)\s+encapsulation\s+(\S+)\s+secondary/mi )
	{
	    push(
		@{ $interface->{interfaceIPX}->{interfaceIPXAddress} },
		{ ipxAddress => $1, encapsulation => $2, ipxOrder => 2 }
		);
	}
	if ( $blob =~ /^\s+ipx\s+network\s+(\S+)\s+encapsulation\s+(\S+)/mi ) {
	    push(
		@{ $interface->{interfaceIPX}->{interfaceIPXAddress} },
		{ ipxAddress => $1, encapsulation => $2, ipxOrder => 1 }
		);
	}
	if ( $blob =~ /^\s+ipx\s+network\s+(\S+)\s+secondary/mi ) {
	    push(
		@{ $interface->{interfaceIPX}->{interfaceIPXAddress} },
		{ ipxAddress => $1, ipxOrder => 2 }
		);
	}
	if ( $blob =~ /^\s+ipx\s+network\s+(\S+)/mi ) {
	    push(
		@{ $interface->{interfaceIPX}->{interfaceIPXAddress} },
		{ ipxAddress => $1, ipxOrder => 1 }
		);
	}

	# FRAME RELAY
	if ( $blob =~ /\s+encapsulation\s+frame-relay\s+(\S+)\s*$/mi ) {
	    $interface->{interfaceFrameRelay}->{encapsulation} = $1;
	}
	if ( $blob =~ /^\s+frame-relay\s+lmi-type\s+(\S+)\s*$/mi ) {
	    $interface->{interfaceFrameRelay}->{lmiType} = $1;
	}
	if ( $blob =~ /^\s+frame-relay\s+interface-dlci\s+(\d+)/mi ) {
	    push(
		@{ $interface->{interfaceFrameRelay}->{virtualFrameRelay} },
		{ dlci => $1 }
		);
	}

	# VLAN trunk references
	if ( $blob =~ /^\s+switchport\s+trunk\s+allowed\s+vlan\s+(\S+)/mi ) {
	    my $vlandata = $1;
	    if ( $vlandata =~ /^\d+$/ ) {
		push(
		    @{ $interface->{interfaceVlanTrunks} },
		    { startVlan => $vlandata, }
		    );
	    }
	    else {
		foreach my $trunkedvlan ( split /,/, $vlandata ) {
		    if ( $trunkedvlan =~ /-/ ) {
			my ( $start, $end ) = split /-/, $trunkedvlan;
			push(
			    @{ $interface->{interfaceVlanTrunks} },
			    { startVlan => $start, endVlan => $end, }
			    );
		    }
		    else {
			push(
			    @{ $interface->{interfaceVlanTrunks} },
			    { startVlan => $trunkedvlan, }
			    );
		    }
		}
	    }
	}

	# EIGRP
	
	# parse for AS-level EIGRP properties - hello-interval and hold-time.  Can be different
	# for different AS'es.
	if ($blob =~ /ip h[a-z\-]+ eigrp (\d+)\s+\d+/gm) {
	    my $eigrpAS = $1;
	    $interface->{"cisco:eigrp"}->{"cisco:asNumber"}      = $eigrpAS;						
	    if ( $blob =~ /hello-interval eigrp $eigrpAS (\d+)/ ) {
		$interface->{"cisco:eigrp"}->{"cisco:helloInterval"} = $1;
	    }
	    if ( $blob =~ /hold-time eigrp $eigrpAS (\d+)/ ) {
		$interface->{"cisco:eigrp"}->{"cisco:holdTime"} = $1;
	    }
	    # see if we can detect this as passive on this AS.  Or active, if interfaces default to passive.
	    if ($in->{routingProcesses}->{eigrp}->{$eigrpAS}) {
		if ($in->{routingProcesses}->{eigrp}->{$eigrpAS}->{'cisco:passiveInterfaceDefault'} eq 'true') {
		    # start by assuming it *is* passive
		    $interface->{"cisco:eigrp"}->{"cisco:passive"} = 1;								
		    foreach my $activeEigrpIf (@{$in->{routingProcesses}->{eigrp}->{$eigrpAS}->{'cisco:activeInterface'}}) {
			if ($activeEigrpIf eq $name) {
			    $interface->{"cisco:eigrp"}->{"cisco:passive"} = 0;								
			}
		    }
		}
		
		foreach my $passiveEigrpIf (@{$in->{routingProcesses}->{eigrp}->{$eigrpAS}->{'cisco:passiveInterface'}}) {
		    if ($passiveEigrpIf eq $name) {
			$interface->{"cisco:eigrp"}->{"cisco:passive"} = 1;								
		    }
		}					
	    }
	    # out of all of that....
	    if (not defined $interface->{"cisco:eigrp"}->{"cisco:passive"}) {
		$interface->{"cisco:eigrp"}->{"cisco:passive"} = 0;
	    }
	}

	# process this particular interface from "show ip ospf interface"
	if ( $in->{ospf_ints} =~ /^($name\s.+?)(?=^\S)/msi ) {
	    my $blob = $1;
	    if ( $blob =~ /Area (\S+)/i ) {
		$interface->{interfaceOspf}->{area} = $1;
	    }
	    if ( $blob =~
		 /Process ID\s+(\d+),\s*Router ID\s+\S+,\s*Network Type\s+(\S+),\s*Cost:\s+(\d+)/i
		)
	    {
		$interface->{interfaceOspf}->{processId}   = $1;
		$interface->{interfaceOspf}->{networkType} = $2;
		$interface->{interfaceOspf}->{cost}        = $3;
	    }
	    if ( $blob =~
		 /Transmit Delay is (\d+) sec,\s*State (\S+),\s*(Priority (\d+))?/i
		)
	    {
		$interface->{interfaceOspf}->{transmitDelay}  = $1;
		$interface->{interfaceOspf}->{routerState}    = $2;
		$interface->{interfaceOspf}->{routerPriority} = $4 if ($4);
	    }
	    if ( $blob =~
		 /Hello\s*(\d+),\s*Dead\s*(\d+),\s*Wait\s*(\d+),\s*Retransmit\s*(\d+)/i
		)
	    {
		$interface->{interfaceOspf}->{helloInterval}      = $1;
		$interface->{interfaceOspf}->{deadInterval}       = $2;
		$interface->{interfaceOspf}->{waitInterval}       = $3;
		$interface->{interfaceOspf}->{retransmitInterval} = $4;
	    }
	}

	# process this particular interface from "show interfaces"
	if ( $in->{interfaces} =~ /^($name\s.+?)(?=^\S)/msi ) {
	    my $intBlob = $1;
	    if ( $intBlob =~ /$name\s+is\s+(up|\S+\s+down|down),/mi ) {
		$interface->{adminStatus} =
		    ( $1 =~ /down/i ) ? 'down' : 'up';    # Defaults to 'up'.
	    }
	    if ( $intBlob =~
		 /address is ([a-f0-9A-F]{4}\.[a-f0-9A-F]{4}\.[a-f0-9A-F]{4})/mi
		)
	    {
		my $macAddress = $1;
		$interface->{interfaceEthernet}->{macAddress} =
		    strip_mac($macAddress);
	    }
	    if ( $intBlob =~ /MTU\s+(\d+)\s+bytes,\s+BW\s+(\d+)\s+(\S+),/mi ) {
		my $mtu   = $1;
		my $speed = $2;
		my $units = $3;
		if ( $units =~ /Mb/i ) { $speed = $speed * 1000 * 1000; }
		if ( $units =~ /Kb/i ) { $speed = $speed * 1000; }
		$interface->{speed} = $speed;
		$interface->{mtu}   = $mtu;
	    }
	    if ( $intBlob =~
		 /MTU\s+(\d+)\s+bytes,\s+sub\s+MTU\s+\d+,\s+BW\s+(\d+)\s+(\S+),/
		)
	    {
		my $mtu   = $1;
		my $speed = $2;
		my $units = $3;
		if ( $units =~ /Mbit/i ) { $speed = $speed * 1000; }
		$interface->{speed} = $speed;
		$interface->{mtu}   = $mtu;
	    }
	    if ( $intBlob =~ /Encapsulation\s([^\s^,]+)/ ) {
		my $encaps = $1;
		if (   ( defined $interface->{interfaceEthernet} )
		       || ( $interface->{interfaceType} eq "ethernet" ) )
		{
		    $interface->{interfaceEthernet}->{encapsulation} = $encaps;
		}
	    }
	    if ( $intBlob =~
		 /(Half|Full|Auto)[- ]duplex,\s+(100Mb|10Mb|Auto).+,\s+(media type is |)(\S+)/i
		)
	    {
		my $duplex    = $1;
		my $autospeed = $2;
		my $media     = $4;
		if ( $duplex =~ /half/i ) {
		    $interface->{interfaceEthernet}->{operationalDuplex} =
			"half";
		}
		elsif ( $duplex =~ /full/i ) {
		    $interface->{interfaceEthernet}->{operationalDuplex} =
			"full";
		}
		elsif ( $duplex =~ /auto/i ) {
		    $interface->{interfaceEthernet}->{autoDuplex} = "true";
		}
		if ( $autospeed =~ /auto/i ) {
		    $interface->{interfaceEthernet}->{autoSpeed} = "true";
		}
		$interface->{interfaceEthernet}->{mediaType} = $media;
	    }

	    # Last chance for an internet address here if we don't have one (for example, an ISDN interface only activated
	    # on demand where no IP address would be in running-config, but might be in 'show interfaces' state)
	    if (not defined $ipConfiguration)
	    {
	    	$order = 1;
			while ( $intBlob =~ /Internet\s+address\s+is\s+([\d.:a-fA-F]+)\/(\d+)/gi )
			{
	    		$ipConfiguration = {
					ipAddress => $1,
					mask      => mask_to_bits($2),
					precedence => $order++,
	    		};
		    	push( @{ $interface->{"interfaceIp"}->{"ipConfiguration"} }, $ipConfiguration );

			    my $subnet =
				new ZipTie::Addressing::Subnet( $ipConfiguration->{ipAddress},
								$ipConfiguration->{mask} );
			    push( @{ $subnets->{$name} }, $subnet );

	    	}
		}
	}

	if ( $in->{"running_config"} =~
	     /^ntp\s+(server|peer)/mi )  #i.e. not running sntp AND ntp is active
	{
	    $interface->{ntpServerEnabled} =
		( $blob =~ /ntp\s+disable/mi ) ? "false" : "true";
	}

	my @failovers = ( 'standby', 'vrrp', 'glbp' );
	foreach my $failover (@failovers) {
	    if ( $blob =~ $failover ) {
		my $fail;

		my $group;

# Call routine to parse failover groups.  The "fully qualified" case has a number
# following "standby" (or whatever), and the "default" group case has no failover
# group number, in which case it defaults to group 0.
#
		if ( $blob =~
		     /(\s+(?:$failover)\s+[a-z]+.*\n(?:\s+(?:$failover)\s+[a-z]+.*\n)*)/mig
		    )
		{
		    $group =
			_process_if_failover_group( $in, $failover, $name, $1,
						    0 );

		    if ( defined $group ) {
			push( @{ $fail->{"failoverGroup"} }, $group );
		    }
		}

		while ( $blob =~
			/(\s+(?:$failover)\s+(\d+)\s+.*\n(?:\s+(?:$failover)\s+\2\s+.*\n)*)/mig
		    )
		{
		    $group =
			_process_if_failover_group( $in, $failover, $name, $1,
						    $2 );

		    if ( defined $group ) {
			push( @{ $fail->{"failoverGroup"} }, $group );
		    }
		}

		if ( $failover eq "standby" ) {
		    $fail->{icmpRedirect} = "true"
			if ( $blob =~ /standby\s+redirect/mi );

		    my ( $min_del, $reload_del ) =
			$blob =~
			/standby\s+delay\s+minimum\s+(\d+)\s+reload\s+(\d+)/mi;
		    $fail->{interfaceDelay} = $min_del if ($min_del);

		    if ( $blob =~ /standby\s+mac-refresh\s+(\d+)/mi ) {
			$fail->{macRefresh} = $1;
		    }

		    $fail->{reloadDelay} = $reload_del if ($reload_del);

		    $fail->{version} =
			( $blob =~ /standby\s+version\s+(\d+)/ ) ? $1 : "1";
		}

		$failover = "hsrp" if ( $failover eq "standby" );

		# Make sure at least one failover group was specified and successfully
		# parsed before assigning the owner failover element.
		$interface->{failover}->{$failover} = $fail
		    if ($fail)
		    && ( defined $fail->{"failoverGroup"} );
	    }
	}

	if ( $blob =~ /service-policy\s+(input|output)\s+(\S+)\s*/mi ) {
	    my $direction  = $1;
	    my $policyName = $2;

	    $direction = "Inbound"  if ( $direction eq "input" );
	    $direction = "Outbound" if ( $direction eq "output" );

	    my $appliedQos = {
		name      => $policyName,
		direction => $direction,
	    };

	    $interface->{interfaceQOS}->{policyMap} = $appliedQos;
	}

	while ( $blob =~
		/rate-limit\s+(input|output)(?:\s+dscp\s(\d+))?(?:\s+access-group\s+(?:rate-limit\s+)?(\d+))?\s+(\d+)\s+(\d+)\s+(\d+)\s+conform-action\s+(.*?)\s+exceed-action\s+(.*?)\s*$/mig
	    )
	{
	    my $direction = $1;
	    my $dscp      = $2;
	    my $acl       = $3;
	    my $bps       = $4;
	    my $normal    = $5;
	    my $max       = $6;
	    my $conform   = $7;
	    my $exceed    = $8;

	    my $rateLimit;

	    $direction = "Inbound"  if ( $direction eq "input" );
	    $direction = "Outbound" if ( $direction eq "output" );

	    $rateLimit->{direction}            = $direction;
	    $rateLimit->{dscp}                 = $dscp if ($dscp);
	    $rateLimit->{accessGroup}          = $acl if ($acl);
	    $rateLimit->{bps}->{bandwidthBits} = $bps;
	    $rateLimit->{normalBurst}          = $normal;
	    $rateLimit->{maxBurst}             = $max;
	    $rateLimit->{conformAction}        = _set_action($conform);
	    $rateLimit->{exceedAction}         = _set_action($exceed);

	    push( @{ $interface->{"interfaceQOS"}->{"rateLimit"} },
		  $rateLimit );
	}
	
		$out->print_element( "cisco:interface", $interface )  if (defined $out);

		if (!defined($in->{_remoteGatewayResolution}))
		{
			$in->{_remoteGatewayResolution} = ZipTie::RemoteGatewayResolution->new();
		}
		
		if ( defined $interface->{name} )
		{
			# Add the interfaces to the remoteGatewayResolution object to help resolve static routes with remote gateways.
			if (defined($interface->{interfaceIp}->{ipConfiguration}))
			{
				foreach my $ipConfiguration (@{$interface->{interfaceIp}->{ipConfiguration}})
				{
					$in->{_remoteGatewayResolution}->add_local_subnet($ipConfiguration->{ipAddress}, $ipConfiguration->{mask}, $interface->{name});
				}
			}
		}
    }

    $out->close_element("interfaces")  if (defined $out);
	
    return ($subnets, $filterAssignments);
}

sub _process_if_failover_group {
    my ( $in, $failover_kwd, $if_name, $failover_cfg, $failover_group_number ) =
	@_;
    my $group_zed_cfg;    # a reference to the config hash we will build up

    if ( $failover_cfg =~ /weighting\s+(\d+.*)$/mi )    #only applies to GLBP
    {
	my $values = $1;

	my ( $max, $val1, $val2 ) =
	    $values =~ /(\d+)(?:\s+(lower\s+\d+))?(?:\s+(upper\s+\d+))?/mi;

	$group_zed_cfg->{glbpWeighting}->{maximum} = $max;
	if ( !$val2 ) {
	    if ( $val1 =~ /upper/mi ) {
		$val2 = $val1;
		undef $val1;
	    }
	}

	$val1 =~ s/lower\s+//mi;
	$val2 =~ s/upper\s+//mi;

	$group_zed_cfg->{glbpWeighting}->{lower} = $val1 if ($val1);
	$group_zed_cfg->{glbpWeighting}->{upper} = $val2 if ($val2);
    }

    $group_zed_cfg->{groupID} = $failover_group_number;

    if ( $failover_cfg =~ /(?:name|description)\s+(.*)$/mi ) {
	$group_zed_cfg->{groupName} = $1;
    }

    if ( $failover_cfg =~ /mac-address\s+(\d{4}\.\d{4}\.\d{4})/mi ) {
	$group_zed_cfg->{macAddress} = strip_mac($1);
    }
    elsif ( $in->{$failover_kwd}
	)    # try to get the virtual mac from the show commands
    {
	my ($mac_blob) =
	    $in->{$failover_kwd} =~
	    /($if_name\s+-\s+Group\s+$failover_group_number.*\n(?:\S+.*\n)?(?:\s+.*\n)*)/mi;
	if ($mac_blob) {
	    if ( $mac_blob =~
		 /^\s+(?:local\s+)?virtual\s+MAC\s+address\s+is\s+(\S+)/mi
		)    #VRRP and HSRP only
	    {
		$group_zed_cfg->{macAddress} = strip_mac($1);
	    }
	}
    }

    if ( $failover_cfg =~ /preempt\s*(delay\s+.*)$/mi ) {
	my $preempt_blob = $1;

	$group_zed_cfg->{preempt}->{state} = "true";
	if ( $preempt_blob =~ /minimum\s+(\d+)/mi ) {
	    $group_zed_cfg->{preempt}->{delay} = $1;
	}
	if ( $preempt_blob =~ /reload\s+(\d+)/mi ) {
	    $group_zed_cfg->{preempt}->{reloadDelay} = $1;
	}
	if ( $preempt_blob =~ /sync\s+(\d+)/mi ) {
	    $group_zed_cfg->{preempt}->{syncDelay} = $1;
	}
    }

    if ( $failover_cfg =~ /ip\s+(\d+\.\d+\.\d+\.\d+)\s*$/mi ) {
	$group_zed_cfg->{primaryIPaddress} = $1;
    }

    if ( $failover_cfg =~ /priority\s+(\d+)/mi ) {
	$group_zed_cfg->{priority} = $1;
    }

    while ( $failover_cfg =~ /ip\s+(\d+\.\d+\.\d+\.\d+)\s+secondary\s*$/mig ) {
	push( @{ $group_zed_cfg->{"secondaryIPaddress"} }, $1 );
    }

    if ( $failover_cfg =~ /authentication\s+(.*)/mi ) {
	my $authentication_line = $1;

	if ( $authentication_line =~
	     /md5\s+(key-chain|key-string)\s+(\S+|\d+\s+\S+)/mi )
	{
	    my $type     = $1;
	    my $key_blob = $2;

	    $group_zed_cfg->{security}->{encryption} = "MD5";
	    if ( $authentication_line =~ /key-string\s+\d+\s+(\S+)/mi ) {
		$group_zed_cfg->{security}->{keyString} = $1;
	    }
	    else {
		( $group_zed_cfg->{security}->{keyChain}->{keyChainName} ) =
		    $authentication_line =~ /key-chain\s+(\S+)/mi;
	    }
	}
	else {
	    $group_zed_cfg->{security}->{encryption} = "Plain text";
	    $group_zed_cfg->{security}->{keyString}  = $authentication_line;
	}
    }

    if ( $failover_cfg =~
	 /timers\s+(?:((?:msec\s+)?\d+)\s+((?:msec\s+)?\d+)|advertise\s+((?:msec\s+)?\d+))/mi
	)    # glbp&hsrp | vrrp
    {
	my $hello =
	    ($3)
	    ? $3
	    : $1
	    ; #if $3 is filled then it's an "advertise xxx" else it's a "xxx yyy".
	my $hold = $2;    #only for hsrp & glbp
	$group_zed_cfg->{timers}->{helloTimer} =
	    ( $hello =~ /msec\s+(\d+)/mi ) ? $1 : $hello * 1000;
	$group_zed_cfg->{timers}->{holdTimer} =
	    ( $hold =~ /msec\s+(\d+)/mi ) ? $1 : $hold * 1000
	    if ($hold);
    }

    while ( $failover_cfg =~ /track\s+(\S+)\s*(?:(?:decrement\s+)?(\d+))?/mig )
    {
	my $track;

	my $tracker   = $1;
	my $decrement = $2;

	if ( $tracker =~ /^(?:(?:[1-4]?[0-9]?[0-9])|500)$/mi
	    ) #if the track object is just the number between 1 and 500 (Cisco Tracker ID range)
	{
	    $track->{objectID} = $tracker;
	}
	else {
	    $track->{interface} = $tracker;
	}

	$track->{decrement} = ($decrement) ? $decrement : 10;

	push( @{ $group_zed_cfg->{"track"} }, $track );
    }

    # Wayne: Bug 29646: perform validation on the group looking
    # for "required" and essential assignments and element value
    # (or cross-element) semantics.  Do not emit into ZED if
    #  validation fails here.
    if ( not defined $group_zed_cfg->{groupID} ) {
	$LOGGER->debug(
	    "Bypassing generation of failover group with no group ID");
    }
    elsif ( not defined $group_zed_cfg->{primaryIPaddress} ) {

	# No ipaddr "shared" with other members of
	# failover group--will KO QRM topology
	$LOGGER->debug( "Bypassing generation of entry for failover group "
			. $group_zed_cfg->{groupID}
			. "; no primary IP address found" );
    }
    else {

	# SUCCESSFUL FAILOVER GROUP PARSE.

	return $group_zed_cfg;
    }

    # Fallthrough condition--failure to parse or unsuccessful validation.
    return undef;
}

sub parse_stp {

    # parses Spanning Tree information
    my ( $in, $out ) = @_;
    my $spanningTree;

    # instances from "show spanning-tree"
    while ( $in->{stp} =~ /^(VLAN|MST)0*(\d+)(.+?)^-+/msig ) {
	my $blob = $3;
	my $instance = { vlan => $2, };
	if ( $blob =~ /^\s*Root\s+ID(.+?)^\s*$/msi ) {
	    my $rootSection = $1;
	    if ( $rootSection =~ /Priority\s+(\d+)/i ) {
		$instance->{designatedRootPriority} = $1;
	    }
	    if ( $rootSection =~ /Address\s+([0-9a-f.]+)/i ) {
		$instance->{designatedRootMacAddress} = strip_mac($1);
	    }
	    if ( $rootSection =~ /Cost\s+(\d+)/i ) {
		$instance->{designatedRootCost} = $1;
	    }
	    if ( $rootSection =~ /Port\s+\d+\s+\((\S+)\)/i ) {
		$instance->{designatedRootPort} = $1;
	    }
	    if ( $rootSection =~
		 /Hello Time\s+(\d+) sec\s+Max Age\s+(\d+) sec\s+Forward Delay\s+(\d+) sec/i
		)
	    {
		$instance->{designatedRootHelloTime}    = $1;
		$instance->{designatedRootMaxAge}       = $2;
		$instance->{designatedRootForwardDelay} = $3;
	    }
	}
	if ( $blob =~ /^\s*Bridge\s+ID(.+?)^\s*$/msi ) {
	    my $bridgeSection = $1;
	    if ( $bridgeSection =~ /Priority\s+(\d+)/i ) {
		$instance->{priority} = $1;
	    }
	    if ( $bridgeSection =~ /Address\s+([0-9a-f.]+)/i ) {
		$instance->{systemMacAddress} = strip_mac($1);
	    }
	    if ( $bridgeSection =~
		 /Hello Time\s+(\d+) sec\s+Max Age\s+(\d+) sec\s+Forward Delay\s+(\d+) sec/i
		)
	    {
		$instance->{helloTime}    = $1;
		$instance->{maxAge}       = $2;
		$instance->{forwardDelay} = $3;
	    }
	}
	push( @{ $spanningTree->{spanningTreeInstance} }, $instance );
    }
    $out->print_element( "spanningTree", $spanningTree ) if ($spanningTree);
}

sub parse_routing {
    my ( $in, $out ) = @_;

    # call common scalar routine to set up %IOSRoutingProcesses
    _parse_routing_processes($in, $out);
    
    if ( scalar keys %{$in->{routingProcesses}} ) {

	# the generic routing protocols must come before the cisco:eigrp definition
	my $routInstId;
	$out->open_element("cisco:routing");
	foreach  $routInstId  ( keys %{$in->{routingProcesses}->{bgp}}  ) {
	    $out->print_element( "bgp", $in->{routingProcesses}->{bgp}->{$routInstId} );
	}
	foreach  $routInstId ( keys %{$in->{routingProcesses}->{ospf}} ) {
	    $out->print_element( "ospf", $in->{routingProcesses}->{ospf}->{$routInstId} );
	}
	foreach my $routInstId (  keys %{$in->{routingProcesses}->{rip}} ) {
	    $out->print_element( "rip", $in->{routingProcesses}->{rip}->{$routInstId} );
	}
	foreach my $routInstId (  keys %{$in->{routingProcesses}->{eigrp}} ) {
	    $out->print_element( "cisco:eigrp", $in->{routingProcesses}->{eigrp}->{$routInstId} );
	}
	$out->close_element("cisco:routing");
    }
}

#  This interface parses the running config to determine the current configured state 
#  of routing by dynamic routing protocol.  It returns a hash instance compatible with the
#  IOSRoutingProcesses hash format.
#
#  It sets up the context of what it encounters in the $in->{routingProcesses} hash, which is 
#  directly keyed by protocol name.
sub _parse_routing_processes {
	my ($in, $out) = @_;    	

	# If this has already been called, they bypass...state is in $in->{routingProcesses}
	# HACK!  For now, differentiate whether it was set up by calling this param with a
	# $out, which indicates (sigh) it's to be used directly for SED assignment.  Otherwise,
	# do regenerate.
	if (scalar keys %{$in->{routingProcesses}}) {
		if (not (defined $out xor $in->{routingProcesses}->{outputZed} )) {
			return;
		}
	}

	$in->{routingProcesses}->{outputZed} = defined $out;

	# Parse out BGP, OSPF and EIGRP details
	while ( $in->{running_config} =~ /^router\s+(eigrp|ospf|bgp|rip)(?:\s+(\d+))?\s*\n((?:\s+.*?\n)*)/msig ) {
		my $type = $1;

		# Essential that "id" is assigned somehow
		my $id = ($2) ? $2 : 'global';
		my $blob = $3;

		# Used for OSPF
		my $areas;

		my $protocol;


		# BUG 48074: As it stands now, the only real processing of route redistribution is to "declare" it in the 
		# routing process declaration.  Took out a ton of logic in parse_dynamic_routes to "do" redistribution
		# directly into the SED.  At the point where we really care about routing toplogy (and where a route 
		# actually comes from) it makes a lot more sense to have this handled in QRM topology who can not only
		# assess the dynamic route, but also the filters which will be inevitably applied to the redistribution.
		# OTOH, as it pertains to path-path topology, the redistribution will show up in the route base for 
		# the target redistribution in the SED of the routing neighbor, right?
		while ($blob =~ /^\s*redistribute\s*(bgp|ospf|eigrp|rip|static|connected)(\s(\d+))?(\s+route\-map\s+(\S+))?/img) {
			my $redistribution = { targetProtocol => $1, };
			$redistribution->{processId} = $3 if ($3);

			# We may not always handle things like this, but right now, there really isn't manifest "support" for routing
			# from other instances of the same routing protocol.  This will be reexamined, in a huge way, when we support
			# OSPF and BGP for which this is a common practice.  It's more exotic (and explicitly discouraged) by Cisco.
			# The end result is that each EIGRP process gets a full instance of *all* EIGRP routes as it *is*, so supporting
			# "redistribute eigrp y" into a definition for eigrp x is really just doubling the size of that route base,
			# which can be huge as it is (let alone the sheer redundancy, which is a bug in its own right--*given the current
			# implementation*, which will be inadequate for BGP or OSPF).
			#
			# As a result, we just compare "protocol" keywords in the if statement below.
			if ($redistribution->{targetProtocol} ne $type) {
				push( @{ $protocol->{ _get_routing_element_namespace($type) . "redistribution" } }, $redistribution );		
			}
		}
		if ( $blob =~ /synchronization/i ) {
			$protocol->{ _get_routing_element_namespace($type) . "synchronization" } =
			( $blob =~ /no synch/i ) ? "false" : "true";
		}
		if ( $blob =~ /auto-summary/i ) {
			$protocol->{ _get_routing_element_namespace($type) . "autoSummarization" } =
			( $blob =~ /no auto-sum/i ) ? "false" : "true";
		}

		if ( $type eq 'ospf' ) {
			while ( $blob =~ /^\s*area\s(\d+\S*)\s(nssa|stub)/img ) {
				my $area     = $1;
				my $areaType = $2;
				if ( !defined $areas->{$1} ) {
					$areaType = "SA" if ($areaType eq 'stub');
					$areas->{$1} = { areaId => $area, areaType => uc($areaType) };
				}
			}
			while ($blob =~ /^\s*network\s*(\d+\.\d+\.\d+\.\d+)\s*(\d+\.\d+\.\d+\.\d+).+?area\s(\d+\S*)/img) {
				my $area    = $3;
				my $network = create_network_element( $1, $2,  1 );

				if ( !defined $areas->{$area} ) {
					$areas->{$area} = {
						areaId   => $area,
						areaType => 'normal'
					};
				}
				push( @{ $areas->{$area}->{network} }, $network );
			}
		} elsif ( $type eq 'rip' ) {
			while ( $blob =~ /^\s*network\s*((\d+)\.\d+\.\d+\.\d+)\s*/mig ) {
				my $net = { address => $1 };
				my $octet1 = $2;

				my $mask;
				$mask = '255.0.0.0'     if ( $octet1 < 127 );
				$mask = '255.255.0.0'   if ( $octet1 >= 127 && $octet1 < 192 );
				$mask = '255.255.255.0' if ( $octet1 >= 192 );

				$net->{mask} = mask_to_bits($mask);
				push( @{ $protocol->{ _get_routing_element_namespace($type) . "network" } }, $net);
			}
		} else {
			# This regex is compatible between IOS v.12 and IOS v.15, where "mask" is put into the network declaration.
			while ($blob =~ /^\s*network\s*(\d+\.\d+\.\d+\.\d+)(\s+mask)? ?(\d+\.\d+\.\d+\.\d+)?/mig) {
				my $mask = $3;
				if ( !$mask ) {
					$mask = net_addr_to_mask($1);
				} elsif ( $mask eq "0.0.0.0") {
					# Tend towards wildcard here, that's a /32
					$mask = invert_mask($mask);
				}
				push( @{ $protocol->{ _get_routing_element_namespace($type) . "network" } }, 
				create_network_element($1, $mask, not is_contiguous_mask($mask)) );
			}
		}

		# Areas and summarizedAddress element only for OSPF...now.
		if ($type eq 'ospf') {
			foreach my $key ( keys %$areas ) {
				push( @{ $protocol->{"area"} }, $areas->{$key} );
			}

			while ($blob =~	/summary-address\s+(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+)/g) {
				my $summarizedAddress = create_network_element( $1, $2, 0 );
				push(
					@{ $protocol->{ _get_routing_element_namespace($type) . "summarizedAddress" } },
					$summarizedAddress
				);
			}
		}

		if ( $blob =~ /^\s*passive-interface default/ ) {
			if ( $type =~ /ospf|rip/mi ) {
				$protocol->{allInterfacesEnabled} = 'true';
			} elsif ( $type eq 'eigrp' ) {
				$protocol->{'cisco:passiveInterfaceDefault'} = 'true';
			}
		} else {
			if ( $type =~ /ospf|rip/mi ) {
				$protocol->{allInterfacesEnabled} = 'false';
			} elsif ( $type eq 'eigrp' ) {
				$protocol->{'cisco:passiveInterfaceDefault'} = 'false';
			}
		}

		while ( $blob =~ /^\s*no passive-interface (\S+)/mig ) {
			my $int = $1;
			if ( $int ne 'default' ) {
				if ( $type =~ /ospf|rip/mi ) {
					push( @{ $protocol->{enabledInterface} }, $int );
				} elsif ( $type eq 'eigrp' ) {
					push( @{ $protocol->{'cisco:activeInterface'} }, $int );
				}
			}
		}
		while ( $blob =~ /^\s*passive-interface (\S+)/mig ) {
			my $int = $1;
			if ( $int ne 'default' ) {
				if ( $type =~ /ospf|rip/mi ) {
					push( @{ $protocol->{disabledInterface} }, $int );
				} elsif ( $type eq 'eigrp' ) {
					push( @{ $protocol->{'cisco:passiveInterface'} }, $int );
				}
			}
		}

		if ( $type =~ /eigrp|bgp/i ) {
			$protocol->{ _get_routing_element_namespace($type) . "asNumber" } = $id;
		} elsif ( $type eq "ospf" ) {
			$protocol->{processId} = $id;
		}

		# Get the OSPF or EIGRP router ID
		if ( $type =~ /ospf|eigrp/mi ) {
			if ( $blob =~ /^\s*router-id\s*(\S+)/mi ) {
				$protocol->{_get_routing_element_namespace($type) . "routerId"} = $1;
			} elsif ( $in->{protocols} =~ /^Routing Protocol is \"$type $id\"(.+?)(?=^\S)/msi ) {
				my $protBlob = $1;
				if ( $protBlob =~ /^\s*Router ID\s+(\S+)/im ) {
					$protocol->{_get_routing_element_namespace($type) . "routerId"} = $1;
				}
			}
		}

		# Get the BGP neighbor definitions
		if ( $type eq 'bgp' ) {
			my $peerGroupAsNumbers = {};
			while ( $blob =~ /^\s*neighbor\s(\S+)\s+remote-as\s+(\d+)/mig ) {
				my $def = $1;
				my $as  = $2;
				if ( $def =~ /\d+\.\d+\.\d+\.\d+/ ) {
					push(
						@{ $protocol->{neighbor} },
						{ address => $def, asNumber => $as, }
					);
				} else {
					$peerGroupAsNumbers->{$def} = $as;
				}
			}
			while ( $blob =~ /^\s*neighbor\s(\d+\.\d+\.\d+\.\d+)\s+peer-group\s+(\S+)/mig )	{
				if ( $peerGroupAsNumbers->{$2} ) {
					push(@{ $protocol->{neighbor} },
						{
							address  => $1,
							asNumber => $peerGroupAsNumbers->{$2},
						}
					);
				}
			}
		}

		# Check RIP version
		if ( $type eq 'rip' ) {
			my $version = 2;
			$version = 1 if ( $blob =~ /version\s1/mi );
			$protocol->{version} = $version;
		}
		$in->{routingProcesses}->{$type}->{$id} = $protocol;
	}
}

# _parse_interface_groupings
#
#  Does a pass of running-config to establish the interfaces which are "grouped" along several criteria--
#  deferral of an unnumbered interface to an interface whose address it is using as endpoint, channeling
#  of multiple interfaces, etc.
#
#  This sets up a hash, $in->{interfaceGroupings}, whose index is the "primary" interface of such a 
#  relationship (or the channel name in the case of EtherChannel, for example).
#
sub _parse_interface_groupings
{
    my ($in, $out) = @_;
    
    while ( $in->{"running_config"} =~ /^interface\s+(\S+)(.+?)^!/smig ) 
    {
	my $ifName = $1;
	my $ifBlob = $2;
	my $primaryIf = undef;
		
	
		# One case  deals with unnumbered interfaces
		if ( $ifBlob =~ /^\s*ip\s+unnumbered\s+(\S+)\s*$/mi ) 
		{
			$primaryIf = $1;
			
			# Assert presence of a hash bucket with the primary name of the interface
			$in->{interfaceGroupings}->{"$primaryIf"}->{"primaryInterface"} = 
			{ 
				ifName => $primaryIf,
				  ifGroupCharacteristic => "proxyAuthority"
			};
			my $proxyAffiliationDesc = { ifName => $ifName,
								ifGroupCharacteristic => "proxyAffiliation"};
			
			push (@{$in->{interfaceGroupings}->{"$primaryIf"}->{"secondaryInterfaces"}}, $proxyAffiliationDesc);
			# Rounding this up, We'll use an extra ref off of $in to store an index *by secondary interface*
			if (defined $primaryIf)
			{
				push (@{$in->{interfacesData}->{"$ifName"}->{ifGroupings}}, $primaryIf);
			}		
	
		}
		
		# Another deals independently with subinterfaces, which for IOS, we'll take from the interface name
		if ($ifName =~ /^(\S+)\.\d+/)
		{
			$primaryIf = $1;
	
			$in->{interfaceGroupings}->{"$primaryIf"}->{"primaryInterface"} = 
			{ 
				ifName => $primaryIf,
				  ifGroupCharacteristic => "primaryInterface"
			};
	
			my $subInterfaceDesc = { ifName => $ifName,
								ifGroupCharacteristic => "subInterface"};
			push (@{$in->{interfaceGroupings}->{"$primaryIf"}->{"secondaryInterfaces"}}, $subInterfaceDesc);									
			# Rounding this up, We'll use an extra ref off of $in to store an index *by secondary interface*
			if (defined $primaryIf)
			{
				push (@{$in->{interfacesData}->{"$ifName"}->{ifGroupings}}, $primaryIf);
			}		
		}
		
	}		
}

sub _full_int_name {

    # given a short name like "Fa3/15" return "FastEthernet3/15"
    my $name = trim(shift);
    for ($name) {
	s/^Fa(?=\d)/FastEthernet/;
	s/^Eth(?=\d)/Ethernet/;
	s/^Et(?=\d)/Ethernet/;
	s/^Gig?(?=\d)/GigabitEthernet/;
	s/^Po(?=\d)/PortChannel/;
    }
    return $name;
}

# just to make things easy for parse_routing, if this is EIGRP, then prepend the elements with
# the namespace for all enclosed elements.

sub _get_routing_element_namespace {
    my $rProt = shift;
    
    return (  $rProt eq "eigrp" ) ? "cisco:" : "" ;
    
}

# This routine is used to create a mapping of NAT "zones" to their corresponding interfaces
sub _parse_nat_interfaces 
{
	my $in = shift;
	my $natZoneMapping = undef;

	# Commence a scan of the interface sections, looking for the "ip nat" indicator
	while ( $in->{"running_config"} =~ /^interface\s+(\S+)(.+?)^!/smig ) {
		my $name      = $1;
		my $blob      = $2;
		
		if ($blob =~ /^\s*ip nat\s+(\S+)/mi)
		{
			push (@{$natZoneMapping->{"$1"}}, $name);
		}
	}
	return $natZoneMapping;
}

sub _parse_file_storage {

    # populate the deviceStorage elements of the chassis
    my ( $in, $out ) = @_;
    while ( $in->{show_fs} =~
	    /^\*?\s+(\d+)\s+(\d+)\s+\b(\S+)\b\s+[A-Za-z]+\s+(\S+):/mig )
    {
	my $storage = {
	    name        => $4,
	    storageType => $3,
	    size        => $1,
	    freeSpace   => $2,
	};

	if ( $storage->{storageType} !~ /opaque|nvram/i ) {
	    my $show = $in->{file_systems}->{ $storage->{name} };

	    if ( defined $show ) {
		$storage->{rootDir} = { name => "root", };
		if ( $show =~ /^\s*File\s+Length\s+Name\/status\s*$/ms
		    )    # 'show flash' format output
		{
		    while ( $show =~ /^\s*(\d+)\s+(\d+)\s+(\S+)\s*$/mig ) {
			my $file = {
			    size => $2,
			    name => $3,
			};

			push( @{ $storage->{rootDir}->{file} }, $file );
		    }

		    # this is the more accurate physical size
		    if ( $show =~ /^\s*(\d+)K bytes of processor board/mi ) {
			$storage->{size} = $1 * 1024;
		    }
		}
		elsif ( $show =~
			/^\s*-#-\s--length--\s-----date\/time------\spath\s*$/ms
		    )    # 'show disk' format output
		{
		    while (
			$show =~ /^\s*\d+\s+(\d+)\s(?:.+?\s){4}(\S+)\s*$/mig )
		    {
			my $file = {
			    size => $1,
			    name => $2,
			};

			push( @{ $storage->{rootDir}->{file} }, $file );
		    }
		}
		elsif ( $show =~
			/^\s*-#-\sED\s.+?type.+?crc.+?seek.+?nlen.+?length.+?date\/time.+name\s*$/ms
		    )    # 'show bootflash format output
		{
		    while ( $show =~
			    /^\s*\d+\s+(?:\S+\s+){5}(\d+)\s(?:\S+\s+){5}(\S+)\s*$/mig
			)
		    {
			my $file = {
			    size => $1,
			    name => $2,
			};

			push( @{ $storage->{rootDir}->{file} }, $file );
		    }
		}
		elsif ( $show =~
			/^\s*\d+\s+[-drwx]{4}\s+(\d+)\s+(?:.*?)\s+\d+:\d+:\d+\s+(?:.*?)\s+(\S+)\s*\n/ms
		    )    #show flash output that has no headers
		{
		    while ( $show =~
			    /^\s*\d+\s+[-drwx]{4}\s+(\d+)\s+(?:.*?)\s+\d+:\d+:\d+\s+(?:.*?)\s+(\S+)\s*\n/mig
			)
		    {
			my $file = {
			    size => $1,
			    name => $2,
			};

			push( @{ $storage->{rootDir}->{file} }, $file );
		    }
		}
	    }
	    $out->print_element( "deviceStorage", $storage );
	}
    }
}

sub _parse_cards {

    # populate the card and daughter card elements of the chassis
    my ( $in, $out ) = @_;

    if ( defined $in->{show_inventory} ) {
	my $first_chassis = 1;
	while ( $in->{show_inventory} =~ /(NAME:.+?)SN:\s+(\S+)/migs ) {
	    if ( !$first_chassis ) {
		my $chassisBlob = $1;
		my $chassisSN   = $2;
		my ($desc) = $chassisBlob =~ /DESCR:(.+)$/mi;
		$desc =~ s/"//g;
		my $card = { "core:description" => trim($desc) };
		my ($model) = $chassisBlob =~ /PID:\s+(\S+)/mi;

		$card->{"core:asset"}->{"core:assetType"} = "Card";
		$card->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
		    "Cisco";
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:modelNumber"} = $model;
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:serialNumber"} = $chassisSN;

		$out->print_element( "card", $card );
	    }
	    $first_chassis = 0;
	}
    }

    if ( defined $in->{show_module} ) {

#Mod Ports Card Type                              Model              Serial No.
#--- ----- -------------------------------------- ------------------ -----------
	while ( $in->{show_module} =~
		/^\s*(\d+)\s+(\d+)\s+\b(.+?)\s+((?:WS|7600)-\S+)\s+(\S+)\s*$/mig )
	{
	    my $no   = $1;
	    my $card = {
		slotNumber         => $no,
		portCount          => $2,
		"core:description" => $3,
	    };

	    $card->{"core:asset"}->{"core:assetType"} = "Card";
	    $card->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
		"Cisco";
	    $card->{"core:asset"}->{"core:factoryinfo"}->{"core:modelNumber"} =
		"Unknown";
	    $card->{"core:asset"}->{"core:factoryinfo"}->{"core:partNumber"} =
		$4;
	    $card->{"core:asset"}->{"core:factoryinfo"}->{"core:serialNumber"} =
		$5;

	    # now get the card status
	    if ( $in->{show_module} =~
		 /^\s*$no\s+[a-f0-9.]{14}\s+to\s+[a-f0-9.]{14}\s+(?:[\w.()]+)\s+(?:[\w.()]+)\s+(?:\S+)\s+(\S+)/mi
		)
	    {
		$card->{"status"} = $1;
	    }

	    # now get the HW, FW and SW versions from 'show mod ver' output
	    ($_) = $in->{show_mod_ver} =~ /^\s+$no(.+?)(?:^\s+\d+|^\S+)/mis;
	    if ( defined $_ ) {
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:hardwareVersion"} = trim($1)
		    if (/\bHw\s+:(.+)$/mi);
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:firmwareVersion"} = trim($1)
		    if (/\bFw\s+:(.+)$/mi);
		$card->{"softwareVersion"} = trim($1) if (/\bSw\s+:(.+)$/mi);
	    }

	    # now get any daughter cards (sub-mod)
	    if ( $in->{show_module} =~ /^Mod\s+Sub-Mod(.+?)^\s*$/msi ) {
		my $blob = $1;
		while ( $blob =~
			/^\s*$no(?:\/\d+)?\s+\b(.+?)\s+((?:WS|SPA)-\S+)\s+(\S+)\s+([\d.]+)\s+(\S+)/mig
		    )
		{
		    my $daughterCard = { "core:description" => $1, };
		    $daughterCard->{"core:asset"}->{"core:assetType"} = "Card";
		    $daughterCard->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:make"} = "Cisco";
		    $daughterCard->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $daughterCard->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $2;
		    $daughterCard->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $3;
		    $daughterCard->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $4;
		    $daughterCard->{"status"} = $5;

		    push( @{ $card->{daughterCard} }, $daughterCard );
		}
	    }
	    $out->print_element( "card", $card );
	}
    }
    elsif ( defined $in->{show_diag} ) {
	my $slotnum   = 0;
	my $slotcount = -1;
	my $cardref;
	my $desc             = "";
	my $hwVer            = "";
	my $snum             = "";
	my $pnum             = "";
	my $frupnum          = "";
	my $dslotnum         = -1;
	my $cardtype         = "";
	my $type             = "";
	my $mslotType        = "";
	my $mslotnum         = -1;
	my $mslotname        = "";
	my $dslottype        = "";
	my $nodesc           = 0;
	my $nodescencryptaim = 0;
	my $nodescaimmodule  = 0;
	my $nodescaimcarrier = 0;

	foreach my $line ( split( /\n/, $in->{show_diag} ) ) {

	    # mother card
	    if ( $line =~ /^Slot\s+(\d+):/i ) {
		$slotnum          = $1;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "";
		$dslottype        = "";
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # AIM mother card
	    elsif ( $line =~ /^AIM\s+slot\s+(\d+):\s+([A-Za-z-\/ ]+)/i ) {
		$slotnum          = $1 + 100;
		$mslotnum         = $1;
		$desc             = $2;
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "AIM";
		$dslottype        = "";
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # Encryption AIM mother card
	    elsif ( $line =~ /Encryption\s+AIM\s+(\d+):/i ) {
		$slotnum          = $1 + 300;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "Encryption AIM";
		$dslottype        = "";
		$nodescencryptaim = 1;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # Compression AIM mother card
	    elsif ( $line =~ /Compression\s+AIM\s+(\d+):/i ) {
		$slotnum          = $1 + 500;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "Compression AIM";
		$dslottype        = "";
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # ATM AIM mother card
	    elsif ( $line =~ /ATM\s+AIM\s*[: ]\s*(\d+)/i ) {
		$slotnum          = $1 + 900;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "ATM AIM";
		$dslottype        = "";
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # AIM module mother card
	    elsif ( $line =~ /AIM\s+Module\s+in\s+slot[: ]\s*(\d+)/i ) {
		$slotnum          = $1 + 800;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "physical";
		$mslotname        = "AIM Module";
		$dslottype        = "";
		$nodescencryptaim = 0;
		$nodescaimmodule  = 1;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # virtual mother card
	    elsif ( $line =~ /^Slot\s+(\d+)\s+\(virtual\):/i ) {
		$slotnum          = $1 + 200;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "virtual";
		$mslotname        = "(virtual)";
		$dslottype        = "";
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # logical mother card
	    elsif ( $line =~ /^Slot\s+(\d+):Logical_index\s+(\d+)/i ) {
		$slotnum          = $2 + 400;
		$mslotnum         = $1;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "logical";
		$mslotname        = "Logical_index";
		$dslottype        = "";
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "mother";
	    }

	    # WIC/VIC daughter card
	    elsif ( ( $line =~ /([A-Z\/]*IC) Slot\s+(\d+):/i ) ) {

		# handle no stop state
		if ($cardtype) {
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
			"Cisco";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:assetType"} = "Card";

		    if ( $dslotnum >= 0 ) {
			my $actualslotnum = $dslotnum;
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotNumber} =
			    $actualslotnum > 99
			    ? $actualslotnum -= 100
			    : $actualslotnum += 0;
		    }
		    if ($dslottype) {
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotType} = $dslottype;  						
		    }
		    elsif ($mslotType)
		    {
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotType} = $mslotType;
			
		    }
		}
		$dslotnum         = $2;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "";
		$dslottype        = $1;
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "daughter";
	    }

	    # (on Carrier Card) AIM daughter card
	    elsif ( $line =~ /AIM\s+on\s+Carrier\s+Card\s+(\d+):/i ) {
		if ($cardtype) {
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
			"Cisco";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:assetType"} = "Card";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{slotType} = $mslotType
			if ($mslotType);
		}
		$dslotnum         = $1;
		$desc             = "AIM on Carrier Card " . $1;
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "";
		$dslottype        = "AIM on Carrier Card";
		$type             = "";
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 1;
		$cardtype         = "daughter";
	    }

	    # PVDM daughter card
	    elsif (( $line =~ /(PVDM) Slot\s+(\d+):/i )
		   or ( $line =~ /(Packet Voice DSP Module) Slot\s+(\d+):/i ) )
	    {

		# handle no stop state
		if ( $cardtype ne "" ) {
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
			"Cisco";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:assetType"} = "Card";

		    if ( $dslotnum >= 0 ) {
			my $actualslotnum = $dslotnum;
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotNumber} =
			    $actualslotnum > 99
			    ? $actualslotnum -= 100
			    : $actualslotnum += 0;
		    }
		    if ($dslottype) {
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotType} = $dslottype;						
		    }
		    elsif ($mslotType) {
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotType} = $mslotType;				
		    }
		}
		$dslotnum         = $2 + 100;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "";
		$dslottype        = $1;
		$nodesc           = 1;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "daughter";
	    }

	    # PA Bay daughter card
	    elsif (( $line =~ /(PA\s+Bay)\s+(\d+)\s+Information:/i )
		   or ( $line =~ /(PA)\s+(\d+)\s+Information:/i ) )
	    {

		$dslotnum         = $2;
		$desc             = "";
		$hwVer            = "";
		$snum             = "";
		$pnum             = "";
		$frupnum          = "";
		$mslotType        = "";
		$dslottype        = $1;
		$nodesc           = 0;
		$nodescencryptaim = 0;
		$nodescaimmodule  = 0;
		$nodescaimcarrier = 0;
		$cardtype         = "daughter";
	    }

	    # description
	    elsif (
		($cardtype)
		and ( $desc eq "" )
		and (  ( $line =~ /\s*(.*\s+Port\s+adapter,\s+\S+\s+\S+)/i )
		       or ( $line =~ /\s*(.*\s+[a-z]+.\s+Port\s+adapter)/i )
		       or ( $line =~ /\s*(.*daughter\s+card)/i ) )
		)
	    {
		$desc = trim($1);
	    }

	    # description and Hardware revision
	    elsif (
		($cardtype)
		and ( $line =~
		      /\s+([^,]+),\s+HW\s+rev\s+([^,]+),\s+board\s+revision\s+(\S+)/i
		)
		)
	    {
		$desc = $1;
		$hwVer = $2 if ($hwVer);
	    }

	    # Hardware revision and board revision
	    elsif (
		($cardtype)
		and ( $line =~
		      /Hardware revision\s([^ ]+)\s+Board revision\s+([^ ]+)/i )
		)
	    {
		$hwVer = $1;
	    }

	    # description -- 2nd line after "slot" as default case
	    elsif ( ( $desc eq "" )
		    and ( $nodesc == 0 )
		    and ( $nodescencryptaim == 0 )
		    and ( $nodescaimmodule == 0 )
		    and ( $nodescaimcarrier == 0 )
		    and ( $line =~ /\s+(.*)/ ) )
	    {
		$desc = $1;
	    }

	    elsif (
		($cardtype)
		and (
		    (
		     $line =~
		     /\s+(Hardware|(.*),\s+HW)\s+(Revision|rev)\s+(\d+.\d+)[,\s+]\s+Board\s+Revision\s+(\S+)/i
		    )
		    or
		    ( $line =~ /((HW)\s+(rev))\s+(\S+),\s+Board\s+revision/i )
		)
		)
	    {
		$hwVer = $4 if ( !$hwVer );
	    }

	    # hardwareVersion
	    elsif ( ($cardtype)
		    and ( $line =~ /\s+Hardware\s+Revision\s+:\s+(\d+.\d+)/i ) )
	    {
		$hwVer = $1 if ( !$hwVer );
	    }

	    # serialNumber must precede the next serial number and part number
	    elsif (
		($cardtype)
		and ( $line =~
		      /\s+Serial\s+Number[:\s+]\s+(\S+|\d+)\s+Part\s+Number[:\s+]\s+(\S+|\d+).*/i
		)
		)
	    {
		$snum = $1 if ( !$snum );
		$pnum = $2 if ( !$pnum );

		if ( ( $dslottype =~ /PA/ ) and ( $cardtype eq "daughter" ) ) {
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
			"Cisco";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:assetType"} = "Card";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{slotNumber} = $dslotnum
			if ( $dslotnum >= 0 );
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{slotType} = $mslotType
			if ($mslotType);
		}
	    }

	    # serialNumber
	    elsif ( ( $cardtype ne "" )
		    and ( $line =~ /\s+\S+\s+Serial\s+Number\s+:\s+(\w+)/i ) )
	    {
		$snum = $1 if ( !$snum );
	    }

	    # partNumber
	    elsif ( ( $cardtype ne "" )
		    and ( $line =~ /\s+Part\s+Number\s+:\s+(\S+)/i ) )
	    {
		if ( $pnum eq "" ) {
		    $pnum = $1;
		}
	    }

	    # FRU partNumber
	    elsif (
		( $cardtype ne "" )
		and (  ( $line =~ /FRU Part Number[: ]\s*([^ ]+)/i )
		       or ( $line =~ /Product \(FRU\) Number\s+:\s+([^ ]+)/i ) )
		)
	    {
		$frupnum = $1 if ( !$frupnum );
	    }

	    # connectorType
	    elsif ( ( $cardtype ne "" )
		    and ( ( $line =~ /Connector\s+Type\s+[: ]\s+(.*)/i ) ) )
	    {
		$type = $1;
	    }

	    elsif ( ( $cardtype ne "" )
		    and ( $line =~ /EEPROM\s+contents\s+\(hex\):/i ) )
	    {
		if ( $cardtype eq "mother" ) {
		    $cardref->{$slotnum}->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:make"} = "Cisco";
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{"core:asset"}->{"core:assetType"} =
			"Card";
		    $cardref->{$slotnum}->{slotNumber} = $mslotnum;
		    $cardref->{$slotnum}->{slotType}   = $mslotType;
		}
		elsif ( $cardtype eq "daughter" ) {
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:description"} = $desc
			if ($desc);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
			"Cisco";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:modelNumber"} = "Unknown";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:hardwareVersion"} = $hwVer
			if ($hwVer);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $snum
			if ($snum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:partNumber"} = $pnum
			if ($pnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:fruPartNumber"} = $frupnum
			if ($frupnum);
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{"core:asset"}->{"core:assetType"} = "Card";
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{slotNumber} = $dslotnum
			if ( $dslotnum >= 0 );
		    $cardref->{$slotnum}->{daughterCard}->{$dslotnum}
		    ->{slotType} = $mslotType
			if ( $mslotType and $mslotType >= 0 );

		    if ( $dslotnum >= 0 ) {
			my $actualslotnum = $dslotnum;
			$cardref->{$slotnum}->{daughterCard}->{$dslotnum}
			->{slotNumber} =
			    $actualslotnum > 99
			    ? $actualslotnum -= 100
			    : $actualslotnum += 0;
		    }
		}
		$cardtype = "";
	    }
	    elsif ( $line =~ /Controller\s+Memory\s+Size:\s+(\d+)\s+MBytes/i ) {
		my $cardmem = $1 * 1024 * 1024;
		$cardref->{$slotnum}->{memory}->{size} = $cardmem;
	    }
	}

	foreach my $key ( keys %$cardref ) {
	    my @dcs;

	    # pull off any daughter cards
	    if ( defined $cardref->{$key}->{daughterCard} ) {
		foreach my $dc ( keys %{ $cardref->{$key}->{daughterCard} } ) {
		    push( @dcs, $cardref->{$key}->{daughterCard}->{$dc} );
		}
		$cardref->{$key}->{daughterCard} = \@dcs if ( @dcs > 0 );
	    }
	    $out->print_element( "card", $cardref->{$key} );
	}
    }
    elsif ( defined $in->{version} ) {
	my ($stack) =
	    $in->{version} =~
	    /^switch\s+Ports\s+Model\s+SW\sVersion\s+SW\sImage\s*\n--.*\n((?:\*?\s+\d+\s+\d+\s+\S+\s+\S+\s+\S+\s*\n)*)/mi;

	if ($stack) {

#			my $members = 0; #count number of member devices (the ones without a *). If > 0 then populate stack information else ignore as it is a stack only of the master.
#			while ( $stack =~ /^(\s+)(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/mig )
#			{
#				$members++;
#			}

	    #			if ( $members > 0 )
	    #			{
	    while ( $stack =~
		    /^(\*?\s+)(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/mig )
	    {
		my $member = $2;
		$member = "0" . $member if ( $member < 10 );
		my $ports   = $3;
		my $model   = $4;
		my $sw_ver  = $5;
		my $sw_imag = $6;
		my $master  = $1 =~ /\*/mi ? 1 : 0;

		my $card;

		$card->{"core:asset"}->{"core:assetType"} = "Card";
		$card->{"core:asset"}->{"core:factoryinfo"}->{"core:make"} =
		    "Cisco";
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:partNumber"} = "$model";
		$card->{"core:asset"}->{"core:factoryinfo"}
		->{"core:modelNumber"} = "$model";

		if ( $master == 1 ) {
		    my ($serial) =
			$in->{version} =~ /^System\sserial\snumber\s+:\s+(\S+)/mi;
		    $card->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $serial
			if ($serial);

		    $card->{"core:description"} = "Master Switch";
		}
		else {
		    my ($stack_memdef) =
			$in->{version} =~
			/^Switch\s+$member\s*\n-+\n((?:\S+.*\n)*)\s*\n/mi;

		    my ($serial) =
			$stack_memdef =~ /^System\sserial\snumber\s+:\s+(\S+)/mi;
		    $card->{"core:asset"}->{"core:factoryinfo"}
		    ->{"core:serialNumber"} = $serial
			if ($serial);

		    $card->{"core:description"} = "Member Switch";
		}

		$card->{portCount}         = $ports;
		$card->{slotNumber}        = $member;
		$card->{slotType}          = "Stack";
		$card->{"softwareVersion"} = "$sw_ver ($sw_imag)";

		$out->print_element( "card", $card );
	    }

	    #			}
	}
    }
}

sub parse_services {
    my ( $input, $output ) = @_;

    $output->open_element("services");

    parse_ntp( $input, $output );
    parse_jailer( $input, $output );

    #other service parsers go here

    $output->close_element("services");
}

sub parse_ntp {
    my ( $in, $out ) = @_;

    my $ntp;

    my $cfg = $in->{running_config};

    if ( $cfg =~ /^s?ntp/mi ) {
	while ( $cfg =~ /^ntp\s+access-group\s+(\S+)\s+(\d+)/mig ) {

	    #ntp access-group (query-only) (98)

	    my $access;

	    $access->{filterList} = $2;
	    $access->{groupType}  = $1;

	    push( @{ $ntp->{"accessGroup"} }, $access );
	}

#IOS has no specific NTP enable command - search for ntp master, ntp server or ntp peer to see if it is active
	$ntp->{enabled} =
	    ( $cfg =~
	      /^(?:ntp\s+master\s+\d+)|(?:ntp\s+server\s+\d+\.\d+\.\d+\.\d+)|(?:ntp\s+peer\s+\d+\.\d+\.\d+\.\d+)/mi
	    )
	    ? "true"
	    : "false";

	my ($trusted_key) = $cfg =~ /^ntp\s+trusted-key\s+(\d+)/mi;

	while ( $cfg =~
		/^ntp\s+authentication-key\s+(\d+)\s+(md5)\s+(\S+)(?:\s+.*$)/mig )
	{

	    #ntp authentication-key 2 md5 0948480E150E1D160D000321212024353E281701 7

	    my $key;

	    $key->{keyNumber} = $1;
	    my $the_key  = $3;
	    my $the_rest = $4;

	    $key->{keyType}   = ( $2 =~ /md5/mi ) ? "MD5" : "Plain text";
	    $key->{serverKey} = $the_key;
	    $key->{trusted}   =
		( $key->{keyNumber} == $trusted_key ) ? "true" : "false";

	    push( @{ $ntp->{"localKeySettings"} }, $key );
	}

	if ( $cfg =~ /^ntp\s+master\+(\d+)/mi ) {
	    $ntp->{masterStratum} = $1;
	}

	if ( $cfg =~ /^ntp\s+max-associations\s+(\d+)/mi ) {
	    $ntp->{maxAssociations} = $1;
	}

	while ( $cfg =~
		/^(sntp|ntp)\s+(server|peer)\s+(\d+\.\d+\.\d+\.\d+)\s*(.*)$/mig )
	{

#(ntp) (peer|server) (1.2.3.4) (key <number> prefer source <interfaceName> version <1|2|3>)

	    my $server;

	    my $protocol  = $1;
	    my $mode      = $2;
	    my $server_ip = $3;
	    my $rest      = $4;

	    if ( $protocol eq "sntp" ) {
		$server->{mode} = "client";
	    }
	    else {
		$server->{mode} = ( $mode =~ /peer/ ) ? "peer" : "server";
	    }

	    $server->{serverIPaddress} = $server_ip;

	    $server->{protocolVersion}->{protocol} = lc $protocol;
	    $server->{protocolVersion}->{version}  =
		( $rest =~ /version\s+(\d+)/ ) ? "$1" : "1";

	    if ( $rest =~ /key\s+(\d+)/mi ) {
		$server->{keyNumber} = $1;
	    }

	    if ( $rest =~ /source\s+(\S+)/mi ) {
		$server->{sourceInterface} = $1;
	    }

	    $server->{preference} = ( $rest =~ /prefer/ ) ? "true" : "false";

	    push( @{ $ntp->{"server"} }, $server );
	}

	if ( $cfg =~ /^ntp\s+source(?:-interface)\s+(\S+)/mi ) {
	    $ntp->{sourceInterface} = $1;
	}

	$ntp->{useAuthentication} =
	    ( $cfg =~ /^ntp authenticate$/mi ) ? "true" : "false";
    }

    $out->print_element( "ntp", $ntp ) if ($ntp);
}

sub parse_jailer {
    my ( $in, $out ) = @_;

    my $cfg = $in->{running_config};

    my $jailer;

    #first go check any configured key-chains
    if ( $cfg =~ /^key\s+chain/mi ) {
	my $keyFob;

	while ( $cfg =~ /^key\s+chain\s+(\S+)\s*\n((?:\s+.*\n)*)/mig ) {
	    my $keyRing;

	    $keyRing->{ringName} = $1;
	    my $key_blobs = $2;

	    while ( $key_blobs =~ /^\skey\s+(\d+)\s*\n((?:\s\s.*\n)*)/mig ) {
		my $aKey;

		$aKey->{number} = $1;
		my $remainder = $2;

		my ($keyString) =
		    $remainder =~ /^\s+key-string\s+(\d+)\s+(\S+)\s*$/mi;
		$aKey->{string} = $2;

		push( @{ $keyRing->{"key"} }, $aKey ) if ($aKey);
	    }

	    push( @{ $keyFob->{"keyring"} }, $keyRing ) if ($keyRing);
	}

	$jailer->{keyfob} = $keyFob if ($keyFob);
    }

    #place other Jailer related parsing here (eg. SSL certs/RSA keys)

    $out->print_element( "jailer", $jailer ) if ($jailer);
}

sub parse_mpls {
    my ( $in, $out ) = @_;

    my $mpls;

    $out->print_element( 'mpls', $mpls ) if ( defined $mpls );
}

sub parse_qos {
    my ( $in, $out ) = @_;

    my $precedence_vals = {
	'0' => 'routine',
	'1' => 'priority',
	'2' => 'immediate',
	'3' => 'flash',
	'4' => 'flash-override',
	'5' => 'critical',
	'6' => 'internet',
	'7' => 'network',
    };

    my $qos;

    #parse class_maps

    my $class_maps;

    while ( $in->{running_config} =~
	    /^class-map\s+match-(any|all)\s+(\S+)\n((?:\s+.*\n)*)/mig )
    {
	my $class_map;

	my $type = $1;
	my $name = $2;
	my $rest = $3;

	$class_map->{domain} = lc $name;
	$class_map->{match}  = lc $type;

	while ( $rest =~ /^\s+match\s+(.*)$/mig ) {
	    my $line = $1;

	    my $match;

	    $match->{reverseLogic} = 'true' if ( $line =~ /^not\s+/ );

	    if ( $line =~ /access\-group\s+(?:name\s+)?(\S+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchAccessGroup"} }, $match );
	    }
	    elsif ( $line =~ /class\-map\s+(\S+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchClassMap"} }, $match );
	    }
	    elsif ( $line =~ /destination\-address\s+mac\s+(\S+)/mi ) {
		$match->{value} = strip_mac($1);
		push( @{ $class_map->{"matchDestinationMAC"} }, $match );
	    }
	    elsif ( $line =~ /discard\-class\s+(\d+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchDiscardClass"} }, $match );
	    }
	    elsif ( $line =~ /fr\-de/mi ) {
		$match->{value} = 'true';
		push( @{ $class_map->{"matchFrDE"} }, $match );
	    }
	    elsif ( $line =~ /flow\s+(\S+)/mi ) {
		$match->{value} = lc $1;
		push( @{ $class_map->{"matchFlow"} }, $match );
	    }
	    elsif ( $line =~ /fr\-dlci\s+(\d+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchFrDLCI"} }, $match );
	    }
	    elsif ( $line =~ /input\-interface\s+(\S+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchInputInterface"} }, $match );
	    }
	    elsif ( $line =~ /cos((?:\s+\d+)+)/mi ) {
		my $blob = $1;
		while ( $blob =~ /\s+(\d+)/mig ) {
		    push( @{ $match->{"value"} }, $1 );
		}

		push( @{ $class_map->{"matchIpCOS"} }, $match );
	    }
	    elsif ( $line =~ /ip dscp((?:\s+\S+)+)/mi )    #dscp string
	    {
		my $blob       = $1;
		my $foundMatch = 0;
		while ( $blob =~ /\s+(\d+)/mig ) {
		    $foundMatch = 1;
		    push( @{ $match->{"value"} }, $1 );
		}
		push( @{ $class_map->{"matchIpDSCP"} }, $match ) if $foundMatch;
	    }
	    elsif ( $line =~ /ip\s+precedence((?:\s+\S+)+)/mi )   #ip precedence
	    {
		my $blob = $1;
		while ( $blob =~ /\s+(\d+)/mig ) {
		    my $new_val = $precedence_vals->{$1};
		    push( @{ $match->{"value"} }, $new_val );
		}

		push( @{ $class_map->{"matchIpPrecedence"} }, $match );
	    }
	    elsif ( $line =~
		    /packet\s+length((?:\s+(?:min|max)\s+\d+)*)/mi )  #packet length
	    {
		my $blob = $1;
		my ($min) = $blob =~ /min\s+(\d+)/mi;
		my ($max) = $blob =~ /max\s+(\d+)/mi;

		$match->{minValue} = $min if $min;
		$match->{maxValue} = $max if $max;

		push( @{ $class_map->{"matchPacketLength"} }, $match );
	    }
	    elsif ( $line =~ /protocol\s+(\S+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchProtocol"} }, $match );
	    }
	    elsif ( $line =~ /qos\-group\s+(\d+)/mi ) {
		$match->{value} = $1;
		push( @{ $class_map->{"matchQosGroup"} }, $match );
	    }
	    elsif ( $line =~ /ip\s+rtp\s+(\d+\s+\d+)/mi )    #real-time protocol
	    {
		my $blob = $1;
		my ( $start, $range ) = $blob =~ /(\d+)\s+(\d+)/mi;

		$match->{startPort} = $start;
		$match->{endPort}   = $start + $range;

		push( @{ $class_map->{"matchRTPport"} }, $match );
	    }
	    elsif ( $line =~ /source\-address\s+mac\s+(\S+)/mi ) {
		$match->{value} = strip_mac($1);
		push( @{ $class_map->{"matchSourceMAC"} }, $match );
	    }
	    elsif ( $line =~ /vlan((?:\s+\d+(?:\-\d+)?)+)/mi ) {
		my $vlan_string = $1;
		while ( $vlan_string =~ /((?:\d+)(?:\-\d+)?)/mig ) {
		    push( @{ $match->{"value"} }, $1 );
		}

		push( @{ $class_map->{"matchVlan"} }, $match );
	    }

	    #parsing for any other class map elements to go here

	}

	push( @{ $class_maps->{"classMap"} }, $class_map );
    }

    push( @{ $qos->{"qosClasses"} }, $class_maps ) if ($class_maps);

    #parse policy_maps

    my $policy_maps;

    while ( $in->{running_config} =~ /^policy-map\s+(\S+)\n((?:\s.*?\n)*)/mig )
    {
	my $policy_map;

	my $name = $1;
	my $rest = $2;

	$policy_map->{name} = $name;

	while ( $rest =~ /^\sclass\s+(\S+)\n((?:\s\s+.*?\n)*)/mig ) {
	    my $class_name = $1;
	    my $class_blob = $2;

	    my $class;

	    $class->{classDomain} = $class_name;

	    while ( $class_blob =~ /^\s+(.*)$/mig ) {
		my $action = $1;

		if ( $action =~ /^set atm-clp/mi ) {
		    $class->{atmCLP} = 'true';
		}
		elsif ( $action =~ /^set\s+cos\s+(\d+)/mi ) {
		    $class->{cosValue} = $1;
		}
		elsif ( $action =~ /^set\s+discard\-class\s+(\d+)/mi ) {
		    $class->{discardClass} = $1;
		}
		elsif ( $action =~ /^set\s+ip\s+dscp\s+(\d+)/mi ) {
		    $class->{ipDSCP} = $1;
		}
		elsif ( $action =~ /^set\s+fr\-de/mi ) {
		    $class->{frDE} = 'true';
		}
		elsif ( $action =~ /^set\s+ip\s+precedence\s+(\d+)/mi ) {
		    $class->{ipPrecedence} = $precedence_vals->{$1};
		}
		elsif ( $action =~ /^set\s+qos\-group\s+(\S+)/mi ) {
		    $class->{qosGroup} = $1;
		}
		elsif ( $action =~ /^drop/mi ) {
		    $class->{drop} = 'true';
		}
		elsif ( $action =~ /^bandwidth\s+(.+)/mi ) {
		    my $line = $1;
		    my $result;

		    if ( $line =~ /^(\d+)$/mi ) {
			$result->{bandwidthBits} = $1;
		    }
		    elsif ( $line =~ /^remaining\s+percent\s+(\d+)/mi ) {
			$result->{remainingPercent} = $1;
		    }
		    elsif ( $line =~ /^percent\s+(\d+)/mi ) {
			$result->{bandwidthPercent} = $1;
		    }

		    $class->{bandwidth} = $result if ($result);
		}
		elsif ( $action =~ /^shape\s+(average|peak)\s+(.*)/mi ) {
		    my $type = $1;
		    my $rest = $2;

		    my $shape;

		    $shape->{type} = $type;

		    if ( $rest =~ /percent\s+(.*)/mi ) {
			my $vals = $1;

			my ( $per, $sus, $exc ) =
			    $vals =~
			    /(\d+)(?:\s+(\d+)\s+ms)?(?:\s+(\d+)\s+ms)?/mi;

			$shape->{percent}->{cirPercentage}   = $per;
			$shape->{percent}->{sustainedPeriod} = $sus if ($sus);
			$shape->{percent}->{excessPeriod}    = $exc if ($exc);
		    }
		    else {
			my ( $bps, $sus, $exc ) =
			    $rest =~ /(\d+)(?:\s+(\d+))?(?:\s+(\d+))?/mi;

			$shape->{rate}->{targetRate}    = $bps;
			$shape->{rate}->{sustainedRate} = $sus if ($sus);
			$shape->{rate}->{excessRate}    = $exc if ($exc);
		    }

		    $class->{shaping} = $shape;
		}
		elsif ( $action =~
			/^police\s+(\d+)\s+(\d+)(?:\s+(\d+))?\s+conform-action\s+(.*?)\s+exceed-action\s+(.*?)(?:\sviolate-action\s+(.*?))\s*$/mi
		    )
		{
		    my $bps         = $1;
		    my $burst_norm  = $2;
		    my $burst_max   = $3;
		    my $conf_action = $4;
		    my $exce_action = $5;
		    my $viol_action = $6;

		    my $cir = { bandwidthBits => $bps };

		    my $police = {
			bps         => $cir,
			normalBurst => $burst_norm,
		    };

		    $police->{maxBurst}      = $burst_max if ($burst_max);
		    $police->{conformAction} = _set_action($conf_action);
		    $police->{exceedAction}  = _set_action($exce_action);
		    $police->{violateAction} = _set_action($viol_action)
			if ($viol_action);

		    $class->{police} = $police;
		}

		#TODO: priority bandwidth <value> - is this used by anyone?

	    }

	    push( @{ $policy_map->{"class"} }, $class );
	}

	push( @{ $policy_maps->{"policyMap"} }, $policy_map );
    }

    push( @{ $qos->{"qosPolicies"} }, $policy_maps ) if ($policy_maps);

    #do any other QOS parsing here

    $out->print_element( 'qos', $qos ) if ( defined $qos );
}

sub _set_action {
    my $action_string = shift;

    my $action;

    if ( $action_string =~ /^continue$/mi ) {
	$action->{continue} = "true";
    }
    elsif ( $action_string =~ /^drop$/mi ) {
	$action->{drop} = "true";
    }
    elsif ( $action_string =~ /^set-dscp-continue\s+(\S+)\s*$/mi ) {
	$action->{setDSCPContinue} = $1;
    }
    elsif ( $action_string =~ /^set-dscp-transmit\s+(\S+)\s*$/mi ) {
	$action->{setDSCPTransmit} = $1;
    }
    elsif ( $action_string =~ /^set-mpls-exp-continue\s+(\S+)\s*$/mi ) {
	$action->{setMPLSexpContinue} = $1;
    }
    elsif ( $action_string =~ /^set-mpls-exp-transmit\s+(\S+)\s*$/mi ) {
	$action->{setMPLSexpTransmit} = $1;
    }
    elsif ( $action_string =~ /^set-prec-continue\s+(\S+)\s*$/mi ) {
	$action->{setPrecedenceContinue} = $1;
    }
    elsif ( $action_string =~ /^set-prec-transmit\s+(\S+)\s*$/mi ) {
	$action->{setPrecedenceTransmit} = $1;
    }
    elsif ( $action_string =~ /^set-qos-continue\s+(\S+)\s*$/mi ) {
	$action->{setQosContinue} = $1;
    }
    elsif ( $action_string =~ /^set-qos-transmit\s+(\S+)\s*$/mi ) {
	$action->{setQosGroupTransmit} = $1;
    }
    elsif ( $action_string =~ /^transmit$/mi ) {
	$action->{transmit} = "true";
    }

    return $action;
}

# Scans for a route map, and begins the parsing to create a Ziptie::Addressing::RouteFilter
# from it if found.

sub _parse_route_map
{
	my ($in, $rmName) = @_;
	my $rfOut = ZipTie::Addressing::RouteFilter->new(0);
	
	return $rfOut;
	
}

sub _is_physical {
    my $name = shift;
    if ( $name =~ /\.\d+$/ ) {
	return "false";
    }
    elsif ( $name =~ /seri|eth|gig|^fe|^fa|token|[bp]ri/i ) {
	return "true";
    }
    else {
	return "false";
    }
}

sub _apply_masks {

    # remove any trivial pieces of the configuration that
    # may cause improper diffs.

    my $text = shift;
    $text =~
	s/\n^ntp clock-period.+$//m;   # remove the variable ntp clock-period line
    $text =~ s/\n^!\s*NVRAM config last updated at.+$//m
	;                              # remove the NVRAM updated time comment
    $text =~ s/\n^!\s*Last configuration change at.+$//m
	;                              # remove the Last config change comment
    return $text;
}

#  Presuming the interface -> subnet structure returned from parse_interfaces to find a primary host
#  address associated with the interface.

sub _mapIfNameToPrimaryAddr
{
	my ($ifName, $ifSubnets) = @_;
	my $iMapAddr = undef;
	
	if (defined $ifSubnets->{$ifName})
	{
		# +++wayne this only supports one IP per interface
		$iMapAddr = $ifSubnets->{$ifName}->[0]->getHostIp();
	}
	
	return $iMapAddr;
}

1;
